/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Humberto Longo                                    */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

///////////////////////////////////////////////////////////////////////////////
// FILE : CarpIO.cpp                                                         //
//---------------------------------------------------------------------------//
// Class that read a Carp instance data and converts it to a Cvrp instance   //
//---------------------------------------------------------------------------//
// AUTHOR : Humberto Jos Longo                                              //
//---------------------------------------------------------------------------//
// CREATED ON : 27/02/2004                                                   //
// LAST MODIFIED :                                                           //
///////////////////////////////////////////////////////////////////////////////

//---------------------------------------------------------------------------//

#include "CarpIO.h"

//---------------------------------------------------------------------------//

CarpIO::CarpIO()
{
  nNodes      = 0;
  nArcs       = 0;
  reqArcs     = 0;
  cvrpSize    = 0;
  nVehicles   = 0;
  capacity    = 0;
  carpLB      = 0;
  carpUB      = 0;
  nodeTabSize = 0;
  sumDem      = 0;
  sumDist     = 0;
  edgeTab     = NULL;
  nodeTab     = NULL;
  carpDem     = NULL;
  cvrpDem     = NULL;
  carpDist    = NULL;
  cvrpDist    = NULL;
}

//---------------------------------------------------------------------------//

CarpIO::~CarpIO()
{
  register int i = 0;

  if (edgeTab != NULL)
  {
    for (i = 0; i <= nArcs; i++)
      delete [] edgeTab[i];

    delete [] edgeTab;
  }

  if (nodeTab != NULL)
  {
    for (i = 0; i < nodeTabSize; i++)
      delete [] nodeTab[i];

    delete [] nodeTab;
  }

  if (carpDem != NULL)
  {
    delete [] carpDem;
  }

  if (cvrpDem != NULL)
  {
    delete [] cvrpDem;
  }

  if (carpDist != NULL)
  {
    for (i = 0; i < nNodes; i++)
      delete [] carpDist[i];

    delete [] carpDist;
  }

  if (cvrpDist != NULL)
  {
    for (i = 0; i < nNodes; i++)
      delete [] cvrpDist[i];

    delete [] cvrpDist;
  }
}

//---------------------------------------------------------------------------//

void CarpIO::readData( char* _fname )
{
  register int   i     = 0;
  register int   j     = 0;
           FILE* fdata = NULL;

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  strcpy( fname, _fname );

  fdata = fopen( fname, "rt" );
  if (fdata == NULL)
  {
    printf( "Unable to open file %s\n", fname );
    throw(1);
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  fscanf( fdata, "%d %d", &nNodes, &nArcs );
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  carpDem = new double [nArcs + 1];

  edgeTab = new int* [nArcs + 1];
  for (i = 0; i <= nArcs; i++)
    edgeTab[i] = new int [2];

  carpDist = new double* [nNodes];
  for (i = 0; i < nNodes; i++)
    carpDist[i] = new double [nNodes];

  for (i = 0; i < nNodes; i++)
  {
    carpDist[i][i] = 0.0;
    for (j = i + 1; j < nNodes; j++)
    {
      carpDist[i][j] = INFINITY_VALUE;
      carpDist[j][i] = INFINITY_VALUE;
    }
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  edgeTab[0][0] = 0;
  edgeTab[0][1] = 0;
  for (i = 1; i <= nArcs; i++)
  {
    fscanf( fdata, "%d %d", &edgeTab[i][0], &edgeTab[i][1] );
    fscanf( fdata, "%lf %lf", &carpDist[edgeTab[i][0]][edgeTab[i][1]], &carpDem[i] );

    carpDist[edgeTab[i][1]][edgeTab[i][0]] = carpDist[edgeTab[i][0]][edgeTab[i][1]];

    if (carpDem[i] > 0)
      reqArcs++;

    sumDem  += carpDem[i];
    sumDist += carpDist[edgeTab[i][0]][edgeTab[i][1]];
  }

  printf( "\nCarp : |E| = %d  |R| = %d  total dist. = %.0lf  total dem. = %.0lf\n\n",
          nArcs, reqArcs, sumDem, sumDist );
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // - - - - - - - - - - - - - - - - - - - - - - -
  fscanf( fdata, "%d %d", &nVehicles, &capacity );
  // - - - - - - - - - - - - - - - - - - - - - - -

  // - - - - - - - - - - - - - - - - - - - - -
  fscanf( fdata, "%d %d", &carpLB, &carpUB );
  carpUB++;
  // - - - - - - - - - - - - - - - - - - - - -

  // - - - - - - -
  fclose( fdata );
  // - - - - - - -
}

//---------------------------------------------------------------------------//

// - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Transforming Arc Routing int Node Routing Problems
// W.-L. Pearn, A. Assad and B. L. Golden
// Computers and Operations Research 14 (4), 285-288, 1987
// ==> Replaces each arc (i,j) of the carp graph instance
//     with three nodes (s_ij, m_ij, s_ji).
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
void CarpIO::toCvrpByPearn()
{
  register int i = 0;
  register int j = 0;
  register int k = 0;
  register int x = 0;
  register int y = 0;

  int aux1 = 0;
  int aux2 = 0;

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  cvrpSize    = (3 * reqArcs) + 1;
  nodeTabSize = (3 * nArcs) + 1;
  carpLB     *= 4;
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  cvrpDist = new double* [cvrpSize];
  for (i = 0; i < cvrpSize; i++)
    cvrpDist[i] = new double [cvrpSize];

  for (i = 0; i < cvrpSize; i++)
  {
    cvrpDist[i][i] = 0.0;
    for (j = 0; j < i; j++)
    {
      cvrpDist[i][j] = INFINITY_VALUE;
      cvrpDist[j][i] = INFINITY_VALUE;
    }
  }

  cvrpDem = new double [cvrpSize];

  nodeTab = new int* [nodeTabSize];
  for (i = 0; i < nodeTabSize; i++)
    nodeTab[i] = new int [4];
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  nodeTab[0][0] = 0;
  nodeTab[0][1] = 0;
  nodeTab[0][2] = 0;
  nodeTab[0][3] = 0;

  for (i = 1; i <= nArcs; i++)
  {
    carpDist[edgeTab[i][0]][edgeTab[i][1]] *= 4;
    carpDist[edgeTab[i][1]][edgeTab[i][0]]  = carpDist[edgeTab[i][0]][edgeTab[i][1]];

    j = (3 * i) - 2;
    nodeTab[j][0] = 1;
    nodeTab[j][1] = edgeTab[i][0];
    nodeTab[j][2] = edgeTab[i][1];
    nodeTab[j][3] = i;

    j++;
    nodeTab[j][0] = 2;
    nodeTab[j][1] = edgeTab[i][0];
    nodeTab[j][2] = edgeTab[i][1];
    nodeTab[j][3] = i;

    j++;
    nodeTab[j][0] = 3;
    nodeTab[j][1] = edgeTab[i][1];
    nodeTab[j][2] = edgeTab[i][0];
    nodeTab[j][3] = i;
  }

  fw.shortestPaths( nNodes, carpDist );
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // Distance between depot node and nodes s_ij
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  for (x = 1, j = 1; j < nodeTabSize; j++)
  {
    if (carpDem[nodeTab[j][3]] == 0)
      continue;

    if (nodeTab[j][0] != 2)
    {
      cvrpDist[0][x] = (carpDist[nodeTab[j][1]][nodeTab[j][2]] / 4) +
                       (fw.minDist[nodeTab[j][1]][0]);
//      cvrpDist[0][x] = 0 * (carpDist[nodeTab[j][1]][nodeTab[j][2]] / 4) +
//                       (fw.minDist[nodeTab[j][1]][0]);
      cvrpDist[x][0] = cvrpDist[0][x];
    }

    x++;
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // Distance between nodes m_ij and nodes s_ij and s_ji
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  for (x = 2, j = 2; j < nodeTabSize; j += 3)
  {
    if (carpDem[nodeTab[j][3]] == 0)
      continue;

    cvrpDist[x][x-1] = (carpDist[nodeTab[j][1]][nodeTab[j][2]] / 4);
//    cvrpDist[x][x-1] = 2 * (carpDist[nodeTab[j][1]][nodeTab[j][2]] / 4);
    cvrpDist[x-1][x] = cvrpDist[x][x-1];
    
    cvrpDist[x][x+1] = (carpDist[nodeTab[j][1]][nodeTab[j][2]] / 4);
//    cvrpDist[x][x+1] = 2 * (carpDist[nodeTab[j][1]][nodeTab[j][2]] / 4);
    cvrpDist[x+1][x] = cvrpDist[x][x+1];

    x += 3;
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // Distance between two nodes s_ij
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  for (x = 1, i = 1; i < nodeTabSize; i++)
  {
    if (carpDem[nodeTab[i][3]] == 0)
      continue;

    if (nodeTab[i][0] == 2)
    {
      x++;
      continue;
    }
    
    k = 0;

    for (y = 1, j = 1; j < nodeTabSize; j++)
    {
      if (carpDem[nodeTab[j][3]] == 0)
        continue;

      if (nodeTab[j][0] == 2)
      {
        y++;
        continue;
      }

      k = 1;

      if ((nodeTab[i][1] == nodeTab[j][1]) && (nodeTab[i][2] == nodeTab[j][2]))
      {
        cvrpDist[x][y] = 0.0;
      }
      else
        if ((nodeTab[i][1] == nodeTab[j][2]) && (nodeTab[i][2] == nodeTab[j][1]))
        {
          cvrpDist[x][y] = INFINITY_VALUE;
        }
        else
        {
          cvrpDist[x][y] = ((carpDist[nodeTab[i][1]][nodeTab[i][2]] + carpDist[nodeTab[j][1]][nodeTab[j][2]]) / 4) + 
                            (fw.minDist[nodeTab[i][1]][nodeTab[j][1]]);
//          cvrpDist[x][y] = 0 * ((carpDist[nodeTab[i][1]][nodeTab[i][2]] + carpDist[nodeTab[j][1]][nodeTab[j][2]]) / 4) + 
//                            (fw.minDist[nodeTab[i][1]][nodeTab[j][1]]);
        }

      y++;
    }
    
    x += k;
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  k = 3;
  for (i = 1; i <= nArcs; i++)
    if ((carpDem[i] > 0) && (carpDem[i] < k))
      k = (int) carpDem[i];

  cvrpDem[0] = 0;
 
  if (k == 1)
  {
    for (x = 0, i = 1; i <=nArcs; i++)
    {
      if (carpDem[i] == 0)
        continue;

      j = 3 * x;
      cvrpDem[j + 1] = carpDem[i];
      cvrpDem[j + 2] = carpDem[i];
      cvrpDem[j + 3] = carpDem[i];

      x++;
    }
  }
  else
    if (k == 2)
    {
      for (x = 0, i = 1; i <= nArcs; i++)
      {
        if (carpDem[i] == 0)
          continue;

        j = 3 * x;
        aux1 = (int) (carpDem[i] * 2);
        aux2 = aux1 % 3;
        
        cvrpDem[j + 1] = (int) aux1 / 3;
        cvrpDem[j + 2] = (int) aux1 / 3;
        cvrpDem[j + 3] = (int) aux1 / 3;
        
        if (aux2 > 0)
          cvrpDem[j + 1]++;
        if (aux2 > 1)
          cvrpDem[j + 3]++;

        x++;
      }
    }
    else
    {
      for (x = 0, i = 1; i <= nArcs; i++)
      {
        if (carpDem[i] == 0)
          continue;

        j = 3 * x;
        aux1 = (int) carpDem[i];
        aux2 = aux1 % 3;
        
        cvrpDem[j + 1] = (int) aux1 / 3;
        cvrpDem[j + 2] = (int) aux1 / 3;
        cvrpDem[j + 3] = (int) aux1 / 3;
        
        if (aux2 > 0)
          cvrpDem[j + 1]++;
        if (aux2 > 1)
          cvrpDem[j + 3]++;

        x++;
      }
    }

  capacity *= (4 - k);

  printf( "k = %d, cap = %d\n", k, capacity );
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/*
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  cvrpDem[0] = 0;
  for (i = 0; i < nArcs; i++)
  {
    k = 3 * i;
    dem = carpDem[i] / 3;
    for (j = 1; j <= 3; j++)
      cvrpDem[k + j] = dem;
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
}

//---------------------------------------------------------------------------//


// - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// ==> Replaces each arc (i,j) of the carp graph instance
//     with one node (m_ij).
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
void CarpIO::toCvrpByLongo()
{
  register int i  = 0;
  register int j  = 0;
  register int k  = 0;
  register int x  = 0;
  register int y  = 0;
           int d1 = 0;
           int d2 = 0;
           int d3 = 0;
           int d4 = 0;

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  cvrpSize    = (reqArcs + 1);
  nodeTabSize = (nArcs = 1);
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  cvrpDist = new double* [cvrpSize];
  for (i = 0; i < cvrpSize; i++)
    cvrpDist[i] = new double [cvrpSize];

  for (i = 0; i < cvrpSize; i++)
  {
    cvrpDist[i][i] = 0.0;
    for (j = i+1; j < cvrpSize; j++)
    {
      cvrpDist[i][j] = INFINITY_VALUE;
      cvrpDist[j][i] = INFINITY_VALUE;
    }
  }

  nodeTab = new int* [nodeTabSize];
  for (i = 0; i < nodeTabSize; i++)
    nodeTab[i] = new int [4];
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  nodeTab[0][0] = 0;
  nodeTab[0][1] = 0;
  nodeTab[0][2] = 0;
  nodeTab[0][3] = 0;

  for (i = 1; i <= nArcs; i++)
  {
    nodeTab[i][0] = 1;
    nodeTab[i][1] = edgeTab[i][0];
    nodeTab[i][2] = edgeTab[i][1];
    nodeTab[i][3] = i;
  }

  fw.shortestPaths( nNodes, carpDist );
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // Distance between depot node and nodes m_ij
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  for (x = 0, i = 1; i < nodeTabSize; i++)
  {
    if (carpDem[nodeTab[i][3]] == 0)
      continue;

    d1 = ((int) floor(carpDist[nodeTab[i][1]][nodeTab[i][2]] / 2)) +
         ((int) fw.minDist[nodeTab[i][1]][0]);

    d2 = ((int)  ceil(carpDist[nodeTab[i][1]][nodeTab[i][2]] / 2)) +
         ((int) fw.minDist[nodeTab[i][2]][0]);
    
    d1 = (d1 < d2) ? d1 : d2;

    cvrpDist[0][x] = cvrpDist[x][0] = d1;

    x++;
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // Distance between two nodes m_ij
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  for (x = 0, i = 1; i < cvrpSize; i++)
  {
    if (carpDem[nodeTab[i][3]] == 0)
      continue;

    k = 0;

    for (y = 0, j = 1; j < cvrpSize; j++)
    {
      if (carpDem[nodeTab[j][3]] == 0)
        continue;

      k = 1;

      if (((nodeTab[i][0] == nodeTab[j][1]) && (nodeTab[i][2] == nodeTab[j][2])) ||
          ((nodeTab[i][0] == nodeTab[j][2]) && (nodeTab[i][2] == nodeTab[j][1])))
      {
        cvrpDist[i][j] = 0.0;
      }
      else
      {
        d1 = ((int) floor(carpDist[nodeTab[i][1]][nodeTab[i][2]] / 2)) + 
             ((int) floor(carpDist[nodeTab[j][1]][nodeTab[j][2]] / 2)) + 
             ((int) fw.minDist[nodeTab[i][1]][nodeTab[j][1]]);

        d2 = ((int) floor(carpDist[nodeTab[i][1]][nodeTab[i][2]] / 2)) + 
             ((int)  ceil(carpDist[nodeTab[j][1]][nodeTab[j][2]] / 2)) + 
             ((int) fw.minDist[nodeTab[i][1]][nodeTab[j][2]]);

        d3 = ((int)  ceil(carpDist[nodeTab[i][1]][nodeTab[i][2]] / 2)) + 
             ((int) floor(carpDist[nodeTab[j][1]][nodeTab[j][2]] / 2)) + 
             ((int) fw.minDist[nodeTab[i][2]][nodeTab[j][1]]);

        d4 = ((int)  ceil(carpDist[nodeTab[i][1]][nodeTab[i][2]] / 2)) + 
             ((int)  ceil(carpDist[nodeTab[j][1]][nodeTab[j][2]] / 2)) + 
             ((int) fw.minDist[nodeTab[i][2]][nodeTab[j][2]]);

        d1 = (d1 < d2) ? d1 : d2;
        d3 = (d3 < d4) ? d3 : d4;
        d1 = (d1 < d3) ? d1 : d3;

        cvrpDist[i][j] = d1;

        y++;
      }

      x += k;
    }
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  cvrpDem[0] = 0;
  for (i = 1; i <= nArcs; i++)
  {
    cvrpDem[i] = carpDem[i];
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
}

//---------------------------------------------------------------------------//

// - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// ==> Replaces each arc (i,j) of the carp graph instance
//     with two nodes (s_ij, s_ji).
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
void CarpIO::toCvrpByUchoa( int option )
{
  register int i  = 0;
  register int j  = 0;
  register int k  = 0;
  register int x  = 0;
  register int y  = 0;
           int d1 = 0;
           int d2 = 0;

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  cvrpSize    = (2 * reqArcs) + 1;
  nodeTabSize = (2 * nArcs) + 1;
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  cvrpDist = new double* [cvrpSize];
  for (i = 0; i < cvrpSize; i++)
    cvrpDist[i] = new double [cvrpSize];

  for (i = 0; i < cvrpSize; i++)
  {
    cvrpDist[i][i] = 0.0;
    for (j = 0; j < i; j++)
    {
      cvrpDist[i][j] = INFINITY_VALUE;
      cvrpDist[j][i] = INFINITY_VALUE;
    }
  }

  cvrpDem = new double [cvrpSize];

  nodeTab = new int* [nodeTabSize];
  for (i = 0; i < nodeTabSize; i++)
    nodeTab[i] = new int [4];
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  nodeTab[0][0] = 0;
  nodeTab[0][1] = 0;
  nodeTab[0][2] = 0;
  nodeTab[0][3] = 0;

  for (j = 0, i = 1; i <= nArcs; i++)
  {
    // node s_ij
    j++;
    nodeTab[j][0] = 1;
    nodeTab[j][1] = edgeTab[i][0];
    nodeTab[j][2] = edgeTab[i][1];
    nodeTab[j][3] = i;

    // node s_ji
    j++;
    nodeTab[j][0] = 2;
    nodeTab[j][1] = edgeTab[i][1];
    nodeTab[j][2] = edgeTab[i][0];
    nodeTab[j][3] = i;
  }

  fw.shortestPaths( nNodes, carpDist );
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // Distance between depot node and nodes s_ij and s_ji
  // Direct distance between node s_ij and node s_ji, arc (s_ij,s_ji)
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  switch (option)
  {
    case 1 : // i__a__b__j : c(a,b) = c(i,j)
      // - - - - - - - - - - - - - - - - - -
      for (x = 1, j = 1; j < nodeTabSize; j += 2)
      {
        if (carpDem[nodeTab[j][3]] == 0)
          continue;

        // Distance between depot node and node s_ij
        cvrpDist[0][x] = fw.minDist[nodeTab[j][1]][0];
        cvrpDist[x][0] = cvrpDist[0][x];

        // Distance between depot node and node s_ji
        x++;
        cvrpDist[0][x] = fw.minDist[nodeTab[j][2]][0];
        cvrpDist[x][0] = cvrpDist[0][x];

        // Direct distance between node s_ij and node s_ji, arc (s_ij,s_ji)
        cvrpDist[x-1][x] = carpDist[nodeTab[j][1]][nodeTab[j][2]];
        cvrpDist[x][x-1] = cvrpDist[x-1][x];

        x++;
      }
      break;
      // - - - - - - - - - - - - - - - - - -

    case 2 : // i__a__b__j : c(a,b) = 0
      // - - - - - - - - - - - - - - - - - -
      for (x = 1, j = 1; j < nodeTabSize; j += 2)
      {
        if (carpDem[nodeTab[j][3]] == 0)
          continue;

        // Distance between depot node and node s_ij
        cvrpDist[0][x] = floor(carpDist[nodeTab[j][1]][nodeTab[j][2]] / 2) +
                         (fw.minDist[nodeTab[j][1]][0]);
        cvrpDist[x][0] = cvrpDist[0][x];
        
        // Distance between depot node and node s_ji
        x++;
        cvrpDist[0][x] = ceil(carpDist[nodeTab[j][1]][nodeTab[j][2]] / 2) +
                         (fw.minDist[nodeTab[j][2]][0]);
        cvrpDist[x][0] = cvrpDist[0][x];
        
        // Direct distance between node s_ij and node s_ji, arc (s_ij,s_ji)
        cvrpDist[x-1][x] = 0.0;
        cvrpDist[x][x-1] = 0.0;

        x++;
      }
      break;
      // - - - - - - - - - - - - - - - - - -

    case 3 : // i__a__b__j : c(a,b) = c(i,j) / 3
      // - - - - - - - - - - - - - - - - - -
      for (x = 1, j = 1; j < nodeTabSize; j += 2)
      {
        if (carpDem[nodeTab[j][3]] == 0)
          continue;

        // Distance between depot node and node s_ij
        if (((int) carpDist[nodeTab[j][1]][nodeTab[j][2]] % 3) < 2)
          d1 = (int) floor(carpDist[nodeTab[j][1]][nodeTab[j][2]] / 3);
        else
          d1 = (int)  ceil(carpDist[nodeTab[j][1]][nodeTab[j][2]] / 3);
        
        cvrpDist[0][x] = d1 + fw.minDist[nodeTab[j][1]][0];
        cvrpDist[x][0] = cvrpDist[0][x];
        
        // Distance between depot node and node s_ji
        x++;
        
        cvrpDist[0][x] = d1 + fw.minDist[nodeTab[j][2]][0];
        cvrpDist[x][0] = cvrpDist[0][x];
        
        // Direct distance between node s_ij and node s_ji, arc (s_ij,s_ji)
        if (((int) carpDist[nodeTab[j][1]][nodeTab[j][2]] % 3) < 2)
          cvrpDist[x-1][x] = (int)  ceil(carpDist[nodeTab[j][1]][nodeTab[j][2]] / 3);
        else
          cvrpDist[x-1][x] = (int) floor(carpDist[nodeTab[j][1]][nodeTab[j][2]] / 3);
        
        cvrpDist[x][x-1] = cvrpDist[x-1][x];

        x++;
      }
      break;
      // - - - - - - - - - - - - - - - - - -
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -


  // Distance between two nodes s_ij 
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  switch (option)
  {
    case 1 : // i__a__b__j : c(a,b) = c(i,j)
      // - - - - - - - - - - - - - - - - - -
      for (x = 1, i = 1; i < nodeTabSize; i += 2)
      {
        if (carpDem[nodeTab[i][3]] == 0)
          continue;

        k = 0;

        for (y = x + 2, j = i + 2; j < nodeTabSize; j += 2)
        {
          if (carpDem[nodeTab[j][3]] == 0)
            continue;

          k = 2;

          cvrpDist[x][y]     = fw.minDist[nodeTab[i][1]][nodeTab[j][1]];
          cvrpDist[x][y+1]   = fw.minDist[nodeTab[i][1]][nodeTab[j+1][1]];
          cvrpDist[x+1][y]   = fw.minDist[nodeTab[i+1][1]][nodeTab[j][1]];
          cvrpDist[x+1][y+1] = fw.minDist[nodeTab[i+1][1]][nodeTab[j+1][1]];

          cvrpDist[y][x]     = cvrpDist[x][y];
          cvrpDist[y+1][x]   = cvrpDist[x][y+1];
          cvrpDist[y][x+1]   = cvrpDist[x+1][y];
          cvrpDist[y+1][x+1] = cvrpDist[x+1][y+1];

          y += 2;
        }

        x += k;
      }

      break;
      // - - - - - - - - - - - - - - - - - -

    case 2 : // i__a__b__j : c(a,b) = 0
      // - - - - - - - - - - - - - - - - - -
      for (x = 1, i = 1; i < nodeTabSize; i += 2)
      {
        if (carpDem[nodeTab[i][3]] == 0)
          continue;

        k = 0;

        for (y = x + 2, j = i + 2; j < nodeTabSize; j += 2)
        {
          if (carpDem[nodeTab[j][3]] == 0)
            continue;

          k = 2;

          cvrpDist[x][y]     = floor(carpDist[nodeTab[i][1]][nodeTab[i][2]] / 2) +
                               floor(carpDist[nodeTab[j][1]][nodeTab[j][2]] / 2) + 
                               fw.minDist[nodeTab[i][1]][nodeTab[j][1]];
          cvrpDist[x][y+1]   = floor(carpDist[nodeTab[i][1]][nodeTab[i][2]] / 2) +
                               ceil(carpDist[nodeTab[j+1][1]][nodeTab[j+1][2]] / 2) + 
                               fw.minDist[nodeTab[i][1]][nodeTab[j+1][1]];
          cvrpDist[x+1][y]   =  ceil(carpDist[nodeTab[i+1][1]][nodeTab[i+1][2]] / 2) +
                               floor(carpDist[nodeTab[j][1]][nodeTab[j][2]] / 2) + 
                               fw.minDist[nodeTab[i+1][1]][nodeTab[j][1]];
          cvrpDist[x+1][y+1] =  ceil(carpDist[nodeTab[i+1][1]][nodeTab[i+1][2]] / 2) +
                               ceil(carpDist[nodeTab[j+1][1]][nodeTab[j+1][2]] / 2) + 
                               fw.minDist[nodeTab[i+1][1]][nodeTab[j+1][1]];
          
          cvrpDist[y][x]     = cvrpDist[x][y];
          cvrpDist[y+1][x]   = cvrpDist[x][y+1];
          cvrpDist[y][x+1]   = cvrpDist[x+1][y];
          cvrpDist[y+1][x+1] = cvrpDist[x+1][y+1];

          y += 2;
        }

        x += k;
      }
      break;
      // - - - - - - - - - - - - - - - - - -

    case 3 : // i__a__b__j : c(a,b) = c(i,j) / 3
      // - - - - - - - - - - - - - - - - - -
      for (x = 1, i = 1; i < nodeTabSize; i += 2)
      {
        if (carpDem[nodeTab[i][3]] == 0)
          continue;

        k = 0;

        for (y = x + 2, j = i + 2; j < nodeTabSize; j += 2)
        {
          if (carpDem[nodeTab[j][3]] == 0)
            continue;

          k = 2;

          // -----------------------------
          if (((int) carpDist[nodeTab[i][1]][nodeTab[i][2]] % 3) < 2)
            d1 = (int) floor(carpDist[nodeTab[i][1]][nodeTab[i][2]] / 3);
          else
            d1 = (int)  ceil(carpDist[nodeTab[i][1]][nodeTab[i][2]] / 3);
          
          if (((int) carpDist[nodeTab[j][1]][nodeTab[j][2]] % 3) < 2)
            d2 = (int) floor(carpDist[nodeTab[j][1]][nodeTab[j][2]] / 3);
          else
            d2 = (int)  ceil(carpDist[nodeTab[j][1]][nodeTab[j][2]] / 3);
          
          cvrpDist[x][y] = d1 + d2 + fw.minDist[nodeTab[i][1]][nodeTab[j][1]];
          // -----------------------------
          
          // -----------------------------
          if (((int) carpDist[nodeTab[j+1][1]][nodeTab[j+1][2]] % 3) < 2)
            d2 = (int) floor(carpDist[nodeTab[j+1][1]][nodeTab[j+1][2]] / 3);
          else
            d2 = (int)  ceil(carpDist[nodeTab[j+1][1]][nodeTab[j+1][2]] / 3);
          
          cvrpDist[x][y+1] = d1 + d2 + fw.minDist[nodeTab[i][1]][nodeTab[j+1][1]];
          // -----------------------------
          
          // -----------------------------
          if (((int) carpDist[nodeTab[i+1][1]][nodeTab[i+1][2]] % 3) < 2)
            d1 = (int) floor(carpDist[nodeTab[i+1][1]][nodeTab[i+1][2]] / 3);
          else
            d1 = (int)  ceil(carpDist[nodeTab[i+1][1]][nodeTab[i+1][2]] / 3);
          
          if (((int) carpDist[nodeTab[j][1]][nodeTab[j][2]] % 3) < 2)
            d2 = (int) floor(carpDist[nodeTab[j][1]][nodeTab[j][2]] / 3);
          else
            d2 = (int)  ceil(carpDist[nodeTab[j][1]][nodeTab[j][2]] / 3);
          
          cvrpDist[x+1][y] = d1 + d2 + fw.minDist[nodeTab[i+1][1]][nodeTab[j][1]];
          // -----------------------------
          
          // -----------------------------
          if (((int) carpDist[nodeTab[i+1][1]][nodeTab[i+1][2]] % 3) < 2)
            d1 = (int) floor(carpDist[nodeTab[i+1][1]][nodeTab[i+1][2]] / 3);
          else
            d1 = (int)  ceil(carpDist[nodeTab[i+1][1]][nodeTab[i+1][2]] / 3);
          
          if (((int) carpDist[nodeTab[j+1][1]][nodeTab[j+1][2]] % 3) < 2)
            d2 = (int) floor(carpDist[nodeTab[j+1][1]][nodeTab[j+1][2]] / 3);
          else
            d2 = (int)  ceil(carpDist[nodeTab[j+1][1]][nodeTab[j+1][2]] / 3);
          
          cvrpDist[x+1][y+1] = d1 + d2 + fw.minDist[nodeTab[i+1][1]][nodeTab[j+1][1]];
          // -----------------------------
          
          cvrpDist[y][x]     = cvrpDist[x][y];
          cvrpDist[y+1][x]   = cvrpDist[x][y+1];
          cvrpDist[y][x+1]   = cvrpDist[x+1][y];
          cvrpDist[y+1][x+1] = cvrpDist[x+1][y+1];

          y += 2;
        }

        x += k;
      }
      break;
      // - - - - - - - - - - - - - - - - - -
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -


  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  k = 2;
  for (i = 1; (i <= nArcs) && (k > 1); i++)
    k = (carpDem[i] == 1) ? 1 : 2;

  cvrpDem[0] = 0;
  
  if (k == 1)
  {
    for (x = 1, i = 1; i <= nArcs; i++)
    {
      if (carpDem[i] == 0)
        continue;

      j = 2 * x;
      cvrpDem[j-1] = carpDem[i];
      cvrpDem[j]   = 0;
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//      cvrpDem[j-1] = carpDem[i];
//      cvrpDem[j]   = carpDem[i];
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

      x++;
    }
  }
  else
  {
    for (x = 1, i = 1; i <= nArcs; i++)
    {
      if (carpDem[i] == 0)
        continue;

      j = 2 * x;
//      cvrpDem[j-1] = carpDem[i];
//      cvrpDem[j]   = 0;
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      cvrpDem[j-1] = (int) floor((double) carpDem[i] / 2);
      cvrpDem[j]   = (int)  ceil((double) carpDem[i] / 2);
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

      x++;
    }
  }

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//  capacity *= (3 - k);
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
}

//---------------------------------------------------------------------------//

void CarpIO::writeCvrpFile()
{
           char  cvrpName[256];
           char  instanceName[256];
           char* fileName = NULL;
           char* fileType = NULL;
  register int   i        = 0;
  register int   j        = 0;
           FILE* fout     = NULL;

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  // Find the file name 
  fileName = strrchr( (char *) fname, '\\' );
  
  if (fileName == NULL)
    fileName = fname;
  else
    fileName = fileName + 1;

  // Find the file type
  fileType = strrchr( (char *) fileName, '.' );
  if (fileType != NULL)
    fileType[0] = '\0';

  // Instance name
  strcpy( instanceName, fileName );

  // Sets the file name type
  sprintf( cvrpName, "%s.vrp", instanceName );
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  fout = fopen( cvrpName, "wt" );

  if (fout == NULL)
  {
    printf( "Unable to open file %s\n", cvrpName );
    throw(2);
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  fprintf( fout, "NAME : %s\n", instanceName );
  fprintf( fout, "COMMENT :\n" );
  fprintf( fout, "TYPE : CVRP\n" );
  fprintf( fout, "DIMENSION : %d\n", cvrpSize );
  fprintf( fout, "EDGE_WEIGHT_TYPE : EXPLICIT\n" );
  fprintf( fout, "EDGE_WEIGHT_FORMAT: LOWER_ROW \n" );
  fprintf( fout, "DISPLAY_DATA_TYPE : NO_DISPLAY\n" );
  fprintf( fout, "CAPACITY : %d\n", capacity );

  fprintf( fout, "EDGE_WEIGHT_SECTION\n" );
  for (i = 2; i < cvrpSize; i++)
  {
    for (j = 1; j < i; j++)
    {
      fprintf( fout, "%d ", (int) cvrpDist[i][j] );
    }
    fprintf( fout, "\n" );
  }

  for (j = 1; j < cvrpSize; j++)
  {
    fprintf( fout, "%d ", (int) cvrpDist[0][j] );
  }
  fprintf( fout, "\n" );

  fprintf( fout, "DEMAND_SECTION\n" );
  for (i = 0; i < cvrpSize; i++)
    fprintf( fout, "%d %d\n", i+1, (int) cvrpDem[i] );

  fprintf( fout, "DEPOT_SECTION\n" );
  fprintf( fout, "1\n" );
  fprintf( fout, "-1\n" );
  fprintf( fout, "EOF\n" );
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  fclose( fout );
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
}

//---------------------------------------------------------------------------//

void CarpIO::writeCvrpDist()
{
  register int i  = 0;
  register int j  = 0;
  register int x  = 0;
  register int y  = 0;

//  FILE* fp = freopen( "tabdist.txt", "w", stdout );
  FILE* fp = stdout;

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  fprintf( fp, "         " );
  for (i = 0; i < nodeTabSize; i++)
  {
    if ((i > 0) && (carpDem[nodeTab[i][3]] == 0))
      continue;

    fprintf( fp, "(%2d,%2d)  ", nodeTab[i][1]+1, nodeTab[i][2]+1 );
  }
  fprintf( fp, "\n" );
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  for (x = 0, i = 0; i < nodeTabSize; i++)
  {
    if ((i > 0) && (carpDem[nodeTab[i][3]] == 0))
      continue;

    fprintf( fp, "(%2d,%2d) ", nodeTab[i][1]+1, nodeTab[i][2]+1 );
    for (y = 0, j = 0; j < nodeTabSize; j++)
    {
      if ((j > 0) && (carpDem[nodeTab[j][3]] == 0))
        continue;

      fprintf( fp, "%8d ", (int) cvrpDist[x][y] );

      y++;
    }
    fprintf( fp, "\n" );

    x++;
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

//  fclose( fp );
}

//---------------------------------------------------------------------------//

void CarpIO::writeCvrpDem()
{
  register int i = 0;

  for (i = 0; i < cvrpSize; i++)
  {
    printf( "%2d  %lf\n", i, cvrpDem[i] );
  }
}

//---------------------------------------------------------------------------//

void CarpIO::writePearnDist()
{
  register int i  = 0;
  register int j  = 0;
           int n  = 0;

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  printf( "       " );
  for (i = 0; i < cvrpSize; i++)
    printf( "(%d,%d)  ", nodeTab[i][1]+1, nodeTab[i][2]+1 );
  printf( "\n" );
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  for (i = 0; i < cvrpSize; i++)
  {
    printf( "(%d,%d) ", nodeTab[i][1]+1, nodeTab[i][2]+1 );
    for (j = 0; j < cvrpSize; j++)
    {
      printf( "%6d ", (int) cvrpDist[i][j] );
    }
    printf( "\n" );
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
}

//---------------------------------------------------------------------------//

void CarpIO::writeLongoDist()
{
  register int i  = 0;
  register int j  = 0;
           int n  = 0;

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  printf( "       " );
  for (i = 0; i <= nArcs; i++)
    if (i == 0)
      printf( "(%d,%d)  ", edgeTab[i][0], edgeTab[i][1] );
    else
      printf( "(%d,%d)  ", edgeTab[i][0]+1, edgeTab[i][1]+1 );
  printf( "\n" );
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  for (i = 0; i <= nArcs; i++)
  {
    if (i == 0)
      printf( "(%d,%d) ", edgeTab[i][0], edgeTab[i][1] );
    else
      printf( "(%d,%d) ", edgeTab[i][0]+1, edgeTab[i][1]+1 );
    for (j = 0; j <= nArcs; j++)
    {
      printf( "%6d ", (int) cvrpDist[i][j] );
    }
    printf( "\n" );
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
}

//---------------------------------------------------------------------------//

void CarpIO::writeUchoaDist()
{
  register int i  = 0;
  register int j  = 0;
           int n  = 0;

  // freopen( "tabdist.txt", "w", stdout );

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  printf( "        " );
  for (i = 0; i < cvrpSize; i++)
    printf( "(%2d,%2d) ", nodeTab[i][0]+1, nodeTab[i][1]+1 );
  printf( "\n" );
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  for (i = 0; i < cvrpSize; i++)
  {
    printf( "(%2d,%2d)  ", nodeTab[i][0]+1, nodeTab[i][1]+1 );
    for (j = 0; j < cvrpSize; j++)
    {
      printf( "%6d  ", (int) cvrpDist[i][j] );
    }
    printf( "\n" );
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
}




/*      ____    _______           ____    _______           ____    _______ 
   |   |    \  |       |  _  |   |    \  |       |  _  |   |    \  |       |
   |______|  \____| |__|     |______|  \____| |__|     |______|  \____| |__|  */
