/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Humberto Longo                                    */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

///////////////////////////////////////////////////////////////////////////////
// FILE : CFloydWarshall.cpp                                                 //
//---------------------------------------------------------------------------//
// Class to compute shortest paths using the Floyd-Warshall algorithm        //
//---------------------------------------------------------------------------//
// AUTHOR : Humberto Jos Longo                                              //
//---------------------------------------------------------------------------//
// CREATED ON : 26/02/2004                                                   //
// LAST MODIFIED :                                                           //
///////////////////////////////////////////////////////////////////////////////

//---------------------------------------------------------------------------//

#include "FloydWarshall.h"

//---------------------------------------------------------------------------//

FloydWarshall::FloydWarshall()
{
  nNodes  = 0;
  path    = NULL;
  dist    = NULL;
  minPath = NULL;
  minDist = NULL;
}

//---------------------------------------------------------------------------//

FloydWarshall::FloydWarshall( int _nNodes )
{
  register int i = 0;
  register int j = 0;

  nNodes  = _nNodes;
  minPath = NULL;
  minDist = NULL;

  path = new int** [2];
  for (i = 0; i < 2; i++)
  {
    path[i] = new int* [nNodes];
    for (j = 0; j < nNodes; j++)
       path[i][j] = new int [nNodes];
  }

  dist = new double** [2];
  for (i = 0; i < 2; i++)
  {
    dist[i] = new double* [nNodes];
    for (j = 0; j < nNodes; j++)
       dist[i][j] = new double [nNodes];
  }
}

//---------------------------------------------------------------------------//

FloydWarshall::~FloydWarshall()
{
  register int i = 0;
  register int j = 0;

  if (dist != NULL)
  {
    for (i = 0; i < 2; i++)
    {
      for (j = 0; j < nNodes; j++)
        delete [] dist[i][j];

      delete [] dist[i];
    }
    
    delete [] dist;
  }

  if (path != NULL)
  {
    for (i = 0; i < 2; i++)
    {
      for (j = 0; j < nNodes; j++)
        delete [] path[i][j];
      delete [] path[i];
    }
    
    delete [] path;
  }
}

//---------------------------------------------------------------------------//

void FloydWarshall::allocDataStructs( int graphSize )
{
  register int i = 0;
  register int j = 0;

  nNodes = graphSize;

  if (minPath == NULL)
  {
    path = new int** [2];
    for (i = 0; i < 2; i++)
    {
      path[i] = new int* [nNodes];
      for (j = 0; j < nNodes; j++)
        path[i][j] = new int [nNodes];
    }
  }

  if (minDist == NULL)
  {
    dist = new double** [2];
    for (i = 0; i < 2; i++)
    {
      dist[i] = new double* [nNodes];
      for (j = 0; j < nNodes; j++)
        dist[i][j] = new double [nNodes];
    }
  }
}

//---------------------------------------------------------------------------//

void FloydWarshall::shortestPaths( int      graphSize,
                                   double** edgeCost )
{
  register int    i    = 0;
  register int    j    = 0;
  register int    k    = 0;
           int    curr = 1;
           int    prev = 0;
           double min  = 0.0;

  allocDataStructs( graphSize );


  // Shortest distances initial matrix
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  for (i = 0; i < nNodes; i++)
  {
    for (j = 0; j < nNodes; j++)
    {
      dist[0][i][j] = edgeCost[i][j];
    }
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -


  // Shortest paths initial matrix
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  for (i = 0; i < nNodes; i++)
  {
    for (j = 0; j < nNodes; j++)
    {
      if (i == j)
      {
        path[0][i][j] = -1;
      }
      else
      {
        path[0][i][j] = i;
      }
    }
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -


  // Floyd-Warshall algorithm
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  for (k = 0; k < nNodes; k++)
  {
    prev = 1 - curr;
    for (i = 0; i < nNodes; i++)
    {
      for (j = 0; j < nNodes; j++)
      {
        min = dist[prev][i][k] + dist[prev][k][j];
        dist[curr][i][j] = (dist[prev][i][j] < min) ? dist[prev][i][j] : min;
        if (dist[prev][i][j] <= min)
        {
          dist[curr][i][j] = dist[prev][i][j];
          path[curr][i][j] = path[prev][i][j];
        }
        else
        {
          dist[curr][i][j] = min;
          path[curr][i][j] = path[prev][k][j];
        }
      }
    }
    curr = prev;
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  
  minDist = dist[1 - curr];
  minPath = path[1 - curr];
}

//---------------------------------------------------------------------------//
