/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CBBNode.h                        //
////////////////////////////////////////////
// Class that represents a node of the    //
// branch tree                            //
////////////////////////////////////////////

//---------------------------------------------------------------------------//

class CBBNode;

//---------------------------------------------------------------------------//

#ifndef CBBNODE_H_
#define CBBNODE_H_

//---------------------------------------------------------------------------//

#include "CBasis.h"
#include "CMaster.h"
#include "CColumn.h"
#include "CRow.h"

#ifdef WIN32
#include <vector.h>
#elif defined LINUX
#include <vector>
using namespace std;
#endif

//---------------------------------------------------------------------------//

/**
 * Class that a node of the branch tree
 */
class CBBNode
{
public:

   /**
    * Creates a new node
    * @param m master of the node
    * @param aID the identifier of the node
    */
   CBBNode(CMaster *m, double);

   /**
    * Creates a new node
    * @param f node father of this node
    * @param aID the identifier of the node
    * @param or information to calculate skewness
    */
   CBBNode(CBBNode *f, double, int);


   /**
    * Free memory allocated
    */
   ~CBBNode();

   /**
    * Returns the origin of the node
    * @return the origin of the node
    */
   inline int getOrigin(){ return origin; }

   /**
    * Returns the lower bound of the node
    * @return the lower bound of the node
    */
   double getLB();

   /**
    * Set the lower bound of the node
    * @param l lower bound of the node
    */
   inline void setLB(double l){ lbVal = l; }

   /**
    * Returns the upper bound of the node
    * @return the upper bound of the node
    */
   inline double getUB(){ return ubVal; }

   /**
    * Set the upper bound of the node
    * @param u upper bound of the node
    */
   inline void setUB(double u){ ubVal = u; }
   
   /**
    * Returns the identifier of the father node
    * @return the identifier of the father node
    */
   double getFather();

   /**
    * Tells if the node is is descendant of the node with the specified id
    * @param pid identifier to be verified
    * @return true, if the node is descendant; false, otherwise
    */
   bool isDescendant(double pid);

   /**
    * Returns the depth of the node
    * @return the depth of the node
    */
   int getDepth();

   /**
    * Columng generation procedure (to be implemented)
    * @return the number of columns generated
    */
   //virtual int columnGen() = 0;

   /**
    * Cut generation procedure (to be implemented)
    * @return the number of cuts generated
    */
   virtual int cutGen() = 0;

   /**
    * Creates the master problem (to be implemented)
    * @return 
    */
   virtual int createMaster() = 0;

   /**
    * Optimizes the node (to be implemented)
    * @return the status of the solved node
    */
   virtual NodeStatus optimize() = 0;

   /** 
    * Verifies if a solution is feasible (to be implemented)
    * @param x solution to be verified
    * @return true, if the solution is feasible; false, otherwise
    */
   virtual bool isFeasible( double* x ) = 0;

   /**
    * Procedure to do the branch (to be implemented)
    */
   virtual void doBranch() = 0;

   /**
    * Procedure to choose the variable to branch (to be implemented)
    * @param f auxiliary return variable
    * @return the variable to branch
    */
   virtual int chooseBranchArc(int &f) = 0;

   /**
    * Solves the local lp by the primal simplex method
    */
   void solvePrimal();

   /**
    * Solves the local lp by the dual simplex method
    */
   void solveDual();

   /**
    * Returns the solution of the local lp
    * @return an array with the solution of the local lp
    */
   inline double* getLPSol(){ return lpSol; }

   /**
    * Returns the dual solution of the local lp
    * @return an array with the dual solution of the local lp
    */
   inline double* getLPDual(){ return lpDual; }

   /** 
    * Copies the specified basis to the node's basis
    * @param bsrc the basis to be copied
    */
   void copyBasis(const CBasis &bsrc);

   /** 
    * Adds the row to the pool and to the list of rows of the node
    * @param r the row to be added
    * @return a pointer to the location of the row in the pool
    */
   CRow* addRow(const CRow &r);

   /** 
    * Adds the column to the pool and to the list of columns of the node
    * @param c the column to be added
    * @return a pointer to the location of the column in the pool
    */
   CColumn* addCol(const CColumn &c);

   /**
    * Returns the identifier of the node
    * @return the identifier of the node
    */
   inline double getID() { return id; }

   /**
    * Calculates the objective function value using the dual solution
    * @return the objective function value
    */
   double checkDualObj();

   /**
    * Verifies if the value yields a fathomed by bound
    * @param p_lb Must have an added value of 3.0 when a carp instance is 
    *              solved using cvrp techniques
   *              and Pearn's graph transformation, otherwise it should be 0.0
    * @return true, if the node is fathomed by bound; false, otherwise
    */
   bool isFathomed( double p_lb );

   /**
    * Add the basis to the stored to the local lp
    */
   virtual void addBasisToLP();

   /**
    * Gets the basis of the local lp
    * @param nMC number of columns of the basis
    * @param nMR number of rows of the basis
    */
   void obtainBasisFromLP(int nMC, int nMR);

   /**
    * Separate a maximum number of columns from the pool
    * @param nC the maximum number of columns to be separated
    * @return the number of columns separated from the pool
    */
   virtual int separateColsFromPool(int nC);

   /**
    * Separate a maximum number of rows from the pool
    * @param nR the maximum number of rows to be separated
    * @return the number of rows separated from the pool
    */
   virtual int separateRowsFromPool(int nR);


   /**
    * Process the solved node
    * @param lbAdd must be 3.0 when a carp instance is solved using cvrp techniques
   *              and Pearn's graph transformation, otherwise it should be 0.0
    * @param maxObjVal the maximum value of the objective function
    * @return the status of the solved node
    */
   NodeStatus processSolvedNode( double lbAdd, double maxObjVal );

   /**
    * Adds a column to the locallp of the node
    * @param col the column to be added
    */
   virtual void addColToLP(CColumn &col);

   /**
    * Adds a row to the locallp of the node
    * @param r the row to be added
    */
   virtual void addRowToLP(CRow &r);

   /**
    * Set the size of the solution of this node
    * @param s size of the solution
    */
   void setSolSize(int s);

   /**
    * Returns the size of the solution of this node
    * @return the size of the solution of this node
    */
   inline int getSolSize() { return solSize; }

   /**
    * Set the size of the dual solution of this node
    * @param s size of the dual solution
    */
   void setDualSize(int s);

   /**
    * Returns the size of the dual solution of this node
    * @return the size of the dual solution of this node
    */
   inline int getDualSize() { return dualSize; }

   inline double getParentLB(){ return parentLB; };

protected:

   /** Basis of the lp of this node */
   CBasis basis;

   /** Identifier of the node */
   double id;

   /** Local lp of the node */
   LPptr locallp;

   /** Size of the solution of locallp */
   int solSize;

   /* Size of the vector lpsol (has to be >= solSize) */
   int maxSolSize;

   /* Size of the vector lpdual (has to be >= dualSize) */
   int maxDualSize;

   /** Size of the dual solution of locallp */
   int dualSize;

   /** Lower bound of the father of the node */
   double parentLB;

   /** Lower bound of the node */
   double lbVal;

   /** Upper bound of the node */
   double ubVal;

   /** Solution of the locallp */
   double* lpSol;

   /** Dual solution of the locallp */
   double* lpDual;

   /** List of the columns of locallp */
   vector< CColumn* > lpCols;

   /** List of the columns of locallp */
   vector< CRow* > lpRows;

   /** Master of this node */
   CMaster* master;

   /** Used to calculate skewness */
   int origin;

   /** Number of cols before column generation starts. Used to transfer basis from
       one node to another */
   int masterCols;

   /** Number of rows before cut generation starts. Used to transfer basis from
       one node to another */
   int masterRows;

};

//---------------------------------------------------------------------------//

#endif
