/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CBasis.h                         //
////////////////////////////////////////////
// Class that represents a basis of a lp  //
////////////////////////////////////////////

//---------------------------------------------------------------------------//

#include "CBBNode.h"
#include "CColumn.h"
#include "CRow.h"
#ifdef WIN32
#include <vector.h>
#elif defined LINUX
#include <vector>
using namespace std;
#endif

//---------------------------------------------------------------------------//

class CBasis;

//---------------------------------------------------------------------------//

#ifndef CBASIS_H_
#define CBASIS_H_

//---------------------------------------------------------------------------//

/**
 * Class that represents the basis of a lp
 */
class CBasis
{

friend class CBBNode;

public:

   /**
    * Default constructor
    */
   CBasis();

   /**
    * Free memory allocated
    */
   ~CBasis();

   /**
    * Allocates memory to a basis
    * @param nC number of columns of the basis
    * @param nR number of rows of the basis
    */
   void allocBasis(int nC,int nR);

   /**
    * Deallocates memory of basis
    */
   void deallocBasis();
   
   /**
    * Returns the number of empty variables that are basic
    * @return the number of empty variables that are basic
    */
   inline int getNStaticCols(){ return nStaticCols; }

   /**
    * Set the number of empty variables that are basic
    * @param n the number of empty variables that are basic
    */
   inline void setNStaticCols(int n){ nStaticCols = n; }

   /** 
    * Returns the number of columns in masterCols that are basic
    * @return the number of columns in masterCols that are basic
    */
   inline int getNBasicDinamic(){ return nBasicDinamic; }

   /** 
    * Set the number of columns in masterCols that are basic
    * @param n number of columns in masterCols that are basic
    */
   inline void setNBasicDinamic(int n){ nBasicDinamic = n; }

   /**
    * Returns the number of columns stored in the basis
    * @return the number of columns stored in the basis
    */
   inline int getNMasterCols() { return nMasterCols; }

   /**
    * Returns the number of rows stored in the basis
    * @return the number of rows stored in the basis
    */
   inline int getNMasterRows(){ return nMasterRows; }

   /**
    * Returns the list of basic columns
    * @return the list of basic columns
    */
   inline vector<CColumn*>* getBasicCols(){ return &basicCols; }

   /**
    * Returns the list of basic rows
    * @return the list of basic rows
    */
   inline vector<CRow*>* getBasicRows(){ return &basicRows; }

   /**
    * Returns the number of the basic column in the lp
    * @param i index of the vector of basic columns
    * @return the number of the basic column in the lp
    */
   int getColStat( int i );

   /**
    * Returns the number of the basic row in the lp
    * @param i index of the vector of basic rows
    * @return the number of the basic row in the lp
    */
   int getRowStat( int i );

   /**
    * Make a copy of the basis
    * @param bs basis to be copied
    */
   void copyBasis(const CBasis &bs);


   // Adds a basic column (different from master) to lp

   /**
    * Adds a column to the list of basic columns
    * @param col column to be added
    */
   void addBasicCol(CColumn* col);

   /**
    * Adds a row to the list of basic rows
    * @param row row to be added
    */
   void addBasicRow(CRow* row);

private:

   /** Stores the status of the master columns of the lp */
   int* masterCols;

   /** Stores the status of the master rows of the lp */
   int* masterRows;

   /** Number of master columns of the lp */
   int nMasterCols;

   /** Number of master rows of the lp */
   int nMasterRows;

   /** Tells if the basis was allocated */
   bool allocated;

   /** List of pointers to the basic columns */
   vector<CColumn*> basicCols;

   /** List of pointers to the basic rows */
   vector<CRow*> basicRows;

   /* Associate empty variable to the base */
   int nStaticCols;

   /* Number of variables of nMasterCols that are basic */
   int nBasicDinamic;

};

//---------------------------------------------------------------------------//

#endif
