/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CColumn.h                        //
////////////////////////////////////////////
// Class that represents a column.        //
////////////////////////////////////////////

//---------------------------------------------------------------------------//

class CColumn;

//---------------------------------------------------------------------------//

#ifndef CCOLUMN_H_
#define CCOLUMN_H_

//---------------------------------------------------------------------------//

#include <stdio.h>
#include "Constants.h"
#include "CParameters.h"

//---------------------------------------------------------------------------//

/**
 * Different types of the column: C ( continuous ), B ( binary ) or I ( integer )
 */
enum ColType { C, B, I, NONE };

//---------------------------------------------------------------------------//

/**
 * Class that represents an item of a 
 * row                           
 */
class CColElem
{

public:

   /** Identifier of the element */
   int idx;

   /** Coefficient of the element */
   double coeff;

   /** Pointer to the next element */
   CColElem* next;

};

//---------------------------------------------------------------------------//

/**
 * Class that represents a column
 */
class CColumn
{

public:

   /** 
    * Default constructor
    */
   CColumn();

   /** 
    * Constructs a column with the parameters specified
    * @param p parameters for the column
    */
   CColumn(CParameters *p,int __maxk);

   /** 
    * Free the memory allocated
    */
   ~CColumn();

   /**
    * Copy constructor
    * @param c column to be copied
    */
   CColumn( const CColumn &c );

   /**
    * Set the parameters for the column
    * @param p parameters for the column
    */
   inline void setParameters(CParameters *p) { params = p; }

   /**
    * Returns the parameters for the column
    * @return the parameters for the column
    */
   inline CParameters* getParameters() { return params; }

   /**
    * Modify the coefficient of the column in the objective function
    * @param v value of the objective function coefficient
    */
   inline void setObj( double v  ){ obj = v; }

   /**
    * Returns the coefficient of the column in the objective function
    * @return value of the objective function coefficient
    */
   inline double getObj(){ return obj; }

   /**
    * Modify the column upper bound
    * @param u the column upper bound
    */
   inline void setUB( double u ){ ub = u; }

   /**
    * Returns the column upper bound
    * @return the column upper bound
    */
   inline double getUB(){ return ub; }

   /**
    * Modify the column lower bound
    * @param l the column lower bound
    */
   inline void setLB( double l ){ lb = l; }

   /**
    * Returns the column lower bound
    * @return the column lower bound
    */
   inline double getLB(){ return lb; }

   /**
    * Modify the column type
    * @param t the column type
    */
   inline void setType( ColType t ){ type = t; }

   /**
    * Returns the column type
    * @return the column type
    */
   inline ColType getType(){ return type; }

   /**
    * Tests if two columns are equal
    * @param cc column to be compared
    * @return true if the two columns are equal; false otherwise
    */
   bool isColumnEqual( const CColumn &cc );
   
   /**
    * Make a copy of the column
    * @param cc column to be copied
    */
   void copy( const CColumn &cc );

   /**
    * Returns the reduced cost of the column
    * @param lpdual array with the dual solution
    * @return the reduced cost of the column
    */
   double getRedCost(double* lpdual);

   /** 
    * Set if the column is in the pool or not
    * @param p true if the column is in pool, false otherwise
    */
   inline void setInPool( bool p ){ inPool = p; }

   /** 
    * Returns if the column is in the pool or not
    * @return true if the column is in pool, false otherwise
    */
   inline bool getInPool(){ return inPool; };

   /** 
    * Verifies if the column contains an element with
    * identifier idx
    * @param idx identifier to be search
    * @return true if the column contains idx, false otherwise
    */
   bool contains( int idx );

   /** 
    * Returns if the column can be separated from the pool
    * @return true if the column can be separated, false otherwise
    */
   inline bool getSepFromPool(){ return sepFromPool; }

   /** 
    * Set if the column can be separated from the pool
    * @param p true if the column can be separated from the pool, false otherwise
    */
   inline void setSepFromPool(bool l){ sepFromPool = l; }

   /** 
    * Returns the number of LPs that use this column
    * @return the number of LPs that use this column
    */
   inline int getNLPs() { return nLPs; }

   /** 
    * Set the number of LPs that use this column
    * @param n the number of LPs that use this column
    */
   inline void setNLPs(int n){ nLPs = n; }

   /** 
    * Increase the number of LPs that use this column
    */
   inline void increaseNLPs(){ nLPs++; }

   /** 
    * Decrease the number of LPs that use this column
    */
   inline void decreaseNLPs(){ if( nLPs > 0 ) nLPs--; }

   /** 
    * Returns a pointer to the first element of the column
    * @return a pointer to the first element of the column
    */
   inline CColElem* getFirstElem(){ return first; }

   /**
    * Add an element to the column
    * @param idx index of the element
    * @param coeff coefficient of the element
    */
   void addElem( int idx, double coeff );

   /**
    * Add an element to the column (sum the coeffs if 
    * the element already exists)
    * @param idx index of the element
    * @param coeff coefficient of the element
    */
   void addElemSum( int idx, double coeff );

   /**
    * Create the list with the coeffs using the an array
    * @param coeffs array with the coefficients to be added
    * @param n number of elements of the array
    */
   void createColumnFromArray(double *coeffs, int n);

   /**
    * Create the list with the coeffs using the an array
    * @param coeffs array with the coefficients to be added
    * @param index index of each coefficient in coeffs
    * @param n number of elements of the array
    */
   void createColumnFromArray(double *coeffs, int *index, int n);

   /** 
    * Returns the number of non-zero elements of the column
    * @return number of non-zero elements of the column
    */
   inline int getNNZeros() { return nnZeros; }


   /**
      REturns if the column has a cycle of length k
      */
   inline bool getHasKCycle(int k ){
      if( ( k >=2 ) &&( k <= maxk ) )
         return hasKCycles[k-2];
      else
      {   printf("Invalid value of kcycle in get\n");
          throw(INVALIDKCYCLEVALUE);
      }
      return false;
    };

   /**
     Sets if the column has a cycle of length k
    */
   inline void setHasKCycle( int k ){ 
      if( ( k >=2 ) &&( k <= maxk ) )
         hasKCycles[k-2] = true;
      else
      {   printf("Invalid value of kcycle in set\n");
          throw(INVALIDKCYCLEVALUE);
      }
    };
	 
private:

   /** Parameters for the column */
   CParameters *params;

   /** Coefficient of the column in the objective function */
   double obj;

   /** Lower bound to the value of the column */
   double lb;

   /** Upper bound to the value of the column */
   double ub;

   /** Type of the column */
   ColType type;

   /** Number of LPs that use this row/column */
   int nLPs;

   /** Pointer to the first element of the row/column */
   CColElem* first;

   /** Pointer to the last element of the row/column */
   CColElem* last;

   /** Number of elements different of zero in this row/column */
   int nnZeros;

   /** Tells if the row/column can be separated from the pool */
   bool sepFromPool;

   /** Tells if the row/column is or not in the pool */
   bool inPool;

   /** Stores information if the column has k-cycles.
   Position 0 corresponds to k=2 and so on until maxk 
   The size of the vector is VSIZE (defined in CColumn.cpp)
   */
   bool* hasKCycles;

   /** Stores the maximum value of k-cycles that will be eliminated */
   int maxk;
};

//---------------------------------------------------------------------------//

#endif

