/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CLPSolver.h                      //
////////////////////////////////////////////
// Class that represents the linear       //
// programming solver.                    //
////////////////////////////////////////////

//---------------------------------------------------------------------------//

class CLPSolver;

//---------------------------------------------------------------------------//

#ifndef CLPSOLVER_H_
#define CLPSOLVER_H_

//---------------------------------------------------------------------------//

#include <cplex.h>

//---------------------------------------------------------------------------//

// Define the LP variable
#define LPptr CPXLPptr
#define BASIC_VAR CPX_BASIC
#define LOWER_ROW CPX_AT_LOWER

// Define objective sense
#define LP_MIN CPX_MIN 
#define LP_MAX CPX_MAX 

// Solver infinity
#define LP_INFINITY CPX_INFBOUND

// Display parameter
#define LP_MIPDISPLAY CPX_PARAM_MIPDISPLAY

// Mip optimal code
#define LP_MIPOPTIMAL CPXMIP_OPTIMAL

// Mip optimal code
#define LP_MIPOPTIMAL_TOL CPXMIP_OPTIMAL_TOL

// Mip optimal code
// Added by Marcelo for CPLEX 9
#define LP_MIPOPTIMAL_INFEAS CPXMIP_OPTIMAL_INFEAS

//---------------------------------------------------------------------------//

class CLPSolver
{
public:

   /**
    * Default constructor
    */
   CLPSolver();

   /**
    * Free memory allocated
    */
   ~CLPSolver();

   /**
    * Initialize the solver license
    * @return true, if the solver license is initiated; false, otherwise
    */
   bool initLicense();

   /** 
    * Solve the linear program using the primal simplex method
    * @param lp the linear program to be optimized
    * @param obj the objective function value (return)
    * @param x the solution array (return)
    * @param pi the dual array (return)
    * @param xsize size of the solution array
    * @param pisize size of the dual array
    */
   void solvePrimal(LPptr lp,double &obj,double *x,double *pi,int xsize,int pisize);

   /** 
    * Solve the linear program using the dual simplex method
    * @param lp the linear program to be optimized
    * @param obj the objective function value (return)
    * @param x the solution array (return)
    * @param pi the dual array (return)
    * @param xsize size of the solution array
    * @param pisize size of the dual array
    */
   void solveDual(LPptr lp,double &obj,double *x,double *pi,int xsize,int pisize);

   /**
    * Solve a MIP problem
    * @param lp the mip problem
    */
   void solveMIP(LPptr lp);

   /**
    * Returns the status of the problem
    * @param lp the linear program
    * @return the status of the problem
    */
   int getStatus(LPptr lp);

   /** 
    * Initialize the lp
    * @param lp the linear program to be initialized
    * @return true, if the LP was initiated; false, otherwise
    */
   bool initLP(LPptr &lp);

   /**
    * Free the lp variable
    * @param lp the linear program to be deleted
    */
   void freeLP(LPptr &lp);

   /**
    * Return the number of rows of the lp
    * @param lp the linear program
    * @return the number of rows of the lp
    */
   int getNRows(LPptr lp);

   /**
    * Returns the number of columns of the lp
    * @param lp the linear program
    * @return the number of columns of the lp
    */
   int getNColumns(LPptr lp);

   /**
    * Returns the number of nonzero coeffs in the lp
    * @param lp the linear program
    * @return the number of nonzero coeffs in the lp
    */
   int getNNZeros(LPptr lp);

   /**
    * Gets the solution of the lp
    * @param lp the lp to get the solution
    * @param x an array with the solution
    */
   void getX(LPptr lp, double* x);

   /**
    * Gets the dual solution of the lp
    * @param lp the lp to get the dual solution
    * @param y an array with the dual solution
    */
   void getPi(LPptr lp, double* y);

   /**
    * Gets the RHS array of the lp
    * @param lp the lp to get the RHS array
    * @param rhs an array which receives the RHS
    */
   void getRHS(LPptr lp, double* rhs);

   /**
    * Gets the RHS array of the lp from index begin to index end
    * @param lp the lp to get the RHS array
    * @param rhs an array which receives the RHS
    * @param begin the index of the first row for which to get the rhs
    * @param end the index of the last row for which to get the rhs
    */
   void getRHS(LPptr lp, double* rhs, int begin, int end);

   /**
    * Change the coefficient of the variable in the row
    * @param lp the linear program
    * @param idxRow index of the row
    * @param idxCol index of the column
    * @param coeff the coefficient value
    */
   void setCoef(LPptr lp, int idxRow, int idxCol, double coeff);

   // FUKA - DUALSTAB - BEGIN
   /**
   Changes the upper bound on a variable
   @param lp The linear program
   @param idxCol The index of the column
   @param ub   The upper bound
   */
   void setUB( LPptr lp, int idxCol, double ub );
   // FUKA - DUALSTAB - END
   
   /**
    * Sets the sense of the problem (minimization or maximization)
    * @param lp the linear program
    * @param sense the sense (minimization or maximization)
    */
   void setObjSense(LPptr lp, int sense); 

   /**
    * Sum the coefficient of the variable in the row with a number
    * @param lp the linear program
    * @param idxRow index of the row
    * @param idxCol index of the column
    * @param coeff the value to be added to the coefficient of the variable
    */
   void sumCoef(LPptr lp, int idxRow, int idxCol, double coeff);

   /**
    * Add a number of columns to the lp
    * @param lp the linear program
    * @param nc number of columns to be added
    * @param cJ coefficients of the columns in the objective function
    * @param lb lower bounds of the columns
    * @param ub upper bounds of the columns
    * @param tp type of the columns
    */
   void addNewCols(LPptr lp, int nc, double *cJ, double *lb, double *ub, char *tp);

   /**
    * Add a number of rows to the lp
    * @param lp the linear program
    * @param nr number of rows to be added
    * @param rhs RHS of the rows
    * @param rt type of the rows
    */
   void addNewRows(LPptr lp, int nr, double *rhs, char *rt);

   /**
    * Delete the rows in the specified range
    * @param lp the linear program
    * @param begin the index of the first row to be deleted
    * @param end the index of the last row to be deleted
    */
   void delRows(LPptr lp, int begin, int end);

   /**
    * Delete the columns in the specified range
    * @param lp the linear program
    * @param begin the index of the first column to be deleted
    * @param end the index of the last column to be deleted
    */
   void delCols(LPptr lp, int begin, int end);

   /**
    * Delete the rows in the specified range
    * @param lp the linear program
    * @param delstat incidence vector that tells if the row will deleted
    */
   void delRowSet(LPptr lp, int *delstat);

   /**
    * Delete the columns in the specified range
    * @param lp the linear program
    * @param delstat incidence vector that tells if the column will deleted
    */
   void delColSet(LPptr lp, int *delstat);

   /**
    * Gets the basis of the lp
    * @param cols the columns of the basis (return)
    * @param rows the rows of the basis (return)
    */
   void getBasis(LPptr lp, int *cols, int *rows);

   /**
    * Copy the basis of the lp
    * @param cols the columns of the basis (return)
    * @param rows the rows of the basis (return)
    */
   void copyBasis(LPptr lp, int *cols, int *rows);

   /**
    * Writes the problem in the LP format
    * @param lp the linear program to be printed
    * @param fName the name of the file
    */
   void writeLPProblem(LPptr lp, char *fName);

   /**
    * Try to set an integer parameter
    * @param param the parameter to set
    * @param val the value to be set
    */
   void setIntParameter(int param, int val);

   /**
    * Return the MIP problem solution
    * @param lp the MIP problem
    * @param xsol the mip solution
    * @param begin the first index of the solution
    * @param end the last index of the solution
    */
   void getMIPSol(LPptr lp, double *xsol, int begin, int end);

   /**
    * Get the MIP objective value
    * @param lp the MIP problem
    * @param obj the objective value
    */
   void getMIPObj(LPptr lp, double *obj);

   /**
    * Get the simplex number of iterations
    * @param lp the MIP problem
    */
   int getSimplexIter(LPptr lp);


   /**
    * Get the coefficient of a column in a row
    * @param lp the LP problem
    * @param idxRow the Index of the row
    * @param idxCol the Index of the column
    */
   double getCoef( LPptr lp, int idxRow, int idxCol);

protected:

   /** Variable that tells if the license was initiated */
   bool initialized;

   /** Cplex environment variable */
   CPXENVptr env;

};

//---------------------------------------------------------------------------//

#endif
