/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CMaster.h                        //
////////////////////////////////////////////
// Class that represents the control of   //
// the branch-and-cut-and-price           //
////////////////////////////////////////////

//---------------------------------------------------------------------------//

class CMaster;

//---------------------------------------------------------------------------//

#ifndef CMASTER_H_
#define CMASTER_H_

//---------------------------------------------------------------------------//

#include <stdio.h>
#include <stdarg.h>
#include "CColPool.h"
#include "CCutPool.h"
#include "CRow.h"
#include "CColumn.h"
#include "CLPSolver.h"
#include "CParameters.h"
#include "CBBNodeList.h"

//---------------------------------------------------------------------------//

#define SIZEINCREASE 10000


/**
 * Class that represents the control of the 
 * branch-and-cut-and-price algorithm
 */ 
class CMaster
{
protected:

   /** Best upper bound found during the execution of the algorithm */
   double bestUB;

   /** Best lower bound found during the execution of the algorithm */
   double bestLB;

   /** Log file for the resolution */
   FILE *logFile;

   /** Cut pool for the B & C & P */
   CCutPool* cutPool;

   /** Column pool for the B & C & P */
   CColPool* colPool;

   /** List with the open nodes of the tree */
   CBBNodeList openNodes;

   /** Size of the representation of the best solution */
   int solSize;

   /* Maximum size of the solution vector (>= solSize) */
   int maxSolSize;

   /** Best solution found during the execution */
   double* bestSol;

   /** The most deep node explored in the tree during the execution */
   int maxDepth;

   /** Total openned nodes in the tree  */
   int totalNodes;

   /** Total nodes of the tree solved */
   int totalSolvedNodes;

   /** Estimative of the percent of the tree solved */
   double elapsedBB;

   /** Number of nodes fixed to one */
   int numIntNodes;

   /** Number of nodes fixed to zero */
   int numIntZeroNodes;


   /** This is the final log file name */
   char endlogfname[200];
   
   /** This is the initial log file name */
   char initiallogfname[200];


public:

   /** 
    * Constructs a CMaster object
    */
   CMaster();

   /** 
    * Free memory allocated 
    */
   ~CMaster();

   /** 
    * Write the buffer information to the log file 
    * @param stdOutput true if buffer should also be printed on stdout
    */
//   void writeBufferToLog( bool stdOutput );

   /** 
    * Write a string to the log file 
    * @param stdOutput true if buffer should also be printed on stdout
    * @param format 
    */
   void writeLog( bool stdOutput, char* format, ... );

   /** 
    * Calculate the skewness of the resolution tree 
    * @return the skewness of the tree
    */
   double getSkew();

   /** 
    * Returns a pointer to the cut pool
    * @return a pointer to the cut pool
    */
   inline CCutPool* getCutPool(){ return cutPool; }

   /** 
    * Returns a pointer to the column pool
    * @return a pointer to the column pool
    */
   inline CColPool* getColPool(){ return colPool; }

   /**
    * Set the log file name
    * @param logfile log file name
    */
   void setLogFile(char* logfile);

   /**
    * Reads the data of the problem (to be implemented)
    */
   virtual void readData( int argc, char** argv ) = 0;

   /** 
    * Optimize the problem (to be implemented by the class) 
    */
   virtual void optimize() = 0;

   /**
    * Update the bound values
    */
   void updateBounds();

   /**
    * Add an open node to the resolution tree
    * @param newnode pointer to the node to be added
    */
   void addOpenNode( CBBNode* newnode );

   /**
    * Returns the best upper bound value
    * @return the best upper bound value
    */
   inline double getUB(){ return bestUB; }

   /**
    * Returns the best lower bound value
    * @return the best lower bound value
    */
   inline double getLB(){ return bestLB; }

   /**
    * Add a feasible solution to the problem
    * @param xsol the solution to be added
    * @param solval the value of the solution to be added
    */
   void addFeasibleSol( double* xsol, double solval );

   /**
    * Select breadth first node selection
    * @return a pointer to the selected node
    */
   CBBNode* selectBFSNode();

   /**
    * Select depth first node selection
    * @return a pointer to the selected node
    */
   CBBNode* selectDFSNode();

   /**
    * Select best bound node selection
    * @return a pointer to the selected node
    */
   CBBNode* selectBestBndNode();

   /**
    * Compute the node
    * @param a pointer to the node to the computed
    * @param the status of the node to be computed
    */
   void computeNode( CBBNode* node, NodeStatus stat );

   /** 
    * Checks if the best solution is optimal
    * @return true, if the solution is optimal; false otherwise
    */
   bool checkOptimal();

   /**
    * Clean the cut pool if the size of the pool exceed the maximum 
    * value especified in maxCutPoolSize
    * @return the number of cuts eliminated from the pool
    */
   int cleanCutPool();

   /**
    * Clean the column pool if the size of the pool exceed the maximum 
    * value especified in maxColPoolSize
    * @return the number of columns eliminated from the pool
    */
   int cleanColPool();

   /**
    * Empty the cut pool
    * @return the number of cuts eliminated from the pool
    */
   int emptyCutPool();

   /**
    * Empty the column pool
    * @return the number of columns eliminated from the pool
    */
   int emptyColPool();

   /**
    * Add a column to the pool of columns
    * @param c the column to be added
    * @return a pointer to the column added
    */
   CColumn* addColToPool( const CColumn &c );

   /**
    * Add a row to the pool of cuts
    * @param c the row to be added
    * @return a pointer to the row added
    */
   CRow* addRowToPool( const CRow &r );

   /**
    * Set the size of the best sol
    * @param s the size of the solution
    */
   void setSolSize(int s);

   /**
    * Returns the size of the solution
    * @return the size of the solution
    */
   inline int getSolSize(){ return solSize; }

   /** Buffer used to the write to the log file */
   char buffer[1000];

   /** Parameters for the branch and cut and price  */
   CParameters *params;

   /** Linear programming solver */
   CLPSolver *solver;
};

//---------------------------------------------------------------------------//

#endif
