/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CParameters.h                    //
////////////////////////////////////////////
// Class with the parameters of the       //
// branch-and-cut-and-price algorithm     //
////////////////////////////////////////////

//---------------------------------------------------------------------------//

class CParameters;

//---------------------------------------------------------------------------//

#ifndef CPARAMETERS_H_
#define CPARAMETERS_H_

//---------------------------------------------------------------------------//

class CParameters
{
public:

   /**
    * Default constructor. Initialize the default values.
    */ 
   CParameters();

   /**
    * Free memory. Restores original values.
    */
   ~CParameters();

   /**
    * Set if the objective function is integer or not
    * @param b true, if the objective function is integer; false, otherwise
    */
   inline void setObjInteger( bool b ){ objInteger = b; }

   /**
    * Returns if the objective function is integer or not
    * @return true, if the objective function is integer; false, otherwise
    */
   inline bool isObjInteger(){ return objInteger; }

   /** 
    * Returns the tolerance of the objective function 
    * @return the tolerance of the objective function 
    */
   inline double getObjEps(){ return objEps; }

   /** 
    * Set the tolerance of the objective function 
    * @param o the tolerance of the objective function 
    */
   inline void setObjEps(double o){ objEps = o; }

   /** 
    * Returns the tolerance of integrality 
    * @return the tolerance of integrality 
    */
   inline double getIntEps(){ return intEps; }

   /** 
    * Set the tolerance of integrality 
    * @param o the tolerance of integrality 
    */
   inline void setIntEps(double o){ intEps = o; }

   /** 
    * Returns the tolerance of cut violation
    * @return the tolerance of cut violation
    */
   inline double getCutEps(){ return cutEps; }

   /** 
    * Set the tolerance of cut violation
    * @param o the tolerance of cut violation
    */
   inline void setCutEps(double o){ cutEps = o; }

   /** 
    * Returns the maximum size of the column pool
    * @return the maximum size of the column pool
    */
   inline int getMaxColPoolSize(){ return maxColPoolSize; }

   /** 
    * Set the maximum size of the column pool
    * @param ms the maximum size of the column pool
    */
   inline void setMaxColPoolSize(int ms){ maxColPoolSize = ms; }

   /** 
    * Returns the maximum size of the cut pool
    * @return the maximum size of the cut pool
    */
   inline int getMaxCutPoolSize(){ return maxCutPoolSize; }

   /** 
    * Set the maximum size of the cut pool
    * @param ms the maximum size of the cut pool
    */
   inline void setMaxCutPoolSize(int ms){ maxCutPoolSize = ms; }

   /** 
    * Returns the infinity value
    * @return the infinity value
    */
   inline double getInfinity(){ return infinity; }

   /** 
    * Set the infinity value
    * @param iv the infinity value
    */
   inline void setMaxCutPoolSize(double iv){ infinity = iv; }

   /**
    * Set the fractional precision
    * @param p Fractional precision
    */
   inline void setFracPrec(int pv){ fracprec = pv; }

   /** Returns the fractional precision
   *
   */
   inline int getFracPrec(){ return fracprec; }

   /** Sets the node selection rule
    * @param n Node selection rule
    */
   inline void setNodeSel(int s){nodeSel=s;}

   /**
    * Returns the node Selection rule
    */
   inline int getNodeSel(void){ return nodeSel;}
private:

   /** Tells if the objective function is integer */
   bool objInteger;

   /** Fractional precision of the objective function */
   int fracprec;

   /** Tolerance to the objective function */
   double objEps;

   /** Tolerance to the integrality of the variables */
   double intEps;

   /** Tolerance to the cut violation */
   double cutEps;

   /** Infinity value */
   double infinity;

   /** The maximum size of the column pool */
   int maxColPoolSize;

   /** The maximum size of the cut pool */
   int maxCutPoolSize;

   /** Node selection rule */
   int nodeSel;

};

//---------------------------------------------------------------------------//

#endif
