/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CRow.h                           //
////////////////////////////////////////////
// Class that represents a row.           //
////////////////////////////////////////////

//---------------------------------------------------------------------------//

class CRow;

//---------------------------------------------------------------------------//

#ifndef CROW_H_
#define CROW_H_

//---------------------------------------------------------------------------//

#include <stdio.h>
#include "Constants.h"
#include "CParameters.h"

//---------------------------------------------------------------------------//

/**
 * Diferent types of rows (L -> <= ; E -> == ; G -> >=)
 */
enum RowType { L,G,E };

//---------------------------------------------------------------------------//

/**
 * Class that represents an item of a 
 * row                           
 */
class CRowElem
{

public:

   /** Identifier of the element */
   int idx;

   /** Coefficient of the element */
   double coeff;

   /** Pointer to the next element */
   CRowElem* next;

};

//---------------------------------------------------------------------------//

/**
 * Class that represents a row
 */
class CRow
{
public:

   /** 
    * Default constructor
    */
   CRow();

   /** 
    * Constructs a row with the parameters specified
    * @param p parameters for the row
    */
   CRow(CParameters *p);

   /** 
    * Free the memory allocated
    */
   ~CRow();

   /**
    * Copy constructor
    * @param r row to be copied
    */
   CRow( const CRow& r );

   /**
    * Set the parameters for the row
    * @param p parameters for the row
    */
   inline void setParameters(CParameters *p) { params = p; }

   /**
    * Returns the parameters for the row
    * @return the parameters for the row
    */
   inline CParameters* getParameters() { return params; }

   /**
    * Modify the RHS value of the row
    * @param v the value of the RHS
    */
   inline void setRHS( double v  ){ rhs = v; }

   /**
    * Returns the RHS value of the row
    * @return the value of the RHS
    */
   inline double getRHS(){ return rhs; }

   /**
    * Modify the row type
    * @param t the type of the row (L,G,E)
    */
   inline void setType( RowType t ){ type = t; }

   /**
    * Returns the row type
    * @return the type of the row (L,G,E)
    */
   inline RowType getType(){ return type; }

   /**
    * Tests if two rows are equal
    * @param r row to be compared
    * @return true if the two rows are equal; false otherwise
    */
   bool isRowEqual( const CRow &r );

   /**
    * Make a copy of the row
    * @param r row to be copied
    */
   void copy( const CRow &r );

   /**
    * Check if the solution violates the constraint
    * @param xsol array with the solution to be checked
    * @return true if the solution violates the constraint, false otherwise
    */
   bool isViolated( double* xsol );

   /**
    * Check if the solution violates the constraint
    * @param xsol array with the solution to be checked
    * @param value the value of the violation (return)
    * @return true if the solution violates the constraint, false otherwise
    */
   bool isViolated(double *xsol, double &value);

   /** 
    * Set if the row is in the pool or not
    * @param p true if the row is in pool, false otherwise
    */
   inline void setInPool( bool p ){ inPool = p; }

   /** 
    * Returns if the row is in the pool or not
    * @return true if the row is in pool, false otherwise
    */
   inline bool getInPool(){ return inPool; };

   /** 
    * Verifies if the row contains an element with
    * identifier idx
    * @param idx identifier to be search
    * @return true if the row contains idx, false otherwise
    */
   bool contains( int idx );

   /** 
    * Returns if the row can be separated from the pool
    * @return true if the row can be separated, false otherwise
    */
   inline bool getSepFromPool(){ return sepFromPool; }

   /** 
    * Set if the row can be separated from the pool
    * @param p true if the row can be separated from the pool, false otherwise
    */
   inline void setSepFromPool(bool l){ sepFromPool = l; }

   /** 
    * Returns the number of LPs that use this row
    * @return the number of LPs that use this row
    */
   inline int getNLPs() { return nLPs; }

   /** 
    * Set the number of LPs that use this row
    * @param n the number of LPs that use this row
    */
   inline void setNLPs(int n){ nLPs = n; }

   /** 
    * Increase the number of LPs that use this row
    */
   inline void increaseNLPs(){ nLPs++; }

   /** 
    * Decrease the number of LPs that use this row
    */
   inline void decreaseNLPs(){ if( nLPs > 0 ) nLPs--; }

   /** 
    * Returns a pointer to the first element of the row
    * @return a pointer to the first element of the row
    */
   inline CRowElem* getFirstElem(){ return first; }

   /**
    * Add an element to the row
    * @param idx index of the element
    * @param coeff coefficient of the element
    */
   void addElem( int idx, double coeff );

   /**
    * Add an element to the row (sum the coeffs if 
    * the element already exists)
    * @param idx index of the element
    * @param coeff coefficient of the element
    */
   void addElemSum( int idx, double coeff );

   /**
    * Create the list with the coeffs using the an array
    * @param coeffs array with the coefficients to be added
    * @param n number of elements of the array
    */
   void createRowFromArray(double *coeffs, int n);

   /**
    * Create the list with the coeffs using the an array
    * @param coeffs array with the coefficients to be added
    * @param index index of each coefficient in coeffs
    * @param n number of elements of the array
    */
   void createRowFromArray(double *coeffs, int *index, int n);

   /** 
    * Returns the number of non-zero elements of the row
    * @return number of non-zero elements of the row
    */
   inline int getNNZeros() { return nnZeros; }

private:

   /** Parameters for the row */
   CParameters *params;

   /** RHS of the row */
   double rhs;

   /** Row type */
   RowType type;

   /** Number of LPs that use this row/column */
   int nLPs;

   /** Pointer to the first element of the row/column */
   CRowElem* first;

   /** Pointer to the last element of the row/column */
   CRowElem* last;

   /** Number of elements different of zero in this row/column */
   int nnZeros;

   /** Tells if the row/column can be separated from the pool */
   bool sepFromPool;

   /** Tells if the row/column is or not in the pool */
   bool inPool;
};

//---------------------------------------------------------------------------//

#endif

