/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CVRPBBNode.h                     //
////////////////////////////////////////////
// Class that represents a node of the    //
// branch tree for the CVRP               //
////////////////////////////////////////////

#ifndef CVRPBBNODE_H_
#define CVRPBBNODE_H_

//---------------------------------------------------------------------------//

class CVRPBBNode;

//struct VElemStr;
//struct PredElemStr;
//struct dynPrgElem;
//struct IODist;

//---------------------------------------------------------------------------//

#include "CVRPMaster.h"
#include "CBBNode.h"
#include "CVRPFullGraph.h"
#include "CPUTimer.h"
#include "CVRPHash.h"
#include "cnstrmgr.h"


//---------------------------------------------------------------------------//



/**
 * Class that represents a node of the
 * branch tree for the CVRP
 */
class CVRPBBNode : public CBBNode
{

public:

   /**
    * Creates a new node
    * @param m master of the node
    * @param aID the identifier of the node
    */
   CVRPBBNode( CVRPMaster* m, double aID);

   /**
    * Creates a new node
    * @param f node father of this node
    * @param aID the identifier of the node
    * @param or information to calculate skewness
    */
   CVRPBBNode( CVRPBBNode *, double, int );


   /**
    * Free memory allocated
    */
   ~CVRPBBNode();

   /**
    * Initializes the node properties
    */
   void initializeNode();

   /**
    * Try to separate columns from the pool
    * @param maxcols maximum number of columns to be separated
    * @return the number of columns added to lp
    */
   virtual int separateColsFromPool(int maxcols);

   /**
    * Try to separate cuts from the pool
    * @param maxrows maximum number of cuts to be separated
    * @return the number of cuts added to lp
    */
   virtual int separateRowsFromPool(int maxrows);

   /**
    * Column generation for the CVRP
      Second parameter is: Do we eliminate columns with k-cycles 
      for k < kmax (for -itercg mode)
    * @return number of columns generated
    */
   int columnGen(int);


   /**
    * Column generation for the CVRP
      Modified version to evaluate dynamically which k to use in 
      the whole tree
    * @return number of columns generated
    */
   int columnGenKEval(int,double, double);
  
   /* Modified version of the fastDinProg to evaluate k's 
   */
   int fastDinProgKEval(int);
   

   /**
    * Dinamic programming for generate columns without 2-cycles
    * @return number of columns generated
    */
   int dinProg();

   /**
    * Dinamic programming for generate columns without k-cycles and
    * not generating all the states
    * @return number of columns generated
    */
   // FUKA - DUALSTAB - BEGIN
   int fastDinProg(int,int);
   // FUKA - DUALSTAB - END

   /**
    * Dinamic programming for generate columns to carp instances without 2-cycles
    * @return number of columns generated
    */
   int carpDynProg();

   /**
    * Create the initial master problem
    * @return a status code
    */
   int createMaster();

   /**
    * Separate exact the capacity cuts
    * @return the number of cuts separated
    */
   int exactCapacityCutGen();

   /**
    * Separate capacity cuts
    * @param noOfEdges number of non-zero edges 
    * @param maxNoOfCuts maximum number of cuts to separate
    * @param feasible tells if the solution is feasible after cuts (return)
    * @param maxViol max violation of the cuts (return)
    */
    void capacityCuts(int noOfEdges, int maxNoOfCuts, char &feasible, double &maxViol);

   /**
    * Separate multistar and partial multistar cuts
    * @param noOfEdges number of non-zero edges 
    * @param maxNoOfCuts maximum number of cuts to separate
    * @param maxViol max violation of the cuts (return)
    */
    void multiStarCuts(int noOfEdges, int maxNoOfCuts, double &maxViol);

   /**
    * Separate framed capacity cuts
    * @param noOfEdges number of non-zero edges 
    * @param maxNoOfTreeNodes maximum number of tree nodes to explored
    * @param maxNoOfCuts maximum number of cuts to separate
    * @param maxViol max violation of the cuts (return)
    */
    void framedCapacityCuts(int noOfEdges, int maxNoOfTreeNodes, int maxNoOfCuts, double &maxViol);

    /**
     * Separate strengthened comb cuts
     * @param qMin lower bound on the demand that must be delivered
     * @param noOfEdges number of non-zero edges 
     * @param maxNoOfCuts max number of cuts to be separated
     * @param maxViol max violation of the cuts (return)
     */
    void combCuts(int qMin, int noOfEdges, int maxNoOfCuts, double &maxViol);

    /**
     * Separate hypotour cuts
     * @param noOfEdges number of non-zero edges 
     * @param maxNoOfCuts max number of cuts to be separated
     * @param maxViol max violation of the cuts (return)
     */
    void hypotourCuts(int noOfEdges, int maxNoOfCuts, double &maxViol);

    /**
     * Separate generalized large multistar cuts
     * @param noOfEdges number of non-zero edges 
     * @param custList list of customers in the nucleous
     * @param custListSize size of list of customers in the nucleous
     * @param maxViol max violation of the cuts (return)
     */
    void generalizedLargeMSCuts(int noOfEdges, int *custList, int &custListSize, double maxViol);

    /**
     * Add the cuts separated with the CAPSEP package to LP
     * @param custListSize size of the list of customers of the nucleous of GLMS cut 
     * @param custList list of customers of the nucleous of GLMS cut 
     * @return number of cuts added to lp
     */
    int addCapSepCutsToLP(int custListSize, int *custList);

   /**
    * Cut generation for the CVRP
    * @return the number of cuts added
    */
   int cutGen();

   /**
    * Generate the carp bound cuts
    * @return the number of cuts added
    */
   int carpBoundCut();

   /**
    * Generate the bound cuts
    * @return the number of cuts added
    */
   int boundCut();

   /**
    * Try to fix edges by reduced cost using dinamic programming eliminating 2-cycles
    * @return number of edges fixed
    */
   int fixByRedCost();

   /**
    * Solves the node
    * @return the node status
    */
   NodeStatus optimize();

   /**
    * Verifies if the solution is feasible
    * @param the solution to be verified
    * @return true, if the solution is feasible; false, otherwise
    */
   bool isFeasible( double* x );

   /**
    * Do the branch in the node
    */
   void doBranch();

   /**
    * Choose an arc to do the branch
    * @param flagTwo aditional information (return)
    * @retunr the arc to do the branch
    */
   int chooseBranchArc(int &flagTwo);

   /**
    * Adds the basis stored to LP
    */
   virtual void addBasisToLP();

   /**
    * Sets the local graph of the node
    * @param g the graph to be set
    */
   inline void setGraph( CVRPFullGraph* g ) { localg = g; }

   /**
    * Returns the maximum number of artificial columns
    * @return the maximum number of artificial columns
    */
   inline int getMaxArtificials() { return maxArtificials; }

   /**
    * Sets the maximum number of artificial columns
    * @param ma the maximum number of artificial columns
    */
   inline void setMaxArtificials (int ma) { maxArtificials = ma; }

   /**
    * Converts solution to the edge format
    * @param lpe the solution in the edge format (return)
    * @param lp the solution in the node format
    */
   void expandSolution (double *lpe, double *lp);

   /**
    * Converts dual solution to the edge format
    * @param lpe the dual solution in the edge format (return)
    * @param lp the dual solution in the node format
    */
   void expandPi (double *lpe, double *lp);

   /**
    * Converts dual solution to the edge format
    * @param lpe the dual solution in the edge format (return)
    * @param lp the dual solution in the node format
    * @param lp the dual solution (edge format) to carp instances
    */
   void expandPi (double *lpe, double *lp, double *lpcarp);

   /**
    * Create an artificial column to a new row
    * @param r the index of the row
    * @param p the type of the row
    * @return true, if the column was created; false, otherwise
    */
   bool createArtificialColumn(int r,int p);

   /**
    * Fix edges of the graph
    */
   void masterFixEdges ();

   /**
    * Adds an expanded row to the linear program
    * @param listIdx list of index of the coeficients
    * @param listCoef list of coefficients to be added
    * @param nIdx size of the lists of the listIdx e listCoef
    * @param aRHS RHS of the new row
    * @param aType type of the new row
    */
   void addExpandedRowToLP(int *listIdx, double *listCoef, int nIdx, double aRHS, char aType);

   /**
    * Adds an expanded row to the linear program
    * @param aRow the row to be added
    */
   void addExpandedRowToLP(CRow &aRow);

   /**
    * Adds an expanded column to the linear program
    * @param listIdx list of index of the coeficients
    * @param listCoef list of coefficients to be added
    * @param nIdx size of the lists of the listIdx e listCoef
    * @param aObj coefficient of the objective function of the new column
    * @param aLB lower bound of the column
    * @param aUB upper bound of the column
    */
   void addExpandedColToLP(int *listIdx, double *listCoef, int nIdx, double aObj, 
                           double aLB, double aUB);

   /**
    * Adds an expanded column to the linear program
    * @param aCol the column to be added
    */
   void addExpandedColToLP(CColumn &aCol);

   /** 
    * Sets the number of fixed sets in the graph
    * @param a number of fixed sets in the graph
    */
   inline void setNFixedSets(int a) { nFixedSets=a; }

   /** 
    * Sets the list of fixed sets in the graph
    * @param a the list of fixed sets in the graph
    */
   inline void setFixedSets (CRow **a) { fSets = a; }

   /**
    * Print information about the routes
    * @param fout output file
    */
   void printll (FILE* &fout);

   /**
    * Generate columns with just one edge
    */
   void generateColumnsOneEdge();

   /**
    * Eliminate a number of non-used columns of the LP
    * @param maxCols maximum number of columns to be eliminated
    * @return the number of columns eliminated from LP
    */
   int eliminateLPColumns(int maxCols);

   /**
    * Eliminate a number of non-used rows of the LP
    * @param maxRows maximum number of rows to be eliminated
    * @return the number of rows eliminated from LP
    */
   int eliminateLPRows(int maxRows);

   /**
    * Remove the rows from the lpRows vector
    * @param delstat incidence vector that tells the rows to be removed
    * @param nRows number of rows of the LP
    */
   void removeRowsFromLPRows(int *delstat, int nRows);

   /**
    * Remove the columns from the lpCols vector
    * @param delstat incidence vector that tells the columns to be removed
    * @param nCols number of columns of the LP
    */
   void removeColsFromLPCols(int *delstat, int nCols);

   /**
    * Select the branch candidate sets
    */
   void getBranchCandidateSets();

   /**
     Eliminates columns from the LP that have cycles of length
     k or less
   */
   int eliminateKCycleCols(int k);

   /** Cronometer to store time to add cuts to LP */
   CPUTimer cronoAddCutsToLP;

   /** Cronometer to store time to add columns to LP */
   CPUTimer cronoAddColsToLP;

   /** Cronometer to store time to generate columns */
   CPUTimer cronoColumnGeneration;

   /** Cronometer to store time to generate cuts */
   CPUTimer cronoCutGeneration;

   /** Cronometer to store time solving LP */
   CPUTimer cronoSolveLP;

   /** Total Node solving time */
   CPUTimer nodetimer; 



   /** Used to eliminate all columns in the range from the hash, from
   the lpcols list and from the lp itself  */
   int eliminateColRange( int begin, int end );

   /** Used to eliminate all rows in the range from the hash, from
   the lprows list and from the lp itself  */
   int eliminateRowRange(int begin, int end);


   inline double getRootProximity(){ return rootproximity; };

   // 0 -> comes from branch close to root
   // 1 -> comes from branch in mid
   // 2 -> comes from branch close to leaf
   int branchorg;


   int SBtries;

private:

   
   // FUKA2 - BEGIN
   /** Controls if CG should stop after reaching maxcgiter */
   bool cgiterlimit;
   int maxcgiter;

   /** Controls if CG should be stopped by time or bound (used while evaluating
       which value of k is best */
   bool cgtimelimit;
   bool cgbndlimit;

   // Flag to control whether evaluation of k is being performed or not
   // (used inside the cg procedure)
   bool inevalk;

   // Flag to control whether SB is being performed
   bool inSB;

   /** These are auxiliary variables for the purpose above
       The actual rule is defined on the .cpp code */
   double bestktime;
   double bestkbound;

   // Stores if the node has been previously initialized, so it doesn't have to be 
   // again
   // (a node can be previously initalized if SB is made)
   bool initialized;   
   // FUKA2 - END

   // FUKA - DUALSTAB - BEGIN
   /** Used to check whether it is the first time dual stabilization is called
   (and therefore a heuristic has to be performed to estimate dual variables) */
   bool firststab;

   /** Array used to store the dual estimates used for the dual stabilization process */
   double* dualestimate;


   /**
   Stores the number of rows in the master (before any cutgen)
   */
   int nMasterRows;

   /** Stores the number of master columns that are not artificial
    (before any colgen) */
   int nMasterCols; 

   /** Maximum number of dual stabilization 1 columns */
   int maxStab1;
   
   /** Actual number of dual stabilization 1 columns */
   int nStab1;
   
   /** Maximum number of dual stabilization 2 columns */
   int maxStab2;
   
   /** Actual number of dual stabilization 2 columns */
   int nStab2;

   // FUKA - DUALSTAB - END
  
  
   /** Local graph that represents the CVRP instance */
   CVRPFullGraph* localg;

   /** Maximum number of artificial columns */
   int maxArtificials;

   /** Actual number of artificial columns */
   int nArtificials;

   /** Solution of the problem in the edge format */
   double *lpSolExp;

   /** Dual solution of the problem in the edge format */
   double *lpDualExp;

   /** Dual solution (edge format) of carp instances */
   double *lpDualCarp;

   /** Fixed sets of the node */
   CRow **fSets;

   /** Number of sets fixed  */
   int nFixedSets;


   /** Stores the edges fixed by reduced cost for this node */
   int* fixedEdges;

   /** Stores the number of edges fixed by reduced cost for this node */
   int nFixedEdges;

   /** Stores the value of the best path using edge (depot,i) */
   double* bestpath;

   /** Matrix for dinamic programming */
   struct VElemStr **vMatrix;

   /** Matrix for dinamic programming */
   struct PredElemStr **predMatrix;

   /** Matrix for fix by reduced cost */
   struct VElemStr **vMatrix2;

   /** Matrix for fix by reduced cost */
   struct PredElemStr **predMatrix2;

   /** Matrix for carp dynamic programming */
//   struct dynProgElem ***dpCarp;

   /** Auxiliary distance matrix for carp dynamic programming */
//   struct IODist **dpDist;

   /** Hash of rows */
   CVRPHash *rowHash;

   /** Hash of columns */
   CVRPHash *colHash;

   /** Variable to store the cuts of generated by CVRPSEP */
   CnstrMgrPointer myCutsCMP;

   /** Variable to store the old cuts of generated by CVRPSEP */
   CnstrMgrPointer myOldCutsCMP;

   /** Variable to store the sets used to branch */
   CnstrMgrPointer setsCMP;

   /** Store the origins of the edges (for CVRPSEP) */
   int *edgeHead;

   /** Store the destiny of the edges (for CVRPSEP) */
   int *edgeTail;

   /** Store the solution variables (for CVRPSEP) */
   double *edgeX;

   /** List of customers in the S set */
   int *listS;

   /** Node incidence array */
   int *nodeP;

   /** First time that a point is visited */
   bool firstTime;

   /** Path generated in the dynamic programming */
   int *pathDinProg;

   /** Tells to separate or not capacity cuts */
   bool sepCapCuts;

   /** Tells to separate or not framed capacity cuts */
   bool sepFCCuts;

   /** Tells to separate or not multistar cuts */
   bool sepMSCuts;

   /** Tells to separate or not hypotour cuts */
   bool sepHypCuts;

   /** Tells to separate or not comb cuts */
   bool sepCombCuts;

   /** Tells to separate or not gen large multistar cuts */
   bool sepGLMSCuts;


   /** Flag to control if it is the first time columns are being
   generated */
   bool firstcolgen;

   /** Flag to control if any bound has been set to zero on boundCut() */
   bool fixzero;



   /** Weights on the edges -> FixByRedCost of the sons of this node
   will be performed in order of the edgeweights */
   double* edgeweights;


   // Returns the estimate of how close this current node
   // is from being a leaf of the tree
   // The higher the value, the closest to the root
   // it is estimated.
   // The lower the value, the closest to be
   // a leaf
   // 
   // It is actually a calculation of the gap
   double calcRootProximity(double);

   // Estimates the number of branches
   // that will be performed starting with a LB of val
   double estimateNumBranches( double val );


   double rootproximity;



   // sbmode = 0 -> No SB
   // sbmode = 1 -> max-min evaluation rule with sbcgiterlimit
   // sbmode = 2 -> tree estimate eval rule without sbcgiterlimit
   int sbmode;

   

};



#endif


