/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CVRPFullGraph.h                  //
////////////////////////////////////////////
// Class that represents the graph for    //
// CVRP problem (full matrix)             //
////////////////////////////////////////////

#ifndef CVRPFULLGRAPH_H_
#define CVRPFULLGRAPH_H_

//---------------------------------------------------------------------------//

class CVRPFullGraph;

//---------------------------------------------------------------------------//

#include <stdio.h>

//---------------------------------------------------------------------------//

/**
 * Class that represents a vertex of the graph
 */
class CVRPFullGraphVertex
{
   friend class CVRPFullGraph;

private:

   /** Coordinate x of the vertex */
   double x;

   /** Coordinate y of the vertex */
   double y;
};

//---------------------------------------------------------------------------//

/**
 * Class that represents an edge of the graph
 */
class CVRPFullGraphEdge
{
   friend class CVRPFullGraph;

private:

   /** Cost of the edge */
   double cost;

   /** Origin vertex of the edge */
   int org;

   /** Destiny vertex of the edge */
   int dst;

   /** Tells if the edge is fixed */
   int fixed;

   /** Index of the edge */
   int index;

   /** Tells if the can be fixed in two */
   bool twoEdge;  
};

//---------------------------------------------------------------------------//

/**
 * Represents a full graph for the CVRP problem
 */
class CVRPFullGraph
{
public:

   /**
    * Default constructor
    */
   CVRPFullGraph();

   /**
    * Copy constructor
    * @param g the graph to be copied
    */
   CVRPFullGraph(const CVRPFullGraph &g);

   /**
    * Free memory allocated
    */
   ~CVRPFullGraph();

   /**
    * Allocate memory for the graph
    */
   void allocate(int n, int m);

   /**
    * Deallocate memory for the graph
    */
   void deAllocate();

   /**
    * Copy the specified graph
    * @param g the graph to be copied
    */
   void copy(const CVRPFullGraph &g);

   /**
    * Returns the number of vertex of the graph
    * @return the number of vertex of the graph
    */
   inline int getN() { return n; }

   /**
    * Returns the number of edges of the graph
    * @return the number of edges of the graph
    */
   inline int getM() { return m; }

   /**
    * Returns the cost of the edge
    * @param i origin of the edge
    * @param j destiny of the edge
    * @return the cost of the edge
    */
   inline double getEdgeCost (int i, int j) { return ((i>j)?(distMatrix[i][j]->cost):((i!=j)?(distMatrix[j][i]->cost):0)); }

   /**
    * Set the cost of the edge
    * @param i origin of the edge
    * @param j destiny of the edge
    * @param w the cost of the edge
    */
   inline void setEdgeCost (int i, int j, double w) { (i>j)?(distMatrix[i][j]->cost=w):((i!=j)?(distMatrix[j][i]->cost=w):0); }

   /**
    * Returns the cost of the edge
    * @param i index of the edge
    * @return the cost of the edge
    */
   inline double getEdgeCost (int i) { return listEdges[i].cost; } 

   /**
    * Set the cost of the edge
    * @param i index of the edge
    * @param w the cost of the edge
    */
   inline void setEdgeCost (int i, double w) { listEdges[i].cost=w; }

   /**
    * Returns the value of the edge
    * @param i origin of the edge
    * @param j destiny of the edge
    * @return -1, if the edge is not fixed; otherwise, the value
    */
   inline int getEdgeFixed (int i, int j) { return ((i>j)?(distMatrix[i][j]->fixed):((i!=j)?(distMatrix[j][i]->fixed):0)); }

   /**
    * Fix the value of the edge
    * @param i origin of the edge
    * @param j destiny of the edge
    * @param f the value to fix the edge
    */
   inline void setEdgeFixed (int i, int j, int f) { (i>j)?(distMatrix[i][j]->fixed=f):((i!=j)?(distMatrix[j][i]->fixed=f):0); }

   /**
    * Returns the value of the edge
    * @param i index of the edge
    * @return -1, if the edge is not fixed; otherwise, the value
    */
   inline int getEdgeFixed (int i) { return listEdges[i].fixed; }

   /**
    * Fix the value of the edge
    * @param i index of the edge
    * @param f the value to fix the edge
    */
   inline void setEdgeFixed (int i, int f) { listEdges[i].fixed=f; }

   /** 
    * Returns if the edge can be two
    * @param i origin of the edge
    * @param j destiny of the edge
    * @return true, if the edge can have value two; false, otherwise
    */
   inline bool canBeTwoEdge (int i, int j) { return ((i>j)?(distMatrix[i][j]->twoEdge):((i!=j)?(distMatrix[j][i]->twoEdge):0)); }

   /** 
    * Sets if the edge can be two
    * @param i origin of the edge
    * @param j destiny of the edge
    * @param f true, if the edge can have value two; false, otherwise
    */
   inline void setCanBeTwoEdge (int i, int j, bool f) { (i>j)?(distMatrix[i][j]->twoEdge=f):((i!=j)?(distMatrix[j][i]->twoEdge=f):0); }

   /** 
    * Returns if the edge can be two
    * @param i index of the edge
    * @return true, if the edge can have value two; false, otherwise
    */
   inline bool canBeTwoEdge (int i) { return listEdges[i].twoEdge; }

   /** 
    * Sets if the edge can be two
    * @param i index of the edge
    * @param f true, if the edge can have value two; false, otherwise
    */
   inline void setCanBeTwoEdge (int i, bool f) { listEdges[i].twoEdge=f; }

   /**
    * Returns the edge index
    * @param i origin of the edge
    * @param j destiny of the edge
    * @return the edge index
    */
   inline int getEdgeIndex (int i, int j) { return ((i>j)?(distMatrix[i][j]->index):((i!=j)?(distMatrix[j][i]->index):0)); } 

   /**
    * Sets the edge index
    * @param i origin of the edge
    * @param j destiny of the edge
    * @param k the edge index
    */
   inline void setEdgeIndex (int i, int j, int k) { (i>j)?(distMatrix[i][j]->index=k):((i!=j)?(distMatrix[j][i]->index=k):0); }

   /**
    * Returns the origin of the edge
    * @param i index of the edge
    * @return the origin of the edge
    */
   inline int getEdgeOrg (int i) { return listEdges[i].org; }

   /**
    * Sets the origin of the edge
    * @param i index of the edge
    * @param o the origin of the edge 
    */
   inline void setEdgeOrg (int i, int o) { listEdges[i].org=o; }

   /**
    * Returns the destiny of the edge
    * @param i index of the edge
    * @return the destiny of the edge
    */
   inline int getEdgeDst (int i) { return listEdges[i].dst; }

   /**
    * Sets the destiny of the edge
    * @param i index of the edge
    * @param o the destiny of the edge 
    */
   inline void setEdgeDst (int i, int d) { listEdges[i].dst=d; }

   /**
    * Returns the coordinate of x
    * @param i the index of the vertex
    * @return the coordinate of x
    */
   inline double getVertexX (int i) { return listVertex[i].x; }

   /**
    * Sets the coordinate of x
    * @param i the index of the vertex
    * @param x the coodinate of x
    */
   inline void setVertexX (int i, double x) { listVertex[i].x = x; }

   /**
    * Returns the coordinate of y
    * @param i the index of the vertex
    * @return the coordinate of y
    */
   inline double getVertexY (int i) { return listVertex[i].y; }

   /**
    * Sets the coordinate of y
    * @param i the index of the vertex
    * @param y the coodinate of y
    */
   inline void setVertexY (int i, double y) { listVertex[i].y=y; }

private:

   /** Distance matrix */
   CVRPFullGraphEdge ***distMatrix; 

   /** Number of vertex */
   int n;

   /** Number of edges */
   int m;

   /** List of edges of the graph */
   CVRPFullGraphEdge *listEdges;

   /** List of vertices of the graph */
   CVRPFullGraphVertex *listVertex;
};

//---------------------------------------------------------------------------//

#endif


