/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CVRPHash.h                       //
////////////////////////////////////////////
// Class that a hash used to calculate    //
// the coefficients of the compact CVRP   //
// formulation                            //
////////////////////////////////////////////

#ifndef CVRPHASH_H_
#define CVRPHASH_H_

//---------------------------------------------------------------------------//

#include <stdio.h>
#ifdef WIN32
#include <vector.h>
#elif defined LINUX
#include <vector>
using namespace std;
#endif

#include "Constants.h"

//---------------------------------------------------------------------------//

class CVRPHash;

//---------------------------------------------------------------------------//

/**
 * Class that represents an element of the hash
 */
class CVRPHashElem
{
   friend class CVRPHash;

private:

   /** Index of the element */
   int idx;

   /** Coefficient of the element */
   double coeff;
};

//---------------------------------------------------------------------------//

/**
 * Class that represents the hash used to calculate
 * the coefficients of the compact CVRP formulation
 */
class CVRPHash
{
public:

   /**
    * Constructs an hash with n positions
    */
   CVRPHash(int n);

   /**
    * Free memory allocated
    */
   ~CVRPHash();

   /** 
    * Add an item to the hash
    * @param key the key of the item
    * @param idx the index of the item
    * @param coeff the coefficient of the item
    */ 
   void add(int key, int idx, double coeff);

   /** 
    * Remove the last item of the hash with the specified key
    * @param key the key of the item
    */
   void removeLast(int key);

   /** 
    * Returns the size of the hash
    * @return the size of the hash
    */
   inline int getSize() { return size; }

   /**
    * Starts the iterator for the specified key
    * @param key the key of the hash
    */
   inline void startIterator(int key) { position = 0; }

   /**
    * Starts the iterator for the specified key
    * @param key the key of the hash
    * @param idx the index of the element (return)
    * @param coeff the coefficient of the element (return)
    * @return true, if there are more elements to iterate; false, otherwise
    */
   bool getNext(int key, int &idx, double &coeff);

   /**
    * Returns the last element of the hash
    * @param key the key of the element
    * @return the index of the last element of the hash
    */
   int getLast(int key);

   /**
    * Sum the value of the coefficient to the last element of the hash
    * @param key the key of the element
    * @param coeff the coefficient to be added
    */
   void sumLast(int key, double coeff);

   /**
    * Delete the elements of the specified index of the hash
    * @param elements incidence of the elements to be deleted
    */
   void delElements(int *elements);

   /** 
    * Returns the number of bytes used by the hash
    * @return the number of bytes used by the hash
    */
   int getNumBytes();

   /** 
    * Returns the number of bytes needed by the hash
    * @return the number of bytes needed by the hash
    */
   int getMinNumBytes();

private:

   /** Hash vector */
   vector< vector< CVRPHashElem > > hash;

   /** Size of the hash */
   int size;

   /** Position of the hash list with a specified key */
   int position;

};

//---------------------------------------------------------------------------//

#endif
