/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CVRPMaster.h                     //
////////////////////////////////////////////
// Class that represents the control of   //
// the branch-and-cut-and-price for the   //
// the CVRP                               //
////////////////////////////////////////////

#ifndef CVRPMASTER_H_
#define CVRPMASTER_H_

//---------------------------------------------------------------------------//

class CVRPMaster;

//---------------------------------------------------------------------------//

#include "CMaster.h"
#include "CVRPFullGraph.h"
#include "CVRPBBNode.h"
#include "CPUTimer.h"
#include "CParameters.h"
#include "CarpIO.h"

//---------------------------------------------------------------------------//

// Auxiliary structures for the dynamic programming (carp instances)
typedef struct dynPrgElemStr
{
//  int    c1;
//  int    c2;
  int    y1;
  int    y2;
  int    p1;
  int    p2;
  double v1;
  double v2;
} dynPrgElem;


typedef struct IODistStr
{
  double d[2][2];
} IODist;


typedef struct PredElemStr
{
   int p1,p2;
} PredElem;

typedef struct VElemStr
{
   double v1,v2;
} VElem;

//---------------------------------------------------------------------------//

/**
 * Class that represents the control of
 * the branch-and-cut-and-price for the
 * the CVRP
 */
class CVRPMaster: public CMaster
{
public:


   /**
    * Default constructor
    */
   CVRPMaster();

   /**
    * Free memory allocated
    */
   ~CVRPMaster();

   /** 
    * Tells that an output file in the TSPview format will be printed
    * @param val true, if the TSPView format will be used; false, otherwise
    */
   inline void setTspView(bool val) { tspView = val; }

   /** 
    * Returns if an output file in the TSPview format will be printed
    * @return true, if the TSPView format will be used; false, otherwise
    */
   inline bool getTspView() { return tspView; }

   /**
    * Returns the capacity of each vehicle
    * @return the capacity of each vehicle
    */
   inline int getVehicleCap(){ return vehicleCapacity; }

   /**
    * Sets the capacity of each vehicle
    * @param c the capacity of each vehicle
    */
   inline void setVehicleCap(int c) { vehicleCapacity = c; }

   /**
    * Returns the number of vehicles that must be used
    * @return the number of vehicles that must be used
    */
   inline int getK() { return numberOfVehicles; }

   /**
    * Sets the number of vehicles that must be used
    * @param mk the number of vehicles that must be used
    */
   inline void setK(int mk) { numberOfVehicles = mk; }

   /**
    * Sets the graph of the instance of the CVRP
    * @param g the graph to be used
    */
   inline void setGraph(CVRPFullGraph* g ){ gMaster = g; }

   /**
    * Returns the graph of the instance of the CVRP
    * @return the graph of the instance of the CVRP
    */
   inline CVRPFullGraph* getGraph(){ return gMaster; }

   /** 
    * Returns the minimum capacity of the vehicle that must me used
    * @return the minimum capacity of the vehicle that must me used
    */
   inline int getMinCapVehicle() { return minCapVehicle; }

   /** 
    * Sets the minimum capacity of the vehicle that must me used
    * @param mcv the minimum capacity of the vehicle that must me used
    */
   inline void setMinCapVehicle(int mcv) { minCapVehicle = mcv; }

   /**
    * Returns the number of the depot vertex
    * @return the number of the depot vertex
    */
   inline int getDepot() { return depot; }

   /**
    * Sets the number of the depot vertex
    * @param d the number of the depot vertex
    */
   inline void setDepot(int d) { depot = d; }

   /**
    * Returns the demand of the node
    * @param i the node to get the demand
    * @return the demand of the node
    */
   inline int getDemand(int i) { return listDemands[i]; }

   /**
    * Sets the demand of the node
    * @param i the node to set the demand
    * @param x the demand of the node
    */
   inline void setDemand(int i, int x) { listDemands[i]=x; }

   /**
    * Returns the list of node demands
    * @return the list of node demands
    */
   inline int* getAllDemands() { return listDemands; }

   /** 
    * Returns the sum of all edges of the graph
    * @return the sum of all edges of the graph
    */
   inline double getSumAllEdges() { return sumAllEdges; }

   /** 
    * Sets the sum of all edges of the graph
    * @param the sum of all edges of the graph
    */
   inline void setSumAllEdges (long s) {sumAllEdges = s; }

   /** 
    * Returns the sum of all demands of the nodes
    * @return the sum of all demands of the nodes
    */
   inline long getSumAllDemands () { return sumAllDemands; }

   /** 
    * Sets the sum of all demands of the nodes
    * @param s the sum of all demands of the nodes
    */
   inline void setSumAllDemands (long s) { sumAllDemands = s; }

   /**
    * Read the cvrp data file 
    * @param fname the name of the input file
    * @param g the graph to be charged
    */
   void readDataCvrp(char* fname, CVRPFullGraph* g);

   /**
    * Read the carp data file 
    * @param fname the name of the input file
    * @param g the graph to be charged
    */
   void readDataCarp(char* fname, CVRPFullGraph* g);

   /**
    * Read the data for the instance and the parameters of prompt command
    * @param argc the counter of the number of parameters of prompt command
    * @param argv the list of parameters of prompt command
    */
   void readData(int argc, char** argv);

   /**
    * Optimizes the problem
    */
   void optimize();

   /**
    * Preprocess the depot edges
    */
   void prepDepotEdges();

   /**
    * Write the solution to a file in the format of TSP view
    * @param node the node of the solution
    * @param fname the file name
    * @param xsol the solution to be printed
    * @param g the graph of the problem
    */
   void writeSolTSPView(CVRPBBNode *node, char *fname, double* xsol, CVRPFullGraph* g);

   /**
    *  Reads the solution from a fig file
    *  @param fname Name of the file to be read.
    * */
   void readFig( char* fname );

   /**
    * Write the solution to a file in the format of Ted Ralphs
    * @param node the node of the solution
    * @param prefix prefix to the solution file name
    * @param xsol the solution to be printed
    * @param g the graph of the problem
    */
   void writeSolTedRalphs(CVRPBBNode *node, char *prefix, double* xsol, CVRPFullGraph* g);

   /**
    * Computes the auxiliary distance matrix for carp dynamic programming
    */
   void computeDPDist( CarpIO* carp );

   /**
    * Returns the carp dynamic programming matrix
    * @return the carp dynamic programming matrix
    */
   inline dynPrgElem*** getdpCarp() { return dpCarp; }

   /**
    * Returns the auxiliary distance matrix for the carp dynamic programming
    * @return the auxiliary distance matrix for the carp dynamic programming
    */
   inline IODist** getdpDist() { return dpDist; }

   /**
    * Returns the cost matrix
    * @return the cost matrix
    */
   inline VElem** getVMatrix() { return vMatrix; }

   /**
    * Returns the predecessor matrix
    * @return the predecessor matrix
    */
   inline PredElem** getPredMatrix() { return predMatrix; }

   /**
    * Returns the cost matrix
    * @return the cost matrix
    */
   inline VElem** getVMatrix2() { return vMatrix2; }

   /**
    * Returns the predecessor matrix
    * @return the predecessor matrix
    */
   inline PredElem** getPredMatrix2() { return predMatrix2; }

   /**
    * Returns the edge head array
    * @return the edge head array
    */
   inline int* getEdgeHead() { return edgeHead; }

   /**
    * Returns the edge tail array
    * @return the edge tail array
    */
   inline int* getEdgeTail() { return edgeTail; }

   /**
    * Returns the edge x array
    * @return the edge x array
    */
   inline double* getEdgeX() { return edgeX; }

   /**
    * Returns the list with the S set
    * @return the list with the S set
    */
   inline int* getListS() { return listS; } 

   /**
    * Returns the array with node incidence
    * @return the array with node incidence
    */
   inline int* getNodeP() { return nodeP; }

   /**
    * Returns the array to store the dinprog path
    * @return the array to store the dinprog path
    */
   inline int* getPathDinProg() { return pathDinProg; }

   /**
    * Returns the instance file name
    * @return the instance file name
    */
   inline char* getInstanceFileName() { return instanceFileName; }

   /**
    * Returns the flag indicating if optimization should be done only on
    * root node
    * @return root only flag
    */
   inline bool getRootOnly(){ return rootonly; };


   /** Returns the value of kcycle parameter
   @return kcycleparam
   */
   inline int getKCycleParam(){ return kcycleparam; };


   /** Sets the value of kcycle parameter
   If kcycleparam < 2 then do only BC
   */
   inline void setKCycleParam( int kparam ){ kcycleparam = kparam; };

   /** Returns the value of itercg
   @return itercg
   */
   inline bool getIterCG(){ return itercg; };

   /**
      Returns the value of the revbb parameter
   */
   inline bool getRevBB(){ return revbb; };

   /**
      Returns the value of the vrptw parameter
   */
   inline bool getVrptw(){ return vrptw; };

   /**
      Returns the value of the carp parameter
   */
   inline bool getCarp(){ return carp; };

   /**
      Returns the value of the pearn parameter
   */
   inline bool getPearn(){ return pearn; };

   /**
      Returns the value of the longo parameter
   */
   inline bool getLongo(){ return longo; };

   /**
      Returns the value of the uchoa parameter
   */
   inline bool getUchoa(){ return uchoa; };

   /**
    Outputs to screen the usage of the program
   */
   void printUsage(char** argv);

   /** Cronometer to store the total time doing SB */ 
   CPUTimer sbtimer; 

   /** Cronometer to store the total time doing redcost fixing */
   CPUTimer fixredcosttimer;

   /** Cronometer to store time to add cuts to LP */
   CPUTimer mAddCutsToLPTimer;

   /** Cronometer to store time to add columns to LP */
   CPUTimer mAddColsToLPTimer;

   /** Cronometer to store time to generate columns */
   CPUTimer mColGenTimer;

   /** Cronometer to store time to generate cuts */
   CPUTimer mCutGenTimer;

   /** Cronometer to store time solving LP */
   CPUTimer mSolveLPTimer;


   int ncolgen;

   int ncutgen;


   /* Writes feasible solution to file */
   void writeFeasibleSol( char*, double*, CVRPFullGraph*); 
    
   /** Dynamic choice of k to be performed or not */
   inline bool getDynk() {return dynk;};

   /** Minimum number of sets to do SB on */
   inline int getPMin() {return pmin;};
   
   /** Maximum number of sets to do SB on */
   inline int getPMax() {return pmax;};
   
   /** Number of SBCGITERS  */
   inline int getSBCGIter() { return sbcgiter; };

   /** Matrix for carp dynamic programming */
   dynPrgElem*** dpCarp;

   /** Auxiliary distance matrix for carp dynamic programming */
   IODist** dpDist;


   inline double getBImp(){ return bimp; };

   inline double getRootProx(){ return rootprox; };

   inline double getLeafProx(){ return leafprox; };

   inline double getBdTarget(){ return bdtarget; };

   inline double getGoodFactor(){ return goodfactor; };

   //inline double getGoodGain(){ return goodgain; };

   inline bool getDoStab(){ return dostab; };

   inline bool getDoSB(){ return dosb; };



   int nmiddleleftSBtries;
   int nmiddlerightSBtries;
   int maxmiddleleftSBtries;
   int minmiddleleftSBtries;
   int maxmiddlerightSBtries;
   int minmiddlerightSBtries;
   
private:

   /** Determine if SB is to be done */
   bool dosb;

   /** Determine if dualstabilization is supposed to be done */
   bool dostab;

   /** Defines how close to goodgain a branch has to be to be considered
   good */
   double goodfactor;

   /** Defines the standard of how much a good branch should improve
   (in percentage of UB) */
   //double goodgain;

   /** Defines target to define order in which to try to do SB */
   double bdtarget;

   /** Parameter which defines what is to be close to the root */
   double rootprox;

   /** Parameter which defines what is to be close to the leafs */
   double leafprox;

   /** Stop after first node is fathomed */
   bool stopfath;

   /** Estimate on the improvement each branching does */
   double bimp;

   /** Dynamic choice of k to be performed or not */
   bool dynk;

   /** true if vrp with time windows instance will be solved */
   bool vrptw;

   /** true if carp instance will be solved */
   bool carp;

   /** true if Pearns graph transformation is used */
   bool pearn;

   /** true if Longoss graph transformation is used */
   bool longo;

   /** true if Uchoas graph transformation is used */
   bool uchoa;

   /** number of if Uchoas graph transformation used */
   int nUchoa;
   
   /** Minimum number of sets to do SB on */
   int pmin;
   
   /** Maximum number of sets to do SB on */
   int pmax;
   
   /** Number of SBCGITERS  */
   int sbcgiter;

   /** true if iterative kcycle CG is to be performed */
   bool itercg;

   /** true if BB nodes have to be chosen in the predicted code order. */
   bool revbb;
   
   /** Parameter that controls the length of k-cycles to eliminate 
       in column generation */
   int kcycleparam;
   
   /** Parameter that says if you only need to solve the root node */
   bool rootonly;
   
   /** Graph of the instance of CVRP */
   CVRPFullGraph *gMaster;

   /** Name of the file with the CVRP instance */
   char instanceFileName[100];

   /** Name of the solution file (tspView) */
   char prefixFileName[100];

   /** Tells if the file will be printed in the format of tspview */
   bool tspView; 

   /** Sum of the value of all edges (auxiliary) */
   double sumAllEdges;

   /** Sum of all demands of the problem (auxiliary) */
   long sumAllDemands;

   /** Capacity of each vehicle */
   int vehicleCapacity;

   /** Number of vehicles of the problem */
   int numberOfVehicles; 

   /** List of each node demand */
   int *listDemands;

   /** Minimum capacity of the vehicle that must be used */
   int minCapVehicle;

   /** The number of the depot vertex */
   int depot;

   /** Matrix for dinamic programming */
   VElem **vMatrix;

   /** Matrix for dinamic programming */
   PredElem **predMatrix;

   /** Matrix for fix by reduced cost */
   VElem **vMatrix2;

   /** Matrix for fix by reduced cost */
   PredElem **predMatrix2;

   /** Store the origins of the edges (for CVRPSEP) */
   int *edgeHead;

   /** Store the destiny of the edges (for CVRPSEP) */
   int *edgeTail;

   /** Store the solution variables (for CVRPSEP) */
   double *edgeX;

   /** List of vertices in the S set */
   int *listS;

   /** Node incidence array */
   int *nodeP;

   /** Path generated in the dynamic programming */
   int *pathDinProg;

   /**
    * Calculates the mdc of two numbers
    * @param m first number
    * @param n second number
    * @return the mdc of m and n
    */
   int MDC(int m, int n);

   /**
    * Compare the first n caracters of two strings ignoring the case 
    * @param a the first string
    * @param b the second string
    * @param n the number of caracters to be compared
    * @return true, if the strings are equal; false, otherwise
    */
   bool strNCaseCmp (char *a, char *b, int n);


   /**
    * Stores the best lp expanded solution
    * */
   double* bestlpe; 

   

};

//---------------------------------------------------------------------------//

#endif
