/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CBBNode.cpp                      //
////////////////////////////////////////////
// Class that represents a node of the    //
// branch tree                            //
////////////////////////////////////////////

#include "CBBNode.h"
#include "Constants.h"
#include <stdlib.h>
#include <math.h>

//---------------------------------------------------------------------------//

CBBNode::CBBNode(CMaster *m, double aID)
{
   if ( !m )
      throw(MASTERNOTINITEXCEPTION);

   master = m;
   id = aID;
   locallp = NULL;

   lbVal = - master->params->getInfinity();
   ubVal = master->params->getInfinity();

   parentLB = lbVal;
   lpSol = NULL;
   lpDual = NULL;
   origin = -1;
   solSize = 0;
   dualSize = 0;

   maxSolSize = 0;
   maxDualSize = 0;
}

//---------------------------------------------------------------------------//

CBBNode::CBBNode( CBBNode* father, double aID, int p_or)
{
   if ( !father )
      throw(NODENOTINITEXCEPTION);

   if (!(father->master))
      throw(MASTERNOTINITEXCEPTION);

   master = father->master;
   id = aID;
   locallp = NULL;
   
   lbVal = father->lbVal;
   ubVal = father->ubVal;
   parentLB = lbVal;

#ifndef LPONLY
   copyBasis((father->basis));
#endif

   lpSol = NULL;
   lpDual = NULL;
   origin = p_or;
   solSize = 0;
   dualSize = 0;
   maxSolSize = 0;
   maxDualSize = 0;
}

//---------------------------------------------------------------------------//

CBBNode::~CBBNode()
{
   register int i; // counter

   for( i = 0; i < (int)lpRows.size(); i++ )
   {
      if (lpRows[i])
         lpRows[i]->decreaseNLPs();
      lpRows[i]=NULL;
   }
   for( i = 0; i < (int)lpCols.size(); i++ )
   {
      if ( lpCols[i] )
         lpCols[i]->decreaseNLPs();
      lpCols[i]=NULL;
   }

   lpCols.clear();
   lpRows.clear();

   if ( locallp )
      master->solver->freeLP(locallp);

   locallp = NULL;

   if( lpSol )
      delete[] lpSol;
   if( lpDual )
      delete[] lpDual;

   master = NULL;
}

//---------------------------------------------------------------------------//

void CBBNode::copyBasis(const CBasis &bsrc)
{
   basis.copyBasis(bsrc);
}

//---------------------------------------------------------------------------//

void CBBNode::solvePrimal()
{
   master->solver->solvePrimal(locallp,lbVal,lpSol,lpDual,solSize,dualSize);
}

//---------------------------------------------------------------------------//

void CBBNode::solveDual()
{
   master->solver->solveDual(locallp,lbVal,lpSol,lpDual,solSize,dualSize);
}

//---------------------------------------------------------------------------//

bool CBBNode::isDescendant( double pid )
{
   double auxid; // auxiliary identifier
   
   auxid = id;

   while( auxid >= pid - master->params->getIntEps() )
   {
      if( fabs( auxid - pid ) < master->params->getIntEps() )
         return true;

      auxid = floor( auxid / 2.0 + master->params->getIntEps() );
   }
   return false;
}

//---------------------------------------------------------------------------//

int CBBNode::getDepth()
{
   double auxid; // auxiliary identifier
   int dpt; // depth of the node

   auxid = id;
   dpt = 0;

   while( auxid > 1.0 + master->params->getIntEps() )
   {
      dpt++;
      auxid = floor( auxid / 2.0 + master->params->getIntEps() );
   }
   return dpt;
}

//---------------------------------------------------------------------------//

void CBBNode::addColToLP(CColumn &col)
{
   int idxcol; // index of the column to be added
   double auxLB; // auxiliary lower bound
   double auxUB; // auxiliary upper bound
   double auxCJ; // auxiliary coefficient of obj function
   char auxTP; // auxiliary type
   CColElem *auxColElem; // auxiliary column element
   
   idxcol = master->solver->getNColumns(locallp);

   auxCJ = col.getObj();
   auxLB = col.getLB();
   auxUB = col.getUB();
   
   if (col.getType()==NONE || col.getType()==C)
   {
      master->solver->addNewCols(locallp, 1, &auxCJ, &auxLB, &auxUB, NULL);
   }
   else
   {
      if ( col.getType() == B )
         auxTP = 'B';
      else
         auxTP = 'I';

      master->solver->addNewCols(locallp, 1, &auxCJ, &auxLB, &auxUB, &auxTP);
   }

   auxColElem = col.getFirstElem();

   while( auxColElem )
   {
      master->solver->setCoef(locallp,auxColElem->idx,idxcol,auxColElem->coeff);
      auxColElem = auxColElem->next;
   }
}

//---------------------------------------------------------------------------//

void CBBNode::addRowToLP(CRow &r)
{
   int idxrow; // index of the row to be added
   double auxRHS; // auxiliary RHS of the row
   char auxRT; // auxiliary row type
   CRowElem *auxRowElem; // auxiliary row element
    
   idxrow = master->solver->getNRows(locallp);

   auxRHS = r.getRHS();

   if ( r.getType() == G )
      auxRT = 'G';
   else if (r.getType() == L )
      auxRT = 'L';
   else 
      auxRT = 'E';

   master->solver->addNewRows(locallp, 1, &auxRHS, &auxRT);

   auxRowElem = r.getFirstElem();

   while( auxRowElem )
   {
      master->solver->setCoef(locallp,idxrow,auxRowElem->idx,auxRowElem->coeff);

      auxRowElem = auxRowElem->next;
   }
}

//---------------------------------------------------------------------------//

CRow* CBBNode::addRow(const CRow &r)
{
   CRow* auxrow; // auxiliary row
   
   auxrow = master->addRowToPool(r);
   lpRows.push_back( auxrow );
   return auxrow;
}

//---------------------------------------------------------------------------//

CColumn* CBBNode::addCol(const CColumn &c)
{
   CColumn* auxcol; // auxiliary column
   
   auxcol = master->addColToPool(c);
   lpCols.push_back( auxcol );

   return auxcol;
}

//---------------------------------------------------------------------------//

int CBBNode::separateRowsFromPool(int nR)
{
   int ncutpool; // number of cuts separated from the pool
   CCutPoolElem *auxCutElem; // auxiliary cut pool element
   CRow* auxRow; // auxiliary row

   ncutpool = 0;

   auxCutElem = master->getCutPool()->getFirst();

   while( auxCutElem )
   {
       auxRow = auxCutElem->row;

      if( auxRow->isViolated(lpSol) )
      {
         addRowToLP((*auxRow));
         lpRows.push_back(auxRow);
         auxRow->increaseNLPs();

         ncutpool++;

         if( nR > 0 && ncutpool >= nR )
            return ncutpool;
      }
      auxCutElem = auxCutElem->next;
   }

   return ncutpool;
}

//---------------------------------------------------------------------------//

int CBBNode::separateColsFromPool(int nC)
{
   int ncolpool; // number of columns separated
   CColPoolElem *auxColElem; // auxiliary col pool element
   CColumn* auxCol; // auxiliary column

   ncolpool = 0;

   auxColElem = master->getColPool()->getFirst();

   while( auxColElem )
   {
      auxCol = auxColElem->column;

      if( auxCol->getRedCost(lpDual) > master->params->getObjEps() )
      {
         addColToLP((*auxCol));

         lpCols.push_back(auxCol);
         auxCol->increaseNLPs();
         ncolpool++;

         if( nC > 0 && ncolpool >= nC )
            return ncolpool;
      }
      auxColElem = auxColElem->next;
   }

   return ncolpool;
}

//---------------------------------------------------------------------------//

NodeStatus CBBNode::processSolvedNode( double lbAdd, double maxObjVal )
{
   NodeStatus retval; // status of the solved node

   if(isFeasible(lpSol))
   {
      if( master->params->isObjInteger() )
         ubVal = ceil( lbVal - master->params->getObjEps() );
      else if( master->params->getFracPrec() > 0 )
      {
         double precision = pow(10.0, master->params->getFracPrec() );
         ubVal = ceil(  precision * 
            lbVal - master->params->getObjEps() ) / precision;
      }
      else
         ubVal = lbVal;
      

      retval = FEASIBLE;

      master->addFeasibleSol(lpSol, ubVal);
   }
   else if(isFathomed( lbVal + lbAdd ))
   {
      retval = FATHOMED;
   }
   else if( lbVal < maxObjVal + master->params->getObjEps() )
   {
      retval = BRANCH;
   }
   else
   {
      retval = INFEASIBLE;
   }

   return retval;
}

//---------------------------------------------------------------------------//

double CBBNode::checkDualObj()
{
   int nrows; // number of rows of the locallp
   double* auxdual; // dual solution of the locallp
   double* rhs; // rhs of the locallp
   double auxsum; // auxiliary sum variable
   register int i; // counter

   nrows = master->solver->getNRows(locallp);
   auxdual = new double[ nrows ];

   if ( !auxdual )
      throw(LOWMEMORYEXCEPTION);

   rhs = new double[ nrows ];

   if ( !rhs )
      throw(LOWMEMORYEXCEPTION);

   master->solver->getRHS(locallp,rhs);
   master->solver->getPi(locallp,auxdual);
   
   // Calculates the dual objective
   auxsum= 0.0;
   for(i = 0; i < nrows; i++)
   {
      auxsum += auxdual[i] * rhs[i];
   }

   if ( auxdual )
      delete[] auxdual;

   if ( rhs )
      delete[] rhs;

   return auxsum;
}

//---------------------------------------------------------------------------//

void CBBNode::obtainBasisFromLP(int nMC, int nMR)
{
   register int i; // counter
   int* cols; // columns of the basis
   int* rows; // rows of the basis
   int ncols; // number of columns
   int nrows; // number of rows

   cols = NULL;
   rows = NULL;

   basis.allocBasis(nMC, nMR);

   ncols = master->solver->getNColumns(locallp);
   nrows = master->solver->getNRows(locallp);

   cols = new int[ncols];

   if (!cols)
      throw(LOWMEMORYEXCEPTION);

   rows = new int[nrows];

   if (!rows)
      throw(LOWMEMORYEXCEPTION);

   master->solver->getBasis(locallp,cols,rows);

   for(i=0; i<ncols; i++)
   {
      if(i < nMC )
      {
         basis.masterCols[i] = cols[i];

         if(i >= basis.getNStaticCols())
         {
            if(cols[i] == BASIC_VAR) 
            {
               basis.setNBasicDinamic( basis.getNBasicDinamic() + 1 );
            }
         }
      }
      else
      {
         if(i-nMC >= (int)lpCols.size())
            throw(BASISCOLSIZEEXCEPTION);

         if( cols[i] == BASIC_VAR ) 
         {
            basis.addBasicCol(lpCols[i-nMC] );
         }
      }
   }
   
   for( i = 0; i < nrows; i++ )
   {  
      if( i < nMR )
      {
         basis.masterRows[i] = rows[i];
      }
      else
      {
         if( i-nMR >= (int)lpRows.size() )
            throw(BASISROWSIZEEXCEPTION);

         if( rows[i] == LOWER_ROW)
         {
            basis.addBasicRow(lpRows[i-nMR]);
         }
      }
   }

   if( cols )
      delete[] cols;

   if( rows )
      delete[] rows;   
}

//---------------------------------------------------------------------------//

void CBBNode::addBasisToLP()
{
   int* cols; // cols of the basis
   int* rows; // rows of the basis
   int ncols; // number of columns of the basis
   int nrows; // number of rows of the basis
   register int i; // counter
   CColumn *auxCol; // auxiliary column
   CRow *auxRow; // auxiliary row

   cols = NULL;
   rows = NULL;
   ncols = basis.nMasterCols + basis.basicCols.size();

   cols = new int[ncols];

   if ( !cols )
      throw(LOWMEMORYEXCEPTION);

   for(i = 0; i < ncols; i++)
   {
      if( i < basis.nMasterCols )
      {
         cols[i] = basis.masterCols[i];
      }
      else
      {
         auxCol = basis.basicCols.at(i);
         addColToLP((*auxCol)); 

         lpCols.push_back(auxCol);
         auxCol->increaseNLPs();

         cols[i] = BASIC_VAR;
      }
   }

   nrows = basis.nMasterRows + basis.basicRows.size();

   rows = new int[nrows];

   if ( !rows )
      throw(LOWMEMORYEXCEPTION);

   for(i = 0; i < nrows; i++)
   {
      if( i < basis.nMasterRows )
      {
         rows[i] = basis.masterRows[i];
      }
      else
      {
         auxRow = basis.basicRows.at(i);
         addRowToLP( (*auxRow) ); 

         lpRows.push_back(auxRow);
         auxRow->increaseNLPs();
         rows[i] = LOWER_ROW;
      }
   }

   master->solver->copyBasis(locallp,cols,rows);

   if( cols )
      delete [] cols;
   if( rows )
      delete [] rows;
}

//---------------------------------------------------------------------------//
// Parameter lbAdd must be 3.0 when a carp instance is solved using cvrp
// techniques and Pearn's graph transformation, otherwise it should be 0.0
bool CBBNode::isFathomed( double p_lb )
{
   if (master->params->isObjInteger())
   {
      return (ceil( p_lb - master->params->getObjEps() ) >= 
         (master->getUB() - master->params->getObjEps()));
   }
   else if( master->params->getFracPrec() > 0 )
   {
      double precision = pow( 10.0, master->params->getFracPrec() );
      return( ceil( precision * p_lb - master->params->getObjEps() ) / precision >=
         master->getUB() - master->params->getObjEps() );
      
   }
   else
   {
      return ((p_lb) >= (master->getUB() - master->params->getObjEps()));
   }
}

//---------------------------------------------------------------------------//

double CBBNode::getFather()
{
   if( id < 1.1 )
   {
      return -1.0;
   }
   else
   {
      return( floor( id / 2.0 ) );
   }
}

//---------------------------------------------------------------------------//

void CBBNode::setSolSize( int s )
{
   solSize = s;
   
   if( lpSol )
   {
      if( maxSolSize < s )
      {
         delete [] lpSol;
         lpSol = NULL;
	 maxSolSize = s + SIZEINCREASE;
	 lpSol = new double[ maxSolSize ];
      }
   }
   else
   {
      maxSolSize = s + SIZEINCREASE;
      lpSol = new double[ maxSolSize ];
   }
   

   master->setSolSize(s);
}

//---------------------------------------------------------------------------//

void CBBNode::setDualSize( int s )
{
   dualSize = s;
   if( lpDual )
   {
      if( maxDualSize < s )
      {
         delete [] lpDual;
	 lpDual = NULL;
	 maxDualSize = s + SIZEINCREASE;
	 lpDual = new double[ maxDualSize ];
      }
   }
   else
   {
      maxDualSize = s + SIZEINCREASE;
      lpDual = new double[ maxDualSize ];
   }
   
}

//---------------------------------------------------------------------------//

double CBBNode::getLB()
{ 
   return lbVal; 
}
