/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CBasis.h                         //
////////////////////////////////////////////
// Class that represents a basis of a lp  //
////////////////////////////////////////////

//---------------------------------------------------------------------------//

#include "Constants.h"
#include "CBasis.h"

//---------------------------------------------------------------------------//

CBasis::CBasis()
{
   masterCols = NULL;
   masterRows = NULL;
   allocated = false;
   nBasicDinamic = 0;
   nStaticCols = 0;
   nMasterCols = 0;
   nMasterRows = 0;
}

//---------------------------------------------------------------------------//

CBasis::~CBasis()
{
   if( allocated )
   {
      deallocBasis();
   }
}

//---------------------------------------------------------------------------//

void CBasis::deallocBasis()
{
   register int i; // counter
   CColumn *auxCol; // auxiliary column
   CRow *auxRow; // auxiliary row

   for (i=0;i<(int)basicCols.size();i++)
   {
      auxCol = basicCols[i];
      if (auxCol)
         auxCol->decreaseNLPs();
      basicCols[i]=NULL;
   }

   basicCols.clear();

   

   for (i=0;i<(int)basicRows.size();i++)
   {
      auxRow = basicRows[i];
      if (auxRow)
         auxRow->decreaseNLPs();
      basicRows[i]=NULL;
   }

   basicRows.clear();


   if( masterCols )
   {
      delete[] masterCols;
   
      masterCols = NULL;
   }
   if( masterRows )
   {
      delete[] masterRows;
      masterRows = NULL;
   }

   nMasterCols = 0;
   nMasterRows = 0;
   nStaticCols = 0;
   nBasicDinamic = 0;
   allocated = false;
}

//---------------------------------------------------------------------------//

void CBasis::allocBasis(int nC, int nR)
{
   if( allocated )
   {
      deallocBasis();  
   }

   masterCols = new int[ nC ];

   if ( !masterCols )
      throw(LOWMEMORYEXCEPTION);

   masterRows = new int[ nR ];

   if ( !masterRows )
      throw(LOWMEMORYEXCEPTION);


   nMasterCols = nC;
   nMasterRows = nR;

   allocated = true;
}

//---------------------------------------------------------------------------//

void CBasis::copyBasis(const CBasis &bs)
{
   register int i; // counter
   CColumn *auxCol; // auxiliary column
   CRow *auxRow; // auxiliary row

   // Basis to be copied was not allocated
   if ( ! bs.allocated )
      return;

   // Basis are already allocated
   if ( allocated )
   {
      deallocBasis();
   }
   
   // Allocate memory for the new basis
   allocBasis( bs.nMasterCols, bs.nMasterRows );
   
   // Copy the master columns
   for( i = 0; i < nMasterCols; i++ )
   {
      masterCols[i] = bs.masterCols[i];
   }

   basicCols.resize(bs.basicCols.size());

   for (i=0;i < (int)bs.basicCols.size();i++)
   {
      auxCol = bs.basicCols.at(i);
      if ( auxCol )
         auxCol->increaseNLPs();
      basicCols.at(i) = auxCol;
   }

   // Copy the master rows
   for( i = 0; i < nMasterRows; i++ )
   {
      masterRows[i] = bs.masterRows[i];
   }

   basicRows.resize(bs.basicRows.size());

   for (i=0;i < (int)bs.basicRows.size();i++)
   {
      auxRow = bs.basicRows.at(i);
      if ( auxRow )
         auxRow->increaseNLPs();
      basicRows.at(i) = auxRow;
   }

   nStaticCols = bs.nStaticCols;
   nBasicDinamic = bs.nBasicDinamic;
}

//---------------------------------------------------------------------------//

int CBasis::getColStat(int i )
{
   if( ( i >= nMasterCols ) || ( i < 0 ) )
   {
      throw(ARRAYINVALIDPOSITION);
   }

   return masterCols[i];
}

//---------------------------------------------------------------------------//

int CBasis::getRowStat( int i )
{
   if( ( i >= nMasterRows ) || ( i < 0 ) )
   {
      throw(ARRAYINVALIDPOSITION);
   }
   return masterRows[i];
}

//---------------------------------------------------------------------------//

void CBasis::addBasicCol(CColumn* col)
{
   col->increaseNLPs();
   basicCols.push_back(col);
}

//---------------------------------------------------------------------------//

void CBasis::addBasicRow(CRow* row)
{
   row->increaseNLPs();
   basicRows.push_back(row);
}

//---------------------------------------------------------------------------//
