/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CColPool.cpp                     //
////////////////////////////////////////////
// Class that represents a pool of cols.  //
////////////////////////////////////////////

#include "CColPool.h"
#include <stdio.h>

//---------------------------------------------------------------------------//

CColPoolElem::CColPoolElem()
{
   next = NULL;
   prev = NULL;
   column = NULL;
}

//---------------------------------------------------------------------------//

CColPoolElem::~CColPoolElem()
{
   next = NULL;
   prev = NULL;

   if ( column )
      delete column;
}

//---------------------------------------------------------------------------//

CColPool::CColPool()
{ 
   first = NULL;
   last = NULL;
   nCols = 0;
}

//---------------------------------------------------------------------------//

CColPool::~CColPool()
{
   empty();
}

//---------------------------------------------------------------------------//

bool CColPool::isInPool(const CColumn &aCol)
{
   CColPoolElem *aux; // auxiliary element

   // Search for the row
   aux = first;
   while ( aux )
   {
      if ( aux->column )
      {
         if ( aux->column->isColumnEqual(aCol) )
            return true;
      }

      aux = aux->next;
   }

   return false;
}

//---------------------------------------------------------------------------//

void CColPool::empty()
{
   CColPoolElem *aux,*aux2; // auxiliary elements

   aux = first;

   while ( aux )
   {
      aux2 = aux;

      aux = aux->next;

      delete aux2;
      nCols--;
   }

   first = NULL;
   last = NULL;

   if (nCols > 0)
      throw( MEMNOTDEALLOCEXCEPTION );

}

//---------------------------------------------------------------------------//

int CColPool::clean()
{
   int nRemoved; // number of cols removed from the pool
   CColPoolElem *aux,*aux2; // auxiliary elements

   aux = first;
   nRemoved = 0;

   while( aux )
   {
      aux2 = aux->next;

      if ( aux->column )
      {
         if ( aux->column->getNLPs()==0 )
         {
            // Remove the column
            if ( aux == first )
            {
               first = aux->next;

               if (first)
                  first->prev = NULL;
               else
                  last = NULL;

               delete aux;

               nCols--;
            }
            else if ( aux == last )
            {
               last = last->prev;
               if ( last )
                  last->next = NULL;
               else
                  first = NULL;
               

               delete aux;

               nCols--;
            }
            else
            {
               aux->next->prev = aux->prev;
               aux->prev->next = aux->next;

               delete aux;

               nCols--;
            }


            nRemoved++;
         }
      }

      aux = aux2;
   }

   return nRemoved;
}

//---------------------------------------------------------------------------//

CColumn* CColPool::addColumn( const CColumn &aCol )
{
   CColPoolElem *aux; // auxiliary element
   CColumn *newCol; // new column created in the pool

   aux = first;

   while( aux )
   {
      if ( aux->column )
      {
         // Column is already in the pool
         if (aux->column->isColumnEqual(aCol))
         {
            aux->column->increaseNLPs();

            return (aux->column);
         }
      }

      aux = aux->next;
   }

   // Add new row to the pool
   newCol = NULL;
   newCol = new CColumn( aCol );

   if ( !newCol )
      throw(LOWMEMORYEXCEPTION);

   newCol->increaseNLPs();
   newCol->setInPool(true);

   // Create new element in the pool
   aux = NULL;
   aux = new CColPoolElem;

   if (!aux)
      throw(LOWMEMORYEXCEPTION);

   if ( first )
   {
     aux->column = newCol;
     aux->prev = last;
     aux->next = NULL;
     last->next = aux;

     last = aux;
     nCols++;
   }
   else
   {
     aux->column = newCol;
     aux->prev = NULL;
     aux->next = NULL;

     first = aux;
     last = aux;

     nCols++;
   }

   return newCol;
}

//---------------------------------------------------------------------------//
