/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CColumn.cpp                      //
////////////////////////////////////////////
// Class that represents a column.        //
////////////////////////////////////////////

#include "CColumn.h"
#include <stdio.h> 
#include <stdlib.h>
#include <string.h>

//---------------------------------------------------------------------------//

// Define used only to make sure the vector hasKCycles' 
//  size is always the same and
// we don't forget to use the -1
#define VSIZE maxk-1

CColumn::CColumn()
{
   // Init members
   first = NULL;
   last = NULL;
   nnZeros = 0;
   sepFromPool = true;
   inPool = false;
   lb = -INFINITY_VALUE;
   ub = INFINITY_VALUE;
   obj = 0.0;
   type = C;
   params = NULL;
   nLPs = 0;

   hasKCycles = NULL;
   maxk = 0;
}

//---------------------------------------------------------------------------//

CColumn::CColumn(CParameters *p, int __maxk)
{
   // Init members
   first = NULL;
   last = NULL;
   nnZeros = 0;
   sepFromPool = true;
   inPool = false;
   params = p;
   if ( !p )
   {
      lb = -INFINITY_VALUE;
      ub = INFINITY_VALUE;
   }
   else
   {
      lb = - params->getInfinity();
      ub = params->getInfinity();
   }
   obj = 0.0;
   type = C;
   nLPs = 0;

   maxk = __maxk;
   if( VSIZE > 0 )
   {
      hasKCycles = new bool[ VSIZE ];
      memset( hasKCycles, 0, VSIZE * sizeof(bool) );
   }
   else
      hasKCycles = NULL;

}

//---------------------------------------------------------------------------//

CColumn::CColumn( const CColumn &c )
{
   // Init members
   first = NULL;
   last = NULL;
   nnZeros = 0;
   sepFromPool = true;
   inPool = false;
   lb = -INFINITY_VALUE;
   ub = INFINITY_VALUE;
   obj = 0.0;
   type = C;
   params = NULL;
   nLPs = 0;

   maxk = c.maxk;
   if( VSIZE > 0 )
   {
      hasKCycles = new bool[VSIZE];
      memset( hasKCycles, 0, VSIZE * sizeof(bool) );
   }
   else
      hasKCycles = NULL;

   copy( c );
}

//---------------------------------------------------------------------------//

CColumn::~CColumn()
{
   CColElem *next, *iter; // auxiliary column elements

   // First element of the row
   iter = first;
   
   // Erase elements
   while( iter )
   {
      next = iter->next;
      delete iter;
      iter = next;

      nnZeros--;
   }

   first = NULL;
   last = NULL;
   sepFromPool = true;
   inPool = false;
   lb = -INFINITY_VALUE;
   ub = INFINITY_VALUE;
   obj = 0.0;
   type = C;
   params = NULL;

   if (nnZeros > 0)
     throw(MEMNOTDEALLOCEXCEPTION);

   if( hasKCycles )
      delete [] hasKCycles;
}

//---------------------------------------------------------------------------//
 
bool CColumn::contains( int idx )
{
   CColElem* iter; // auxiliary col element

   // First element of the col
   iter = first;

   // Search for idx
   while( iter )
   {
      if( iter->idx == idx )
         return true;

      iter = iter->next;
   }

   return false;
}

//---------------------------------------------------------------------------//

void CColumn::addElem( int idx, double coeff )
{
   CColElem* aux; // auxiliary element

   if ( nnZeros )
   {
      aux = new CColElem;

      if ( !aux )
         throw(LOWMEMORYEXCEPTION);

      aux->next = NULL;
      aux->idx = idx;
      aux->coeff = coeff;

      last->next = aux;
      last = aux;

      nnZeros++;
   }
   else
   {
      aux = new CColElem;

      if ( !aux )
         throw(LOWMEMORYEXCEPTION);

      aux->next = NULL;
      aux->idx = idx;
      aux->coeff = coeff;

      first = aux;
      last = aux;
      nnZeros++;
   }
}

//---------------------------------------------------------------------------//

void CColumn::addElemSum( int idx, double coeff )
{
   CColElem *aux; // auxiliary element
   bool flag; // flag that tells if idx was found

   if ( nnZeros ) // list with elements
   {
      aux = first;
      flag = false;

      while ( aux )
      {
         if ( aux->idx == idx ) // idx was found
         {
            aux->coeff+=coeff;
            flag = true;
         }

         aux = aux->next;
      }

      if ( !flag ) // idx was not found
      {
         aux = new CColElem;
         
         if ( !aux )
            throw(LOWMEMORYEXCEPTION);
         
         aux->next = NULL;
         aux->idx = idx;
         aux->coeff = coeff;
         
         last->next = aux;
         last = aux;
         
         nnZeros++;
      }
   }
   else // list empty
   {
      aux = new CColElem;

      if ( !aux )
         throw(LOWMEMORYEXCEPTION);

      aux->next = NULL;
      aux->idx = idx;
      aux->coeff = coeff;

      first = aux;
      last = aux;
      nnZeros++;
   }
}

//---------------------------------------------------------------------------//

void CColumn::createColumnFromArray(double *coeffs, int n)
{
   CColElem *aux; // auxiliary element
   register int i; // counter

   // Row is already created
   if ( first )
      return;

   // Add coeffs to the row
   for (i = 0; i < n ; i++)
   {
      if ( (!params && coeffs[i] > INTEPS) || ( params && coeffs[i] > params->getIntEps() ) )
      {
         if ( nnZeros )
         {
            aux = new CColElem;

            if ( !aux )
               throw(LOWMEMORYEXCEPTION);

            aux->next = NULL;
            aux->idx = i;
            aux->coeff = coeffs[i];

            last->next = aux;
            last = aux;

            nnZeros++;
         }
         else
         {
            aux = new CColElem;

            if ( !aux )
               throw(LOWMEMORYEXCEPTION);

            aux->next = NULL;
            aux->idx = i;
            aux->coeff = coeffs[i];

            first = aux;
            last = aux;
            nnZeros++;
         }
      }
   }
}

//---------------------------------------------------------------------------//

void CColumn::createColumnFromArray(double *coeffs, int *index, int n)
{
   CColElem *aux; // auxiliary element
   register int i; // counter

   // Row is already created
   if ( first )
      return;

   // Add coeffs to the row
   for (i = 0; i < n ; i++)
   {
      if ( (!params && coeffs[i] > INTEPS) || ( params && coeffs[i] > params->getIntEps() ) )
      {
         if ( nnZeros )
         {
            aux = new CColElem;

            if ( !aux )
               throw(LOWMEMORYEXCEPTION);

            aux->next = NULL;
            aux->idx = index[i];
            aux->coeff = coeffs[i];

            last->next = aux;
            last = aux;

            nnZeros++;
         }
         else
         {
            aux = new CColElem;

            if ( !aux )
               throw(LOWMEMORYEXCEPTION);

            aux->next = NULL;
            aux->idx = index[i];
            aux->coeff = coeffs[i];

            first = aux;
            last = aux;
            nnZeros++;
         }
      }
   }
}

//---------------------------------------------------------------------------//

void CColumn::copy( const CColumn &cc )
{
   CColElem *aux,*aux2; // auxiliary column elements
   int i;
   aux = cc.first;
   lb = cc.lb;
   ub = cc.ub;
   type = cc.type;
   obj = cc.obj;
   nLPs = cc.nLPs;
   sepFromPool = cc.sepFromPool;
   inPool = cc.inPool;
   params = cc.params;

   while ( aux )
   {
      aux2 = new CColElem;

      if ( !aux2 )
         throw(LOWMEMORYEXCEPTION);

      aux2->next = NULL;
      aux2->idx = aux->idx;
      aux2->coeff = aux->coeff;

      if ( !first )
      {
         first = aux2;
         last = aux2;
         nnZeros++;
      }
      else
      {
         last->next = aux2;
         last = aux2;
         nnZeros++;
      }

      aux = aux->next;
   }

   if( maxk != cc.maxk )
   {
      if( hasKCycles )
         delete [] hasKCycles;
      maxk = cc.maxk;
      if( VSIZE > 0 )
      {
         hasKCycles = new bool[ VSIZE ];
         memset( hasKCycles, 0, VSIZE * sizeof(bool) );
      }
      else
         hasKCycles = NULL;
   }
   else
   {
      if( ! hasKCycles )
      {
         if( VSIZE > 0 )
         {
            hasKCycles = new bool[ VSIZE ];
            memset( hasKCycles, 0, VSIZE * sizeof(bool) );
         }
      }
   }
   for( i = 0; i < VSIZE; i++ )
   {
      hasKCycles[i] = cc.hasKCycles[i];
   }
}

//---------------------------------------------------------------------------//

bool CColumn::isColumnEqual( const CColumn &cc )
{
   CColElem *auxR1,*auxR2; // auxiliary column elements
   double *r1, *r2; // auxiliary arrays to store the cols
   int maxIdx1, maxIdx2; // max idx of each col
   register int i; // counter
   bool flagEqual; // tells if the cols are equal

   // Tests if the RHS or the type are different
   if ( cc.type != type || cc.lb != lb ||
        cc.ub != ub || obj != cc.obj )
        return false;

   // Get the first element
   auxR1 = first;
   auxR2 = cc.first;

   // Tests if the first element is NULL
   if ( !auxR1 && !auxR2 )
      return true;
   else if ( !auxR1 || !auxR2 )
      return false;

   // Get the max idx of each row
   maxIdx1 = 0;
   while ( auxR1 )
   {
      if ( auxR1->idx > maxIdx1 )
         maxIdx1 = auxR1->idx;

      auxR1 = auxR1->next;
   }

   maxIdx2 = 0;
   while ( auxR2 )
   {
      if ( auxR2->idx > maxIdx2 )
         maxIdx2 = auxR2->idx;

      auxR2 = auxR2->next;
   }

   // The max idx is different
   if ( maxIdx1 != maxIdx2 )
      return false;

   // Allocate memory
   r1 = new double[maxIdx1+1];
   r2 = new double[maxIdx1+1];

   // Initialize the elements
   for (i = 0; i <= maxIdx1; i++)
   {
      r1[i]=0;
      r2[i]=0;
   }

   // Fill the arrays
   auxR1 = first;
   auxR2 = cc.first;

   while ( auxR1 )
   {
      r1[auxR1->idx]+=auxR1->coeff;
      auxR1 = auxR1->next;
   }

   while ( auxR2 )
   {
      r2[auxR2->idx]+=auxR2->coeff;
      auxR2 = auxR2->next;
   }

   // Tests if the cols are equal
   flagEqual = true;
   for (i = 0; i <= maxIdx1; i++)
   {
      if ( r1[i]!=r2[i] )
      {
         flagEqual=false;
         break;
      }
   }

   // Free memory
   if (r1)
      delete[] r1;

   if (r2)
      delete[] r2;

   return flagEqual;
}

//---------------------------------------------------------------------------//

double CColumn::getRedCost(double* lpdual)
{
   double redcost; // reduced cos of the column
   CColElem* auxelem; // auxiliary column element

   // Init variables
   redcost=0.0;
   auxelem = first;

   // Calculate the reduced cost of the column
   while( auxelem )
   {
      redcost += lpdual[ auxelem->idx ] * auxelem->coeff;
      auxelem = auxelem->next;
   }

   redcost -= obj;

   return redcost;
}

//---------------------------------------------------------------------------//
