/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CCutPool.cpp                     //
////////////////////////////////////////////
// Class that represents a pool of cuts.  //
////////////////////////////////////////////

#include "CCutPool.h"
#include <stdio.h>

//---------------------------------------------------------------------------//

CCutPoolElem::CCutPoolElem()
{
   next = NULL;
   prev = NULL;
   row = NULL;
}

//---------------------------------------------------------------------------//

CCutPoolElem::~CCutPoolElem()
{
   next = NULL;
   prev = NULL;

   if ( row )
      delete row;
}

//---------------------------------------------------------------------------//

CCutPool::CCutPool()
{ 
   first = NULL;
   last = NULL;
   nCuts = 0;
}

//---------------------------------------------------------------------------//

CCutPool::~CCutPool()
{
   empty();
}

//---------------------------------------------------------------------------//

bool CCutPool::isInPool(const CRow &aRow)
{
   CCutPoolElem *aux; // auxiliary element

   // Search for the row
   aux = first;
   while ( aux )
   {
      if ( aux->row )
      {
         if ( aux->row->isRowEqual(aRow) )
            return true;
      }

      aux = aux->next;
   }

   return false;
}

//---------------------------------------------------------------------------//

void CCutPool::empty()
{
   CCutPoolElem *aux,*aux2; // auxiliary element elements

   aux = first;

   while ( aux )
   {
      aux2 = aux;

      aux = aux->next;

      delete aux2;
      nCuts--;
   }

   first = NULL;
   last = NULL;

   if (nCuts > 0)
      throw( MEMNOTDEALLOCEXCEPTION );

}

//---------------------------------------------------------------------------//

int CCutPool::clean()
{
   int nRemoved; // number of rows removed from the pool
   CCutPoolElem *aux,*aux2; // auxiliary elements

   aux = first;
   nRemoved = 0;

   while( aux )
   {
      aux2 = aux->next;

      if ( aux->row )
      {
         if ( aux->row->getNLPs()==0 )
         { 
            // Remove the row
            if ( aux == first )
            {
               first = aux->next;

               if (first)
                  first->prev = NULL;
               else
                  last = NULL;

               delete aux;

               nCuts--;
            }
            else if ( aux == last )
            {
               last = last->prev;
               if ( last )
                  last->next = NULL;
               else
                  first = NULL;
               

               delete aux;

               nCuts--;
            }
            else
            {
               aux->next->prev = aux->prev;
               aux->prev->next = aux->next;

               delete aux;

               nCuts--;
            }


            nRemoved++;
         }
      }

      aux = aux2;
   }

   return nRemoved;
}

//---------------------------------------------------------------------------//

CRow* CCutPool::addRow( const CRow &aRow )
{
   CCutPoolElem *aux; // auxiliary element
   CRow *newRow; // new row create in the pool

   aux = first;

   while( aux )
   {
      if ( aux->row )
      {
         // Cut is already in the pool
         if (aux->row->isRowEqual(aRow))
         {
            aux->row->increaseNLPs();

            return (aux->row);
         }
      }

      aux = aux->next;
   }

   // Add new row to the pool
   newRow = NULL;
   newRow = new CRow( aRow );

   if ( !newRow )
      throw(LOWMEMORYEXCEPTION);

   newRow->increaseNLPs();
   newRow->setInPool(true);

   // Create new element in the pool
   aux = NULL;
   aux = new CCutPoolElem;

   if (!aux)
      throw(LOWMEMORYEXCEPTION);

   if ( first )
   {
     aux->row = newRow;
     aux->prev = last;
     aux->next = NULL;
     last->next = aux;

     last = aux;
     nCuts++;
   }
   else
   {
     aux->row = newRow;
     aux->prev = NULL;
     aux->next = NULL;

     first = aux;
     last = aux;

     nCuts++;
   }

   return newRow;
}

//---------------------------------------------------------------------------//






