/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CLPSolver.cpp                    //
////////////////////////////////////////////
// Class that represents the linear       //
// programming solver.                    //
////////////////////////////////////////////

#include "CLPSolver.h"
#include "Constants.h"

#ifdef WIN32
#include <windows.h>
#endif

//---------------------------------------------------------------------------//

CLPSolver::CLPSolver()
{
   initialized = false;
   env = NULL;
}

//---------------------------------------------------------------------------//

CLPSolver::~CLPSolver()
{
   if ( initialized )
   {
      if ( env )
      {
         CPXcloseCPLEX( & env );
      }
   }

   initialized = false;
}

//---------------------------------------------------------------------------//

bool CLPSolver::initLicense()
{
   int status; // status returned by cplex

   // Tests if it is already initiated
   if( initialized )
      return true;

   // Try to initialize the cplex environment
   env = CPXopenCPLEX( &status );
   
   if( !env )
   {
      throw(SOLVERNOTINITIALIZEDEXCEPTION);
   }

   initialized = true;

   return true;
}

//---------------------------------------------------------------------------//

void CLPSolver::solvePrimal(LPptr lp,double &obj,double *x,double *pi,int xsize,int pisize)
{
   if ( !env )
   {
      throw( SOLVERNOTINITIALIZEDEXCEPTION );
   }

   if ( !lp )
   {
      throw( LPVARIABLENOTINITIALIZED );
   }

   if( CPXprimopt(env,lp) == 0 )
   {
      if( CPXgetobjval( env, lp, &obj ) != 0 )
         throw(LPSOLVEROBJFUNCTIONEXCEPTION);

      if( CPXgetx( env, lp, x, 0, xsize-1 ) != 0 )
         throw(LPSOLVERSOLUTIONEXCEPTION);

      if( CPXgetpi( env, lp, pi, 0, pisize - 1 ) != 0 )
         throw(LPSOLVERDUALSOLEXCEPTION);
   }
   else
   {
      throw(LPHASNOSOLUTIONEXCEPTION);
   }
}

//---------------------------------------------------------------------------//

void CLPSolver::solveDual(LPptr lp,double &obj,double *x,double *pi,int xsize,int pisize)
{
   if ( !env )
   {
      throw( SOLVERNOTINITIALIZEDEXCEPTION );
   }

   if ( !lp )
   {
      throw( LPVARIABLENOTINITIALIZED );
   }

   if( CPXdualopt(env,lp) == 0 )
   {
      if( CPXgetobjval( env, lp, &obj ) != 0 )
         throw(LPSOLVEROBJFUNCTIONEXCEPTION);

      if( CPXgetx( env, lp, x, 0, xsize-1 ) != 0 )
         throw(LPSOLVERSOLUTIONEXCEPTION);

      if( CPXgetpi( env, lp, pi, 0, pisize - 1 ) != 0 )
         throw(LPSOLVERDUALSOLEXCEPTION);
   }
   else
   {
      throw(LPHASNOSOLUTIONEXCEPTION);
   }
}

//---------------------------------------------------------------------------//

bool CLPSolver::initLP(LPptr &lp)
{
   int status; // auxiliary status variable

   if ( lp )
      return false;
   else
   {
      lp = CPXcreateprob(env,&status,"");
      
      if ( lp )
         return true;
      else
         return false;
   }
}

//---------------------------------------------------------------------------//

void CLPSolver::freeLP(LPptr &lp)
{
   if ( lp )
   {
      CPXfreeprob(env,&lp);
   }
}

//---------------------------------------------------------------------------//

int CLPSolver::getNRows(LPptr lp)
{
   if ( !env )
   {
      throw( SOLVERNOTINITIALIZEDEXCEPTION );
   }

   if ( !lp )
   {
      throw( LPVARIABLENOTINITIALIZED );
   }

   return CPXgetnumrows(env,lp);
}

//---------------------------------------------------------------------------//

int CLPSolver::getNColumns(LPptr lp)
{
   if ( !env )
   {
      throw( SOLVERNOTINITIALIZEDEXCEPTION );
   }

   if ( !lp )
   {
      throw( LPVARIABLENOTINITIALIZED );
   }

   return CPXgetnumcols(env,lp);
}

//---------------------------------------------------------------------------//

void CLPSolver::getX(LPptr lp, double *x)
{
   int nC; // number of columns of the lp

   nC = getNColumns(lp);
 
   if( CPXgetx(env, lp, x, 0, nC - 1 ) != 0 )
      throw(LPSOLVERSOLUTIONEXCEPTION);
}

//---------------------------------------------------------------------------//

void CLPSolver::getPi(LPptr lp, double *y)
{
   int nR; // number of rows of the lp

   nR = getNRows(lp);
 
   if( CPXgetpi(env, lp, y, 0, nR - 1 ) != 0 )
      throw(LPSOLVERDUALSOLEXCEPTION);
}

//---------------------------------------------------------------------------//

void CLPSolver::getRHS(LPptr lp, double *rhs)
{
   int nR; // number of rows of the lp

   nR = getNRows(lp);
 
   if( CPXgetrhs(env, lp, rhs, 0, nR - 1 ) != 0 )
      throw(LPSOLVERRHSEXCEPTION);
}


//---------------------------------------------------------------------------//

void CLPSolver::getRHS( LPptr lp, double* rhs, int begin, int end )
{
   if( CPXgetrhs(env, lp, rhs, begin, end ) != 0 )
      throw(LPSOLVERRHSEXCEPTION);
}


//---------------------------------------------------------------------------//

void CLPSolver::setCoef(LPptr lp, int idxRow, int idxCol, double coeff)
{
   if (CPXchgcoef(env, lp, idxRow, idxCol, coeff )!=0)
      throw(LPSOLVERCHGCOEFFEXCEPTION);
}

//---------------------------------------------------------------------------//

void CLPSolver::sumCoef(LPptr lp, int idxRow, int idxCol, double coeff)
{
   double auxCoef; // auxiliary coefficient value

   if( CPXgetcoef(env,lp,idxRow,idxCol,&auxCoef)!=0 )
      throw(LPSOLVERGETCOEFFEXCEPTION);

   if( CPXchgcoef(env,lp,idxRow,idxCol,(auxCoef+coeff) )!=0)
      throw(LPSOLVERCHGCOEFFEXCEPTION);
}

//---------------------------------------------------------------------------//

void CLPSolver::addNewCols(LPptr lp, int nc, double *cJ, double *lb, double *ub, char *tp)
{
   if (CPXnewcols(env, lp, nc, cJ, lb, ub, tp, NULL ) != 0)
      throw(LPSOLVERNEWCOLSEXCEPTION);
}

//---------------------------------------------------------------------------//

void CLPSolver::addNewRows(LPptr lp, int nr, double *rhs, char *rt)
{
   if( CPXnewrows(env, lp, nr, rhs, rt, NULL, NULL) != 0 )
      throw(LPSOLVERNEWROWSEXCEPTION);
}

//---------------------------------------------------------------------------//

void CLPSolver::getBasis(LPptr lp, int *cols, int *rows)
{
   if( CPXgetbase(env, lp, cols, rows)!=0 )
      throw(GETBASISEXCEPTION);
}

//---------------------------------------------------------------------------//

void CLPSolver::copyBasis(LPptr lp, int *cols, int *rows)
{
   if( CPXcopybase(env, lp, cols, rows)!= 0)
      throw(COPYBASISEXCEPTION);
}

//---------------------------------------------------------------------------//

void CLPSolver::writeLPProblem(LPptr lp, char *fName)
{
   CPXwriteprob(env,lp, fName, "LP");
}

//---------------------------------------------------------------------------//

void CLPSolver::setObjSense(LPptr lp, int sense)
{
   CPXchgobjsen(env, lp, sense);
}

//---------------------------------------------------------------------------//

int CLPSolver::getNNZeros(LPptr lp)
{
      if ( !env )
   {
      throw( SOLVERNOTINITIALIZEDEXCEPTION );
   }

   if ( !lp )
   {
      throw( LPVARIABLENOTINITIALIZED );
   }

   return CPXgetnumnz(env,lp);
}

//---------------------------------------------------------------------------//

void CLPSolver::delRows(LPptr lp, int begin, int end)
{
   if ( CPXdelrows(env, lp, begin, end) != 0 )
      throw(SOLVERDELROWSEXCEPTION);
}

//---------------------------------------------------------------------------//

void CLPSolver::delCols(LPptr lp, int begin, int end)
{
   if ( CPXdelcols(env, lp, begin, end) != 0 )
      throw(SOLVERDELCOLSEXCEPTION);
}

//---------------------------------------------------------------------------//

void CLPSolver::delRowSet(LPptr lp, int *delstat)
{
   if ( CPXdelsetrows(env, lp, delstat)!=0 )
      throw(SOLVERDELROWSEXCEPTION);
}

//---------------------------------------------------------------------------//

void CLPSolver::delColSet(LPptr lp, int *delstat)
{
   if ( CPXdelsetcols(env, lp, delstat)!=0 )
      throw(SOLVERDELCOLSEXCEPTION);
}

//---------------------------------------------------------------------------//

void CLPSolver::setIntParameter(int param, int val)
{
   if ( CPXsetintparam(env, param, val) != 0)
      throw(SOLVERSETINTPARAMEXCEPTION);
}

//---------------------------------------------------------------------------//

void CLPSolver::solveMIP(LPptr lp)
{
   CPXmipopt( env, lp );
}

//---------------------------------------------------------------------------//

int CLPSolver::getStatus(LPptr lp)
{
   return CPXgetstat(env, lp );
}

//---------------------------------------------------------------------------//

void CLPSolver::getMIPSol(LPptr lp, double *xsol,int begin, int end)
{
   if ( CPXgetmipx(env, lp, xsol, begin, end ) != 0 )
      throw(SOLVERMIPXEXCEPTION);
}

//---------------------------------------------------------------------------//

void CLPSolver::getMIPObj(LPptr lp, double *obj)
{
   if (CPXgetmipobjval(env, lp, obj ) != 0)
      throw(SOLVERMIPOBJEXCEPTION);
}

//---------------------------------------------------------------------------//

int CLPSolver::getSimplexIter(LPptr lp)
{
  return CPXgetitcnt( env, lp );
}

//---------------------------------------------------------------------------//


// FUKA - DUALSTAB - BEGIN

void CLPSolver::setUB( LPptr lp, int idxCol, double ub )
{
   char lu = 'U';
   if( CPXchgbds(env, lp, 1, &idxCol, &lu, &ub ) )
      throw(LPSOLVERCHGUBEXCEPTION);
}

// FUKA - DUALSTAB - END

//---------------------------------------------------------------------------//

double CLPSolver::getCoef( LPptr lp, int idxRow, int idxCol )
{
   double auxd = 0.0;
   if( CPXgetcoef( env, lp, idxRow, idxCol, & auxd ) != 0 )
      throw( LPSOLVERGETCOEFEXCEPTION );
   return auxd;
}
