/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CMaster.cpp                      //
////////////////////////////////////////////
// Class that represents the control of   //
// the branch-and-cut-and-price           //
////////////////////////////////////////////

#include <stdlib.h>
#include <math.h>
#include "CMaster.h"

//---------------------------------------------------------------------------//

CMaster::CMaster()
{
   params = NULL;
   totalNodes = 0;
   maxDepth = 0;
   elapsedBB = 0.0;
   numIntNodes = 0;
   numIntZeroNodes = 0;

   // Try to allocate the parameters
   params = new CParameters;

   if ( !params )
      throw(LOWMEMORYEXCEPTION);

   // Try to allocate the pools
   cutPool = new CCutPool;

   if ( !cutPool )
      throw(LOWMEMORYEXCEPTION);

   colPool = new CColPool;

   if ( !colPool )
      throw(LOWMEMORYEXCEPTION);

   // Try to allocate the solver
   solver = new CLPSolver;

   if ( !solver )
      throw(LOWMEMORYEXCEPTION);

   if (!solver->initLicense())
      throw(SOLVERNOTINITIALIZEDEXCEPTION);

   bestLB = - params->getInfinity();
   bestUB = + params->getInfinity();

   solSize = 0;
   maxSolSize = 0;
   logFile = NULL;

   strcpy( endlogfname, "");


}

//---------------------------------------------------------------------------//

CMaster::~CMaster()
{
   // Close log file
   if ( logFile )
      fclose( logFile );


   if( strlen( endlogfname ) > 0 )
   {
      rename( initiallogfname, endlogfname );     
   }

   openNodes.empty();

   // Delete the pools
   if ( cutPool )
   {
      delete cutPool;
      cutPool = NULL;
   }

   if ( colPool )
   {
      delete colPool;
      colPool = NULL;
   }

   // Delete the bestsol vector
   if( bestSol != NULL )
   {
      delete [] bestSol;
      bestSol = NULL;
   }
   
   if ( solver )
   {
      delete solver;
      solver = NULL;
   }
   
   solSize = 0;
   maxSolSize = 0;
 
   bestLB = - params->getInfinity();
   bestUB = + params->getInfinity();

   if ( params )
   {
      delete params;
      params = NULL;
   }
   
   totalNodes = 0;
   maxDepth = 0;
   elapsedBB = 0.0;
   numIntNodes = 0;
   numIntZeroNodes = 0;
}

//---------------------------------------------------------------------------//

void CMaster::setLogFile(char* logfilename)
{
   // Open the log file
   strcpy( initiallogfname, logfilename );
   logFile = fopen(logfilename,"wt");

   // Throw exception if file couldn't be open
   if ( !logFile )
      throw(FILENOTOPENEXCEPTION);
}

//---------------------------------------------------------------------------//

CColumn* CMaster::addColToPool(const CColumn &c)
{
   return colPool->addColumn(c);
}

//---------------------------------------------------------------------------//

CRow* CMaster::addRowToPool(const CRow &r)
{
   return cutPool->addRow(r);
}

//---------------------------------------------------------------------------//

void CMaster::updateBounds()
{
   CBBNodeListElem* elem; // auxiliary element of the list
   CBBNode* auxnode; // auxiliary node

   double minLB = params->getInfinity();
   
   // Update the best LB and UB
   elem = openNodes.first;

   while( elem )
   {
      auxnode = elem->node;

      if( auxnode->getUB() < bestUB )
      {
         bestUB = auxnode->getUB();
      }

      if( auxnode->getLB() < minLB )
      {
         minLB = auxnode->getLB();
      }
      elem = elem->next;
   }

   bestLB = minLB;
}

//---------------------------------------------------------------------------//

void CMaster::addOpenNode( CBBNode* newnode )
{
   totalNodes++;

   openNodes.addLast( newnode );
}

//---------------------------------------------------------------------------//

void CMaster::addFeasibleSol( double* xsol, double solval )
{
   register int j; // counter

   if( solval < bestUB )
   {
      for(j = 0; j < solSize; j++)
      {
         bestSol[j] = xsol[j];
      }
      bestUB = solval;
   }
}

//---------------------------------------------------------------------------//

CBBNode* CMaster::selectBFSNode()
{
   CBBNode* auxnode; // auxiliary node
   CBBNode* bestnode; // best node to select 
   int mindpt; // minimum depth
   CBBNodeListElem* auxelem; // auxliary list element

   // Initialization
   auxnode = NULL;
   bestnode = NULL;
   mindpt = -1;

   auxelem = openNodes.first;

   while( auxelem )
   {
      auxnode = (CBBNode*) auxelem->node;
      
      if( auxnode->getDepth() < mindpt )
      {
         mindpt = auxnode->getDepth();
         bestnode = auxnode;
      }
      else if( mindpt == -1 )
      {
         mindpt = auxnode->getDepth();
         bestnode = auxnode;
      }
      auxelem = auxelem->next;
   }
   return bestnode;
}

//---------------------------------------------------------------------------//

CBBNode* CMaster::selectDFSNode()
{
   CBBNode* auxnode; // auxiliary node
   CBBNode* bestnode; // best node to select 
   int maxdpt; // maximum depth
   CBBNodeListElem* auxelem; // auxliary list element

   // Initialization
   auxnode = NULL;
   bestnode = NULL;
   maxdpt = -1;
   auxelem = openNodes.first;

   while( auxelem )
   {
      auxnode = (CBBNode*) auxelem->node;
      
      if( auxnode->getDepth() > maxdpt )
      {
         maxdpt = auxnode->getDepth();
         bestnode = auxnode;
      }
      auxelem = auxelem->next;
   }
   return bestnode;
}

//---------------------------------------------------------------------------//

CBBNode* CMaster::selectBestBndNode()
{
   CBBNode* auxnode; // auxiliary node
   CBBNode* bestnode; // best node to select 
   double bestbnd; // best bound found
   CBBNodeListElem* auxelem; // auxliary list element

   // Initialization
   auxnode = NULL;
   bestnode = NULL;
   auxelem = openNodes.first;
   bestbnd = params->getInfinity();

   if( auxelem )
   {
      bestnode = (CBBNode*) auxelem->node;
      bestbnd = bestnode->getLB();
   }

   while( auxelem )
   {
      auxnode = (CBBNode*) auxelem->node;
      
      if( auxnode->getLB() < bestbnd )
      {
         bestbnd = auxnode->getLB();
         bestnode = auxnode;
      }
      auxelem = auxelem->next;
   }
   return bestnode;
}

//---------------------------------------------------------------------------//

double CMaster::getSkew()
{
   return( numIntNodes > 0 ? (double) numIntZeroNodes / (double) numIntNodes : 0.5 );
}

//---------------------------------------------------------------------------//

void CMaster::computeNode( CBBNode* node, NodeStatus stat )
{
   if( stat != BRANCH )
   {
      elapsedBB += pow( 2.0, - node->getDepth() );
   }
   else
   {
      numIntNodes++;
      if( node->getOrigin() == 0 )
         numIntZeroNodes++;
   }
   if( maxDepth < node->getDepth() )
      maxDepth = node->getDepth();
}

//---------------------------------------------------------------------------//

int CMaster::cleanCutPool()
{
   if( cutPool->getSize() > params->getMaxCutPoolSize() )
   {
      return cutPool->clean();
   }
   else
   {
      return 0;
   }
}

//---------------------------------------------------------------------------//

int CMaster::cleanColPool()
{
   if( colPool->getSize() > params->getMaxColPoolSize() )
   {
      return colPool->clean();
   }
   else
   {
      return 0;
   }
}

//---------------------------------------------------------------------------//

bool CMaster::checkOptimal()
{
   
   if( params->isObjInteger() )
   {
      return( ceil( bestLB - params->getObjEps() ) >= bestUB - params->getObjEps() );
   }
   else if( params->getFracPrec() > 0 )
   {
      double precision = pow( 10.0, params->getFracPrec() );
      return( ceil( precision * bestLB  - params->getObjEps() ) / precision >=
         bestUB - params->getObjEps() );
   }
   else
   {
      return( bestLB >= bestUB - params->getObjEps() );
   }
}

//---------------------------------------------------------------------------//

void CMaster::setSolSize( int s )
{
   solSize = s;
   if( bestSol )
   {
      if( maxSolSize < s )
      {
         delete [] bestSol;
         bestSol = NULL;
	 maxSolSize = s + SIZEINCREASE;
	 bestSol = new double[ maxSolSize ];
      }
   }
   else
   {
      maxSolSize = s + SIZEINCREASE;
      bestSol = new double[ maxSolSize ];
   }
}

//---------------------------------------------------------------------------//

void CMaster::writeLog( bool stdOutput,  char* format, ... )
{
  va_list ptr;

  va_start( ptr, format );

  if (stdOutput)
    vfprintf( stdout, format, ptr );
  
  if (logFile)
  {
    vfprintf( logFile, format, ptr );
    fflush( logFile );
  }
  else
    throw(FILENOTOPENEXCEPTION);

  va_end( ptr );
}

//---------------------------------------------------------------------------//

int CMaster::emptyCutPool()
{
   int nDel; // auxiliary size of the pool

   nDel = cutPool->getSize();
   cutPool->empty();

   return nDel;
}

//---------------------------------------------------------------------------//

int CMaster::emptyColPool()
{
   int nDel; // auxiliary size of the pool

   nDel = colPool->getSize();
   colPool->empty();

   return nDel;
}

//---------------------------------------------------------------------------//





