/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/


#include "CRow.h"
#include <math.h>
#include <stdio.h> 
#include <stdlib.h>

//---------------------------------------------------------------------------//

CRow::CRow()
{
   // Init members
   first = NULL;
   last = NULL;
   nnZeros = 0;
   sepFromPool = true;
   inPool = false;
   rhs = 0.0;
   type = L;
   params = NULL;
   nLPs = 0;
}

//---------------------------------------------------------------------------//

CRow::CRow(CParameters *p)
{
   // Init members
   first = NULL;
   last = NULL;
   nnZeros = 0;
   sepFromPool = true;
   inPool = false;
   rhs = 0.0;
   type = L;
   params = p;
   nLPs = 0;
}

//---------------------------------------------------------------------------//

CRow::CRow( const CRow &r )
{
   // Init members
   first = NULL;
   last = NULL;
   nnZeros = 0;
   sepFromPool = true;
   inPool = false;
   rhs = 0.0;
   type = L;
   params = NULL;
   nLPs = 0;

   copy(r);
}

//---------------------------------------------------------------------------//

CRow::~CRow()
{
   CRowElem *next, *iter; // auxiliary elements

   // First element of the row
   iter = first;
   
   // Erase elements
   while( iter )
   {
      next = iter->next;
      delete iter;
      iter = next;

      nnZeros--;
   }

   first = NULL;
   last = NULL;
   sepFromPool = true;
   inPool = false;
   rhs = 0.0;
   type = L;
   params = NULL;

   if (nnZeros > 0)
     throw(MEMNOTDEALLOCEXCEPTION);
}

//---------------------------------------------------------------------------//
 
bool CRow::contains( int idx )
{
   CRowElem* iter; // auxiliary element

   // First element of the row
   iter = first;

   // Search for idx
   while( iter )
   {
      if( iter->idx == idx )
         return true;

      iter = iter->next;
   }

   return false;
}

//---------------------------------------------------------------------------//

void CRow::addElem( int idx, double coeff )
{
   CRowElem* aux; // auxiliary element

   if ( nnZeros )
   {
      aux = new CRowElem;

      if ( !aux )
         throw(LOWMEMORYEXCEPTION);

      aux->next = NULL;
      aux->idx = idx;
      aux->coeff = coeff;

      last->next = aux;
      last = aux;

      nnZeros++;
   }
   else
   {
      aux = new CRowElem;

      if ( !aux )
         throw(LOWMEMORYEXCEPTION);

      aux->next = NULL;
      aux->idx = idx;
      aux->coeff = coeff;

      first = aux;
      last = aux;
      nnZeros++;
   }
}

//---------------------------------------------------------------------------//

void CRow::addElemSum( int idx, double coeff )
{
   CRowElem *aux; // auxiliary element
   bool flag; // flag that tells if idx was found

   if ( nnZeros ) // list with elements
   {
      aux = first;
      flag = false;

      while ( aux )
      {
         if ( aux->idx == idx ) // idx was found
         {
            aux->coeff+=coeff;
            flag = true;
         }

         aux = aux->next;
      }

      if ( !flag ) // idx was not found
      {
         aux = new CRowElem;
         
         if ( !aux )
            throw(LOWMEMORYEXCEPTION);
         
         aux->next = NULL;
         aux->idx = idx;
         aux->coeff = coeff;
         
         last->next = aux;
         last = aux;
         
         nnZeros++;
      }
   }
   else // list empty
   {
      aux = new CRowElem;

      if ( !aux )
         throw(LOWMEMORYEXCEPTION);

      aux->next = NULL;
      aux->idx = idx;
      aux->coeff = coeff;

      first = aux;
      last = aux;
      nnZeros++;
   }
}

//---------------------------------------------------------------------------//

void CRow::createRowFromArray(double *coeffs, int n)
{
   CRowElem *aux; // auxiliary element
   register int i; // counter

   // Row is already created
   if ( first )
      return;

   // Add coeffs to the row
   for (i = 0; i < n ; i++)
   {
      if ( (!params && coeffs[i] > INTEPS) || ( params && coeffs[i] > params->getIntEps() ) )
      {
         if ( nnZeros )
         {
            aux = new CRowElem;

            if ( !aux )
               throw(LOWMEMORYEXCEPTION);

            aux->next = NULL;
            aux->idx = i;
            aux->coeff = coeffs[i];

            last->next = aux;
            last = aux;

            nnZeros++;
         }
         else
         {
            aux = new CRowElem;

            if ( !aux )
               throw(LOWMEMORYEXCEPTION);

            aux->next = NULL;
            aux->idx = i;
            aux->coeff = coeffs[i];

            first = aux;
            last = aux;
            nnZeros++;
         }
      }
   }
}

//---------------------------------------------------------------------------//

void CRow::createRowFromArray(double *coeffs, int *index, int n)
{
   CRowElem *aux; // auxiliary element
   register int i; // counter

   // Row is already created
   if ( first )
      return;

   // Add coeffs to the row
   for (i = 0; i < n ; i++)
   {
      if ( (!params && coeffs[i] > INTEPS) || ( params && coeffs[i] > params->getIntEps() ) )
      {
         if ( nnZeros )
         {
            aux = new CRowElem;

            if ( !aux )
               throw(LOWMEMORYEXCEPTION);

            aux->next = NULL;
            aux->idx = index[i];
            aux->coeff = coeffs[i];

            last->next = aux;
            last = aux;

            nnZeros++;
         }
         else
         {
            aux = new CRowElem;

            if ( !aux )
               throw(LOWMEMORYEXCEPTION);

            aux->next = NULL;
            aux->idx = index[i];
            aux->coeff = coeffs[i];

            first = aux;
            last = aux;
            nnZeros++;
         }
      }
   }
}

//---------------------------------------------------------------------------//

bool CRow::isRowEqual( const CRow &r )
{
   CRowElem *auxR1,*auxR2; // auxiliary row elements
   double *r1, *r2; // auxiliary arrays to store the rows
   int maxIdx1, maxIdx2; // max idx of each row
   register int i; // counter
   bool flagEqual; // tells if the rows are equal

   // Tests if the RHS or the type are different
   if ( r.rhs != rhs || r.type != type )
      return false;

   // Get the first element
   auxR1 = first;
   auxR2 = r.first;

   // Tests if the first element is NULL
   if ( !auxR1 && !auxR2 )
      return true;
   else if ( !auxR1 || !auxR2 )
      return false;

   // Get the max idx of each row
   maxIdx1 = 0;
   while ( auxR1 )
   {
      if ( auxR1->idx > maxIdx1 )
         maxIdx1 = auxR1->idx;

      auxR1 = auxR1->next;
   }

   maxIdx2 = 0;
   while ( auxR2 )
   {
      if ( auxR2->idx > maxIdx2 )
         maxIdx2 = auxR2->idx;

      auxR2 = auxR2->next;
   }

   // The max idx is different
   if ( maxIdx1 != maxIdx2 )
      return false;

   // Allocate memory
   r1 = new double[maxIdx1+1];
   r2 = new double[maxIdx1+1];

   // Initialize the elements
   for (i = 0; i <= maxIdx1; i++)
   {
      r1[i]=0;
      r2[i]=0;
   }

   // Fill the arrays
   auxR1 = first;
   auxR2 = r.first;

   while ( auxR1 )
   {
      r1[auxR1->idx]+=auxR1->coeff;
      auxR1 = auxR1->next;
   }
 
   while ( auxR2 )
   {
      r2[auxR2->idx]+=auxR2->coeff;
      auxR2 = auxR2->next;
   }

   // Tests if the rows are equal
   flagEqual = true;
   for (i = 0; i <= maxIdx1; i++)
   {
      if ( r1[i]!=r2[i] )
      {
         flagEqual=false;
         break;
      }
   }

   // Free memory
   if (r2)
      delete[] r2;

   if (r1)
      delete[] r1;

   return flagEqual;
}

//---------------------------------------------------------------------------//
   
void CRow::copy( const CRow &r )
{
   CRowElem *aux,*aux2; // auxiliary row element

   aux = r.first;

   rhs = r.rhs;
   type = r.type;
   nLPs = r.nLPs;
   sepFromPool = r.sepFromPool;
   inPool = r.inPool;
   params = r.params;

   while ( aux )
   {
      aux2 = new CRowElem;

      if ( !aux2 )
         throw(LOWMEMORYEXCEPTION);

      aux2->next = NULL;
      aux2->idx = aux->idx;
      aux2->coeff = aux->coeff;

      if ( !first )
      {
         first = aux2;
         last = aux2;
         nnZeros++;
      }
      else
      {
         last->next = aux2;
         last = aux2;
         nnZeros++;
      }

      aux = aux->next;
   }
}

//---------------------------------------------------------------------------//

bool CRow::isViolated( double* xsol )
{
   double lhs; // LHS value
   CRowElem *auxelem; // auxiliary row element

   // Initialization
   lhs = 0.0;
   auxelem = first;

   // Calculate the LHS
   while( auxelem )
   {
      lhs += xsol[ auxelem->idx ] * auxelem->coeff;  
      auxelem = auxelem->next;
   }

   // Verify if solution violates the row
   if( type == L )
   {
      if ( !params )
         return lhs > rhs + INTEPS;
      else
         return lhs > rhs + params->getIntEps();
   }
   else if( type == G )
   {
      if ( !params )
         return lhs < rhs - INTEPS;
      else
         return lhs < rhs - params->getIntEps();
   }
   else if( type == E )
   {
      if ( !params )
         return( fabs( lhs - rhs ) > INTEPS );
      else
         return( fabs( lhs - rhs ) > params->getIntEps() );

   }

   // Invalid type of row
   return false;
}

//---------------------------------------------------------------------------//

bool CRow::isViolated(double *xsol, double &value)
{
   double lhs; // LHS value
   CRowElem *auxelem; // auxiliary row element

   // Initialization
   lhs = 0.0;
   auxelem = first;
   value = 0;

   // Calculate the LHS
   while( auxelem )
   {
      lhs += xsol[ auxelem->idx ] * auxelem->coeff;  
      auxelem = auxelem->next;
   }

   // Verify if solution violates the row
   if( type == L )
   {
      value = lhs - rhs; 
      if ( !params )
         return lhs > rhs + INTEPS;
      else
         return lhs > rhs + params->getIntEps();

   }
   else if( type == G )
   {
      value = rhs - lhs;
      if ( !params )
         return lhs < rhs - INTEPS;
      else
         return lhs < rhs - params->getIntEps();

   }
   else if( type == E )
   {
      value = fabs( lhs - rhs );
      if ( !params )
         return( fabs( lhs - rhs ) > INTEPS );
      else
         return( fabs( lhs - rhs ) > params->getIntEps() );

   }

   // Invalid type of row
   return false;
}

//---------------------------------------------------------------------------//
