/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CVRPBBNode.cpp                   //
////////////////////////////////////////////
// Class that represents a node of the    //
// branch tree for the CVRP               //
////////////////////////////////////////////

//---------------------------------------------------------------------------//

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "CPUTimer.h"
#include "CLPSolver.h"
#include "CVRPBBNode.h"
#include "Constants.h"
#include "cnstrmgr.h"
#include "basegrph.h"
#include "capsep.h"
#include "mstarsep.h"
#include "fcisep.h"
#include "combsep.h"
#include "htoursep.h"
#include "glmsep.h"
#include "brnching.h"
#include "qr_oracle_lp.h"
#include "qrsolver.h"

//---------------------------------------------------------------------------//




// Number of iterations before display information about column generation
#define DISPLAYITERS 20

// Max number of columns and rows of the LP before eliminate rows
#define MAXLPNUMROWS 1500

#define MINLPNUMROWS 500

#define LPNUMDELROWS 500

//#define MAXLPNUMCOLS 40000
#define MAXLPNUMCOLS 20000

//#define MINLPNUMCOLS 20000
#define MINLPNUMCOLS 15000

#define LPNUMDELCOLS 5000

// Value used to select the branch
//#define BOUNDARYTARGET 2.85

// Size of the cycles to be avoided in the dynamic programming
//#define KSIZE 2

// Depth until which to do strong branching with colgen
#define SBCOLGENDEPTH 5000000

//#define NUMSBCOLGENITERS 5

/**********************************************************************
To always do SB:
- set MAXFULLSBDEPTH to a very large number
To never do SB
- set MAXFULLSBDEPTH to -1 and MAXBRANCHSETSROOT to 0
********************************************************************/

// Depth until which to do strong branching on all sets
#define MAXFULLSBDEPTH -3 

// Define maximum number of strong branching tests to do in the root node
//#define MAXBRANCHSETSROOT 10

//#define MINSETSTESTED 5

// Define the minimum obj function improvement that needs to be made
// in order to have colgen done
#define COLGENTHRESHOLD 1.003

// Define the minimum obj function improvement that needs to be made in
// order to increase k
#define KTHRESHOLD 1.001

#define TIMEBLOWTHRESHOLD 10.0

//---------------------------------------------------------------------------//

// Auxiliary structure for sort rows and columns
typedef struct CRowSortStr
{
   CRow *elem;
   double value;
} CRowSort;

typedef struct CColSortStr
{
   CColumn *elem;
   double value;
} CColSort;



typedef struct RowColElimStr
{
   int idx;
   double value;
} RowColElim;


typedef struct SortEdgesStr
{
   int n;
   double w;
   int f;
   int fixedNow;
} SortEdges;


// Auxiliary functions for qsort
int compRows (const void *a, const void *b)
{
   return (int)(-((*((CRowSort*)a)).value-((*((CRowSort*)b)).value))*1000);
}

int compCols (const void *a, const void *b)
{
   return (int)(-((*((CColSort*)a)).value-((*((CColSort*)b)).value))*1000);
}

int compEdges (const void *a, const void *b)
{
   return (int)(-( (*((SortEdges*)a)).w - (*((SortEdges*)b)).w)*1000);
}

int compRowColElim (const void *a, const void *b)
{
   return (int)(-((*((RowColElim*)a)).value-((*((RowColElim*)b)).value))*1000);
}



//---------------------------------------------------------------------------//
CVRPBBNode::CVRPBBNode(CVRPMaster* m, double nid): CBBNode(m, nid)
{
   firststab = true;
   nStab1 = 0;
   maxStab1 = 2000;
   nStab2 = 0;
   maxStab2 = 2000;
   dualestimate = new double[ maxStab1 ];   
   
   // FUKA2 - BEGIN
   initialized = false;
   cgtimelimit = false;
   cgiterlimit = false;
   cgbndlimit = false;
   inevalk = false;
   inSB = false;

   bestpath = NULL;
   edgeweights = NULL;
      // FUKA2 - END
   
   nArtificials=0;
   maxArtificials=2000;
   localg = NULL;
   lpSolExp = NULL;
   lpDualExp = NULL;
   lpDualCarp = NULL;
   nFixedSets = 0;
   fSets = NULL;
   
   
   vMatrix = NULL;
   predMatrix = NULL;
   vMatrix2 = NULL;
   predMatrix2 = NULL;
   
   //   dpCarp = NULL;
   //   dpDist = NULL;
   
   edgeHead = NULL;
   edgeTail = NULL;
   edgeX = NULL;
   listS = NULL;
   nodeP = NULL;
   pathDinProg = NULL;
   rowHash = NULL;
   colHash = NULL;
   
   myCutsCMP = NULL;
   myOldCutsCMP = NULL;
   setsCMP = NULL;
   
   firstcolgen = true;


   nFixedEdges = 0;
   fixedEdges = NULL;

   branchorg = -1;

   SBtries = 0;

   nMasterRows = 0;
}

//---------------------------------------------------------------------------//

CVRPBBNode::CVRPBBNode(CVRPBBNode* n, double nid, int p_or) : CBBNode(n,nid,p_or)
{
   firststab = false;
   nStab1 = 0;
   maxStab1 = 2000;
   nStab2 = 0;
   maxStab2 = 2000;
   dualestimate = new double[ maxStab1 ];
   memcpy( dualestimate, n->dualestimate, maxStab1 * sizeof(double) );

   nFixedEdges = 0;
   fixedEdges = NULL;
   
   // FUKA2 - BEGIN
   initialized = false;
   cgtimelimit = false;
   cgiterlimit = false;
   cgbndlimit = false;
   inevalk = false;
   inSB = false;

   bestpath = NULL;
   edgeweights = NULL;
   // FUKA2 - END
   
   
   nArtificials=0;
   maxArtificials=2000;
   localg = NULL;
   lpSolExp = NULL;
   lpDualExp = NULL;
   lpDualCarp = NULL;
   nFixedSets = 0;
   fSets = NULL;
   
   vMatrix = NULL;
   predMatrix = NULL;
   vMatrix2 = NULL;
   predMatrix2 = NULL;
   //   dpCarp = NULL;
   //   dpDist = NULL;
   
   edgeHead = NULL;
   edgeTail = NULL;
   edgeX = NULL;
   listS = NULL;
   nodeP = NULL;
   pathDinProg = NULL;
   rowHash = NULL;
   colHash = NULL;
   
   myCutsCMP = NULL;
   myOldCutsCMP = NULL;
   setsCMP = NULL;
   
   firstcolgen = true;

   branchorg = -1;

   SBtries = 0;
   nMasterRows = 0;
}

//---------------------------------------------------------------------------//

CVRPBBNode::~CVRPBBNode()
{
   register int i; // counter
   
   if ( lpSolExp )
      delete[] lpSolExp;
   
   if ( lpDualExp )
      delete[] lpDualExp;
   
   if ( lpDualCarp )
      delete[] lpDualCarp;

   if ( fSets )
   {
      for (i=0;i<nFixedSets;i++)
      {
         if ( fSets[i] )
            fSets[i]->decreaseNLPs();
         
         fSets[i] = NULL;
      }
      delete[] fSets;
   }

   nArtificials=0;
   maxArtificials=2000;
   localg = NULL;
   lpSolExp = NULL;
   lpDualExp = NULL;
   nFixedSets = 0;
   fSets = NULL;
   
   vMatrix = NULL;
   predMatrix = NULL;
   vMatrix2 = NULL;
   predMatrix2 = NULL;
   //   dpCarp = NULL;
   //   dpDist = NULL;
   
   edgeHead = NULL;
   edgeTail = NULL;
   edgeX = NULL;
   listS = NULL;
   pathDinProg = NULL;
   nodeP = NULL;
   
   if (rowHash)
   {
      delete rowHash;
      rowHash = NULL;
   }
   
   if (colHash)
   {
      delete colHash;
      colHash = NULL;
   }
   
   rowHash = NULL;
   colHash = NULL;
   
   // Deallocate structures of CVRPSEP
   if ( myCutsCMP )
   {
      CMGR_FreeMemCMgr(&myCutsCMP);
      myCutsCMP = NULL;
   }
   
   if ( myOldCutsCMP )
   {
      CMGR_FreeMemCMgr(&myOldCutsCMP);
      myOldCutsCMP = NULL;
   }
   
   if ( setsCMP )
   {
      CMGR_FreeMemCMgr(&setsCMP);
      setsCMP = NULL;
   }
   
   if( dualestimate )
   {
      delete [] dualestimate;
      dualestimate = NULL;
   }

   if( fixedEdges )
   {
      delete [] fixedEdges;
      fixedEdges = NULL;
   }

   if( bestpath )
   {
      delete [] bestpath;
      bestpath = NULL;
   }


   if( edgeweights ) 
      delete [] edgeweights;
   
}

//---------------------------------------------------------------------------//

NodeStatus CVRPBBNode::optimize()
{
   int addcols, addrows; // number of cuts and columns added
   int ncolpool; // number of columns added from the pool
   double auxsum; // auxiliary variable
   bool go; // tells if the optimization must continue
   NodeStatus auxReturn; // node status return
   int nElimRows; // number of rows eliminated from LP
   int nElimCols; // number of columns eliminated from LP
   int nCols, nRows; // number of columns and rows of the LP
   
   
   int k;
   
   int bestk = 0;
   int masternCols, masternRows;
   CPUTimer kevaltimer;
   CColumn** kcols;
   CRow** krows;
   int* nkcols;
   int* nkrows;
   int iter;
   double kfactor;
   bool timeBlow = false;

   if( ! ((CVRPMaster*) master)->getDoStab() )
      firststab = false;
   
   // Print information
   master->writeLog( true, "Solving node %g\n\n", id );
   
   // Starts the time counter
   nodetimer.start();
   
   if( ! initialized )
   {
      // Initialize node properties
      initializeNode();
      
      // Create the master problem for the CVRP
      createMaster();
   }

   int j;

   // Resets the fixed edges for this node
   for( j = 0; j < localg->getM(); j++ )
   {
      if( localg->getEdgeFixed(j) == 3 ) 
      {
         localg->setEdgeFixed(j, -1 );
      }
   }
   for( j = 0; j < nFixedEdges; j++ )
   {
      localg->setEdgeFixed( fixedEdges[j], 3 );
   }
   int ncheck = 0;
   for( j = 0; j < localg->getM(); j++ )
   {
      if( localg->getEdgeFixed(j) == 3 ) 
      {
         ncheck++;
      }

   }
   if( ncheck != nFixedEdges )
   {
      fprintf(stderr,"ERROR !!!!!!!!!!!!! EDGES FIXED DON'T MATCH\n");
      exit(1);
   }
   
   solvePrimal(); 
   masternCols = master->solver->getNColumns(locallp);
   masternRows = master->solver->getNRows(locallp); 
   
   if (getDepth() == 0)
   {
      if ((!((CVRPMaster *)master)->getCarp()) ||
         (((CVRPMaster *)master)->getPearn()) ||
         (((CVRPMaster *)master)->getUchoa()))
      {
         
#define NEWCODE
#ifdef NEWCODE
         
         if (((CVRPMaster*) master)->getDynk())
         {
            /* Sets parameter to stop column generation by time */
            bestktime = 100000000;

            kcols  = new CColumn* [((CVRPMaster*) master)->getKCycleParam() + 1];
            krows  = new CRow* [((CVRPMaster*) master)->getKCycleParam() + 1];
            nkcols = new int [((CVRPMaster*) master)->getKCycleParam() + 1];
            nkrows = new int [((CVRPMaster*) master)->getKCycleParam() + 1];
            
            memset( nkcols, 0, sizeof(int) * (((CVRPMaster*) master)->getKCycleParam() + 1) );
            memset( nkrows, 0, sizeof(int) * (((CVRPMaster*) master)->getKCycleParam() + 1) );
            memset( krows, 0, sizeof(CRow*) * (((CVRPMaster*) master)->getKCycleParam() + 1) );
            memset( kcols, 0, sizeof(CColumn*) * (((CVRPMaster*) master)->getKCycleParam() + 1) );
            
            for (k = 1; k <= ((CVRPMaster*) master)->getKCycleParam(); k++)
            {
               go = true;
               
               kevaltimer.reset();
               kevaltimer.start();
               while(go)
               {
                  addcols = 0;
                  if( k >= 2 )
                  {
                     // Generate columns
                     cronoColumnGeneration.start();
                     
                     cgtimelimit = true;
                     inevalk = true;
                     addcols = columnGen(k);
                     inevalk = false;
                     cgtimelimit = false;
                     
                     setSolSize(master->solver->getNColumns(locallp));
                     solvePrimal();
                     
                     cronoColumnGeneration.stop();
                     
                     nCols = master->solver->getNColumns(locallp);
                  } 
                  // Then do only cut generation
                  else
                  {
                     if( firstTime )
                     {
                        cronoSolveLP.start();
                        
                        solvePrimal();
                        
                        cronoSolveLP.stop();
                        
                        expandPi(lpDualExp,lpDual);
                        generateColumnsOneEdge();
                        
                        addcols = localg->getM();
                        firstTime = false;
                     }
                  }
                  
                  // Solves by the primal simplex
                  cronoSolveLP.start();
                  solvePrimal();
                  cronoSolveLP.stop();
                  
                  // Generate cuts
                  cronoCutGeneration.start();
                  addrows = cutGen();
                  cronoCutGeneration.stop();
                  nRows = master->solver->getNRows(locallp);
                  
                  // Verify if columns or cuts were added to the problem
                  if( addrows == 0 ) 
                  {
                     go = false;
                  }
                  else
                  {
                     go = true;
                  }
                  
                  kevaltimer.stop();
                  if( kevaltimer.getCPUTotalSecs() > max( TIMEBLOWTHRESHOLD * bestktime, 300.0 ) )
                  {
                     
                     timeBlow = true;
                     go = false;
                  }
                  kevaltimer.start();
                  
               } // end while (go)
               /* resets dual stabilization */
               if( ((CVRPMaster*) master)->getDoStab() )
                  firststab = true;
               else
                  firststab = false;

               kevaltimer.stop();
               
               if( ! timeBlow )
               { 
                  // Copy columns and rows generated
                  kcols[k] = new CColumn[ lpCols.size() ];
                  krows[k] = new CRow[ lpRows.size() ];
                  for( iter = 0; iter < lpCols.size(); iter++ )
                  {
                     kcols[k][iter].copy( *lpCols[iter] );	    
                  }
                  for( iter = 0; iter < lpRows.size(); iter++ )
                  {
                     krows[k][iter].copy( *lpRows[iter] );	    
                  }
                  nkcols[k] = lpCols.size();
                  nkrows[k] = lpRows.size();
                  
                  master->writeLog( true, "---------------------------------------\n\n" );
                  master->writeLog( true, "Evaluating root node with k=%d\n", k );
                  master->writeLog( true,"   objval = %g,   EvalTime = %g\n", lbVal, kevaltimer.getCPUTotalSecs() );
                  master->writeLog( true, "---------------------------------------\n\n" );
                  
                  if( k < 2 )
                  {
                     bestkbound = lbVal;
                     bestktime  = kevaltimer.getCPUTotalSecs();
                     bestk = k;
                  }
                  else
                  {
                     if( k == 2 )
                     {
                        kfactor = COLGENTHRESHOLD;
                     }
                     else
                     {
                        kfactor = KTHRESHOLD;
                     }
                     
                     // Evaluate to see if it is worth doing k
                     if (lbVal >= kfactor * bestkbound)
                     {
                        bestkbound = lbVal;
                        bestktime = kevaltimer.getCPUTotalSecs();
                        bestk = k;
                     }
                     else
                        if ((k > 2) &&
                           (kevaltimer.getCPUTotalSecs() < bestktime) &&
                           (lbVal > bestkbound))
                        {
                           bestkbound = lbVal;
                           bestktime = kevaltimer.getCPUTotalSecs();
                           bestk = k;
                        }
                        else
                        {
                           // Forces end of loop without doing break, so that we go
                           // through the below process of eliminating columns and
                           // rows generated.
                           k = ((CVRPMaster*)master)->getKCycleParam();
                        }
                  }
               } // Timeblow
               else
               {
                  k = ((CVRPMaster*)master)->getKCycleParam();
               } 
               
               // Now take out from the LP the oneedgecolumns
               eliminateColRange( masternCols, master->solver->getNColumns(locallp) - 1); 
               
               eliminateRowRange( masternRows, master->solver->getNRows(locallp)- 1 );
               
        } // end for (k = ...
        
        master->getCutPool()->empty();
        master->getColPool()->empty();
        
        for( iter = 0; iter < nkcols[bestk]; iter++ )
        {
           kcols[bestk][iter].setSepFromPool(true);
           addCol( kcols[bestk][iter] );
           addExpandedColToLP( kcols[bestk][iter] );
        }
        for( iter = 0; iter < nkrows[bestk]; iter++ )
        {
           krows[bestk][iter].setSepFromPool(true);
           addRow( krows[bestk][iter] );
           addExpandedRowToLP( krows[bestk][iter] );
        }
        
        setSolSize( master->solver->getNColumns(locallp) );
        setDualSize( master->solver->getNRows(locallp) );
        solvePrimal(); 
        
        master->writeLog( true, "*******************************************\n" );
        master->writeLog( true, "*******************************************\n" );
        master->writeLog( true, "\nChoice of k made:  %d\n\n", bestk );
        master->writeLog( true, "*******************************************\n" );
        master->writeLog( true, "*******************************************\n" );
        
        delete [] nkcols;
        delete [] nkrows;
        
        for( iter = 0; iter <= ((CVRPMaster*)master)->getKCycleParam();
        iter++ )
        {
           if( kcols[iter] )
              delete[] kcols[iter];
           if( krows[iter] )
              delete[] krows[iter];
        }
        delete [] kcols;
        delete [] krows;
        
        ((CVRPMaster*)master)->setKCycleParam( bestk );
        
        
      } // end if (getDynk())
      
#endif
      
    } // end  if (!getCarp() || getPearn() ||getUchoa)
    
  } // end if (getDepth() == 0)
  else // -> Add columns from the pool
  {
     // Solves problem using dual simplex
     cronoSolveLP.start();
     
     solveDual();
     
     cronoSolveLP.stop();
     
     
     // Separate the colums from the pool
     if ((((CVRPMaster*)master)->getCarp()) && (((CVRPMaster*)master)->getLongo()))
        expandPi(lpDualExp,lpDual,lpDualCarp);
     else
        expandPi(lpDualExp,lpDual);
     
     if( ((CVRPMaster*)master)->getKCycleParam() >= 2 )
     {
        ncolpool = separateColsFromPool(50);
        
        // Print information about the colums added
        master->writeLog( true, "  Total cols added from pool: %d\n", ncolpool );
     }
     
     // Solves problem using dual simplex
     cronoSolveLP.start();
     
     solvePrimal();
     
     cronoSolveLP.stop();
     
     // Separate the rows from the pool
     expandSolution(lpSolExp,lpSol);
  } // end if (getDepth() > 0)
  
  printf("Solving node using k=%d\n",
     ((CVRPMaster*)master)->getKCycleParam() );
  
  go = true;
  
  int nconverge = 0;
  
  
  cgiterlimit = false;
  while (go)
  {
     nconverge++;
     
     addcols = 0;
     
     
     if( ((CVRPMaster*)master)->getKCycleParam() >= 2 )
     {
        
        // Generate columns
        cronoColumnGeneration.start();
        inevalk = false;
        cgtimelimit = false;
        cgiterlimit = false;
        addcols = columnGen(((CVRPMaster*)master)->getKCycleParam());
        setSolSize(master->solver->getNColumns(locallp));
        solvePrimal();
        cronoColumnGeneration.stop();
        
        ((CVRPMaster*)master)->ncolgen += addcols;
        
        nCols = master->solver->getNColumns(locallp);
        
        if ( nCols > MAXLPNUMCOLS )
        {         
           master->writeLog( true, "Try to eliminate columns from LP...\n");
           
           nElimCols = eliminateLPColumns( nCols - MINLPNUMCOLS );
           
           master->writeLog( true, "%d Columns eliminated from LP.\n",nElimCols );
           
           setSolSize(master->solver->getNColumns(locallp));
           setDualSize(master->solver->getNRows(locallp));
           solvePrimal();
           
           expandSolution(lpSolExp,lpSol);
           
           if ((((CVRPMaster*)master)->getCarp()) && (((CVRPMaster*)master)->getLongo()))
              expandPi(lpDualExp,lpDual,lpDualCarp);
           else
              expandPi(lpDualExp,lpDual);
        }
     }
     // Then do only cut generation
     else
     {
        if ( firstTime )
        {
           cronoSolveLP.start();
           
           solvePrimal();
           
           cronoSolveLP.stop();
           
           if ((((CVRPMaster*)master)->getCarp()) && (((CVRPMaster*)master)->getLongo()))
              expandPi(lpDualExp,lpDual,lpDualCarp);
           else
              expandPi(lpDualExp,lpDual);
           
           generateColumnsOneEdge();
           
           addcols = localg->getM();
           firstTime = false;
        }
     }
     
     // Solves by the primal simplex
     cronoSolveLP.start();
     
     solvePrimal();
     
     cronoSolveLP.stop();
     
     // Verify if we can stop
     if( ceil( lbVal - master->params->getObjEps() ) >= master->getUB() - master->params->getObjEps() )
        break;
     
     // Generate cuts
     cronoCutGeneration.start();
     addrows = cutGen();
     cronoCutGeneration.stop();
     
     ((CVRPMaster*)master)->ncutgen += addrows;
     
     nRows = master->solver->getNRows(locallp);
     
     if ( nRows > MAXLPNUMROWS )
     {         
        master->writeLog( true, "Try to eliminate rows from LP...\n" );
        
        nElimRows = eliminateLPRows( nRows - MINLPNUMROWS );
        
        master->writeLog( true, "%d Rows eliminated from LP.\n",nElimRows);
        
        setDualSize(master->solver->getNRows(locallp));
        solveDual();
        expandSolution(lpSolExp,lpSol);
        
        if ((((CVRPMaster*)master)->getCarp()) && (((CVRPMaster*)master)->getLongo()))
           expandPi(lpDualExp,lpDual,lpDualCarp);
        else
           expandPi(lpDualExp,lpDual);
     }
     
     // Verify if columns or cuts were added to the problem
     if( addrows == 0 ) 
     {
        go = false;
     }
     else
     {
        go = true;
     }
   }
   
   // Solves the problem by dual simplex and expand solutions
   cronoSolveLP.start();
   
   solveDual();
   
   cronoSolveLP.stop();
   
   if ((((CVRPMaster*)master)->getCarp()) && (((CVRPMaster*)master)->getLongo()))
      expandPi(lpDualExp,lpDual,lpDualCarp);
   else
      expandPi(lpDualExp,lpDual);
   
   expandSolution(lpSolExp,lpSol);
   
   

   // Gets the dual objetive value
   auxsum = checkDualObj();
   
   
   
   // If lower bound of the son is worst than the father -> error
   if( lbVal < parentLB - master->params->getObjEps() )
   {
      master->solver->writeLPProblem(locallp, ".\\cvrp.lp");
      throw(LOWSONLOWERBOUND);
   }
   

   rootproximity = calcRootProximity(getLB());

   if( rootproximity >= ((CVRPMaster*)master)->getRootProx() )
   {
      master->writeLog( true, "CLOSE TO THE ROOT: Proximity = %g \n", rootproximity );
   }
   else if( rootproximity >= ((CVRPMaster*)master)->getLeafProx() )
   {
      master->writeLog( true, "MIDDLE: ROOT Proximity = %g \n", rootproximity );
   }
   else
   {
      master->writeLog( true, "CLOSE TO THE LEAF: Proximity = %g \n", rootproximity );
   }
   

   // Solves problem and expand solutions
   cronoSolveLP.start();
   
   solveDual();
   
   cronoSolveLP.stop();

   expandSolution(lpSolExp,lpSol);
   
   if ((((CVRPMaster*)master)->getCarp()) && (((CVRPMaster*)master)->getLongo()))
      expandPi(lpDualExp,lpDual,lpDualCarp);
   else
      expandPi(lpDualExp,lpDual);
   
#ifndef LPONLY
   obtainBasisFromLP( maxArtificials , nMasterRows );
#endif
   /* Put here to ensure that anything done in the strong branching won't mess with the lbVal printed */
   double auxlbval = lbVal;
   
   if (((CVRPMaster *) master)->getPearn())
   {
      auxReturn = processSolvedNode( 3.0, ((CVRPMaster*) master)->getSumAllEdges() );
   }
   else
   {
      auxReturn = processSolvedNode( 0.0, ((CVRPMaster*) master)->getSumAllEdges() );
   }
   
   lbVal = auxlbval;
   if ( auxReturn == FEASIBLE )
   {
      master->writeLog( true, "INTEGER FEASIBLE!!!!!!!!!!!!!!!!!\n" );

      
   }
   else if ( auxReturn == BRANCH )
   {
      if( ! ((CVRPMaster*)master)->getRootOnly() )
      {
         // Try to fix edges by reduced cost
         doBranch();
         master->writeLog( true, "Branched\n" );
      }
   }
   else if ( auxReturn == FATHOMED )
   {
      master->writeLog( true, "FATHOMED BY BOUND!!!!!!!!!!!!!!!\n" );
   }
   else
   {
      master->writeLog( true, "INFEASIBLE!!!!!!!!!!!!!!!!!!!!!!\n" );
   }
   
   // Stops the time
   nodetimer.stop();
   
   // Print information about the solved node
   master->writeLog( true, "Total time solving node %d: %g\n",
      (int) id, nodetimer.getCronoTotalSecs());
   master->writeLog( true, "Total CPU time solving node %d: %g\n",
      (int) id, nodetimer.getCPUTotalSecs());

   // Print information about the dual objetive
   if (((CVRPMaster *) master)->getPearn())
   {
      master->writeLog( true, "Node %d LB: %g, %g\n", (int) getID(), auxsum / 4, lbVal / 4 );
   }
   else
   {
      master->writeLog( true, "Node %d LB: %g, %g\n", (int) getID(), auxsum, lbVal );
   }


   if ( getID() == 1 )
   {
      master->writeLog( true, "----------------------------------------\n\n" );
      
      if (((CVRPMaster *) master)->getPearn())
      {
         master->writeLog( true, "Node %g solved. UB = %f, LB= %f\n\n", id, (float) ubVal / 4, (float) lbVal / 4 );
      }
      else
      {
         master->writeLog( true, "Node %g solved. UB = %f, LB= %f\n\n", id, (float) ubVal, (float) lbVal );
      }
      
      master->writeLog( true, "\n----------------------------------------\n\n" );
      
      master->writeLog( true, "Node Time Statistics\n" );
      
      master->writeLog(true,"Total number of passes: %d\n",nconverge);
      master->writeLog(true,"Total number of cols generated:     %d\n", ((CVRPMaster*)master)->ncolgen );
      
      
      master->writeLog( true, "Total Time Generating Columns: CPU: %g - Crono: %g\n",
         cronoColumnGeneration.getCPUTotalSecs(),
         cronoColumnGeneration.getCronoTotalSecs());
      
      master->writeLog(true,"Total number of rows generated:     %d\n", ((CVRPMaster*)master)->ncutgen );
      
      master->writeLog( true, "Total Time Generating Cuts: CPU: %g - Crono: %g\n",
         cronoCutGeneration.getCPUTotalSecs(),
         cronoCutGeneration.getCronoTotalSecs());
      
      master->writeLog( true, "Total Time Adding Columns to LP: CPU: %g - Crono: %g\n",
         cronoAddColsToLP.getCPUTotalSecs(),
         cronoAddColsToLP.getCronoTotalSecs());
      
      master->writeLog( true, "Total Time Adding Cuts to LP: CPU: %g - Crono: %g\n",
         cronoAddCutsToLP.getCPUTotalSecs(),
         cronoAddCutsToLP.getCronoTotalSecs());
      
      master->writeLog( true, "Total Time Solving LP: CPU: %g - Crono: %g\n",
         cronoSolveLP.getCPUTotalSecs(),
         cronoSolveLP.getCronoTotalSecs());
      
      master->writeLog( true, "\n----------------------------------------\n\n" );
   }
   
   
   ((CVRPMaster*)master)->mAddCutsToLPTimer += cronoAddCutsToLP;
   ((CVRPMaster*)master)->mAddColsToLPTimer += cronoAddColsToLP;
   ((CVRPMaster*)master)->mSolveLPTimer += cronoSolveLP;
   ((CVRPMaster*)master)->mColGenTimer += cronoColumnGeneration;   
   ((CVRPMaster*)master)->mCutGenTimer += cronoCutGeneration;
   
#ifdef GENERATE_ROOT_TABLE
   
   if ( id == 1 )
   {
      FILE *fout; // output file
      
      // Open the file
      #ifdef WIN32
      fout = fopen("tex\\roottable.tex","a");
      #else
      fout = fopen("tex/roottable.tex","a" );
      #endif

      if ( !fout )
         throw(FILENOTOPENEXCEPTION);
      
      // Print information about the node
      if (((CVRPMaster *) master)->getPearn())
      {
         fprintf(fout,"%s & %g & %g & %g & %d & %g & %d & %g & %g & %g & %d \\\\ \n",
            ((CVRPMaster*)master)->getInstanceFileName(), 
            getLB() / 4,
            nodetimer.getCronoTotalSecs(),
            cronoColumnGeneration.getCPUTotalSecs(),
            ((CVRPMaster*)master)->ncolgen,
            cronoCutGeneration.getCPUTotalSecs(),
            ((CVRPMaster*)master)->ncutgen,
            cronoAddColsToLP.getCPUTotalSecs(),
            cronoAddCutsToLP.getCPUTotalSecs(),
            cronoSolveLP.getCPUTotalSecs(), 
            nconverge
            );
      }
      else
      {
         fprintf(fout,"%s & %g & %g & %g & %d & %g & %d & %g & %g & %g & %d \\\\ \n",
            ((CVRPMaster*)master)->getInstanceFileName(), 
            getLB(),
            nodetimer.getCronoTotalSecs(),
            cronoColumnGeneration.getCPUTotalSecs(),
            ((CVRPMaster*)master)->ncolgen,
            cronoCutGeneration.getCPUTotalSecs(),
            ((CVRPMaster*)master)->ncutgen,
            cronoAddColsToLP.getCPUTotalSecs(),
            cronoAddCutsToLP.getCPUTotalSecs(),
            cronoSolveLP.getCPUTotalSecs(), 
            nconverge
            );
      }
      
      // Close the file
      if ( fout )
         fclose(fout);
   }
   
#endif
   return auxReturn;
}

//---------------------------------------------------------------------------//

void CVRPBBNode::generateColumnsOneEdge()
{
   register int i; // counter
   
   for (i=0;i<localg->getM();i++)
   {
      CColumn newcol(master->params,((CVRPMaster*)master)->getKCycleParam() );
      
      newcol.addElem(i,1.0);
      newcol.setObj(localg->getEdgeCost(i));
      newcol.setLB(0.0);
      newcol.setUB( LP_INFINITY );
      addCol(newcol);
      addExpandedColToLP(newcol);
   }
   setSolSize(master->solver->getNColumns(locallp));
}

//---------------------------------------------------------------------------//

int CVRPBBNode::columnGen(int kmax)
{
   int      kiter;
   int      startk;
   int      ncols;          // auxiliary variables
   int      improve_lp;     // auxiliary variables
   int      st;             // auxiliary variables
   int      nrows =0;          // auxiliary variables
   int      nnzeros;        // auxiliary variables
   int      ngeneratedcols; // number of columns generated
   int      coliters;       // number of column generation iterations
   CPUTimer timer;          // stores the time
   
   
   int i;                  // counter
   int dualstabiter;       // current dual stabilization iteration
   int startstab;          // iteration for which we should start the dual stab. 
   
   double epsilon = 0.0;         // cost of the dual stabilization variables
   
   
   if (((CVRPMaster*)master)->getIterCG())
   {
      startk = 2;
   }
   else
   {
      startk = kmax;
   }
   
   if ((((CVRPMaster*)master)->getCarp()) && (((CVRPMaster*)master)->getLongo()))
      startk = kmax = 2;
   
   // Initialize variables
   coliters = 0;
   
   ngeneratedcols = 0;
   
   if( ((CVRPMaster*) master)->getDoStab() )
   {
      nrows = master->solver->getNRows( locallp );
      
      if( firststab )
      {
         firststab = false;
         for( i = 0; i < nStab1 - localg->getN(); i++ )
         {
            dualestimate[i] = (double)master->getUB() / (double)(nrows*2);
         }
         startstab = 0;
      }
      else
      {
         startstab = 3;
      }
   }
   else
   {
      startstab = 3;
   }
   
   
   // Generate columns
   for (kiter = startk; kiter <= kmax; kiter++)
   {
      master->writeLog( true, "Generating columns without %d-cycles\n", kiter );
      
      improve_lp = 1;
      
      for( dualstabiter = startstab; dualstabiter < 4; dualstabiter++ )
      {
         if( ((CVRPMaster*) master)->getDoStab() )
         {
            improve_lp = true;
            
            master->writeLog(true,"Dual stab iter = %d\n", dualstabiter );
            
            
            if( dualstabiter < 3 )
            {
               for( i = 0; i < nStab1 - localg->getN(); i++ )
               {
                  master->solver->setCoef( locallp, -1, localg->getN() + i, 
                     - dualestimate[i] );
                  master->solver->setCoef( locallp, -1, maxStab1 + i, 
                     dualestimate[i] );
               }
               cronoSolveLP.start();
               solvePrimal();
               cronoSolveLP.stop();
            }
            
            
            
            if( dualstabiter == 0 )
            {
               epsilon = 1;
            }
            else if( dualstabiter == 1 )
            {
               epsilon = 0.1;
            }
            else if( dualstabiter == 2 )
            {
               epsilon = 0.01;
            }
            else if( dualstabiter == 3 )
            {
               epsilon = 0.0;
            }
            
            for( i = 0; i < nrows; i++ )
            {
               master->solver->setUB(locallp, localg->getN() + i, epsilon );  
               master->solver->setUB(locallp, maxStab1 + i, epsilon ); 
            }
            cronoSolveLP.start();
            solveDual();
            cronoSolveLP.stop();
         }
         
         
         
         while (improve_lp)
         {
            improve_lp = 0;
            
            cronoSolveLP.start();
            
            solvePrimal();
            
            cronoSolveLP.stop();
            
            if ((((CVRPMaster*)master)->getCarp()) && (((CVRPMaster*)master)->getLongo()))
               expandPi(lpDualExp,lpDual,lpDualCarp);
            else
               expandPi( lpDualExp, lpDual );
            
            timer.start();
            if ((((CVRPMaster*)master)->getCarp()) && (((CVRPMaster*)master)->getLongo()))
               st = carpDynProg();
            else
               st = fastDinProg(kiter, dualstabiter);
            timer.stop();

            
            if (st)
            {
               improve_lp = 1;
               ngeneratedcols += st;
            }
            
            ncols   = master->solver->getNColumns( locallp );
            nrows   = master->solver->getNRows( locallp );
            nnzeros = master->solver->getNNZeros( locallp );
            
            
            // Only test this if the CG is going to be performed with a maximum number of 
            // iterations (that happens in SB)
            if( cgiterlimit )
            {
               if( coliters >= maxcgiter )
               {
                  break;
               }
               
            }
            // If it is not in SB, do the standard output
            else
            {
               if (((coliters % DISPLAYITERS) == 0) || (!improve_lp))
               {
                  if (((CVRPMaster*)master)->getPearn())
                  {
                     master->writeLog( true, " Z = %g  ncols=%d  nrows=%d nnzero=%d  iters=%d\n", getLB()  / 4, ncols,nrows,nnzeros, coliters);
                  }
                  else
                  {
                     master->writeLog( true, " Z = %g  ncols=%d  nrows=%d nnzero=%d  iters=%d\n", getLB(), ncols,nrows,nnzeros, coliters);
                  }
               }
            }
         
            if( cgtimelimit )
            {
               if( timer.getCPUTotalSecs() > max( TIMEBLOWTHRESHOLD * bestktime, 300.0 ) )
               {
                  
                  break;
               }
            }
            
            coliters++;  
         }
  
         if( ((CVRPMaster*) master)->getDoStab() )
         {
            memcpy( dualestimate, lpDual, ( nStab1 - localg->getN() ) * sizeof(double) );         
            
            master->writeLog( true, "Partial = %g CPU seconds, %g seconds, %d iterations. Obj= %g\n", 
               timer.getCPUTotalSecs(),timer.getCronoTotalSecs(), coliters, getLB() );
         }
      }
      
      
      if( ! inevalk )
         if (!(((CVRPMaster*)master)->getCarp()) ||
            (((CVRPMaster*)master)->getPearn()) ||
            (((CVRPMaster*)master)->getUchoa()))
            if (kiter < ((CVRPMaster *)master)->getKCycleParam())
            {
               eliminateKCycleCols( kiter + 1 );
            }
            
            setSolSize( master->solver->getNColumns( locallp ) );
  }
  
  master->writeLog( true, "Total time generating columns = %g CPU seconds, %g seconds, %d iterations\n", 
     timer.getCPUTotalSecs(),timer.getCronoTotalSecs(), coliters );
  
  firstcolgen = false;
  
  return ngeneratedcols;
}

//---------------------------------------------------------------------------//

bool CVRPBBNode::createArtificialColumn(int r, int p)
{
   if (nArtificials>=maxArtificials)
      return false;
   
   if (p==0)
      master->solver->setCoef(locallp, r, nArtificials, 1.0);
   else
      master->solver->setCoef(locallp, r, nArtificials, 1.0);
   
   master->solver->setCoef(locallp, -1, nArtificials, ((CVRPMaster*) master)->getUB()+1.0);
   nArtificials++;
   
   return true;
}

//---------------------------------------------------------------------------//

int CVRPBBNode::createMaster()
{
   register int i; // counter
   double ub, cj, lb; // upper and lower bounds and costs
   char type; // type of the row
   int ncols, nrows; // number of columns and rows of the lp
   int nColsDel, nCutsDel; // number of columns and cuts deleted from pool
   int nColsPool, nCutsPool; // number of columns and cuts in the pool
   
   // Initialize
   ncols = 0;
   nrows = 0;
   
   // Constraints of degree of each node
   for (i=1;i<localg->getN();i++)
   {
      if (((i % 2) == 1) && ((CVRPMaster*) master)->getCarp() && ((CVRPMaster*) master)->getUchoa())
      {
         ub = -2.0;
         type = 'G';
      }
      else
      {
         ub = 2.0;
         type = 'G';
      }
      
      master->solver->addNewRows(locallp, 1, &ub, &type);
      nrows++;
      
   }
   
   // Constraint of the number of vehicles
   ub = 2.0 * (((CVRPMaster*) master)->getK());
   if ((((CVRPMaster*) master)->getCarp()) && (((CVRPMaster*) master)->getUchoa()))
      type = 'G';
   else
      type = 'E';
   
   master->solver->addNewRows(locallp, 1, &ub, &type);
   nrows++;

   
   nMasterRows = nrows;

   // Create identity
   // For all the constraints so far
   for( i=1; i <= nMasterRows; i++)
   {
      cj = ((CVRPMaster*) master)->getUB()+1.0;
      
      ub = LP_INFINITY;
      lb = 0;
      type = 'C';
      
      master->solver->addNewCols(locallp, 1, &cj, &lb, &ub, NULL);
      master->solver->setCoef(locallp, i-1, ncols, 1.0);
      ncols++;
   }
   
   nMasterCols = ncols;
   if( ((CVRPMaster*) master)->getDoStab() )
   {
      nMasterCols = ncols;
      // Create dual stab 1 columns
      for( i = nMasterCols; i < maxStab1; i++ )
      {
         ub = 0;
         lb = 0;
         type = 'C';
         cj=0.0;
         
         master->solver->addNewCols(locallp, 1, &cj, &lb, &ub, NULL);
         ncols++;
      }
      
      nMasterCols = ncols;
      
      // Create dual stab 2 columns
      for( i = nMasterCols; i < maxStab2; i++ )
      {
         ub = 0;
         lb = 0;
         type = 'C';
         cj=0.0;
         
         master->solver->addNewCols(locallp, 1, &cj, &lb, &ub, NULL);
         ncols++;
      }
      
      nMasterCols = ncols;
      
      
      // Create coefficients for identity
      for( i=1; i <= localg->getN(); i++)
      {
         master->solver->setCoef(locallp, i-1, localg->getN() + i-1,-1.0);
         
         master->solver->setCoef(locallp, i-1, maxStab1 + i-1,1.0);
      }
   }
   
   
   // Create artificial columns
   for( i = nMasterCols; i < maxArtificials; i++ )
   {
      ub = LP_INFINITY;
      lb = 0;
      type = 'C';
      cj=((CVRPMaster*) master)->getUB()+1.0;
      
      master->solver->addNewCols(locallp, 1, &cj, &lb, &ub, NULL);
      ncols++;
   }
   
   // Set the problem as a minimization problem
   master->solver->setObjSense(locallp, LP_MIN);
   
#ifndef LPONLY
   
   setDualSize(nrows);
   setSolSize(ncols);
   
   if( getDepth() > 0 )
   {
      addBasisToLP();
      
      nColsPool = master->getColPool()->getSize();
      nCutsPool = master->getCutPool()->getSize();
      nColsDel = master->cleanColPool();
      nCutsDel = master->cleanCutPool();
      //nColsDel = 0;
      //nCutsDel = 0;
      
      if ( nCutsDel )
      {
         master->writeLog( true, "Total cuts deleted from the pool: %d of %d\n", nCutsDel,nCutsPool);
      }
      
      if ( nColsDel )
      {
         master->writeLog( true, "Total columns deleted from the pool: %d of %d\n", nColsDel,nColsPool);
      }
   }
   
#endif
   
   masterFixEdges();
   
   ncols = master->solver->getNColumns(locallp);
   nrows = master->solver->getNRows(locallp);
   
   setDualSize(nrows);
   setSolSize(ncols);

   
   
   return 0;
}

//---------------------------------------------------------------------------//

void CVRPBBNode::masterFixEdges()
{
   register int i; // counter
   
   for (i=0; i < nFixedSets; i++)
   {
      fSets[i]->setSepFromPool(false);
      addRow(*fSets[i]);
      addExpandedRowToLP(*fSets[i]);
   }
}

//---------------------------------------------------------------------------//

int CVRPBBNode::cutGen()
{
   int numCutsAdded = 0; // number of cuts added to lp
   int auxcuts; // auxiliary variable
   double auxob; // auxiliary objective values
   int ncutpool; // number of cuts added from the pool
   int noOfEdges; // number of edges different of zero
   char feasible; // feasible status
   double maxViolCap, maxViolMS,maxViolFC,      // max violation
      maxViolHyp, maxViolComb, maxViolGLMS; // of the cuts
   int maxCapCuts, maxMSCuts, maxFCCuts,        // max number
      maxHypCuts, maxCombCuts;                 // of cuts
   register int i; // counter
   int listCustSize; // size of the list of customers
   
   // Initialize
   auxcuts = 0;
   ncutpool = 0;
   
   cronoSolveLP.start();
   
   solvePrimal();
   
   cronoSolveLP.stop();
   
   // Store the obj before the cuts
   auxob = getLB();
   
   if (((CVRPMaster*)master)->getPearn())
   {
      master->writeLog( true, "New obj val: %f\n\n", (double) getLB() / 4 );
   }
   else
   {
      master->writeLog( true, "New obj val: %f\n\n", (double) getLB() );
   }
   
   // Expand the solution
   expandSolution(lpSolExp,lpSol);
   
   // Store information about the solution
   noOfEdges = 0;
   
   for (i=0; i < localg->getM(); i++)
   {
      if ( lpSolExp[i] > master->params->getIntEps() )
      {
         noOfEdges++;
         edgeTail[noOfEdges]= localg->getEdgeOrg(i);
         edgeHead[noOfEdges]= localg->getEdgeDst(i);
         edgeX[noOfEdges]= lpSolExp[i];
      }
   }
   
   // Try to add bound cuts
   fixzero = false;
   auxcuts += boundCut();
   
   //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   if ((((CVRPMaster *)master)->getCarp()) && (((CVRPMaster *)master)->getUchoa()))
      auxcuts += carpBoundCut();
   //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   
   // Try to separate cuts from the pool
   if (getDepth()>=1)
   {
      master->writeLog( true, "Trying to separate cuts from the pool...\n" );
      
      ncutpool = separateRowsFromPool(50);
      auxcuts+=ncutpool;
      
      master->writeLog( true, "Number of cuts added from the pool: %i\n", ncutpool );
   }
   
   // Separate capacity cuts
   maxCapCuts = 50;
   maxViolCap = 0.0;
   capacityCuts(noOfEdges,maxCapCuts,feasible,maxViolCap);
   
   // Separate comb cuts
   maxViolComb = 0.0;
   maxCombCuts = 20;
   
   if (((CVRPMaster *)master)->getCarp())
   {
      if (sepCombCuts )
      {
         combCuts(((CVRPMaster*)master)->getMinCapVehicle(),noOfEdges,maxCombCuts,maxViolComb);
      }
   }
   else
   {
      if ( getID() == 1 && maxViolCap < 0.1 && sepCombCuts )
      {
         combCuts(((CVRPMaster*)master)->getMinCapVehicle(),noOfEdges,maxCombCuts,maxViolComb);
      }
   }
   
   // Separate multistar cuts
   maxViolMS = 0.0;
   maxMSCuts = 30;
   
   if ( getID() == 1 && maxViolCap < 0.1 && maxViolComb < 0.1 && sepMSCuts )
      //   if ( maxViolCap < 0.1 && maxViolComb < 0.1 && sepMSCuts )
   {
      multiStarCuts(noOfEdges,maxMSCuts,maxViolMS);
      sepMSCuts = false;
   }
   
   // Separate framed capacity cuts
   maxViolFC = 0.0;
   maxFCCuts = 10;
   if ( getID()==1 && maxViolMS < 0.1 && maxViolCap < 0.1 &&
      //   if ( maxViolMS < 0.1 && maxViolCap < 0.1 &&
      maxViolComb < 0.1 && sepFCCuts)
   {
      framedCapacityCuts(noOfEdges,20000,maxFCCuts,maxViolFC);
   }
   
   // Separate hypotour cuts
   maxViolHyp = 0.0;
   maxHypCuts = 10;
   
   if (getID()==1 && maxViolCap < 0.1 &&
      //   if (maxViolCap < 0.1 &&
      maxViolMS < 0.1 && maxViolFC < 0.1 && 
      maxViolComb < 0.1 && sepHypCuts)
   {
      hypotourCuts(noOfEdges,maxHypCuts, maxViolHyp);
   }
   
   // Separate generalized large MS Cuts
   maxViolGLMS = 0.0;
   listCustSize = 0;
   maxViolCap = 0;
   
   if (getID()==1 && maxViolCap < 0.1 &&
      //   if (maxViolCap < 0.1 &&
      maxViolMS < 0.1 && maxViolFC < 0.1 && 
      maxViolComb < 0.1 && maxViolHyp < 0.1 && sepGLMSCuts 
      && ! fixzero )
   {
      generalizedLargeMSCuts(noOfEdges,nodeP,listCustSize,maxViolGLMS);
   }
   
   auxcuts += addCapSepCutsToLP(listCustSize,nodeP);
   
   // Separate exact capacity cuts
   if ( getID()==1 && auxcuts == 0 )
   {
      //auxcuts += exactCapacityCutGen();
   }
   
   numCutsAdded += auxcuts;
   
   master->writeLog( true, "Total cuts added: %d\n", numCutsAdded );
   
   setDualSize(master->solver->getNRows(locallp));
   
   cronoSolveLP.start();
   
   solveDual();
   
   cronoSolveLP.stop();
   
   return numCutsAdded;
}

//---------------------------------------------------------------------------//

bool CVRPBBNode::isFeasible(double* xSol)
{
   register int j; // counter
   
   for( j = 0; j < localg->getM(); j++ )
   {
      if( ( ( lpSolExp[ j ] > master->params->getIntEps() ) && ( lpSolExp[ j ] < 1.0 - master->params->getIntEps() ) ) ||
         ( ( lpSolExp[ j ]  > 1.0 + master->params->getIntEps() ) && ( lpSolExp[j] < 2.0-master->params->getIntEps() ) ) )
      {
         return false;
      }
   }
   return true;
}

//---------------------------------------------------------------------------//

void CVRPBBNode::doBranch()
{
   register int i,j,k,iter; // counter
   CRow *newrow1p, *newrow2p; // pointer to the position of the rows
   double maxObj; // best objective among all branch options
   int bestSet; // best set to do branch
   double obj1, obj2; // objective functions
   int numberOfBranchs; // number of brachs to try
   int nRows; // number of rows of the lp
   bool flagBranchEdge; // tells that the branch will be done in the edges
   int flagTwo; // tells if the edge can be fixed at two
   int edgeChoice; // edge choose to be branched
   
   int maxBranchSets;
   double currobjval = lbVal;
   
   /* Used to store the basis of the original problem */
   int* basiccols;
   int* basicrows;
   int numlpcols, numlprows;
   
   int minp = ((CVRPMaster*)master)->getPMin();
   int maxrootp = ((CVRPMaster*)master)->getPMax();
   int sbcgiters = ((CVRPMaster*)master)->getSBCGIter();

   
   
   // Stores the best nodes created by SB
   CVRPBBNode* bestb1 = NULL;
   CVRPBBNode* bestb2 = NULL;
   
   
   
   /* Used to evaluate SB, as in the article by Lysgaard, Letchford
    * and Eglese */
   double bestlbminus, bestlbplus;
   double lbminus, lbplus;
   bestlbminus = -1;
   bestlbplus = -1;

   numlpcols = master->solver->getNColumns(locallp);
   numlprows = master->solver->getNRows(locallp);
   
   basiccols = new int[ numlpcols ];
   if( ! basiccols )
   {
      printf("Error in SB when allocating basiccols\n");
      exit(-1);
   }   
   
   basicrows = new int[ numlprows ];
   if( ! basicrows )
   {
      printf("Error in SB when allocating basicrows\n");
      exit(-1);
   }
   
   master->solver->getBasis(locallp,basiccols,basicrows);
   
   if( ((CVRPMaster*) master)->getRootOnly() )
      return;
   
   flagBranchEdge = false;
   
   // Get the branch candidate sets
   // The sets will already come ordered by absolute difference to
   // BOUNDARYTARGET
   getBranchCandidateSets();
   
   if ( setsCMP->Size <= 0 )
      return;
   
   numberOfBranchs = setsCMP->Size;
   
   nRows = master->solver->getNRows(locallp);
   maxObj = -1;
   

   
   // Default value for when no SB is done
   bestSet = 0;

   CPUTimer sbtimer;

   sbtimer.start();
   
   master->writeLog( true, "Starting Strong branching\n" );
   
   // Old did less branch tests according to depth
   maxBranchSets = max( maxrootp - getDepth(), minp );

   // To make sure that the value maxBranchSets is not above
   // numberOfBranches
   maxBranchSets = min( maxBranchSets, numberOfBranchs );
   

   // If this node is too close to leaves, don't do SB
   if( rootproximity < ((CVRPMaster*)master)->getLeafProx() )
      maxBranchSets = 1;
  
   if( flagBranchEdge )
      maxBranchSets = 1;

   if( ! ((CVRPMaster*) master)->getDoSB() )
      maxBranchSets = 1;
   
   master->writeLog( true, "Doing SB on %d/%d sets\n",maxBranchSets, numberOfBranchs );
   

   for (i = 0; i < maxBranchSets; i++)
   {
      // Get the S set
      for (j = 1; j <= localg->getN(); j++)
         listS[j] = 0;
      
      for (j = 1; j <= setsCMP->CPL[i]->IntListSize; j++)
      {
         listS[setsCMP->CPL[i]->IntList[j]] = 1;
      }
      
      
      CRow newrow1(master->params);
      CRow newrow2(master->params);
      
      if( flagBranchEdge )
      {
         master->writeLog( true, "Doing branching on EDGE\n");
         flagTwo = 0;
         edgeChoice = chooseBranchArc(flagTwo);
         if ( flagTwo == 0 )
         {
            newrow1.addElem(edgeChoice,1.0);
            newrow1.setType(E);
            newrow1.setRHS(0.0);
            newrow2.addElem(edgeChoice,1.0);
            newrow2.setType(E);
            newrow2.setRHS(1.0);
         }
         else
         {
            newrow1.addElem(edgeChoice,1.0);
            newrow1.setType(E);
            newrow1.setRHS(0.0);
            newrow2.addElem(edgeChoice,1.0);
            newrow2.setType(G);
            newrow2.setRHS(1.0);
         }
      }
      else
      {
         for (j = 1; j <= localg->getN(); j++)
         {
            if (listS[j] == 1)
            {
               for (k = 1; k <= localg->getN(); k++)
               {
                  if ((j != k) && (listS[k] == 0))
                  {
                     newrow1.addElem( localg->getEdgeIndex( j, k ), 1.0 );
                     newrow2.addElem( localg->getEdgeIndex( j, k ), 1.0 );
                  }
               }
            }
         }
         newrow1.setType( E );
         newrow1.setRHS( 2.0 );
         newrow2.setType( G );
         newrow2.setRHS( 4.0 );
      }
      
      
      
      newrow1.setSepFromPool(false);
      newrow2.setSepFromPool(false);
      
      newrow1p = addRow(newrow1);
      newrow1p->setSepFromPool(false);
      newrow2p = addRow(newrow2);
      newrow2p->setSepFromPool(false);
      
      // Open the son nodes
      CVRPBBNode* b1 = new CVRPBBNode( this, 2.0 * id, 0 );
      CVRPBBNode* b2 = new CVRPBBNode( this, 2.0 * id + 1.0, 1 );
      
      b1->lbVal = currobjval;
      b2->lbVal = currobjval;
      
      // Set the fixed sets
      b1->fSets=new CRow*[nFixedSets+1];
      b2->fSets=new CRow*[nFixedSets+1];
      
      // Copy the list of fixed sets
      for(iter=0; iter < nFixedSets; iter++)
      {
         b1->fSets[iter]=fSets[iter];
         b2->fSets[iter]=fSets[iter];
         b1->fSets[iter]->increaseNLPs();
         b2->fSets[iter]->increaseNLPs();
      }
      
      b1->fSets[nFixedSets] = newrow1p;
      b2->fSets[nFixedSets] = newrow2p;
      newrow1p->increaseNLPs();
      newrow2p->increaseNLPs();

      b1->nFixedSets=nFixedSets+1;
      b2->nFixedSets=nFixedSets+1;
      
      b1->setGraph( localg );
      b2->setGraph( localg );


      // Close to root
      if( rootproximity >= ((CVRPMaster*)master)->getRootProx() )
      {
         b1->branchorg = 0;
         b2->branchorg = 0;
      }
      else if( rootproximity >= ((CVRPMaster*)master)->getLeafProx() )
      {
         b1->branchorg = 1;
         b2->branchorg = 1;         
      }
      else
      {
         b1->branchorg = 2;
         b2->branchorg = 2;
      }

      
      // Copy the list of fixed edges
      if( nFixedEdges > 0 )
      {
         b1->fixedEdges = new int[ nFixedEdges ];
         b2->fixedEdges = new int[ nFixedEdges ];
         memcpy( b1->fixedEdges, fixedEdges, nFixedEdges * sizeof(int) );
         memcpy( b2->fixedEdges, fixedEdges, nFixedEdges * sizeof(int) );
         b1->nFixedEdges = nFixedEdges;
         b2->nFixedEdges = nFixedEdges;
      }

      

      if( edgeweights )
      {
         b1->edgeweights = new double[localg->getM()];
         b2->edgeweights = new double[localg->getM()];
         if( !b1->edgeweights )
            throw(LOWMEMORYEXCEPTION);
         if( !b2->edgeweights )
            throw(LOWMEMORYEXCEPTION);
         memcpy( b1->edgeweights, edgeweights, localg->getM() * sizeof( double ) );
         memcpy( b2->edgeweights, edgeweights, localg->getM() * sizeof( double ) );
      }

      if( maxBranchSets == 1 )
      {
         b1->initializeNode();
         b1->createMaster();
         
         b2->initializeNode();
         b2->createMaster();
         
         bestb1 = b1;
         bestb2 = b2;
         break;
      }
      
      master->writeLog(true,"\nTesting set1 of %d\n",i);
      b1->initializeNode();
      b1->createMaster();
      b1->solvePrimal();
      if ((getDepth() <= SBCOLGENDEPTH) &&
         (((CVRPMaster*)master)->getKCycleParam() >= 2))
      {
         // If too close to the root, do CG all the way
         // Otherwise, don't

         b1->inSB = true;
         b1->maxcgiter = sbcgiters;
         b1->cgiterlimit = true;
         b1->columnGen( ((CVRPMaster *)master)->getKCycleParam() );
         b1->inSB = false;
         b1->cgiterlimit = false;
      }
      
      obj1 = b1->lbVal;
      b1->lbVal = lbVal;
      
      master->writeLog(true,"Testing set2 of %d\n",i);
      b2->initializeNode();
      b2->createMaster();
      b2->solvePrimal();
      if ((getDepth() <= SBCOLGENDEPTH) &&
         (((CVRPMaster *)master)->getKCycleParam() >= 2))
      {

         b2->inSB = true;
         b2->maxcgiter = sbcgiters;
         b2->cgiterlimit = true;
         b2->columnGen( ((CVRPMaster *)master)->getKCycleParam() );
         
         b2->inSB = false;
         b2->cgiterlimit = false;
      }
      
      obj2 = b2->lbVal;
      b2->lbVal = lbVal;
      
      lbminus = min( obj1, obj2 );
      lbplus = max( obj1, obj2 );

      // IPCO rule
      if( ceil( lbminus ) > ceil( bestlbminus ) )
      {
         bestSet = i;
         bestlbminus = lbminus;
         bestlbplus = lbplus;
      }
      // Tie
      else if( ceil( lbminus ) == ceil( bestlbminus ) )
      {
         if( ceil( lbplus ) > ceil( bestlbplus ) )
         {
            bestSet = i;
            bestlbminus = lbminus;
            bestlbplus = lbplus;
         }
         // Another tie
         else if( ceil( lbplus ) == ceil( bestlbplus ) )
         {
            if( lbminus > bestlbminus )
            {
               bestSet = i;
               bestlbminus = lbminus;
               bestlbplus = lbplus;
            }
            else if( lbminus == bestlbminus )
            {
               if( lbplus > bestlbplus )
               {
                  bestSet = i;
                  bestlbminus = lbminus;
                  bestlbplus = lbplus;
               }
            }
         }
      }

      // Reverse the action performed by addRow, which is to add row to lpRows
      // (don't need to bother about adding to pool for now ->creates some overhead, though)
      lpRows.pop_back();
      lpRows.pop_back();
      
      
      if( bestSet == i )
      {
         if( bestb1 )
            delete bestb1;
         if( bestb2 )
            delete bestb2;
         
         bestb1 = b1;
         bestb2 = b2;
      }
      else
      {
         delete b1;
         delete b2;
      }
      master->writeLog(true, "bestSet = %5d   obj1 = %lf    obj2 = %lf\n", bestSet, obj1, obj2 );      
   }
   SBtries = i + 1;
   if( SBtries > maxBranchSets )
      SBtries = maxBranchSets;
   sbtimer.stop();
   ((CVRPMaster*)master)->sbtimer += sbtimer;
   
   master->writeLog( true, "Finished strong branching\n" );
   
   master->writeLog( true, "Total time doing SB: %g\n",
         sbtimer.getCPUTotalSecs() );
      
   if( getDepth() == 0 )
   {
      // Get the S set
      for (j=1; j <= localg->getN(); j++)
         listS[j] = 0;
      
      for (j=1; j <= setsCMP->CPL[bestSet]->IntListSize; j++)
      {
         listS[setsCMP->CPL[bestSet]->IntList[j]]=1;
      }
#ifdef WIN32
      FILE* fout= fopen("log\\sbset.txt","wt");
#else
      FILE* fout= fopen("log/sbset.txt","wt");
#endif
      fprintf(fout,"%d\n", localg->getN() );
      master->writeLog(true,"SB done on the set:\n");
      for( j=1; j<= localg->getN(); j++ )
      {
         if( listS[j] == 1 )
         {
            master->writeLog(true,"%d,  ",j);
         }
         fprintf(fout,"%d, %d\n",j, listS[j]);
      }
      fclose(fout);
   }
   
   
   if( ( bestb1 == NULL ) || ( bestb2 == NULL ) )
      throw( SBNOBRANCHSETS );

   if( ((CVRPMaster*)master)->getRevBB() )
   {
      master->addOpenNode( bestb2 );
      master->addOpenNode( bestb1 );
   }
   else
   {
      master->addOpenNode( bestb1 );
      master->addOpenNode( bestb2 );
   }
   
   
   delete [] basiccols;
   delete [] basicrows;
}

//---------------------------------------------------------------------------//

int CVRPBBNode::chooseBranchArc(int &flagTwo)
{
   register int i; // counter
   int brancharc; // variable that tells the variable to branch
   double currgrade; // auxiliary variable
   double mingrade; // minimum grade
   int nDepot; // the vertex that is the depot
   
   // Initialize the grades
   brancharc = -1;
   mingrade = 0.0;
   nDepot = ((CVRPMaster*)master)->getDepot();
   
   currgrade = master->params->getInfinity();
   mingrade = master->params->getInfinity();
   
   // Calculates the best grade
   for( i = 0; i < localg->getM(); i++ )
   {
      if( lpSolExp[i] < master->params->getIntEps()  || 
         ( lpSolExp[i] > 1.0 - master->params->getIntEps() && lpSolExp[i] < 1.0+master->params->getIntEps() ) ||
         lpSolExp[i]>2.0-master->params->getIntEps() || localg->getEdgeFixed(i)!=-1)
      {
         currgrade = master->params->getInfinity();
      }
      else
      {
         currgrade=fabs(lpSolExp[i]-0.5);
         currgrade*=(1.0/localg->getEdgeCost(i));
         
         if (localg->getEdgeOrg(i)==nDepot || localg->getEdgeDst(i)==nDepot && 
            localg->canBeTwoEdge(i))
            currgrade*=10000;
      }
      
      if( currgrade < mingrade )
      {
         mingrade = currgrade;
         brancharc = i;
      }
   }
   
   if ((localg->getEdgeOrg(brancharc)==nDepot || localg->getEdgeDst(brancharc)==nDepot) && 
      localg->canBeTwoEdge(brancharc))
      flagTwo=1;
   else
      flagTwo=0;
   
   return brancharc;
}

//---------------------------------------------------------------------------//

// dualstabiter is a parameter that tells in which iteration of the dual
//  stabilization process we are, so that the columns can be generated accordingly
int CVRPBBNode::fastDinProg(int klength, int dualstabiter) 
{
   QROracleLP *oracleGraph; // encapsulates the graph
   QRSolver<QROracleLP> *solverDinProg; // solver for dyn prog
   
   int nCols; // number of columns generated
   register int i,j; // counters
   double routeCost; // cost of the route
   int pathSize; // size of the path generated
   int nDepot; // number of the depot
   double reducedCost; // reduced cost of the column
   
   int* findKCycle = new int[ localg->getN() ];
   int dpt=0;
   
   bool anycycle = false;
   
   int k;
   
   //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   //columnHash cHash;
   //columnHash::iterator cit;
   //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   
   //allocate oracle graph
   oracleGraph = new QROracleLP(localg,(CVRPMaster*)master);
   
   if (!oracleGraph) 
      throw (LOWMEMORYEXCEPTION);
   
   //allocate solver
   solverDinProg = new QRSolver<QROracleLP>( klength, 1 );
   
   if (!solverDinProg) 
      throw(LOWMEMORYEXCEPTION);
   
   oracleGraph->reset(lpDualExp);
   
   if( dualstabiter >=3 )
   {
      // Perform exact separation of the columns
      solverDinProg->solve(oracleGraph);
   }
   else
   {
      solverDinProg->solve(oracleGraph,1);
   }
   
   nDepot = localg->getN(); //depot number
   nCols = 0;               //number of columns generated
   
                            /*--------------------------------------------------
                            | get the paths and generate corresponding columns
   *-------------------------------------------------*/
   
   for (i = 1; i < oracleGraph->getN(); i++) 
   {
      reducedCost = solverDinProg->getBestPath( i, pathDinProg );
      routeCost   = 0;
      pathSize    = pathDinProg[0];

      bestpath[i] = reducedCost;

      if (pathSize > 0) 
      {
         if (pathDinProg[1] == oracleGraph->getDepot())
            continue; //no path through de depot
         
         //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         if ((((CVRPMaster*)master)->getCarp()) && (((CVRPMaster*)master)->getUchoa()))
            reducedCost += oracleGraph->getLength( pathDinProg[1], oracleGraph->getDepot() );
         else
            reducedCost += oracleGraph->getLength( oracleGraph->getDepot(), pathDinProg[1] );
         //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         
         if (reducedCost > -master->params->getIntEps())
         {
            continue;
         }
         
         memset( findKCycle, 0, localg->getN() * sizeof(int) ); 
         
         CColumn newcol( master->params, ((CVRPMaster *) master)->getKCycleParam() );
         newcol.setLB( 0 );
         newcol.setUB( LP_INFINITY );

         
         //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         if ((((CVRPMaster*)master)->getCarp()) && (((CVRPMaster*)master)->getUchoa()))
         {
            if ((pathDinProg[pathSize] % 2) == 1)
            {
               newcol.addElem( localg->getEdgeIndex( nDepot, pathDinProg[pathSize] + 1), 1.0 );
               routeCost += localg->getEdgeCost( nDepot, pathDinProg[pathSize] + 1);

            }
            else
            {
               newcol.addElem( localg->getEdgeIndex( nDepot, pathDinProg[pathSize] - 1), 1.0 );
               routeCost += localg->getEdgeCost( nDepot, pathDinProg[pathSize] - 1);
            }
         }
         else
         {
            newcol.addElem( localg->getEdgeIndex( nDepot, pathDinProg[pathSize] ), 1.0 );
            routeCost += localg->getEdgeCost( nDepot, pathDinProg[pathSize] );


         }
         //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         
         dpt = 1;
         findKCycle[ pathDinProg[pathSize] ] = dpt++;
         for (j = pathSize; j > 1; j--) 
         {
            //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
            if (!((((CVRPMaster*)master)->getCarp()) && (((CVRPMaster*)master)->getUchoa())))
            {
               newcol.addElem(localg->getEdgeIndex(pathDinProg[j],pathDinProg[j-1]),1.0);
               routeCost+=localg->getEdgeCost(pathDinProg[j],pathDinProg[j-1]);


            }
            //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
            
            // Finds if it has kcycles
            if( findKCycle[ pathDinProg[j-1] ] > 0 )
            {
               for( k = dpt - findKCycle[pathDinProg[j-1]]; k <=
                  ((CVRPMaster*)master)->getKCycleParam(); k++ )
               {
                  newcol.setHasKCycle(k);
                  
                  anycycle = true;
               }
            }
            findKCycle[ pathDinProg[j-1] ] = dpt++;
         }
         
         if( !inevalk )
         {
            if( ! firstcolgen )
            {
               if( anycycle )
               {
                  continue;
               }
            }
         }
         
         //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         if ((((CVRPMaster*)master)->getCarp()) && (((CVRPMaster*)master)->getUchoa()))
         {
            for (j = pathSize; j > 0; j--) 
            {
               if ((pathDinProg[j] % 2) == 1)
               {
                  newcol.addElem( localg->getEdgeIndex( pathDinProg[j] + 1, pathDinProg[j] ), 1.0 );
                  routeCost += localg->getEdgeCost( pathDinProg[j] + 1, pathDinProg[j] );

                  
                  if (j < pathSize)
                  {
                     newcol.addElem( localg->getEdgeIndex( pathDinProg[j+1], pathDinProg[j] + 1), 1.0 );
                     routeCost += localg->getEdgeCost( pathDinProg[j+1], pathDinProg[j] + 1 );

                  }
               }
               else
               {
                  newcol.addElem( localg->getEdgeIndex( pathDinProg[j] - 1, pathDinProg[j] ), 1.0 );
                  routeCost += localg->getEdgeCost( pathDinProg[j] - 1, pathDinProg[j] );

                  
                  if (j < pathSize)
                  {
                     newcol.addElem( localg->getEdgeIndex( pathDinProg[j+1], pathDinProg[j] - 1), 1.0 );
                     routeCost += localg->getEdgeCost( pathDinProg[j+1], pathDinProg[j] - 1 );

                  }
               }
            }
         }
         //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         
         newcol.addElem( localg->getEdgeIndex( pathDinProg[1], nDepot ), 1.0 );
         routeCost += localg->getEdgeCost( pathDinProg[1], nDepot );

         {
            newcol.setObj( routeCost );
            
            addCol( newcol );
            addExpandedColToLP( newcol );
            
            nCols++;
         }
     }
   }
   
   setSolSize(master->solver->getNColumns(locallp));
   
   
   //free memory for dynamic programming
   if (solverDinProg) 
      delete solverDinProg;
   
   if (oracleGraph) 
      delete oracleGraph;
   
   delete [] findKCycle;
   
   return nCols;
}

//---------------------------------------------------------------------------//
//---------------------------------------------------------------------------//

int CVRPBBNode::carpDynProg()
{
   bool   flag;              // auxiliary flag
   register int    i           = 0;   // counter
   register int    j           = 0;   // counter
   register int    k           = 0;   // counter
   register int    c           = 0;   // counter
   register int    x           = 0;   // counter
   register int    y           = 0;   // counter
   register int    w           = 0;   // counter
   int    nAdd        = 0;   // number of cols added
   int    nNodes      = 0;   // number of vertex
   int    cRes        = 0;   // residual capacity
   int    vCap        = 0;   // vehicles capacity
   int    nDepot      = 0;   // number of the depot
   double inf         = 0.0; // infinity value
   double v           = 0.0; // auxiliary variable
   double reducedCost = 0.0; // reduced cost
   double routecost   = 0.0; // cost of the route
   
   CVRPMaster* mstr = (CVRPMaster *) master;
   
   
   // Initialize
   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   inf    = LP_INFINITY;
   nNodes = localg->getN();
   vCap   = mstr->getVehicleCap();
   nDepot = mstr->getDepot();
   
   for (i = 1; i <= nNodes; i++)
   {
      for (x = 0; x < 2; x++)
      {
         mstr->dpCarp[i][0][x].y1 = -1;
         mstr->dpCarp[i][0][x].y2 = -1;
         mstr->dpCarp[i][0][x].p1 = -1;
         mstr->dpCarp[i][0][x].p2 = -1;
         mstr->dpCarp[i][0][x].v1 = inf;
         mstr->dpCarp[i][0][x].v2 = inf;
      }
   }
   
   for (c = 0; c <= vCap; c++)
   {
      for (x = 0; x < 2; x++)
      {
         mstr->dpCarp[nDepot][c][x].y1 = -1;
         mstr->dpCarp[nDepot][c][x].y2 = -1;
         mstr->dpCarp[nDepot][c][x].p1 = -1;
         mstr->dpCarp[nDepot][c][x].p2 = -1;
         mstr->dpCarp[nDepot][c][x].v1 = 0.0;
         mstr->dpCarp[nDepot][c][x].v2 = 0.0;
      }
   }
   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   
   // dynamic programming
   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   for (c = 1; c <= vCap; c++)
   {
      for (i = 1; i < nNodes; i++)
      {
         // Cell Initialize
         for (x = 0; x < 2; x++)
         {
            mstr->dpCarp[i][c][x].y1 = -1;
            mstr->dpCarp[i][c][x].y2 = -1;
            mstr->dpCarp[i][c][x].p1 = -1;
            mstr->dpCarp[i][c][x].p2 = -1;
            mstr->dpCarp[i][c][x].v1 = inf;
            mstr->dpCarp[i][c][x].v2 = inf;
         }
         
         cRes = c - mstr->getDemand( i );
         
         if (cRes < 0) 
            continue;
         
         for (x = 0; x < 2; x++)
         {
            for (j = nNodes; j >= 1; j--)
            {
               if ((j == i) || (localg->getEdgeFixed( j, i ) == 0))
                  continue;
               
               for (y = 0; y < 2; y++)
               {
                  if (mstr->dpCarp[j][cRes][y].p1 != i)
                  {
                     v = mstr->dpCarp[j][cRes][y].v1  +
                        mstr->dpDist[j][i].d[1-y][x] -
                        lpDualCarp[localg->getEdgeIndex( j, i )];
                  }
                  else
                     if (mstr->dpCarp[j][cRes][y].p2 != i)
                     {
                        v = mstr->dpCarp[j][cRes][y].v2  +
                           mstr->dpDist[j][i].d[1-y][x] -
                           lpDualCarp[localg->getEdgeIndex( j, i )];
                     }
                     else
                        break;
                     
                     if (v < mstr->dpCarp[i][c][x].v1)
                     {
                        mstr->dpCarp[i][c][x].y2 = mstr->dpCarp[i][c][x].y1;
                        mstr->dpCarp[i][c][x].p2 = mstr->dpCarp[i][c][x].p1;
                        mstr->dpCarp[i][c][x].v2 = mstr->dpCarp[i][c][x].v1;
                        mstr->dpCarp[i][c][x].y1 = y;
                        mstr->dpCarp[i][c][x].p1 = j;
                        mstr->dpCarp[i][c][x].v1 = v;
                     }
                     else
                        if (v < mstr->dpCarp[i][c][x].v2)
                        {
                           mstr->dpCarp[i][c][x].y2 = y;
                           mstr->dpCarp[i][c][x].p2 = j;
                           mstr->dpCarp[i][c][x].v2 = v;
                        }
               } // for (y = ...
            } // for (j = ...
         } // for (x = ...
      } // for (i = ...  
   } // for (c = ...
   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   
   // backtracking
   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   for (k = 1; k < nNodes; k++)
   {
      if (localg->getEdgeFixed( k, nDepot ) == 0)
         continue;
      
      for (w = 0; w < 2; w++)
      {
         x = w;
         j = k;
         c = vCap;
         
         reducedCost = mstr->dpCarp[j][c][x].v1           +
            mstr->dpDist[j][nDepot].d[1-x][0]  -
            lpDualCarp[localg->getEdgeIndex( j, nDepot )];
         
         //      fprintf( fout, "\nRC = %lf\n", reducedCost );
         
         if (reducedCost > -master->params->getIntEps())
            continue;
         
         CColumn newcol( master->params, mstr->getKCycleParam());
         
         newcol.setLB( 0.0 );
         newcol.setUB( LP_INFINITY );
         newcol.addElem( localg->getEdgeIndex( j, nDepot ), 1.0 );
         
         routecost = mstr->dpDist[j][nDepot].d[1-x][0];
         
         flag = false;
         while ((j != nDepot) && (flag == false))
         {
            i = j;
            y = x;
            if (mstr->dpCarp[i][c][y].p1 != i)
            {
               j = mstr->dpCarp[i][c][y].p1;
               x = mstr->dpCarp[i][c][y].y1;
            }
            else
               if (mstr->dpCarp[i][c][y].p2 != i)
               {
                  j = mstr->dpCarp[i][c][y].p2;
                  x = mstr->dpCarp[i][c][y].y2;
               }
               else
               {
                  flag = true;
                  break;
               }
               
               c -= mstr->getDemand( i );
               
               routecost += mstr->dpDist[j][i].d[1-x][y];
               newcol.addElem( localg->getEdgeIndex( j, i ), 1.0 );
               
               flag = (localg->getEdgeFixed( j, i ) == 0);
         }
         
         if (flag)
            continue;
         
         // Add thed column to pool and to lp
         newcol.setObj( routecost );
         addCol( newcol );
         addExpandedColToLP( newcol );
         
         nAdd++;
      } // for (w = ...
   } // for (k = ...
   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   
   //  fclose( fout );
   
   setSolSize( master->solver->getNColumns( locallp ) );
   
   //  master->solver->writeLPProblem( locallp, "teste.lp" );
   
   return nAdd;
}

//---------------------------------------------------------------------------//
//---------------------------------------------------------------------------//

int CVRPBBNode::fixByRedCost()
{
   int n; // number of vertex
   int m; // number of edges
   register int i, j, c, na, nar; // counters
   double inf; // infinity
   double v; // auxiliary variable
   int cRes; // auxiliary variable
   int b; // capacity of the vehicles
   double reducedCost; // reduced cost
   int nfixed; // number of edges fixed by red cost
   int nDepot; // number of the depot
   
   double minRedCost;


   const int maxitertrials = 30;
   int improveiter[2];
   improveiter[0] = -1;
   improveiter[1] = -1;


   inf=LP_INFINITY;
   
   n=localg->getN();
   m=localg->getM();
   b=((CVRPMaster*) master)->getVehicleCap();
   nfixed = 0;
   nDepot = ((CVRPMaster*) master)->getDepot();
   

   /** list of edges to be sorted -> determines order in which to try
    to fix by reduced cost. 
   Order in nondecreasining order of w */
   SortEdges *edges = NULL; 


   edges = new SortEdges[localg->getM()];
   
   if ( !edges )
      throw(LOWMEMORYEXCEPTION);

   // If no edge weights have been assigned before to try the
   // reduced cost fixation, then assign the edge costs
   // to it
   for (i=0;i<localg->getM();i++)
   {
      edges[i].n = i;

      if( getID() == 1 )
         edges[i].w = localg->getEdgeCost(i);
      else
         edges[i].w = edgeweights[i];

      edges[i].f = localg->getEdgeFixed(i);
      edges[i].fixedNow=0;
   }

   if( getID() == 1 )
   {
      edgeweights = new double[ localg->getM() ];
      memset( edgeweights, 0, localg->getM() * sizeof(double) );
   }
   
   qsort(edges,localg->getM(),sizeof(SortEdges),compEdges);

   for (na=0;na<localg->getM();na++)
   {
      if( getID() != 1 )
         if( na - improveiter[1] > maxitertrials )
            break;

      // nar == index of the edge
      nar = edges[na].n;
      
      // If edge is already fixed, no point in trying
      if (edges[na].f!=-1) 
         continue;
      
      // If value is positive, edge is in base, so can't fix by red cost
      if( lpSolExp[nar] > master->params->getIntEps() )
      {
         // If edge weight has never been set, then do it
         if( edgeweights[nar] == 0.0 )
            edgeweights[nar] = - getUB() + localg->getEdgeCost(nar);
         continue;
      }
      
      // If the edge is out of the depot, just do the regular dynamic
      // programming
      if( localg->getEdgeOrg(nar)==nDepot )
      {
         minRedCost = bestpath[ localg->getEdgeDst(nar) ];
      }
      else if( localg->getEdgeDst(nar) == nDepot )
      {
         minRedCost = bestpath[ localg->getEdgeOrg(nar) ];
      }
      else
      {
         // So now, we are treating all edges that are not incident to the
         // depot
         
         // Reset the matrices
         for (i=1;i<=n;i++)
         {
            vMatrix[i][0].v1=inf;
            vMatrix[i][0].v2=inf;
            predMatrix[i][0].p1=-1;
            predMatrix[i][0].p2=-1;
         }
         for (c=0;c<=b;c++)
         {
            vMatrix[nDepot][c].v1=0.0;
            vMatrix[nDepot][c].v2=0.0;
            predMatrix[1][c].p1=-1;
            predMatrix[1][c].p2=-1;
         }
         
         // Now do the dynamic programming without guarantee that we use the edge in question
         for (c=1;c<=b;c++)
         {
            for (i=1;i<n;i++)
            {
               // Reset this part of the matrices
               vMatrix[i][c].v1=inf;
               vMatrix[i][c].v2=inf;
               predMatrix[i][c].p1=-1;
               predMatrix[i][c].p2=-1;
               
               // If capacity left is < 0, continue
               cRes=c-((CVRPMaster*) master)->getDemand(i);
               if (cRes<0) continue;
               
               
               for (j=n;j>=1;j--)
               {
                  if (j==i) continue;
                  
                  if( (localg->getEdgeIndex(i,j)!=nar) &&
                     (localg->getEdgeFixed(i,j)==-1))
                  {
                     if( ( predMatrix[j][cRes].p1==-1 ) || 
                        ( predMatrix[j][cRes].p1!=i  ) || 
                        ( (predMatrix[j][cRes].p1==i) && (predMatrix[j][cRes].p2==-1)))
                     {
                        v=vMatrix[j][cRes].v1 - lpDualExp[localg->getEdgeIndex(i,j)];
                     }
                     else
                     {
                        v=vMatrix[j][cRes].v2 - lpDualExp[localg->getEdgeIndex(i,j)];
                     }
                  }
                  else
                     v=inf;
                  
                  if (v < vMatrix[i][c].v1)
                  {
                     vMatrix[i][c].v2=vMatrix[i][c].v1;
                     vMatrix[i][c].v1=v;
                     predMatrix[i][c].p2=predMatrix[i][c].p1;
                     predMatrix[i][c].p1=j;
                  }
                  else if (v < vMatrix[i][c].v2)
                  {
                     vMatrix[i][c].v2=v;
                     predMatrix[i][c].p2=j;
                  }
               }
            }
         }
         // End of the first DP-> Fills the whole matrix without using edge in question
         
         
         // Now do the second part of the DP
         // Matrix using the edge in question
         
         // Resets the matrices
         for (i=1;i<=n;i++)
         {
            vMatrix2[i][0].v1=inf;
            vMatrix2[i][0].v2=inf;
            predMatrix2[i][0].p1=-1;
            predMatrix2[i][0].p2=-1;
            
         }
         
         for (c=0;c<=b;c++)
         {
            vMatrix2[nDepot][c].v1=inf;
            vMatrix2[nDepot][c].v2=inf;
            predMatrix2[nDepot][c].p1=-1;
            predMatrix2[nDepot][c].p2=-1;
         }
         
         
         // Do the DP
         for (c=1;c<=b;c++)
         {
            for (i=1;i<n;i++)
            {
               // Resets the matrices
               vMatrix2[i][c].v1=inf;
               vMatrix2[i][c].v2=inf;
               predMatrix2[i][c].p1=-1;
               predMatrix2[i][c].p2=-1;
               
               // Makes sure there is enough capacity to hold node i
               cRes=c-((CVRPMaster*) master)->getDemand(i);
               if (cRes<0) continue;
               
               
               // Examine all candidates to connect to node i
               for (j=n;j>=1;j--)
               {
                  if (j==i) continue;
                  
                  
                  // Edge is fixed to zero
                  if( localg->getEdgeFixed(i,j) != -1 )
                  {
                     v=inf;
                  }
                  else if( localg->getEdgeIndex(i,j)!=nar )
                  {  
                     // The second matrix can only use a path within the second
                     //  matrix itself if the path has already used the edge nar
                     // This is made sure by initializing the reduced cost 
                     //   for the depot as infinity (as opposed to zero)
                     if( (predMatrix2[j][cRes].p1==-1) ||
                        (predMatrix2[j][cRes].p1!=i)  || 
                        ( (predMatrix2[j][cRes].p1==i) && (predMatrix2[j][cRes].p2==-1) ))
                     {
                        v=vMatrix2[j][cRes].v1 - lpDualExp[localg->getEdgeIndex(i,j)];
                     }
                     else
                     {
                        v=vMatrix2[j][cRes].v2 - lpDualExp[localg->getEdgeIndex(i,j)];
                     }
                  }
                  // This means that edge (i,j) is the edge in question
                  // So what is considered is the path not using edge
                  // (i,j) 
                  else 
                  {
                     // Get the path from matrix 1
                     if( (predMatrix[j][cRes].p1 == -1) || 
                        (predMatrix[j][cRes].p1 != i)  || 
                        ((predMatrix[j][cRes].p1 == i) && (predMatrix[j][cRes].p2==-1)))
                     {
                        v = vMatrix[j][cRes].v1 - lpDualExp[localg->getEdgeIndex(i,j)];
                     }
                     else
                     {
                        v = vMatrix[j][cRes].v2 - lpDualExp[localg->getEdgeIndex(i,j)];
                     }
                  }
                  
                  if (v < vMatrix2[i][c].v1)
                  {
                     vMatrix2[i][c].v2=vMatrix2[i][c].v1;
                     vMatrix2[i][c].v1=v;
                     predMatrix2[i][c].p2=predMatrix2[i][c].p1;
                     predMatrix2[i][c].p1=j;
                  }
                  else if (v < vMatrix2[i][c].v2)
                  {
                     vMatrix2[i][c].v2=v;
                     predMatrix2[i][c].p2=j;
                  }
                  
               }
            }
         }
         
         minRedCost=inf;
         
         for (j=1;j<=n;j++)
         {
            reducedCost=vMatrix2[j][b].v1 - lpDualExp[localg->getEdgeIndex(j,nDepot)];
            
            if (reducedCost < minRedCost)
               minRedCost = reducedCost;
         }

      }
      
      if (minRedCost==inf)
         continue;
      
      edgeweights[nar] = - getUB() + (minRedCost + getLB());
      if (minRedCost + getLB() > getUB() + master->params->getObjEps())
      {
         nfixed++;
         edges[na].fixedNow=1;

         edgeweights[nar] = -getUB();        
         
         improveiter[1] = improveiter[0];
         improveiter[0] = na;
      }
   }
   // End for edges
   
   if (nfixed>0)
   {     
      if( ! fixedEdges )
      {
         fixedEdges = new int[ nfixed ];
      }
      else
      {
         int* auxfEdges = fixedEdges;
         fixedEdges = new int[ nfixed + nFixedEdges ];
         memcpy( fixedEdges, auxfEdges, nFixedEdges * sizeof(int) );
         delete [] auxfEdges;
      }
      for (i=0;i<localg->getM();i++)
      {
         if (edges[i].fixedNow)
         {
            (((CVRPMaster*)master)->getGraph())->setEdgeFixed(edges[i].n,3);
            localg->setEdgeFixed(edges[i].n,3);

            fixedEdges[nFixedEdges++] = edges[i].n;
         }
      }
   }



   delete [] edges;
   
   return nfixed;
}

//---------------------------------------------------------------------------//
void CVRPBBNode::printll(FILE* &fout)
{
   int nRoute; // number of the route
   register int i,j; // counters
   int ncols; // number of LP cols
   int o,d; // auxiliary variables
   CColumn* auxcol; // auxiliary column
   CColElem* auxcolelem; // auxiliary column element
   int nRealNNZeros; // real number of non-zero of the column
   
   nRoute = 0;
   ncols = master->solver->getNColumns(locallp);
   
   fprintf( fout, "\nBEGINTREE\n" );
   
   for( i = maxArtificials; i < ncols; i++ )
   {
      if( lpSol[ i ] > master->params->getIntEps() )
      {
         nRoute++;
      }
   }
   
   fprintf( fout, "%d\n", nRoute );
   
   nRoute = 0;
   
   for( i = maxArtificials; i < ncols; i++ )
   {
      if( lpSol[ i ] > master->params->getIntEps() )
      {
         // Print the column
         auxcol = lpCols[i-maxArtificials];
         
         for (j=0; j < localg->getM(); j++)
            edgeX[j]=0;
         
         auxcolelem = auxcol->getFirstElem();
         nRealNNZeros = 0;
         
         while ( auxcolelem )
         {
            edgeX[auxcolelem->idx]+=auxcolelem->coeff;
            auxcolelem = auxcolelem->next;
         }
         
         for (j=0; j < localg->getM(); j++)
            if ( edgeX[j] > 0.0000001 )
               nRealNNZeros++;
            
            fprintf(fout,"%d %f %d\n", nRoute, lpSol[ i ], nRealNNZeros );
            
            for (j=0; j < localg->getM(); j++)
            {
               if ( edgeX[j] > 0.0000001 )
               {
                  o = localg->getEdgeOrg(j);
                  d = localg->getEdgeDst(j);
                  
                  if ( o == localg->getN() )
                     fprintf( fout, "0 %d %d\n", d, (int) edgeX[j] );
                  else if ( d == localg->getN() )
                     fprintf( fout, "0 %d %d\n", o, (int) edgeX[j] );
                  else
                     fprintf( fout, "%d %d %d\n", o, d, (int) edgeX[j] );
               }
            }
            nRoute++;
      }
   }
   
   fprintf( fout , "ENDTREE\n" );
}

//---------------------------------------------------------------------------//

int CVRPBBNode::exactCapacityCutGen()
{
   register int i; // counter
   int numcuts; // number of cuts added
   double *xsol; // solution of the problem
   LPptr subgenlp; // lp of the problem
   int nDepot; // number of the depot
   int k; // number of vehicles need to serve S
   int u, v; // origin and destiny of an edge
   char type; // type of column or row
   double lb, ub, cj, rhs; // auxiliary to the solver
   int contEdgesVar; // number of edge variables created
   int cont_lines; // counter of rows
   double sum_dem; // sum of all demands
   double xobj; // auxiliary objective value
   int stat; // status of the solver
   double violExCap; // violation of exact capacity
   
   // Try to allocate the solution array
   xsol = new double[localg->getN() + localg->getM() ];
   
   if ( !xsol )
      throw(LOWMEMORYEXCEPTION);
   
   // Initialize the mip
   subgenlp = NULL;
   master->solver->initLP(subgenlp);
   
   if ( !subgenlp )
      throw(LPVARIABLENOTINITIALIZED);
   
   // Set the display to zero
   master->solver->setIntParameter(LP_MIPDISPLAY,0);
   
   // Get the vertex the is the depot
   nDepot = ((CVRPMaster*)master)->getDepot();
   
   // Create vertex variables
   for( i=1; i <= localg->getN(); i++)
   {
      // cj has the objective function value
      cj = 0.0;
      lb = 0.0;
      
      if( i == nDepot )
         ub = 0.0;
      else
         ub = 1.0;
      
      type = 'B';
      
      // Create a new column
      master->solver->addNewCols(subgenlp,1,&cj,&lb,&ub,&type);
   }
   
   // Create arc variables
   contEdgesVar = 0;
   
   for( i = 0; i < localg->getM(); i++ )
   {
      u = localg->getEdgeOrg(i);
      v = localg->getEdgeDst(i);
      
      if( lpSolExp[ i ] < master->params->getIntEps() )
         continue;
      
      contEdgesVar++;
      
      cj = lpSolExp[ i ];
      
      lb = 0.0;
      ub = LP_INFINITY;
      type = 'C';
      
      // Create a new column
      master->solver->addNewCols(subgenlp,1,&cj,&lb,&ub,&type);
   }
   
   // Create constraints for each edge
   cont_lines=0;
   
   for( i = 0; i < localg->getM(); i++ )
   {
      if( lpSolExp[ i ] < master->params->getIntEps() )
         continue;
      
      u = localg->getEdgeOrg(i);
      v = localg->getEdgeDst(i);
      
      // Set the row RHS
      rhs = 0.0;
      
      // Set the row type
      type = 'G';
      
      // Create the new row
      master->solver->addNewRows(subgenlp,1,&rhs,&type);
      
      // Set the coefficients for the origin vertex
      master->solver->setCoef(subgenlp,cont_lines,u-1,1.0);
      
      // Set the coefficients for the dest vertex
      master->solver->setCoef(subgenlp,cont_lines,v-1,-1.0);
      
      // Set the coefficients for the edge
      master->solver->setCoef(subgenlp,cont_lines, localg->getN()+(cont_lines/2),1.0);
      
      cont_lines++;
      
      // Set the row RHS
      rhs = 0.0;
      
      // Set the row type
      type = 'G';
      
      // Create the new row
      master->solver->addNewRows(subgenlp,1,&rhs,&type);
      
      u = localg->getEdgeDst(i);
      v = localg->getEdgeOrg(i);
      
      // Set the coefficients for the origin vertex
      master->solver->setCoef(subgenlp,cont_lines,u-1,1.0);
      
      // Set the coefficients for the dest vertex
      master->solver->setCoef(subgenlp,cont_lines,v-1,-1.0);
      
      // Set the coefficients for the edge
      master->solver->setCoef(subgenlp,cont_lines, localg->getN()+(cont_lines/2),1.0);
      
      cont_lines++;
   }
   
   // Create the capacity constraint
   k = 0;
   rhs = k * ((CVRPMaster*) master)->getVehicleCap() + 1;
   type = 'G';
   
   master->solver->addNewRows(subgenlp,1,&rhs,&type);
   
   cont_lines++;
   
   // Sets the coefficients
   for( i = 1; i <= localg->getN(); i++ )
   {
      master->solver->setCoef(subgenlp,cont_lines-1,i-1,((CVRPMaster*) master)->getDemand(i));
   }
   
   sum_dem=(double)((CVRPMaster*) master)->getSumAllDemands();
   numcuts = 0;
   for( k = 0; k <= ceil( ( (double) ( sum_dem ) ) / (double) ((CVRPMaster*) master)->getVehicleCap() ); k++ )
   {
      // Set obj function coeff
      master->solver->setCoef(subgenlp,cont_lines-1,-1,k * ((CVRPMaster*) master)->getVehicleCap() + 1);
      
      type = 'L';
      lb = 1.0;
      
      // Solve the MIP
      master->solver->solveMIP(subgenlp);
      
      stat =  master->solver->getStatus(subgenlp);
      
      if ( stat != LP_MIPOPTIMAL && stat != LP_MIPOPTIMAL_TOL && stat != LP_MIPOPTIMAL_INFEAS )
         continue;
      
      // Get the MIP solution
      master->solver->getMIPSol(subgenlp,xsol,0,localg->getN()-1);
      
      // Get MIP obj value
      master->solver->getMIPObj(subgenlp,&xobj);
      
      lb = 0.0;
      if( xobj < (double) 2*(k +1.0) + master->params->getCutEps() )
      {
         // RHS
         rhs = (double) 2 * (k +1.0);
         
         CRow newrow( master->params );
         newrow.setType( G );
         
         for (i=0;i < localg->getM(); i++)
         {
            v = localg->getEdgeDst(i)-1;
            u = localg->getEdgeOrg(i)-1;
            
            if( fabs(xsol[v] - xsol[u]) > 1.0 - master->params->getIntEps() )
            {
               newrow.addElem(i, 1.0);
            }
         }
         
         newrow.setRHS( rhs );
         
         violExCap = 0.0;
         newrow.isViolated(lpSolExp,violExCap);
         
         // Cut is not enough violated
         if ( violExCap < master->params->getCutEps() )
            continue;
         
         // Cut violated
         numcuts++;
         
         addRow(newrow);
         addExpandedRowToLP(newrow);
      }
   }
   if ( numcuts >0 )
   {
      master->writeLog( true, "EXACT CAPACITY CUT ADDED. Total exact capacity cuts: %d\n\n", numcuts );
   }
   
   // Free memory
   if ( xsol )
      delete [] xsol;
   
   // Free the memory for the lp
   master->solver->freeLP(subgenlp);
   
   return numcuts;
}

//---------------------------------------------------------------------------//

void CVRPBBNode::expandSolution (double *lpe, double *lp)
{
   register int i; // counter
   int n;          // number of variables
   int m;          // number of edges
   
   CColumn*  columnAux;
   CColElem* auxelem; 
   
   n = master->solver->getNColumns( locallp );
   m = localg->getM();
   
   for (i = 0; i < m; i++)
      lpe[i] = 0.0;
   
   for (i = maxArtificials; i < n; i++)
   {
      if (lp[i] > master->params->getIntEps())
      {
         columnAux = lpCols[i - maxArtificials];
         
         for (auxelem = columnAux->getFirstElem(); auxelem; auxelem = auxelem->next)
            lpe[auxelem->idx] += lp[i] * (auxelem->coeff);
      }         
   }
}

//---------------------------------------------------------------------------//

void CVRPBBNode::expandPi (double *lpe, double *lp)
{
   register int i; // counter
   int numRows;    // number of rows created
   int m;          // number of edges
   
   CRow*     rowAux;     // auxiliary row variable
   CRowElem* rowAuxElem; // auxiliary row element variable
   
   numRows = master->solver->getNRows(locallp) - nMasterRows;
   m = localg->getM();
   
   for (i = 0; i < m; i++)
   {
      lpe[i] = -localg->getEdgeCost(i) + lp[localg->getEdgeOrg(i) - 1] + lp[localg->getEdgeDst(i) - 1];
   }
   
   for (i = 0; i < numRows; i++)
   {
      rowAux = lpRows[i];
      for (rowAuxElem = rowAux->getFirstElem(); rowAuxElem; rowAuxElem = rowAuxElem->next)
      {
         lpe[rowAuxElem->idx] += rowAuxElem->coeff * lp[i + nMasterRows ];
      }
   }


}

//---------------------------------------------------------------------------//

void CVRPBBNode::expandPi (double *lpe, double *lp, double *lpcarp)
{
   register int i; // counter
   int numRows;    // number of rows created
   int m;          // number of edges
   
   CRow*     rowAux;     // auxiliary row variable
   CRowElem* rowAuxElem; // auxiliary row element variable
   
   numRows = master->solver->getNRows(locallp) - nMasterRows;
   m = localg->getM();
   
   for (i = 0; i < m; i++)
   {
      lpcarp[i] = lp[localg->getEdgeOrg(i) - 1] + lp[localg->getEdgeDst(i) - 1];
      lpe[i]    = -localg->getEdgeCost(i) + lp[localg->getEdgeOrg(i) - 1] + lp[localg->getEdgeDst(i) - 1];
   }
   
   for (i = 0; i < numRows; i++)
   {
      rowAux = lpRows[i];
      
      for (rowAuxElem = rowAux->getFirstElem(); rowAuxElem; rowAuxElem = rowAuxElem->next)
      {
         lpcarp[rowAuxElem->idx] += rowAuxElem->coeff * lp[i + nMasterRows];
         lpe[rowAuxElem->idx]    += rowAuxElem->coeff * lp[i + nMasterRows];
      }
   }
}

//---------------------------------------------------------------------------//

int CVRPBBNode::carpBoundCut()
{
   register int i     = 0; // counter
   int ncuts = 0; // number of cuts added
   
   for (i = 0; i < localg->getM(); i++)
   {
      //    if (lpSolExp[i] < (1.0 - master->params->getIntEps()))
      if (lpSolExp[i] < (1.0 - 1e-3))
      {
         if (localg->getEdgeFixed( i ) == 1)
         {
            CRow newrow(master->params);
            
            newrow.addElem( i, 1.0 );
            newrow.setType( G );
            newrow.setRHS( 1.0 );
            addRow( newrow );
            addExpandedRowToLP( newrow );
            ncuts++;
         }
      }
   }
   
   if (ncuts > 0)
   {
      master->writeLog( true, "CARP BOUND CUTS ADDED. Total carp bound cuts: %d\n\n", ncuts );
   }
   
   if (ncuts > 0)
      setDualSize( master->solver->getNRows( locallp ) );
   
   return ncuts;
}

//---------------------------------------------------------------------------//

int CVRPBBNode::boundCut()
{
   register int i; // counter
   int ncuts; // number of cuts added
   int nDepot; // number of depot vertex
   
   int org, dst;
   
   ncuts=0;
   
   nDepot = ((CVRPMaster*) master)->getDepot();
   
   for (i=0;i<localg->getM();i++)
   {
      if ( lpSolExp[i] > 1.0 + master->params->getIntEps() )
      {
         if ( (localg->getEdgeOrg(i) == nDepot || localg->getEdgeDst(i)==nDepot) &&
            localg->canBeTwoEdge(i))
         {
            if (  lpSolExp[i] > 2.0 + master->params->getIntEps() )
            {
               CRow newrow(master->params);
               
               newrow.addElem(i,1.0);
               newrow.setType(L);
               newrow.setRHS(2.0);
               addRow(newrow);
               addExpandedRowToLP(newrow);
               ncuts++;
            }
            
         }
         else
         {
            CRow newrow(master->params);
            
            newrow.addElem(i,1.0);
            newrow.setType(L);
            newrow.setRHS(1.0);
            addRow(newrow);
            addExpandedRowToLP(newrow);
            ncuts++;
         }
      }
      
      if ( lpSolExp[i] > master->params->getIntEps() )
      {
         org = localg->getEdgeOrg(i);
         dst = localg->getEdgeDst(i);
         
         if( ((CVRPMaster*)master)->getDemand(org) + 
            ((CVRPMaster*)master)->getDemand(dst) > 
            ((CVRPMaster*)master)->getVehicleCap() )
         {
            CRow newrow(master->params);
            
            newrow.addElem(i,1.0);
            newrow.setType(E);
            newrow.setRHS(0.0);
            addRow(newrow);
            addExpandedRowToLP(newrow);
            ncuts++;
            fixzero = true;
         }
      }
   }
   
   if ( ncuts > 0 )
   {
      master->writeLog( true, "BOUND CUTS ADDED. Total bound cuts: %d\n\n", ncuts );
   }
   
   if (ncuts > 0)
      setDualSize(master->solver->getNRows(locallp));
   
   return ncuts;
}

//---------------------------------------------------------------------------//

int CVRPBBNode::separateColsFromPool(int maxcols)
{
   CColPoolElem* auxcolelem; // auxiliary col pool element
   CColElem *colAuxElem; // auxiliary column element variable
   CColSort *listCols; // list of columns to be ordered
   double redCost; // auxiliary reduced cost
   int nSort; // number of columns separated
   register int i; // counter
   CColumn* auxcol; // auxiliary column
   bool flagZero; // auxiliary flag
   
   listCols = NULL;
   nSort = 0;
   auxcol = NULL;
   
   if (master->getColPool()->getSize()>0)
      listCols=new CColSort[master->getColPool()->getSize()];
   else
      return 0;
   
   auxcolelem = master->getColPool()->getFirst();
   
   while( auxcolelem )
   {
      auxcol = auxcolelem->column;
      
      redCost = 0.0;
      if ((((CVRPMaster*)master)->getCarp()) && (((CVRPMaster*)master)->getLongo()))
      {
         //        redCost = auxcol->getRedCost( lpDualCarp ) + auxcol->getObj();
         redCost = auxcol->getRedCost( lpDualCarp );
      }
      else
         redCost = auxcol->getRedCost( lpDualExp );
      
      if( ( redCost > master->params->getObjEps() ) &&
         auxcol->getSepFromPool()  )
      {
         flagZero = false;
         
         for (colAuxElem=auxcol->getFirstElem(); colAuxElem; colAuxElem = colAuxElem->next)
            if (localg->getEdgeFixed(colAuxElem->idx)==0 || localg->getEdgeFixed(colAuxElem->idx)==3) 
               flagZero=true;
            
            if (flagZero)
            {
               auxcolelem = auxcolelem->next;
               continue;
            }
            
            listCols[nSort].elem = auxcol;
            listCols[nSort].value = redCost;
            nSort++;
      }
      auxcolelem = auxcolelem->next;
   }
   
   // Order by reduced cost
   if (nSort>0)
      qsort(listCols,nSort,sizeof(CColSort),compCols);
   
   if (nSort>maxcols)
      nSort=maxcols;
   
   for (i=0;i<nSort;i++)
   {
      addExpandedColToLP (*(listCols[i].elem));
      lpCols.push_back( listCols[i].elem );
      (listCols[i].elem)->increaseNLPs();
      
      listCols[i].elem=NULL;
   }
   
   if (listCols)
      delete[] listCols;
   
   if ( nSort > 0 )
   {
      setSolSize(master->solver->getNColumns(locallp));
   }
   
   return nSort;
}

//---------------------------------------------------------------------------//

void CVRPBBNode::addExpandedRowToLP(int *listIdx, double *listCoef, int nIdx, double aRHS, char aType)
{
   register int i; // counter
   int auxIdx; // auxiliary index
   double auxCoef; // auxiliary coefficient
   int nRows; // number of rows of the lp
   
   // Number of rows
   nRows = master->solver->getNRows(locallp);
   
   // Add the new row
   master->solver->addNewRows(locallp,1,&aRHS,&aType);
   
   for (i=0; i < nIdx; i++)
   {
      colHash->startIterator(listIdx[i]);
      
      while (colHash->getNext(listIdx[i],auxIdx,auxCoef))
      {
         master->solver->sumCoef(locallp, nRows, auxIdx, (auxCoef * listCoef[i]) );
      }
      
      if ( rowHash->getLast(listIdx[i]) != nRows )
         rowHash->add(listIdx[i],nRows,listCoef[i]);
      else
         rowHash->sumLast(listIdx[i],listCoef[i]);
   }
   
   if (aType == 'L')
      createArtificialColumn(nRows,1);
   else
      createArtificialColumn(nRows,0);
}

//---------------------------------------------------------------------------//

void CVRPBBNode::addExpandedRowToLP(CRow &aRow)
{
   int auxIdx; // auxiliary index
   double auxCoef; // auxiliary coefficient
   int nRows; // number of rows of the lp
   CRowElem *auxElem; // auxiliary row element
   double auxRHS; // auxiliary RHS
   char auxType; // auxiliary row type
   double *colsCoef; // coefficients of the columns
   register int i; // counter
   
   cronoAddCutsToLP.start();
   
   if ( aRow.getType() == L )
      auxType = 'L';
   else if ( aRow.getType() == G )
      auxType = 'G';
   else
      auxType = 'E';
   
   auxRHS = aRow.getRHS();
   
   // Number of rows
   nRows = master->solver->getNRows(locallp);
   
   // Add the new row
   master->solver->addNewRows(locallp,1,&auxRHS,&auxType);
   
   auxElem = aRow.getFirstElem();
   
   colsCoef = new double[lpCols.size()];
   
   for (i=0;i<(int)lpCols.size();i++)
      colsCoef[i]=0;
   
   while ( auxElem )
   {
      colHash->startIterator(auxElem->idx);
      while (colHash->getNext(auxElem->idx,auxIdx,auxCoef))
      {
         colsCoef[auxIdx-maxArtificials]+=(auxCoef * auxElem->coeff);
      }
      
      if ( rowHash->getLast(auxElem->idx) != nRows )
         rowHash->add(auxElem->idx,nRows,auxElem->coeff);
      else
         rowHash->sumLast(auxElem->idx,auxElem->coeff);
      
      auxElem = auxElem->next;
   }
   
   for (i=0;i<(int)lpCols.size();i++)
   {
      if ( (colsCoef[i] < -master->params->getIntEps()) ||
         (colsCoef[i] > master->params->getIntEps()) )
         master->solver->setCoef(locallp,nRows,i+maxArtificials,colsCoef[i]);
   }
   
   delete[] colsCoef;
   
   if (auxType == 'L')
      createArtificialColumn(nRows,1);
   else
      createArtificialColumn(nRows,0);
   
   cronoAddCutsToLP.stop();
}

//---------------------------------------------------------------------------//

void CVRPBBNode::addExpandedColToLP(int *listIdx, double *listCoef, int nIdx, double aObj, 
                                    double aLB, double aUB)
{
   register int i; // counter
   int nCols; // number of columns
   int auxIdx; // auxiliary index
   double auxCoef; // auxiliary coefficient
   
   // Number of columns
   nCols = master->solver->getNColumns(locallp);
   
   // Add the new column
   master->solver->addNewCols(locallp,1,&aObj,&aLB,&aUB,NULL);
   
   for (i=0;i<nIdx;i++)
   {
      rowHash->startIterator(listIdx[i]);
      
      master->solver->sumCoef(locallp,localg->getEdgeOrg(listIdx[i])-1,nCols,listCoef[i]);
      master->solver->sumCoef(locallp,localg->getEdgeDst(listIdx[i])-1,nCols,listCoef[i]);
      
      while(rowHash->getNext(listIdx[i],auxIdx,auxCoef))
      {
         master->solver->sumCoef(locallp,auxIdx,nCols,(auxCoef*listCoef[i]));
      }
      
      if ( colHash->getLast(listIdx[i]) != nCols )
         colHash->add(listIdx[i],nCols,listCoef[i]);
      else
         colHash->sumLast(listIdx[i],listCoef[i]);
   }
}

//---------------------------------------------------------------------------//

void CVRPBBNode::addExpandedColToLP(CColumn &aCol)
{
   int nCols; // number of columns
   int auxIdx; // auxiliary index
   double auxCoef; // auxiliary coefficient
   double auxLB, auxUB; // auxiliary upper and lower bounds
   double auxObj; // auxiliary objective function coeff
   CColElem *auxElem; // auxiliary column element
   register int i; // counter
   double *rowsCoef = NULL; // coeficients of each row
   
   double routeDem = 0.0;

   cronoAddColsToLP.start();
   
   // Number of columns
   nCols  = master->solver->getNColumns( locallp );
   
   auxLB  = aCol.getLB();
   auxUB  = aCol.getUB();
   auxObj = aCol.getObj();
   
   // Add the new column
   master->solver->addNewCols( locallp, 1, &auxObj, &auxLB, &auxUB, NULL);
   
   auxElem = aCol.getFirstElem();
   
   for (i = 1; i <= localg->getN(); i++)
      nodeP[i] = 0;
   
   if( lpRows.size() > 0 )
      rowsCoef = new double[lpRows.size()];
   
   for (i = 0; i < (int) lpRows.size(); i++)
      rowsCoef[i] = 0;
   
   while (auxElem)
   {

      rowHash->startIterator(auxElem->idx);
      
      nodeP[localg->getEdgeOrg(auxElem->idx)]++;
      nodeP[localg->getEdgeDst(auxElem->idx)]++;
      
      while (rowHash->getNext( auxElem->idx, auxIdx, auxCoef ))
      {
         rowsCoef[auxIdx-nMasterRows] += (auxCoef * auxElem->coeff);
      }

      
      if (colHash->getLast(auxElem->idx) != nCols)
         colHash->add( auxElem->idx, nCols, auxElem->coeff );
      else
         colHash->sumLast( auxElem->idx, auxElem->coeff );

      routeDem += 0.5 * (double)((CVRPMaster*) master)->getDemand( localg->getEdgeOrg(auxElem->idx) ) + 
         0.5 * (double)((CVRPMaster*) master)->getDemand( localg->getEdgeDst(auxElem->idx) );

      
      auxElem = auxElem->next;
   }
   
   for (i = 1; i <= localg->getN(); i++)
      if (nodeP[i])
         master->solver->setCoef( locallp, i - 1, nCols, nodeP[i] );
      

   for (i = 0; i < (int) lpRows.size(); i++)
   {
      if ((rowsCoef[i] < -master->params->getIntEps()) || 
         (rowsCoef[i] >  master->params->getIntEps()))
         master->solver->setCoef( locallp, i + nMasterRows, nCols,rowsCoef[i] );
   }
   
   if( rowsCoef )
      delete [] rowsCoef;
   
   cronoAddColsToLP.stop();
}

//---------------------------------------------------------------------------//

int CVRPBBNode::separateRowsFromPool(int maxrows)
{
   CCutPoolElem *auxcutelem; // auxiliary element of cut pool
   int m; // number of edges
   CRowSort *listRows; // list of rows to be sorted
   int nSort; // number of rows to be sorted
   double viol; // violation of the row
   register int i; // counter
   CRow* auxrow; // auxiliary row
   double violCut; // violation of the cut
   int nAdd; // number of cuts added
   
   m = localg->getM();
   listRows = NULL;
   nSort = 0;
   viol = 0.0;
   auxrow = NULL;
   nAdd = 0;
   
   if (master->getCutPool()->getSize()>0)
      listRows = new CRowSort[master->getCutPool()->getSize()];
   else 
      return 0;
   
   auxcutelem = master->getCutPool()->getFirst();
   
   while( auxcutelem )
   {
      auxrow= auxcutelem->row;
      
      if( auxrow->getSepFromPool() && auxrow->isViolated( lpSolExp, viol ) )
      {
         listRows[nSort].elem = auxrow;
         listRows[nSort].value = viol;
         nSort++;
      }
      auxcutelem = auxcutelem->next;
   }
   
   // Order by violation
   if (nSort>0)
      qsort(listRows,nSort,sizeof(CRowSort),compRows);
   
   if (nSort>maxrows)
      nSort=maxrows;
   
   // Add separate rows to LP
   for (i=0;i<nSort;i++)
   {
      violCut = 0.0;
      listRows[i].elem->isViolated(lpSolExp,violCut);
      
      // Cut is not enough violated
      if ( violCut < 0.01 )
         break;
      
      addExpandedRowToLP (*(listRows[i].elem));
      nAdd++;
      lpRows.push_back( listRows[i].elem );
      (listRows[i].elem)->increaseNLPs();
      listRows[i].elem=NULL;
   }
   
   if (listRows)
      delete[] listRows;
   
   if ( nAdd > 0 )
      setDualSize(master->solver->getNRows(locallp));
   
   return nAdd;
}

//---------------------------------------------------------------------------//

void CVRPBBNode::addBasisToLP()
{
   int* cols; // cols of the basis
   int* rows; // rows of the basis
   int ncols; // number of columns of the basis
   int nrows; // number of rows of the basis
   register int i; // counter
   CColumn *auxCol; // auxiliary column
   CRow *auxRow; // auxiliary row
   int cntbasic; // auxiliary number of basic variables
   int posRows, posCols; // auxiliary positions
   
   cols = NULL;
   rows = NULL;
   ncols = basis.getNMasterCols() + basis.getBasicCols()->size();
   
   cols = new int[ncols];
   cntbasic = 0;
   
   if ( !cols )
      throw(LOWMEMORYEXCEPTION);
   
   posRows = 0;
   posCols = 0;
   
   for(i = 0; i < ncols; i++)
   {
      if( i < basis.getNMasterCols() )
      {
         cols[i] = basis.getColStat(i);
         
         if( i >= basis.getNStaticCols() )
         {
            if( cntbasic < basis.getNBasicDinamic() )
            {
               cols[i] = BASIC_VAR;
               cntbasic++;
            }
            else
            {  
               cols[i] = LOWER_ROW;
            }
         }
      }
      else
      {
         auxCol = (basis.getBasicCols())->at(posCols);
         addExpandedColToLP((*auxCol)); 
         
         lpCols.push_back(auxCol);
         auxCol->increaseNLPs();
         
         cols[i] = BASIC_VAR;
         posCols++;
      }
   }
   
   nrows = basis.getNMasterRows() + basis.getBasicRows()->size();
   
   rows = new int[nrows];
   
   if ( !rows )
      throw(LOWMEMORYEXCEPTION);
   
   for(i = 0; i < nrows; i++)
   {
      if( i < basis.getNMasterRows() )
      {
         rows[i] = basis.getRowStat(i);
      }
      else
      {
         auxRow = (basis.getBasicRows())->at(posRows);
         addExpandedRowToLP( (*auxRow) ); 
         
         lpRows.push_back(auxRow);
         auxRow->increaseNLPs();
         rows[i] = LOWER_ROW;
         posRows++;
      }
   }
   
   master->solver->copyBasis(locallp,cols,rows);
   
   setSolSize(master->solver->getNColumns(locallp));
   setDualSize(master->solver->getNRows(locallp));
   
   if( cols )
      delete [] cols;
   if( rows )
      delete [] rows;
}

//---------------------------------------------------------------------------//

void CVRPBBNode::initializeNode()
{
   initialized = true;
   
   // Artificial variables
   maxArtificials=2000;
#define MAXCUTS 2000
   
   if( ((CVRPMaster*) master)->getDoStab() )
   {
      nStab1 = localg->getN();
      maxStab1 = nStab1 + MAXCUTS;
      nStab2 = maxStab1;
      maxStab2 = nStab2 + MAXCUTS;
      nArtificials = maxStab2;
      maxArtificials = nArtificials + MAXCUTS;
   }
   else
   {
      nArtificials=localg->getN();
      maxArtificials = nArtificials + MAXCUTS;
   }
   
   // Allocate memory
   lpSolExp=new double[localg->getM()];
   
   if (!lpSolExp)
      throw(LOWMEMORYEXCEPTION);
   
   lpDualExp=new double[localg->getM()];
   
   if (!lpDualExp)
      throw(LOWMEMORYEXCEPTION);
   
   if ((((CVRPMaster*)master)->getCarp()) && (((CVRPMaster*)master)->getLongo()))
   {
      lpDualCarp=new double[localg->getM()];
      
      if (!lpDualCarp)
         throw(LOWMEMORYEXCEPTION);
   }
   
   master->solver->initLP(locallp);
   
   if ( !locallp )
      throw(LPVARIABLENOTINITIALIZED);
   
   vMatrix = ((CVRPMaster*)master)->getVMatrix();
   predMatrix = ((CVRPMaster*)master)->getPredMatrix();
   vMatrix2 = ((CVRPMaster*)master)->getVMatrix2();
   predMatrix2 = ((CVRPMaster*)master)->getPredMatrix2();
   
   edgeHead = ((CVRPMaster*)master)->getEdgeHead();
   edgeTail = ((CVRPMaster*)master)->getEdgeTail();
   edgeX = ((CVRPMaster*)master)->getEdgeX();
   listS = ((CVRPMaster*)master)->getListS();
   nodeP = ((CVRPMaster*)master)->getNodeP();
   pathDinProg = ((CVRPMaster*)master)->getPathDinProg();
   
   rowHash = new CVRPHash(localg->getM());
   
   if ( !rowHash )
      throw(LOWMEMORYEXCEPTION);
   
   colHash = new CVRPHash(localg->getM());
   
   if ( !colHash )
      throw(LOWMEMORYEXCEPTION);
   
   // Initialize the cuts structure
   CMGR_CreateCMgr(&myCutsCMP,100);
   CMGR_CreateCMgr(&myOldCutsCMP,100);
   CMGR_CreateCMgr(&setsCMP,localg->getN()-1);
   
   firstTime = true;
   
   // Which cuts separate
   //   sepCapCuts = true;
   //   sepFCCuts = true;
   //   sepMSCuts = true;
   //   sepHypCuts = true;
   //   sepCombCuts = true;
   //   sepGLMSCuts = true;
   
   sepCapCuts = true;
   sepFCCuts = true;
   sepMSCuts = false;
   sepHypCuts = false;
   sepCombCuts = true;
   sepGLMSCuts = false;



   bestpath = new double[ localg->getN() + 1 ];
}

//---------------------------------------------------------------------------//

void CVRPBBNode::capacityCuts(int noOfEdges, int maxNoOfCuts, char &feasible, double &maxViol)
{
   CAPSEP_SeparateCapCuts(localg->getN()-1,
      ((CVRPMaster*)master)->getAllDemands(),
      ((CVRPMaster*)master)->getVehicleCap(),
      noOfEdges,
      edgeTail,
      edgeHead,
      edgeX,
      myOldCutsCMP,
      maxNoOfCuts,
      master->params->getIntEps(),
      &feasible,
      &maxViol,
      myCutsCMP);
}

//---------------------------------------------------------------------------//

void CVRPBBNode::multiStarCuts(int noOfEdges, int maxNoOfCuts, double &maxViol)
{
   MSTARSEP_SeparateMultiStarCuts(localg->getN()-1,
      ((CVRPMaster*)master)->getAllDemands(),
      ((CVRPMaster*)master)->getVehicleCap(),
      noOfEdges,
      edgeTail,
      edgeHead,
      edgeX,
      myOldCutsCMP,
      maxNoOfCuts,
      &maxViol,
      myCutsCMP);
}

//---------------------------------------------------------------------------//

void CVRPBBNode::framedCapacityCuts(int noOfEdges, int maxNoOfTreeNodes, int maxNoOfCuts, double &maxViol)
{
   FCISEP_SeparateFCIs(localg->getN()-1,
      ((CVRPMaster*)master)->getAllDemands(),
      ((CVRPMaster*)master)->getVehicleCap(),
      noOfEdges,
      edgeTail,
      edgeHead,
      edgeX,
      myOldCutsCMP,
      maxNoOfTreeNodes,
      maxNoOfCuts,
      &maxViol,
      myCutsCMP);
}

//---------------------------------------------------------------------------//

void CVRPBBNode::combCuts(int qMin, int noOfEdges, int maxNoOfCuts, double &maxViol)
{
   COMBSEP_SeparateCombs(localg->getN()-1,
      ((CVRPMaster*)master)->getAllDemands(),
      ((CVRPMaster*)master)->getVehicleCap(),
      qMin,
      noOfEdges,
      edgeTail,
      edgeHead,
      edgeX,
      maxNoOfCuts,
      &maxViol,
      myCutsCMP);
   
}

//---------------------------------------------------------------------------//

void CVRPBBNode::hypotourCuts(int noOfEdges, int maxNoOfCuts, double &maxViol)
{
   HTOURSEP_SeparateHTours(localg->getN()-1,
      ((CVRPMaster*)master)->getAllDemands(),
      ((CVRPMaster*)master)->getVehicleCap(),
      noOfEdges,
      edgeTail,
      edgeHead,
      edgeX,
      myOldCutsCMP,
      maxNoOfCuts,
      &maxViol,
      myCutsCMP);
}

//---------------------------------------------------------------------------//

void CVRPBBNode::generalizedLargeMSCuts(int noOfEdges, int *custList, int &custListSize, double maxViol)
{
   GLMSEP_SeparateGLM(localg->getN()-1,
      ((CVRPMaster*)master)->getAllDemands(),
      ((CVRPMaster*)master)->getVehicleCap(),
      noOfEdges,
      edgeTail,
      edgeHead,
      edgeX,
      custList,
      &custListSize,
      &maxViol);
}

//---------------------------------------------------------------------------//

int CVRPBBNode::addCapSepCutsToLP(int custListSize, int *custList)
{
   register int i,j,k,l; // counters
   int nAddCuts; // number of cuts added
   int nCapCuts; // number of capacity cuts
   int nMSCuts; // number of multistar cuts
   int nFCCuts; // number of framed capacity cuts
   int nCombCuts; // number of comb cuts
   int nHypCuts; // number of hypotour cuts
   int nGLMSCuts; // number of gen large multistar cuts
   int listSize; // size of S set
   int cA, cB, cL; // coefficients of multistar cut
   int maxIdx, minIdx; // auxiliary for FC cuts
   int maxNumberOfTeeth; // maximum number of teeth in the cuts separated
   bool firstComb; // tells if the cut is the first comb
   short **inTooth; // auxiliary for combs
   double rhsGLMS; // auxiliary RHS for gen large MS cuts
   double violCut; // violation of the cut
   
   nAddCuts = 0;
   nCapCuts = 0;
   nMSCuts = 0;
   nFCCuts = 0;
   nCombCuts = 0;
   nHypCuts = 0;
   nGLMSCuts = 0;
   
   firstComb = true;
   inTooth = NULL;
   
   for (i=0; i < myCutsCMP->Size; i++)
   {
      if ( myCutsCMP->CPL[i]->CType == CMGR_CT_CAP ) // capacity cuts
      {
         listSize = 0;
         
         for (j=1; j <= myCutsCMP->CPL[i]->IntListSize; j++)
         {
            listS[++listSize] = myCutsCMP->CPL[i]->IntList[j];
         }
         
         // Add cut to lp
         CRow newrow(master->params);
         
         for ( j=1; j < listSize ; j++ )
         {
            for ( k = j+1; k<=listSize; k++ )
            {
               newrow.addElem(localg->getEdgeIndex(listS[j],listS[k]),1.0);
            }
         }
         
         newrow.setType(L);
         newrow.setRHS(myCutsCMP->CPL[i]->RHS);
         
         violCut = 0.0;
         
         newrow.isViolated(lpSolExp,violCut);
         
         if ( violCut < master->params->getCutEps() )
            continue;
         
         addRow(newrow);
         addExpandedRowToLP(newrow);
         nAddCuts++;
         nCapCuts++;
      }
      else if ( myCutsCMP->CPL[i]->CType == CMGR_CT_MSTAR ) // multistar cuts
      {
         // Coefficients of the cut
         cA = myCutsCMP->CPL[i]->A;
         cB = myCutsCMP->CPL[i]->B;
         cL = myCutsCMP->CPL[i]->L;
         
         // Add cut to lp
         CRow newrow(master->params);
         
         newrow.setType(G);
         newrow.setRHS(cL);
         
         // Nucleus
         for (j=1; j <= localg->getN(); j++)
            listS[j] = 0;
         
         for (j=1; j <= myCutsCMP->CPL[i]->IntListSize; j++)
            listS[myCutsCMP->CPL[i]->IntList[j]]=1;
         
         for (j=1; j<=myCutsCMP->CPL[i]->IntListSize; j++)
         {
            for (k=1; k <= localg->getN(); k++)
               if (k != myCutsCMP->CPL[i]->IntList[j] && listS[k]==0)
                  newrow.addElem(localg->getEdgeIndex(myCutsCMP->CPL[i]->IntList[j],k),cB);
         }
         
         // Satellites and connectors
         for (j=1; j<=myCutsCMP->CPL[i]->ExtListSize; j++)
         {
            for (k=1; k<=myCutsCMP->CPL[i]->CListSize; k++)
               if ( myCutsCMP->CPL[i]->ExtList[j] != myCutsCMP->CPL[i]->CList[k] )
                  newrow.addElem(localg->getEdgeIndex(myCutsCMP->CPL[i]->ExtList[j],myCutsCMP->CPL[i]->CList[k]),-cA);
         }
         
         violCut = 0.0;
         newrow.isViolated(lpSolExp,violCut);
         
         if ( violCut < master->params->getCutEps() )
            continue;
         
         addRow(newrow);
         addExpandedRowToLP(newrow);
         
         nAddCuts++;
         nMSCuts++;
      }
      else if ( myCutsCMP->CPL[i]->CType == CMGR_CT_FCI ) // framed capacity cuts
      {
         // Subsets of the cut
         for (j=1; j <= localg->getN(); j++)
            listS[j] = 0;
         
         maxIdx = 0;
         minIdx = 0;
         
         for (j=1; j <= myCutsCMP->CPL[i]->ExtListSize; j++)
         {
            minIdx = maxIdx + 1;
            maxIdx = minIdx + myCutsCMP->CPL[i]->ExtList[j] - 1;
            
            for (k=minIdx; k <= maxIdx; k++)
               listS[myCutsCMP->CPL[i]->IntList[k]] = j;
         }
         
         CRow newrow(master->params);
         
         newrow.setRHS(myCutsCMP->CPL[i]->RHS);
         newrow.setType(G);
         
         // Edges of the frame
         for (j=1; j <= localg->getN(); j++)
         {
            if ( listS[j] != 0 )
            {
               for (k=1;k <= localg->getN(); k++)
               {
                  if ( k != j && listS[k]==0 )
                     newrow.addElem(localg->getEdgeIndex(j,k),1.0);
               }
            }
         }
         
         // Edges of the subsets
         for (j=1; j <= myCutsCMP->CPL[i]->ExtListSize; j++)
         {
            for (k=1;k <= localg->getN(); k++)
            {
               if ( listS[k] == j )
               {
                  for (l=1;l <= localg->getN(); l++)
                  {
                     if ( l != k && listS[l]!=j)
                        newrow.addElem(localg->getEdgeIndex(k,l),1.0);
                  }
               }
            }
            
         }
         
         violCut = 0.0;
         newrow.isViolated(lpSolExp,violCut);
         
         if ( violCut < master->params->getCutEps() )
            continue;
         
         addRow(newrow);
         addExpandedRowToLP(newrow);
         
         nAddCuts++;
         nFCCuts++;
      }
      else if ( myCutsCMP->CPL[i]->CType == CMGR_CT_STR_COMB ) // comb cuts
      {
         if ( firstComb ) // allocate memory for the combs
         {
            firstComb = false;
            maxNumberOfTeeth = 0;
            
            for (j=0; j < myCutsCMP->Size; j++)
            {
               if ( myCutsCMP->CPL[j]->CType == CMGR_CT_STR_COMB )
               {
                  if ( myCutsCMP->CPL[j]->Key > maxNumberOfTeeth )
                     maxNumberOfTeeth = myCutsCMP->CPL[j]->Key;
               }
            }
            
            inTooth = new short*[localg->getN()+1];
            
            if ( !inTooth )
               throw(LOWMEMORYEXCEPTION);
            
            for (j=0; j <localg->getN()+1; j++)
               inTooth[j] = new short[maxNumberOfTeeth+1];
         }
         
         for (j=1; j <=localg->getN(); j++)
         {
            for (k=0; k <= myCutsCMP->CPL[i]->Key; k++)
               inTooth[j][k]=0;
         }
         
         for (j=1; j <= myCutsCMP->CPL[i]->IntListSize; j++)
         {
            inTooth[myCutsCMP->CPL[i]->IntList[j]][0] = 1;
         }
         
         for (j=1; j <= myCutsCMP->CPL[i]->Key; j++)
         {
            minIdx = myCutsCMP->CPL[i]->ExtList[j];
            
            if ( j == myCutsCMP->CPL[i]->Key )
               maxIdx = myCutsCMP->CPL[i]->ExtListSize;
            else
               maxIdx = myCutsCMP->CPL[i]->ExtList[j+1] - 1;
            
            for (k=minIdx; k <= maxIdx; k++)
            {
               inTooth[myCutsCMP->CPL[i]->ExtList[k]][j] = 1;
            }
         }
         
         // Add cuts to lp
         CRow newrow(master->params);
         
         newrow.setRHS( myCutsCMP->CPL[i]->RHS );
         newrow.setType( G );
         
         // Get the handle
         for (j=1; j <= localg->getN(); j++ )
         {
            if ( inTooth[j][0] == 1 )
            {
               for (k=1; k <= localg->getN(); k++)
               {
                  if ( k!=j && inTooth[k][0]!=1 )
                     newrow.addElem(localg->getEdgeIndex(j,k),1.0);
               }
            }
         }
         
         // Get the teeths
         for (j=1; j <= myCutsCMP->CPL[i]->Key; j++)
         {
            for (k=1; k <=localg->getN(); k++)
            {
               if (inTooth[k][j]==1)
               {
                  for (l=1; l <= localg->getN(); l++)
                  {
                     if ( l!=k && inTooth[l][j]!=1 )
                        newrow.addElem(localg->getEdgeIndex(k,l),1.0);
                  }
               }
            }
         }
         
         violCut = 0.0;
         newrow.isViolated(lpSolExp,violCut);
         
         if ( violCut < master->params->getCutEps() )
            continue;
         
         addRow(newrow);
         addExpandedRowToLP(newrow);
         
         nCombCuts++;
         nAddCuts++;
      }
      else if ( myCutsCMP->CPL[i]->CType == CMGR_CT_TWOEDGES_HYPOTOUR ) // hypotour cuts
      {
         CRow newrow(master->params);
         
         newrow.setRHS(myCutsCMP->CPL[i]->RHS);
         newrow.setType(L);
         
         for (j=1; j <= myCutsCMP->CPL[i]->IntListSize; j++)
         {
            newrow.addElem(localg->getEdgeIndex(myCutsCMP->CPL[i]->IntList[j],
               myCutsCMP->CPL[i]->ExtList[j]),
               myCutsCMP->CPL[i]->CoeffList[j]);
         }
         
         violCut = 0.0;
         newrow.isViolated(lpSolExp,violCut);
         
         if ( violCut < master->params->getCutEps() )
            continue;
         
         addRow(newrow);
         addExpandedRowToLP(newrow);
         
         nHypCuts++;
         nAddCuts++;
      }
   }
   
   // Add generalize large multistar cuts
   if ( custListSize > 0 )
   {
      CRow newrow(master->params);
      
      newrow.setType(L);
      rhsGLMS = ((CVRPMaster*)master)->getVehicleCap()*custListSize;
      
      for (i=1;i<=localg->getN();i++)
      {
         listS[i]=0;
      }
      
      // Calculate RHS
      for (i=1; i <= custListSize; i++)
      {
         rhsGLMS-= ((CVRPMaster*)master)->getDemand(custList[i]);
         listS[ custList[i] ] =1;
      }
      
      newrow.setRHS(rhsGLMS);
      
      // First part of the cut
      for (i=1; i < custListSize; i++)
      {
         for (j=i+1; j <= custListSize; j++)
         {
            if ( custList[i] != custList[j] )
            {
               newrow.addElem(localg->getEdgeIndex(custList[i],custList[j]),
                  ((CVRPMaster*)master)->getVehicleCap());
            }
         }
      }
      
      // Second part of the cut
      for (i=1;i <= localg->getN(); i++)
      {
         if ( listS[i]==1 )
         {
            for (j=1;j <= localg->getN(); j++)
            {
               if ( listS[j] != 1 )
               {
                  //newrow.addElem(localg->getEdgeIndex(i,j),
                  //               ((CVRPMaster*)master)->getDemand(custList[i]));
                  newrow.addElem( localg->getEdgeIndex(i,j),
		                   ((CVRPMaster*)master)->getDemand(j));
               }
            }
         }
      }
      
      violCut = 0.0;
      newrow.isViolated(lpSolExp,violCut);
      
      if ( violCut >= master->params->getCutEps() )
      {
         
         addRow(newrow);
         addExpandedRowToLP(newrow);
         
         nGLMSCuts++;
         nAddCuts++;
      }
   }
   
   // Store old cuts
   for (i=0;i < myCutsCMP->Size; i++)
   {
      CMGR_MoveCnstr(myCutsCMP,myOldCutsCMP,i,0);
   }
   myCutsCMP->Size = 0;
   
   // Free memory for comb
   if ( inTooth )
   {
      for (i=0; i < localg->getN()+1; i++)
         if ( inTooth )
            delete[] inTooth[i];
         
         delete[] inTooth;
   }
   
   if ( nCapCuts > 0 )
   {
      master->writeLog( true, "CAPACITY CUTS ADDED. Total capacity cuts: %d\n\n", nCapCuts );
   }
   
   if ( nMSCuts > 0 )
   {
      master->writeLog( true, "MULTISTAR CUTS ADDED. Total multistar cuts: %d\n\n", nMSCuts);
   }
   
   if ( nFCCuts > 0 )
   {
      master->writeLog( true, "FRAMED CAPACITY CUTS ADDED. Total framed capacity cuts: %d\n\n", nFCCuts);
   }
   
   if ( nCombCuts > 0 )
   {
      master->writeLog( true, "STRENGHENED COMB CUTS ADDED. Total strengthened comb cuts: %d\n\n", nCombCuts);
   }
   
   if ( nHypCuts > 0 )
   {
      master->writeLog( true, "HYPOTOUR CUTS ADDED. Total hypotour cuts: %d\n\n", nHypCuts);
   }
   
   if ( nGLMSCuts > 0 )
   {
      master->writeLog( true, "GL MULTISTAR CUTS ADDED. Total GL multistar cuts: %d\n\n", nGLMSCuts);
   }
   
   return nAddCuts;
}







//---------------------------------------------------------------------------//

int CVRPBBNode::eliminateLPRows(int maxRows)
{
   int nDelRows; // number of rows eliminated from the LP
   RowColElim *rowsDel; // possible rows to be eliminated
   int sizeRowsDel; // number of rows possible to be eliminated
   register int i; // counter
   int *delstat; // deleted status of each row
   int nRows, nCols; // number of rows and columns
   int *bCols, *bRows; // number of columns and rows of the basis
   
   nRows = master->solver->getNRows(locallp);
   nCols = master->solver->getNColumns(locallp);
   
   rowsDel = new RowColElim[nRows];
   
   if ( !rowsDel )
      throw(LOWMEMORYEXCEPTION);
   
   delstat = new int[nRows];
   
   if ( !delstat )
      throw(LOWMEMORYEXCEPTION);
   
   bCols = new int[nCols];
   
   if ( !bCols )
      throw(LOWMEMORYEXCEPTION);
   
   bRows = new int[nRows];
   
   if ( !bRows )
      throw(LOWMEMORYEXCEPTION);
   
   // Get the basis variables
   master->solver->getBasis(locallp,bCols,bRows);
   sizeRowsDel = 0;
   
   // Get rows possible to be eliminated
   for (i=0; i < nRows; i++)
   {
      delstat[i] = 0;
      
      if ( i >= localg->getN())
      {
         if ( bRows[i] != LOWER_ROW)
         {
            rowsDel[sizeRowsDel].idx = i;
            rowsDel[sizeRowsDel].value = lpRows[i-localg->getN()]->getNNZeros();
            sizeRowsDel++;
         }
      }
   }
   
   if ( sizeRowsDel > 0 )
   {
      // Order by non-zero number
      qsort(rowsDel,sizeRowsDel,sizeof(RowColElim),compRowColElim);
      
      if (sizeRowsDel >= maxRows)
         nDelRows = maxRows;
      else
         nDelRows = sizeRowsDel;
      
      for (i=0; i < nDelRows; i++)
      {
         delstat[rowsDel[i].idx] = 1;
      }
      
      // Delete rows from the lp
      master->solver->delRowSet(locallp,delstat);
      
      // Delete rows from hash
      rowHash->delElements(delstat);
      
      // Delete rows list of LP rows
      removeRowsFromLPRows(delstat,nRows);
   }
   else
   {
      nDelRows = 0;
   }
   
   if (rowsDel)
      delete[] rowsDel;
   
   if ( delstat )
      delete[] delstat;
   
   if ( bRows )
      delete[] bRows;
   
   if ( bCols )
      delete[] bCols;
   
   return nDelRows;
}

//---------------------------------------------------------------------------//

void CVRPBBNode::removeRowsFromLPRows(int *delstat, int nRows)
{
   register int i; // counter
   vector<CRow*> auxVec; // auxiliary vector
   
   auxVec.clear();
   
   for (i=localg->getN();i < nRows; i++)
   {
      if (delstat[i]==-1)
      {
         if (lpRows[i-localg->getN()])
            lpRows[i-localg->getN()]->decreaseNLPs();
         
         lpRows[i-localg->getN()]=NULL;
      }
      else
      {
         auxVec.push_back(lpRows[i-localg->getN()]);
         lpRows[i-localg->getN()]=NULL;
      }
   }
   
   lpRows.clear();
   lpRows.resize(auxVec.size());
   
   for (i=0; i < (int)(auxVec.size()); i++)
   {
      lpRows[i] = auxVec[i];
      auxVec[i] = NULL;
   }
   
   auxVec.clear();
}

//---------------------------------------------------------------------------//

int CVRPBBNode::eliminateLPColumns(int maxCols)
{
   int nDelCols; // number of columns eliminated from the LP
   RowColElim *colsDel; // possible rows to be eliminated
   int sizeColsDel; // number of columns possible to be eliminated
   register int i; // counter
   int *delstat; // deleted status of each row
   int nRows, nCols; // number of rows and columns
   int *bCols, *bRows; // number of columns and rows of the basis
   
   nRows = master->solver->getNRows(locallp);
   nCols = master->solver->getNColumns(locallp);
   
   colsDel = new RowColElim[nCols];
   
   if ( !colsDel )
      throw(LOWMEMORYEXCEPTION);
   
   delstat = new int[nCols];
   
   if ( !delstat )
      throw(LOWMEMORYEXCEPTION);
   
   bCols = new int[nCols];
   
   if ( !bCols )
      throw(LOWMEMORYEXCEPTION);
   
   bRows = new int[nRows];
   
   if ( !bRows )
      throw(LOWMEMORYEXCEPTION);
   
   // Get the basis variables
   master->solver->getBasis(locallp,bCols,bRows);
   sizeColsDel = 0;
   
   // Get rows possible to be eliminated
   for (i=0; i < nCols; i++)
   {
      delstat[i] = 0;
      
      if ( i >= maxArtificials)
      {
         if ( bCols[i] != BASIC_VAR)
         {
            colsDel[sizeColsDel].idx = i;
            
            if ((((CVRPMaster*)master)->getCarp()) && (((CVRPMaster*)master)->getLongo()))
               colsDel[sizeColsDel].value = lpCols[i-maxArtificials]->getRedCost(lpDualCarp);
            else
               colsDel[sizeColsDel].value = lpCols[i-maxArtificials]->getRedCost(lpDualExp);
            
            sizeColsDel++;
         }
      }
   }
   
   if ( sizeColsDel > 0 )
   {
      // Order by reduced cost
      qsort(colsDel,sizeColsDel,sizeof(RowColElim),compRowColElim);
      
      if ( sizeColsDel >= maxCols)
         nDelCols = maxCols;
      else
         nDelCols = sizeColsDel;
      
      for (i=0; i < nDelCols; i++)
      {
         delstat[colsDel[i].idx] = 1;
      }
      
      // Delete columns from the lp
      master->solver->delColSet(locallp,delstat);
      
      // Delete columns from hash
      colHash->delElements(delstat);
      
      // Delete columns list of LP cols
      removeColsFromLPCols(delstat,nCols);
   }
   else
   {
      nDelCols = 0;
   }
   
   if (colsDel)
      delete[] colsDel;
   
   if ( delstat )
      delete[] delstat;
   
   if ( bRows )
      delete[] bRows;
   
   if ( bCols )
      delete[] bCols;
   
   return nDelCols;
}

int CVRPBBNode::eliminateRowRange(int begin, int end)
{
   int nDelRows; // number of rows eliminated from the LP
   register int i; // counter
   int *delstat; // deleted status of each row
   int nRows, nCols; // number of rows and columns
   
   nRows = master->solver->getNRows(locallp);
   nCols = master->solver->getNColumns(locallp);
   
   delstat = new int[nRows];
   
   if ( !delstat )
      throw(LOWMEMORYEXCEPTION);
   
   if( begin < localg->getN() )
   {
      printf("Invalid parameter begin in eliminateRowRange\n");
      exit(1);
   }
   if( end >= nRows )
   {
      printf("Invalid parameter end in eliminateRowRange\n");
      exit(1);
   }
   
   // Get the basis variables
   nDelRows = 0;
   
   // Get rows possible to be eliminated
   memset( delstat, 0 , sizeof(int) * nRows );
   for (i=begin; i <= end; i++)
   {
      delstat[i] = 1;
      nDelRows++;
   }
   
   if ( nDelRows > 0 )
   {
      
      // Delete rows from the lp
      master->solver->delRowSet(locallp,delstat);
      
      // Delete rows from hash
      rowHash->delElements(delstat);
      
      // Delete rows list of LP rows
      removeRowsFromLPRows(delstat,nRows);
   }
   
   
   if ( delstat )
      delete[] delstat;
   
   
   setDualSize(master->solver->getNRows(locallp));
   return nDelRows;
}

int CVRPBBNode::eliminateColRange( int begin, int end )
{
   int nDelCols=0; // number of columns eliminated from the LP
   register int i; // counter
   int *delstat; // deleted status of each row
   int nRows, nCols; // number of rows and columns
   

   
   nRows = master->solver->getNRows(locallp);
   nCols = master->solver->getNColumns(locallp);
   
   delstat = new int[nCols];
   
   if ( !delstat )
      throw(LOWMEMORYEXCEPTION);
   
   
   if( begin < maxArtificials )
   {
      printf("Invalid value of begin in eliminateColRange\n");
      exit(1);
   }
   if( end >= nCols )
   {
      printf("Invalid value of end in eliminateColRange\n");
      exit(1);
   }
   
   
   // Get columns possible to be eliminated
   memset( delstat, 0, sizeof(int) * nCols );
   
   for( i = begin; i <= end; i++ )
   {
      nDelCols++;
      delstat[i] = 1;
   }
   
   // Delete columns from the lp
   master->solver->delColSet(locallp,delstat);
   
   // Delete columns from hash
   colHash->delElements(delstat);
   
   // Delete columns list of LP cols
   removeColsFromLPCols(delstat,nCols);
   
   
   if ( delstat )
      delete[] delstat;
   
   
   setSolSize(master->solver->getNColumns(locallp));
   return nDelCols;
}


int CVRPBBNode::eliminateKCycleCols(int maxk)
{
   int nDelCols=0; // number of columns eliminated from the LP
   register int i; // counter
   int *delstat; // deleted status of each row
   int nRows, nCols; // number of rows and columns
   int *bCols, *bRows; // number of columns and rows of the basis
   
   int kiter;
   
   nRows = master->solver->getNRows(locallp);
   nCols = master->solver->getNColumns(locallp);
   
   delstat = new int[nCols];
   
   if ( !delstat )
      throw(LOWMEMORYEXCEPTION);
   
   bCols = new int[nCols];
   
   if ( !bCols )
      throw(LOWMEMORYEXCEPTION);
   
   bRows = new int[nRows];
   
   if ( !bRows )
      throw(LOWMEMORYEXCEPTION);
   
   // Get the basis variables
   master->solver->getBasis(locallp,bCols,bRows);
   
   
   // Get columns possible to be eliminated
   memset( delstat, 0, sizeof(int) * nCols );
   
   // Order by reduced cost
   for( i = maxArtificials; i < nCols; i++)
   {
      for( kiter = 2; kiter <= maxk; kiter++ )
      {
         if( lpCols[i-maxArtificials]->getHasKCycle(kiter) )
         {
            if( bCols[i] != BASIC_VAR )
            {
               nDelCols++;
               delstat[i] = 1;
            }
            else
            {
               //Change the objective function to drive column out of basis
               lpCols[i-maxArtificials]->setObj( ((CVRPMaster*)master)->getSumAllEdges()  );
               master->solver->setCoef( locallp, -1, i, ((CVRPMaster*)master)->getSumAllEdges()  );
            }
            
            break;
         }
      }
   }
   
   // Delete columns from the lp
   master->solver->delColSet(locallp,delstat);
   
   // Delete columns from hash
   colHash->delElements(delstat);
   
   // Delete columns list of LP cols
   removeColsFromLPCols(delstat,nCols);
   
   
   if ( delstat )
      delete[] delstat;
   
   if ( bRows )
      delete[] bRows;
   
   if ( bCols )
      delete[] bCols;
   
   return nDelCols;
}

//---------------------------------------------------------------------------//

void CVRPBBNode::removeColsFromLPCols(int *delstat, int nCols)
{
   register int i; // counter
   vector<CColumn*> auxVec; // auxiliary vector
   
   auxVec.clear();
   
   for (i=maxArtificials;i < nCols; i++)
   {
      if (delstat[i]==-1)
      {
         if (lpCols[i-maxArtificials])
            lpCols[i-maxArtificials]->decreaseNLPs();
         
         lpCols[i-maxArtificials]=NULL;
      }
      else
      {
         auxVec.push_back(lpCols[i-maxArtificials]);
         lpCols[i-maxArtificials]=NULL;
      }
   }
   
   lpCols.clear();
   lpCols.resize(auxVec.size());
   
   for (i=0; i < (int)(auxVec.size()); i++)
   {
      lpCols[i] = auxVec[i];
      auxVec[i] = NULL;
   }
   
   auxVec.clear();
}

//---------------------------------------------------------------------------//

void CVRPBBNode::getBranchCandidateSets()
{
   int noOfEdges; // number of edges in the solution
   register int i; // counter
   
   // solves the LP
   cronoSolveLP.start();
   solvePrimal();
   cronoSolveLP.stop();
   
   // Expand the solution
   expandSolution(lpSolExp,lpSol);
   
   // Store information about the solution
   noOfEdges = 0;
   
   for (i=0; i < localg->getM(); i++)
   {
      if ( lpSolExp[i] > master->params->getIntEps() )
      {
         noOfEdges++;
         edgeTail[noOfEdges]= localg->getEdgeOrg(i);
         edgeHead[noOfEdges]= localg->getEdgeDst(i);
         edgeX[noOfEdges]= lpSolExp[i];
      }
   }
   
   BRNCHING_GetCandidateSets(localg->getN()-1,
      ((CVRPMaster*)master)->getAllDemands(),
      ((CVRPMaster*)master)->getVehicleCap(),
      noOfEdges,
      edgeTail,
      edgeHead,
      edgeX,
      myOldCutsCMP,
      ((CVRPMaster*)master)->getBdTarget(),
      localg->getN()-1,
      1,
      setsCMP);
      /* The value of zero above means order by absolute deviation of
      BOUNDARYTARGET
   The value of 1 means the original ordering, i.e., x(\delta(S)) / q(S) */
}

//---------------------------------------------------------------------------//



double CVRPBBNode::calcRootProximity(double val)
{
   return( ( getUB() - val ) / getUB() );
}


double CVRPBBNode::estimateNumBranches( double val )
{
   return( ( getUB() - val ) / ((CVRPMaster*)master)->getBImp() );
}

