/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CVRPFullGraph.cpp                //
////////////////////////////////////////////
// Class that represents the graph for    //
// CVRP problem (full matrix)             //
////////////////////////////////////////////

//---------------------------------------------------------------------------//

#include "CVRPFullGraph.h"
#include "Constants.h"

//---------------------------------------------------------------------------//

CVRPFullGraph::CVRPFullGraph()
{
   n=0;
   m=0;
   distMatrix=NULL;
   listEdges=NULL;
   listVertex=NULL;
}

//---------------------------------------------------------------------------//

CVRPFullGraph::CVRPFullGraph(const CVRPFullGraph &g)
{
   allocate(g.n,g.m);
   copy(g);
}

//---------------------------------------------------------------------------//

CVRPFullGraph::~CVRPFullGraph()
{
   deAllocate();

   n=0;
   m=0;
   distMatrix = NULL;
   listEdges = NULL;
   listVertex = NULL;
}

//---------------------------------------------------------------------------//

void CVRPFullGraph::copy(const CVRPFullGraph &g)
{
   register int i,j; // counters

   // Size of the graph
   n=g.n;
   m=g.m;

   // Copy list of vertices
   for (i=1;i<=n;i++)
   {
      listVertex[i].x = g.listVertex[i].x;
      listVertex[i].y = g.listVertex[i].y;
   }

   for (i=2;i<=n;i++)
   {
      for (j=1;j<i;j++)
      {
         distMatrix[i][j]->cost = g.distMatrix[i][j]->cost;
         distMatrix[i][j]->fixed = g.distMatrix[i][j]->fixed;
         distMatrix[i][j]->index = g.distMatrix[i][j]->index;
         distMatrix[i][j]->twoEdge = g.distMatrix[i][j]->twoEdge;
         distMatrix[i][j]->org = g.distMatrix[i][j]->org;
         distMatrix[i][j]->dst = g.distMatrix[i][j]->dst;
      }
   }
}

//---------------------------------------------------------------------------//

void CVRPFullGraph::allocate (int n1, int m1)
{
   register int i,j; // counter
   int cont; // auxiliary counter

   n=n1;
   m=m1;

   distMatrix = new CVRPFullGraphEdge**[n+1];

   if ( !distMatrix )
      throw(LOWMEMORYEXCEPTION);

   for (i=1;i<=n;i++)
   {
      distMatrix[i] = new CVRPFullGraphEdge*[i];

      if ( ! distMatrix[i] )
         throw(LOWMEMORYEXCEPTION);
   }

   listEdges=new CVRPFullGraphEdge[m];

   if (!listEdges)
      delete[] listEdges;

   listVertex=new CVRPFullGraphVertex[n+1];

   if (!listVertex)
      delete[] listVertex;

   cont=0;

   for (i=1;i<=n;i++)
   {
      for (j=1;j<=n;j++)
      {
         if (i<j)
         {
            distMatrix[j][i]=&(listEdges[cont]);
            listEdges[cont].index = cont;
            distMatrix[j][i]->fixed = -1;
            distMatrix[j][i]->twoEdge = true;
            distMatrix[j][i]->cost = 0;
            distMatrix[j][i]->org = i;
            distMatrix[j][i]->dst = j;
            
            cont++;
         }
      }
   }

   for (i=1;i<=n;i++)
   {
      listVertex[i].x=0.0;
      listVertex[i].y=0.0;
   }
}

//---------------------------------------------------------------------------//

void CVRPFullGraph::deAllocate()
{
   register int i,j; // counters

   // Deallocate pointers
   for (i=1;i<=n;i++)
   {
      for (j=1;j<=n;j++)
      {
         if (i<j)
         {
            distMatrix[j][i]=NULL;
         }
      }
   }

   // Deallocate distance matrix
   if (distMatrix)
   {
      for (i=1;i<=n;i++)
      {
         if ( distMatrix[i] )
         {
            delete[] distMatrix[i];
         }
      }
      delete[] distMatrix;
   }

   // Free memory
   if (listEdges)
      delete[] listEdges;

   if (listVertex)
      delete[] listVertex;


   // Reset size
   n=0;
   m=0;
}

//---------------------------------------------------------------------------//
