/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CVRPHash.cpp                     //
////////////////////////////////////////////
// Class that a hash used to calculate    //
// the coefficients of the compact CVRP   //
// formulation                            //
////////////////////////////////////////////

#include "Constants.h"
#include "CVRPHash.h"

//---------------------------------------------------------------------------//

CVRPHash::CVRPHash(int n)
{
   hash.resize(n);

   size = n;
}

//---------------------------------------------------------------------------//

CVRPHash::~CVRPHash()
{
   register int i; // counter

   for (i=0;i<size;i++)
      hash[i].clear();

   hash.clear();
}

//---------------------------------------------------------------------------//

void CVRPHash::add(int key, int idx, double coeff)
{
   CVRPHashElem aux; // auxiliary element

   aux.coeff = coeff;
   aux.idx = idx;

   hash[key].push_back(aux);
}

//---------------------------------------------------------------------------//

void CVRPHash::removeLast(int key)
{
   if ( hash[key].size() > 0 )
      hash[key].pop_back();
}

//---------------------------------------------------------------------------//

bool CVRPHash::getNext(int key, int &idx, double &coeff)
{
   if ( (int)(hash[key].size()) <= position )
      return false;
   else
   {
      idx = hash[key].at(position).idx;
      coeff = hash[key].at(position).coeff;
      position++;
      return true;
   }
}

//---------------------------------------------------------------------------//

int CVRPHash::getLast(int key)
{
   int sizeV; // size of the vector

   sizeV = (int)(hash[key].size());

   if ( sizeV > 0 )
      return hash[key][sizeV-1].idx;
   else
      return -1;
}

//---------------------------------------------------------------------------//

void CVRPHash::sumLast(int key, double coeff)
{
   int sizeV; // size of the vector

   sizeV = (int)(hash[key].size());

   if ( sizeV > 0 )
      hash[key][sizeV-1].coeff += coeff;
}

//---------------------------------------------------------------------------//

void CVRPHash::delElements(int *elements)
{
   register int i,j; // counter
   int sizeAux; // auxiliary size
   vector<CVRPHashElem> auxVec; // auxiliary hash element vector

   for (i=0;i < size; i++)
   {
      sizeAux = (int)hash[i].size();
      auxVec.clear();

      for (j=0; j < sizeAux; j++)
      {
         if ( elements[hash[i][j].idx]!= -1 )
         {
            hash[i][j].idx = elements[hash[i][j].idx];
            auxVec.push_back(hash[i][j]);
         }
      }

      hash[i].clear();
      hash[i].resize(auxVec.size());

      for (j=0;j < (int)(auxVec.size()); j++)
      {
         hash[i][j].idx = auxVec[j].idx;
         hash[i][j].coeff = auxVec[j].coeff;
      }
   }
}

//---------------------------------------------------------------------------//

int CVRPHash::getNumBytes()
{
   register int i; // counter
   int sizeBytes; // total bytes size of the hash

   sizeBytes = 0;

   for (i=0; i < size; i++)
   {
      sizeBytes += (int)(hash[i].capacity()) * (int)(sizeof(CVRPHashElem));
   }

   return sizeBytes;
}

//---------------------------------------------------------------------------//

int CVRPHash::getMinNumBytes()
{
   register int i; // counter
   int sizeBytes; // total bytes size of the hash

   sizeBytes = 0;

   for (i=0; i < size; i++)
   {
      sizeBytes += (int)(hash[i].size()) * (int)(sizeof(CVRPHashElem));
   }

   return sizeBytes;
}

//---------------------------------------------------------------------------//
