/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CVRPMaster.cpp                   //
////////////////////////////////////////////
// Class that represents the control of   //
// the branch-and-cut-and-price for the   //
// the CVRP                               //
////////////////////////////////////////////

//---------------------------------------------------------------------------//


#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include "CVRPMaster.h"
#include "CVRPFullGraph.h"
#include "CPUTimer.h"
#include "Constants.h"

//---------------------------------------------------------------------------//

CVRPMaster::CVRPMaster() : CMaster()
{
  tspView = false;
  
  strcpy(instanceFileName, "");
  strcpy(prefixFileName, "");
  
  sumAllDemands = 0;
  sumAllEdges = 0;
  listDemands = NULL;
  numberOfVehicles = 0;
  vehicleCapacity = 0;
  minCapVehicle = 0;
  depot = 0;
  gMaster = NULL;

  vMatrix = NULL;
  predMatrix = NULL;
  vMatrix2 = NULL;
  predMatrix2 = NULL;
  
  dpCarp = NULL;
  dpDist = NULL;

  edgeHead = NULL;
  edgeTail = NULL;
  edgeX = NULL;
  listS = NULL;
  nodeP = NULL;
  pathDinProg = NULL;
  bestSol = NULL;
  
  bestUB = INFINITY_VALUE;
  kcycleparam = DEFAULTKCYCLE;
  rootonly = false;
  
  itercg = false;
  revbb = false;
  
  ncolgen = 0;
  ncutgen = 0;
  mAddCutsToLPTimer.reset();
  mAddColsToLPTimer.reset();
  mColGenTimer.reset();
  mCutGenTimer.reset();
  mSolveLPTimer.reset();
  
  dynk = false;
  pmin = DEFAULTPMIN;
  pmax = DEFAULTPMAX;
  sbcgiter = DEFAULTSBCGITERS;

  vrptw  = false;
  carp   = false;
  pearn  = false;
  longo  = false;
  uchoa  = false;
  nUchoa = 0;

  bimp = 0.0;

  stopfath = false;

  rootprox = 0.005;

  leafprox = 0.002;

  bdtarget = 2.85;

  //goodgain = 0.001;

  goodfactor = 0.8;

  dostab = false;
  dosb = false;


  nmiddleleftSBtries = 0;
  nmiddlerightSBtries = 0;
  maxmiddleleftSBtries = 0;
  minmiddleleftSBtries = 100;
  maxmiddlerightSBtries = 0;
  minmiddlerightSBtries = 100;

  bestlpe = NULL;

}

//---------------------------------------------------------------------------//

CVRPMaster::~CVRPMaster()
{
  register int i;
  register int j;

  if ( listDemands )
    delete[] listDemands;
  
  tspView = false;
  
  strcpy(instanceFileName, "");
  strcpy(prefixFileName, "");
  
  sumAllDemands = 0;
  sumAllEdges = 0;
  listDemands = NULL;
  numberOfVehicles = 0;
  vehicleCapacity = 0;
  minCapVehicle = 0;
  depot = 0;
  
  vMatrix = NULL;
  predMatrix = NULL;
  vMatrix2 = NULL;
  predMatrix2 = NULL;

  if (dpCarp)
  {
    for (i = 0; i < gMaster->getN(); i++)
    {
      if (dpCarp[i])
      {
        for (j = 0; j < vehicleCapacity; j++)
          if (dpCarp[i][j])
            delete [] dpCarp[i][j];
        delete [] dpCarp[i];
      }
    }
    delete [] dpCarp;
  }

  if (dpDist)
  {
    for (i = 0; i < gMaster->getN(); i++)
      delete [] dpDist[i];
    delete [] dpDist;
  }

  edgeHead = NULL;
  edgeTail = NULL;
  edgeX = NULL;
  listS = NULL;
  nodeP = NULL;
  pathDinProg = NULL;

  if( bestlpe )
     delete [] bestlpe;
}

//---------------------------------------------------------------------------//

void CVRPMaster::optimize()
{
#if SHOW_SB_STATS
   double mingainleftroot = 1000000;
   double maxgainleftroot = 0.0;
   double totalgainleftroot = 0.0;
   int nnodesleftroot = 0;
   double mingainrightroot = 1000000;
   double maxgainrightroot = 0.0;
   double totalgainrightroot = 0.0;
   int nnodesrightroot = 0;

   double mingainleftmid = 1000000;
   double maxgainleftmid = 0.0;
   double totalgainleftmid = 0.0;
   int nnodesleftmid = 0;
   double mingainrightmid = 1000000;
   double maxgainrightmid = 0.0;
   double totalgainrightmid = 0.0;
   int nnodesrightmid = 0;

   double mingainleftleaf = 1000000;
   double maxgainleftleaf = 0.0;
   double totalgainleftleaf = 0.0;
   int nnodesleftleaf = 0;
   double mingainrightleaf = 1000000;
   double maxgainrightleaf = 0.0;
   double totalgainrightleaf = 0.0;
   int nnodesrightleaf = 0;


   double maxtimeleftroot = 0.0;
   double mintimeleftroot = CPX_INFBOUND;
   double totaltimeleftroot = 0.0;
   double maxtimerightroot = 0.0;
   double mintimerightroot = CPX_INFBOUND;
   double totaltimerightroot = 0.0;

   double maxtimeleftmid = 0.0;
   double mintimeleftmid = CPX_INFBOUND;
   double totaltimeleftmid = 0.0;
   double maxtimerightmid = 0.0;
   double mintimerightmid = CPX_INFBOUND;
   double totaltimerightmid = 0.0;

   double maxtimeleftleaf = 0.0;
   double mintimeleftleaf = CPX_INFBOUND;
   double totaltimeleftleaf = 0.0;
   double maxtimerightleaf = 0.0;
   double mintimerightleaf = CPX_INFBOUND;
   double totaltimerightleaf = 0.0;
#endif

   



   NodeStatus optstat; // Status of the node
   CVRPBBNode* auxnode; // auxiliary node
   CPUTimer timer; // keeps the time
   char fname[100]; // auxiliary file name
   
   register int i; // counter
   
   int n,b; // number of vertex and capacity
   double *lpe; // expanded solution
   
   
   
   
   int nexplorednodes = 0;
   double roottime;
   
   // Allocate memory for structures
   n=gMaster->getN();
   b=vehicleCapacity;
   
   predMatrix = new PredElem*[n+1];
   
   for (i=0;i<=n;i++)
      predMatrix[i] = new PredElem[b+1];
   
   vMatrix = new VElem*[(n+1)];
   for (i=0;i<=n;i++)
      vMatrix[i] = new VElem[b+1];
   
   predMatrix2 = new PredElem*[n+1];
   
   for (i=0;i<=n;i++)
      predMatrix2[i] = new PredElem[b+1];
   
   vMatrix2 = new VElem*[(n+1)];
   for (i=0;i<=n;i++)
      vMatrix2[i] = new VElem[b+1];
   
   
   edgeHead = new int[gMaster->getM()];
   
   if ( !edgeHead )
      throw(LOWMEMORYEXCEPTION);
   
   edgeTail = new int[gMaster->getM()];
   
   if ( !edgeTail )
      throw(LOWMEMORYEXCEPTION);
   
   edgeX = new double[gMaster->getM()];
   
   if ( !edgeX )
      throw(LOWMEMORYEXCEPTION);
   
   listS = new int[gMaster->getN()+1];
   
   if ( !listS )
      throw(LOWMEMORYEXCEPTION);
   
   nodeP = new int[gMaster->getN()+1];
   
   if ( !nodeP )
      throw(LOWMEMORYEXCEPTION);
   
   pathDinProg = new int[ vehicleCapacity+1 ];
   
   if ( !pathDinProg )
      throw(LOWMEMORYEXCEPTION);
   
   // Create the first node of the tree
   auxnode = new CVRPBBNode( this, 1 );
   
   // Set the graph for the problem
   auxnode->setGraph( gMaster); 
   auxnode->setUB( bestUB );
   auxnode->setNFixedSets(0);
   auxnode->setFixedSets(NULL);
   
   // Insert the node in the tree
   openNodes.addLast(auxnode);
   
   writeLog( true, "STARTING OPTIMIZATION\n" );
   
   // Start the cronometer
   timer.start();
   
   // Branch and bound tree
   while( openNodes.size > 0 )
   {
      // Select the next node to solve
      auxnode = NULL;
      if( params->getNodeSel() == NODE_SELECT_DFS )
         auxnode = (CVRPBBNode*) selectDFSNode();
      else if( params->getNodeSel() == NODE_SELECT_BFS )
         auxnode = (CVRPBBNode*) selectBFSNode();
      else if( params->getNodeSel() == NODE_SELECT_BND )
         auxnode = (CVRPBBNode*) selectBestBndNode();
      
      // Optimize the selected node
      if( ! auxnode->isFathomed( auxnode->getLB() ) )
      {
         nexplorednodes++;
         optstat = auxnode->optimize();
         
         // Root node
         if( auxnode->getID() == 1 )
         {
            roottime = ((CVRPBBNode*)auxnode)->nodetimer.getCPUTotalSecs();
         }
      }	      
      else
      {
         writeLog( true, "\n\nNode %g FATHOMED without exploring!!!!\n\n", auxnode->getID() );
         optstat = FATHOMED;
      }
      
      // Verify the status of the optimization
      if( optstat == FEASIBLE )
      {
         if( auxnode->getUB() <= bestUB )
         {
            strcpy( fname, prefixFileName );
            strcat( fname, "-best");
            lpe = NULL;
            lpe = new double[gMaster->getM()];
            
            if (!lpe)
               throw(LOWMEMORYEXCEPTION);
            
            auxnode->expandSolution(lpe,auxnode->getLPSol());
            writeSolTSPView(auxnode, fname, lpe, gMaster );
            
            writeFeasibleSol( fname, lpe, gMaster );
            memcpy( bestlpe, lpe,  gMaster->getM() * sizeof(double) );
            if ( lpe )
               delete[] lpe;
         }
      }
      else if( optstat == BRANCH )
      {
         if ( auxnode->getID()==1 )
         {
            strcpy( fname, prefixFileName );
            strcat( fname, "-currlp");
            lpe = NULL;
            lpe = new double[gMaster->getM()];
            
            auxnode->expandSolution(lpe,auxnode->getLPSol());
            writeSolTSPView(auxnode, fname, lpe, gMaster );
            
            if ( lpe )
               delete[] lpe;
         }
      }

#if SHOW_SB_STATS
      // Update the SB statistics
      if( ( auxnode->getID() != 1  ) && ( optstat == BRANCH ) )
      {
         double gain = auxnode->getLB() - auxnode->getParentLB();
         double nodetime = ((CVRPBBNode*)auxnode)->nodetimer.getCPUTotalSecs();
         int nSBtries = ((CVRPBBNode*)auxnode)->SBtries;
         if( auxnode->branchorg == 0 )
         {
            // Left branch
            if( auxnode->getOrigin() == 0 )
            {
               mingainleftroot = min( mingainleftroot, gain );
               maxgainleftroot = max( maxgainleftroot, gain );
               totalgainleftroot += gain;

               mintimeleftroot = min( mintimeleftroot, nodetime );
               maxtimeleftroot = max( maxtimeleftroot, nodetime );
               totaltimeleftroot += nodetime;

               nnodesleftroot++;
            }
            // Right branch
            else
            {
               mingainrightroot = min( mingainrightroot, gain );
               maxgainrightroot = max( maxgainrightroot, gain );
               totalgainrightroot += gain;

               mintimerightroot = min( mintimerightroot, nodetime );
               maxtimerightroot = max( maxtimerightroot, nodetime );
               totaltimerightroot += nodetime;

               nnodesrightroot++;
            }
         }
         else if( auxnode->branchorg == 1 )
         {
            // Left branch
            if( auxnode->getOrigin() == 0 )
            {
               mingainleftmid = min( mingainleftmid, gain );
               maxgainleftmid = max( maxgainleftmid, gain );
               totalgainleftmid += gain;

               mintimeleftmid = min( mintimeleftmid, nodetime );
               maxtimeleftmid = max( maxtimeleftmid, nodetime );
               totaltimeleftmid += nodetime;

               nnodesleftmid++;


               nmiddleleftSBtries += nSBtries;
               maxmiddleleftSBtries = max( maxmiddleleftSBtries, nSBtries );
               minmiddleleftSBtries = min( minmiddleleftSBtries, nSBtries );
            }
            // Right branch
            else
            {
               mingainrightmid = min( mingainrightmid, gain );
               maxgainrightmid = max( maxgainrightmid, gain );
               totalgainrightmid += gain;

               mintimerightmid = min( mintimerightmid, nodetime );
               maxtimerightmid = max( maxtimerightmid, nodetime );
               totaltimerightmid += nodetime;

               nnodesrightmid++;

               nmiddlerightSBtries += nSBtries;
               maxmiddlerightSBtries = max( maxmiddlerightSBtries, nSBtries );
               minmiddlerightSBtries = min( minmiddlerightSBtries, nSBtries );
            }
         }
         else if( auxnode->branchorg == 2 )
         {
            // Left branch
            if( auxnode->getOrigin() == 0 )
            {
               mingainleftleaf = min( mingainleftleaf, gain );
               maxgainleftleaf = max( maxgainleftleaf, gain );
               totalgainleftleaf += gain;

               mintimeleftleaf = min( mintimeleftleaf, nodetime );
               maxtimeleftleaf = max( maxtimeleftleaf, nodetime );
               totaltimeleftleaf += nodetime;

               nnodesleftleaf++;
            }
            // Right branch
            else
            {
               mingainrightleaf = min( mingainrightleaf, gain );
               maxgainrightleaf = max( maxgainrightleaf, gain );
               totalgainrightleaf += gain;

               mintimerightleaf = min( mintimerightleaf, nodetime );
               maxtimerightleaf = max( maxtimerightleaf, nodetime );
               totaltimerightleaf += nodetime;

               nnodesrightleaf++;
            }
         }  
      }
#endif
      
      // Compute the node
      computeNode( auxnode, optstat );
      
      // Update the bounds
      updateBounds();
      
      // Print information about the optimization
      writeLog( true, "MASTER: Solved Node:%d, Op=%d/%d, %3.2f dp=%d/%d. sk=%.3f\n", 
         (int) ( auxnode->getID()+0.5 ), 
         openNodes.size - 1, totalNodes, elapsedBB * 100.0,
         auxnode->getDepth(), maxDepth, getSkew() );
      
      if (pearn)
      {
         writeLog( true, "MASTER: Best UB = %f, Best LB = %f, GAP=%.2f\n\n", bestUB / 4, bestLB / 4, 100.0 * (bestUB - bestLB )/ bestUB );
      }
      else
      {
         writeLog( true, "MASTER: Best UB = %f, Best LB = %f, GAP=%.2f\n\n", bestUB, bestLB, 100.0 * (bestUB - bestLB )/ bestUB );
      }
      
      if( stopfath )
      {
         if( optstat != BRANCH )
            break;
      }
      
      // Checks if the solution is optimal
      if( checkOptimal() )
      {
         // Create the solution file name
         strcpy( fname, prefixFileName );
         strcat( fname, "-opt" );
         
         // Print solution
         writeLog( true, "OPTIMAL SOLUTION FOUND!!!!!!!!\n" );
         
         writeSolTSPView(auxnode, fname, bestlpe, gMaster );
         
         writeFeasibleSol( fname, bestlpe, gMaster );
         break;
      }
      
      openNodes.remove( (CBBNode*) auxnode );
      
      if ( auxnode )
         delete auxnode;
   }
   
   // Free the nodes in the open nodes list, since the
   // optimization has ended
   while( openNodes.size > 0 )
   {
      // Select the next node to solve
      auxnode = NULL;
      auxnode = (CVRPBBNode*) selectDFSNode();
      
      openNodes.remove( (CBBNode*) auxnode );
      
      if ( auxnode )
         delete auxnode;
   }
   
   // Stop cronometer
   timer.stop();
   
   // Free memory allocated
   if (pathDinProg)
      delete[] pathDinProg;
   
   if ( nodeP )
      delete[] nodeP;
   
   if (listS)
      delete[] listS;
   
   if (edgeHead)
      delete[] edgeHead;
   
   if ( edgeTail )
      delete[] edgeTail;
   
   if ( edgeX )
      delete[] edgeX;
   
   if (vMatrix)
   {
      for (i=0;i<=n;i++)
         if( vMatrix[i] )
            delete[] vMatrix[i];
         
         delete[] vMatrix;
   }
   
   if (predMatrix)
   {
      for (i=0;i<=n;i++)
         if( predMatrix[i] )
            delete[] predMatrix[i];
         
         delete[] predMatrix;
   }
   
   if (vMatrix2)
   {
      for (i=0;i<=n;i++)
         if( vMatrix2[i] )
            delete[] vMatrix2[i];
         
         delete[] vMatrix2;
   }
   
   if (predMatrix2)
   {
      for (i=0;i<=n;i++)
         if( predMatrix2[i] )
            delete[] predMatrix2[i];
         
         delete[] predMatrix2;
   }
   
   
   if ( bestlpe )
      delete[] bestlpe;
   
   // Write information about the solution
   if (pearn)
   {
      writeLog( true, "Best Solution: %f\n", bestUB / 4 );
   }
   else
   {
      writeLog( true, "Best Solution: %f\n", bestUB );
   }
   
   writeLog( true, "Total number of nodes explored: %d\n", nexplorednodes );
   
   writeLog( true, "Total time: %g\n", timer.getCronoTotalSecs() );
   writeLog( true, "Total CPU time: %g\n", timer.getCPUTotalSecs()) ;
   
   // Fuka B
   writeLog( true, "Total SB time: %g\n", sbtimer.getCronoTotalSecs() );
   writeLog( true, "Total SB CPU time: %g\n", sbtimer.getCPUTotalSecs() );
   
   writeLog( true, "Total time fixing by redcost: %g\n", fixredcosttimer.getCronoTotalSecs() );
   writeLog( true, "Total CPU time fixing by redcost: %g\n", fixredcosttimer.getCPUTotalSecs() );
   
   writeLog( true, "************   STATISTICS ***********************\n\n" );
   writeLog( true, "ST0:  Total time overall:                 %g\n", timer.getCPUTotalSecs()) ;
   writeLog( true, "ST1:  Total number of cols generated:     %d\n", ncolgen );
   writeLog( true, "ST2:  Total number of rows generated:     %d\n", ncutgen );
   writeLog( true, "ST3:  Total time spent in CG:             %g\n", mColGenTimer.getCPUTotalSecs() );
   writeLog( true, "ST4:  Total time spent in CutGen:         %g\n", mCutGenTimer.getCPUTotalSecs() );
   writeLog( true, "ST5:  Total time spent adding cols to LP: %g\n", mAddColsToLPTimer.getCPUTotalSecs() );
   writeLog( true, "ST6:  Total time spent adding cuts to LP: %g\n", mAddCutsToLPTimer.getCPUTotalSecs() );
   writeLog( true, "ST7:  Total time spent solving LP:        %g\n", mSolveLPTimer.getCPUTotalSecs() );
   writeLog( true, "ST8:  Total time doing SB:                %g\n", sbtimer.getCPUTotalSecs() );
   writeLog( true, "ST9:  Total time fixing by redcost:       %g\n", fixredcosttimer.getCPUTotalSecs() );
   writeLog( true, "ST10: Total number of nodes explored:     %d\n", nexplorednodes );
   writeLog( true, "\n****************************************************\n\n");

#if SHOW_SB_STATS
   writeLog( true, "\n************ SB GAIN STATS ********************************\n\n");
   writeLog( true, "               LEFT                            RIGHT                TOTAL\n");
   writeLog( true, "       (max;   min;   avg, n)     (max;   min;   avg, n)      (max;   min;   avg, n)    \n");
   writeLog( true, "root; %.3lf; %.3lf; %.3lf; %d;    %.3lf; %.3lf; %.3lf; %d;   %.3lf; %.3lf; %.3lf; %d\n", 
      maxgainleftroot, mingainleftroot, 
      nnodesleftroot > 0 ? totalgainleftroot / (double) nnodesleftroot : 0.0, 
      nnodesleftroot,
      maxgainrightroot, mingainrightroot, 
      nnodesrightroot > 0 ? totalgainrightroot / (double) nnodesrightroot : 0.0, 
      nnodesrightroot,
      max( maxgainleftroot, maxgainrightroot ),
      min( mingainleftroot, mingainrightroot ),
      nnodesleftroot + nnodesrightroot > 0 ?
         (totalgainleftroot+totalgainrightroot) / (double) (nnodesleftroot+nnodesrightroot)
         : 0,
      nnodesleftroot + nnodesrightroot );
   writeLog( true, "mid;  %.3lf; %.3lf; %.3lf; %d;    %.3lf; %.3lf; %.3lf; %d;   %.3lf; %.3lf; %.3lf; %d\n", 
      maxgainleftmid, mingainleftmid, 
      nnodesleftmid > 0 ? totalgainleftmid / (double) nnodesleftmid : 0.0, 
      nnodesleftmid,
      maxgainrightmid, mingainrightmid, 
      nnodesrightmid > 0 ? totalgainrightmid / (double) nnodesrightmid : 0.0, 
      nnodesrightmid,
      max( maxgainleftmid, maxgainrightmid ),
      min( mingainleftmid, mingainrightmid ),
      nnodesleftmid + nnodesrightmid > 0 ?
         (totalgainleftmid+totalgainrightmid) / (double) (nnodesleftmid+nnodesrightmid)
         : 0,
      nnodesleftmid + nnodesrightmid );
   writeLog( true, "leaf; %.3lf; %.3lf; %.3lf; %d;    %.3lf; %.3lf; %.3lf; %d;   %.3lf; %.3lf; %.3lf; %d\n", 
      maxgainleftleaf, mingainleftleaf, 
      nnodesleftleaf > 0 ? totalgainleftleaf / (double) nnodesleftleaf : 0.0, 
      nnodesleftleaf,
      maxgainrightleaf, mingainrightleaf, 
      nnodesrightleaf > 0 ? totalgainrightleaf / (double) nnodesrightleaf : 0.0, 
      nnodesrightleaf,
      max( maxgainleftleaf, maxgainrightleaf ),
      min( mingainleftleaf, mingainrightleaf ),
      nnodesleftleaf + nnodesrightleaf > 0 ?
         (totalgainleftleaf+totalgainrightleaf) / (double) (nnodesleftleaf+nnodesrightleaf)
         : 0,
      nnodesleftleaf  + nnodesrightleaf );
   double maxgainlefttree;
   double mingainlefttree;
   double maxgainrighttree;
   double mingainrighttree;
   double maxgaintree;
   double mingaintree;
   double avggainlefttree;
   double avggainrighttree;
   double avggaintree;
   maxgainlefttree = max( max( maxgainleftroot, maxgainleftmid ), maxgainleftleaf );
   mingainlefttree = min( min( mingainleftroot, mingainleftmid ), mingainleftleaf );
   maxgainrighttree = max( max( maxgainrightroot, maxgainrightmid ), maxgainrightleaf );
   mingainrighttree = min( min( mingainrightroot, mingainrightmid ), mingainrightleaf );
   maxgaintree = max( maxgainlefttree, maxgainrighttree );
   mingaintree = min( mingainlefttree, mingainrighttree );

   avggainlefttree = ( totalgainleftroot + totalgainleftmid + totalgainleftleaf ) / 
      (double) ( nnodesleftroot + nnodesleftmid + nnodesleftleaf );
   avggainrighttree = ( totalgainrightroot + totalgainrightmid + totalgainrightleaf ) / 
      (double) ( nnodesrightroot + nnodesrightmid + nnodesrightleaf );
   avggaintree = ( totalgainrightroot + totalgainrightmid + totalgainrightleaf + 
                        totalgainleftroot + totalgainleftmid + totalgainleftleaf  ) / 
      (double) ( nnodesrightroot + nnodesrightmid + nnodesrightleaf + 
                 nnodesleftroot + nnodesleftmid + nnodesleftleaf );

   writeLog( true, "tree; %.3lf; %.3lf; %.3lf;    %.3lf; %.3lf; %.3lf;   %.3lf; %.3lf; %.3lf\n", 
      maxgainlefttree, mingainlefttree, avggainlefttree,
      maxgainrighttree, mingainrighttree, avggainrighttree,
      maxgaintree, mingaintree, avggaintree );
   writeLog( true, "\n*********************************************************\n\n");


   writeLog( true, "\n************ NODE TIME STATS ********************************\n\n");
   writeLog( true, "             LEFT                        RIGHT                     TOTAL\n");
   writeLog( true, "       (max;   min;   avg, n)     (max;   min;   avg, n)    (max;   min;   avg, n)    \n");
   writeLog( true, "root; %.3lf; %.3lf; %.3lf; %d;    %.3lf; %.3lf; %.3lf; %d;   %.3lf; %.3lf; %.3lf; %d\n", 
      maxtimeleftroot, mintimeleftroot, 
      nnodesleftroot > 0 ? totaltimeleftroot / (double) nnodesleftroot : 0.0, 
      nnodesleftroot,
      maxtimerightroot, mintimerightroot, 
      nnodesrightroot > 0 ? totaltimerightroot / (double) nnodesrightroot : 0.0, 
      nnodesrightroot,
      max( maxtimeleftroot, maxtimerightroot ),
      min( mintimeleftroot, mintimerightroot ),
      nnodesleftroot + nnodesrightroot > 0 ?
         (totaltimeleftroot+totaltimerightroot) / (double) (nnodesleftroot+nnodesrightroot)
         : 0,
      nnodesleftroot + nnodesrightroot );
   writeLog( true, "mid;  %.3lf; %.3lf; %.3lf; %d;    %.3lf; %.3lf; %.3lf; %d;   %.3lf; %.3lf; %.3lf; %d\n", 
      maxtimeleftmid, mintimeleftmid, 
      nnodesleftmid > 0 ? totaltimeleftmid / (double) nnodesleftmid : 0.0, 
      nnodesleftmid,
      maxtimerightmid, mintimerightmid, 
      nnodesrightmid > 0 ? totaltimerightmid / (double) nnodesrightmid : 0.0, 
      nnodesrightmid,
      max( maxtimeleftmid, maxtimerightmid ),
      min( mintimeleftmid, mintimerightmid ),
      nnodesleftmid + nnodesrightmid > 0 ?
         (totaltimeleftmid+totaltimerightmid) / (double) (nnodesleftmid+nnodesrightmid)
         : 0,
      nnodesleftmid + nnodesrightmid );
   writeLog( true, "leaf; %.3lf; %.3lf; %.3lf; %d;    %.3lf; %.3lf; %.3lf; %d;   %.3lf; %.3lf; %.3lf; %d\n", 
      maxtimeleftleaf, mintimeleftleaf, 
      nnodesleftleaf > 0 ? totaltimeleftleaf / (double) nnodesleftleaf : 0.0, 
      nnodesleftleaf,
      maxtimerightleaf, mintimerightleaf, 
      nnodesrightleaf > 0 ? totaltimerightleaf / (double) nnodesrightleaf : 0.0, 
      nnodesrightleaf,
      max( maxtimeleftleaf, maxtimerightleaf ),
      min( mintimeleftleaf, mintimerightleaf ),
      nnodesleftleaf + nnodesrightleaf > 0 ?
         (totaltimeleftleaf+totaltimerightleaf) / (double) (nnodesleftleaf+nnodesrightleaf)
         : 0,
      nnodesleftleaf + nnodesrightleaf );
   double maxtimelefttree;
   double mintimelefttree;
   double maxtimerighttree;
   double mintimerighttree;
   double maxtimetree;
   double mintimetree;
   double avgtimelefttree;
   double avgtimerighttree;
   double avgtimetree;
   maxtimelefttree = max( max( maxtimeleftroot, maxtimeleftmid ), maxtimeleftleaf );
   mintimelefttree = min( min( mintimeleftroot, mintimeleftmid ), mintimeleftleaf );
   maxtimerighttree = max( max( maxtimerightroot, maxtimerightmid ), maxtimerightleaf );
   mintimerighttree = min( min( mintimerightroot, mintimerightmid ), mintimerightleaf );
   maxtimetree = max( maxtimelefttree, maxtimerighttree );
   mintimetree = min( mintimelefttree, mintimerighttree );

   avgtimelefttree = ( totaltimeleftroot + totaltimeleftmid + totaltimeleftleaf ) / 
      (double) ( nnodesleftroot + nnodesleftmid + nnodesleftleaf );
   avgtimerighttree = ( totaltimerightroot + totaltimerightmid + totaltimerightleaf ) / 
      (double) ( nnodesrightroot + nnodesrightmid + nnodesrightleaf );
   avgtimetree = ( totaltimerightroot + totaltimerightmid + totaltimerightleaf + 
                        totaltimeleftroot + totaltimeleftmid + totaltimeleftleaf  ) / 
      (double) ( nnodesrightroot + nnodesrightmid + nnodesrightleaf + 
                 nnodesleftroot + nnodesleftmid + nnodesleftleaf );

   writeLog( true, "tree; %.3lf; %.3lf; %.3lf;    %.3lf; %.3lf; %.3lf;   %.3lf; %.3lf; %.3lf\n", 
      maxtimelefttree, mintimelefttree, avgtimelefttree,
      maxtimerighttree, mintimerighttree, avgtimerighttree,
      maxtimetree, mintimetree, avgtimetree );
   writeLog( true, "\n*********************************************************\n\n");

   writeLog( true, "\n************ NUM. of SB tries in MIDDLE ********************************\n\n");
   writeLog( true, "             LEFT                        RIGHT                     TOTAL\n");
   writeLog( true, "    (max; min;   avg,  n)   (max; min;   avg,  n)    (max; min;   avg,  n)    \n");
   writeLog( true, "mid; %d;   %d; %.3lf; %d;     %d;  %d; %.3lf; %d;      %d;  %d; %.3lf; %d\n", 
      maxmiddleleftSBtries, 
      minmiddleleftSBtries, 
      nnodesleftmid > 0 ? (double) nmiddleleftSBtries / (double) nnodesleftmid : 0.0, 
      nnodesleftmid,
      maxmiddlerightSBtries, 
      minmiddlerightSBtries, 
      nnodesrightmid > 0 ? (double) nmiddlerightSBtries / (double) nnodesrightmid : 0.0, 
      nnodesrightmid,
      max( maxmiddleleftSBtries, maxmiddlerightSBtries ),
      min( minmiddleleftSBtries, minmiddlerightSBtries ),
      nnodesleftmid + nnodesrightmid > 0 ?
         (double)(nmiddleleftSBtries+nmiddlerightSBtries) / (double) (nnodesleftmid+nnodesrightmid)
         : 0,
      nnodesleftmid + nnodesrightmid );
   writeLog( true, "\n*********************************************************\n\n");


   // Fuka E
#define GENERATE_SB_TABLE
#ifdef GENERATE_SB_TABLE

   FILE *sbfout; // output file
   
   // Open the file
   #ifdef WIN32
   sbfout = fopen("tex\\sbtable.tex","a");
   #else
   sbfout = fopen("tex/sbtable.tex","a");
   #endif
   
   if ( !sbfout )
      throw(FILENOTOPENEXCEPTION);

   fprintf(sbfout,"%s & ", instanceFileName ); 
         
   fprintf(sbfout,"%d & %g & %g & %g & %d & %g & %g & %g  & ",
         
         nnodesleftroot,
         maxgainleftroot, 
         mingainleftroot, 
         nnodesleftroot > 0 ? totalgainleftroot / (double) nnodesleftroot : 0.0, 
         nnodesrightroot,
         maxgainrightroot, 
         mingainrightroot, 
         nnodesrightroot > 0 ? totalgainrightroot / (double) nnodesrightroot : 0.0 ); 
         
  fprintf(sbfout,"%d & %g & %g & %g & %d & %g & %g & %g  & ",        
         nnodesleftmid,
         maxgainleftmid, 
         mingainleftmid, 
         nnodesleftmid > 0 ? totalgainleftmid / (double) nnodesleftmid : 0.0, 
         nnodesrightmid,
         maxgainrightmid, 
         mingainrightmid, 
         nnodesrightmid > 0 ? totalgainrightmid / (double) nnodesrightmid : 0.0);

  fprintf(sbfout,"%d & %g & %g & %g & %d & %g & %g & %g  & ",
         nnodesleftleaf,
         maxgainleftleaf,
         mingainleftleaf, 
         nnodesleftleaf > 0 ? totalgainleftleaf / (double) nnodesleftleaf : 0.0, 
         nnodesrightleaf,
         maxgainrightleaf,
         mingainrightleaf, 
         nnodesrightleaf > 0 ? totalgainrightleaf / (double) nnodesrightleaf : 0.0);

  fprintf( sbfout,"%g & %g & %g & %g & %g & %g  & ",
         maxtimeleftroot,
         mintimeleftroot, 
         nnodesleftroot > 0 ? totaltimeleftroot / (double) nnodesleftroot : 0.0, 
         maxtimerightroot, 
         mintimerightroot, 
         nnodesrightroot > 0 ? totaltimerightroot / (double) nnodesrightroot : 0.0 ); 

  fprintf(sbfout,"%g & %g & %g & %g & %g & %g  & ",
         maxtimeleftmid, 
         mintimeleftmid, 
         nnodesleftmid > 0 ? totaltimeleftmid / (double) nnodesleftmid : 0.0, 
         maxtimerightmid,
         mintimerightmid, 
         nnodesrightmid > 0 ? totaltimerightmid / (double) nnodesrightmid : 0.0 ); 

  fprintf(sbfout,"%g & %g & %g & %g & %g & %g  & ",
         maxtimeleftleaf,
         mintimeleftleaf, 
         nnodesleftleaf > 0 ? totaltimeleftleaf / (double) nnodesleftleaf : 0.0, 
         maxtimerightleaf,
         mintimerightleaf, 
         nnodesrightleaf > 0 ? totaltimerightleaf / (double) nnodesrightleaf : 0.0 ); 

  fprintf(sbfout,"%d & %d & %g & %d & %d & %g   \\\\ \n",
         maxmiddleleftSBtries, 
         minmiddleleftSBtries, 
         nnodesleftmid > 0 ? (double) nmiddleleftSBtries / (double) nnodesleftmid : 0.0, 
         maxmiddlerightSBtries, 
         minmiddlerightSBtries, 
         nnodesrightmid > 0 ? (double) nmiddlerightSBtries / (double) nnodesrightmid : 0.0 ); 

   
  if( sbfout )
     fclose( sbfout );
#endif //GENERATE_SB_TABLE
   
#endif // SHOW_SB_STATS
   
#ifdef GENERATE_BCP_TABLE
   
   FILE *fout; // output file
   
   // Open the file
   #ifdef WIN32
   fout = fopen("tex\\bcptable.tex","a");
   #else
   fout = fopen("tex/bcptable.tex","a");
   #endif
   
   if ( !fout )
      throw(FILENOTOPENEXCEPTION);
   
   // Print information about the node
   // Instance name & bestUB & rootTime & Chosen k & 
   //   tree nodes explored &
   //   Skewness & GCTime & CutTime & SBTime & TotalTime
   if (pearn)
   {
      fprintf(fout,"%s & %g & %g & %d & %d & %g & %g & %g & %g & %g & %d & %d & %g & %g & %g & %g \\\\ \n",
         instanceFileName, bestUB / 4, roottime, kcycleparam, nexplorednodes,
         getSkew(), mColGenTimer.getCPUTotalSecs(), 
         mCutGenTimer.getCPUTotalSecs(), 
         sbtimer.getCPUTotalSecs(), timer.getCPUTotalSecs(),
         ncolgen,
         ncutgen,
         mAddColsToLPTimer.getCPUTotalSecs(),  
         mAddCutsToLPTimer.getCPUTotalSecs(), 
         mSolveLPTimer.getCPUTotalSecs(),
         fixredcosttimer.getCPUTotalSecs() );
   }
   else
   {
      fprintf(fout,"%s & %g & %g & %d & %d & %g & %g & %g & %g & %g & %d & %d & %g & %g & %g & %g  \\\\ \n",
         instanceFileName, bestUB, roottime, kcycleparam, nexplorednodes,
         getSkew(), mColGenTimer.getCPUTotalSecs(), 
         mCutGenTimer.getCPUTotalSecs(), 
         sbtimer.getCPUTotalSecs(), timer.getCPUTotalSecs(),
         ncolgen,
         ncutgen,
         mAddColsToLPTimer.getCPUTotalSecs(),  
         mAddCutsToLPTimer.getCPUTotalSecs(), 
         mSolveLPTimer.getCPUTotalSecs(),
         fixredcosttimer.getCPUTotalSecs() ); 
   }
   
   // Close the file
   if ( fout )
      fclose(fout);
   
#endif
   
}

//---------------------------------------------------------------------------//

void CVRPMaster::computeDPDist( CarpIO* carp )
{
  register int i;
  register int j;

  int      cvrpSize = carp->cvrpSize;
  int      capacity = carp->capacity;
  int**    edgeTab  = carp->edgeTab;
  double** carpDist = carp->carpDist;

  // -------------------------------
  dpDist = new IODist* [cvrpSize+1];
  for (i = 0; i <= cvrpSize; i++)
    dpDist[i] = new IODist [cvrpSize+1];
  // -------------------------------

  // -------------------------------
  dpCarp = new dynPrgElem** [cvrpSize+1];
  for (i = 0; i <= cvrpSize; i++)
  {
    dpCarp[i] = new dynPrgElem* [capacity+1];
    for (j = 0; j <= capacity; j++)
      dpCarp[i][j] = new dynPrgElem [2];
  }
  // -------------------------------

  // Distance between depot node and nodes m_ij
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  dpDist[cvrpSize][cvrpSize].d[0][0] = 0;
  dpDist[cvrpSize][cvrpSize].d[0][1] = 0;
  dpDist[cvrpSize][cvrpSize].d[1][0] = 0;
  dpDist[cvrpSize][cvrpSize].d[1][1] = 0;
  for (i = 1; i < cvrpSize; i++)
  {
    dpDist[cvrpSize][i].d[0][0] = (floor(carpDist[edgeTab[i][0]][edgeTab[i][1]] / 2)) +
                                  (carp->fw.minDist[0][edgeTab[i][0]]);
    dpDist[cvrpSize][i].d[1][0] = dpDist[cvrpSize][i].d[0][0];

    dpDist[cvrpSize][i].d[0][1] = ( ceil(carpDist[edgeTab[i][0]][edgeTab[i][1]] / 2)) +
                                  (carp->fw.minDist[0][edgeTab[i][1]]);
    dpDist[cvrpSize][i].d[1][1] = dpDist[cvrpSize][i].d[0][1];

    dpDist[i][cvrpSize].d[0][0] = dpDist[cvrpSize][i].d[0][0];
    dpDist[i][cvrpSize].d[0][1] = dpDist[cvrpSize][i].d[1][0];
    dpDist[i][cvrpSize].d[1][0] = dpDist[cvrpSize][i].d[0][1];
    dpDist[i][cvrpSize].d[1][1] = dpDist[cvrpSize][i].d[1][1];
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  // Distance between two nodes m_ij
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  for (i = 1; i < cvrpSize; i++)
  {
    dpDist[i][i].d[0][0] = 0;
    dpDist[i][i].d[0][1] = 0;
    dpDist[i][i].d[1][0] = 0;
    dpDist[i][i].d[1][1] = 0;

    for (j = 1; j < i; j++)
    {
      dpDist[i][j].d[0][0] = (floor(carpDist[edgeTab[i][0]][edgeTab[i][1]] / 2)) + 
                             (floor(carpDist[edgeTab[j][0]][edgeTab[j][1]] / 2)) + 
                             (carp->fw.minDist[edgeTab[i][0]][edgeTab[j][0]]);

      dpDist[i][j].d[0][1] = (floor(carpDist[edgeTab[i][0]][edgeTab[i][1]] / 2)) + 
                             ( ceil(carpDist[edgeTab[j][0]][edgeTab[j][1]] / 2)) + 
                             (carp->fw.minDist[edgeTab[i][0]][edgeTab[j][1]]);

      dpDist[i][j].d[1][0] = ( ceil(carpDist[edgeTab[i][0]][edgeTab[i][1]] / 2)) + 
                             (floor(carpDist[edgeTab[j][0]][edgeTab[j][1]] / 2)) + 
                             (carp->fw.minDist[edgeTab[i][1]][edgeTab[j][0]]);

      dpDist[i][j].d[1][1] = ( ceil(carpDist[edgeTab[i][0]][edgeTab[i][1]] / 2)) + 
                             ( ceil(carpDist[edgeTab[j][0]][edgeTab[j][1]] / 2)) + 
                             (carp->fw.minDist[edgeTab[i][1]][edgeTab[j][1]]);

      dpDist[j][i].d[0][0] = dpDist[i][j].d[0][0];
      dpDist[j][i].d[0][1] = dpDist[i][j].d[1][0];
      dpDist[j][i].d[1][0] = dpDist[i][j].d[0][1];
      dpDist[j][i].d[1][1] = dpDist[i][j].d[1][1];
    }
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/*
  for (i = 1; i <= cvrpSize; i++)
    for (j = 1; j <= cvrpSize; j++)
      for (a = 0; a < 2; a++)
        for (b = 0; b < 2; b++)
          printf( "%d | %d | %d %d | %.2lf\n", i, j, a, b, dpDist[i][j].d[a][b] );
*/

}

//---------------------------------------------------------------------------//

void CVRPMaster::readDataCvrp(char* fname, CVRPFullGraph* g)
{
   char   readBuffer[256]; // auxiliary read buffer
   char   weightType[50];  // edge weight type
   register int    i;               // counter
   register int    j;               // counter
   int    n;               // number of vertex
   int    aux1;            // auxiliary variable
   int    aux2;            // auxiliary variable
   int    auxMdc;          // mdc of the demands
   double d;               // auxiliary variable
   FILE*  fin;             // input file
   
  // Try to open the input file
  printf( "Opening file %s\n", fname );
  fin = fopen( fname, "rt" );
  
  if (!fin)
  {
    printf( "Unable to open file %s\n", fname );
    throw(FILENOTOPENEXCEPTION);
    return;
  }
  
  // Find the problem dimension
  fscanf( fin, "%s\n", readBuffer );
  while (!strNCaseCmp( "DIMENSION", readBuffer, 9 ))
    fscanf( fin, "%s\n", readBuffer );
  
  fscanf( fin, ":\n" );
  fscanf( fin, "%i\n", &n );
  
  // Allocates the graph
  depot = n;
  g->allocate( n, (n * (n - 1) / 2) );
  
  // Edge weight type
  fscanf( fin, "%s\n", readBuffer );
  while (!strNCaseCmp( "EDGE_WEIGHT_TYPE", readBuffer, 16 ))
    fscanf( fin, "%s\n", readBuffer );
  
  fscanf( fin, ":\n" );
  fscanf( fin, "%s\n", weightType );
  
  // Capacity of each vehicle
  fscanf( fin, "%s\n", readBuffer );
  while (!strNCaseCmp( "CAPACITY", readBuffer, 8 ))
    fscanf( fin, "%s\n", readBuffer );
  
  fscanf( fin, ":\n" );
  
  // Sets the vehicle capacity
  fscanf( fin, "%i\n", &vehicleCapacity );
  
  if (strNCaseCmp( "EUC_2D", weightType, 6 ))
  {
    // Node coordinates
    fscanf( fin, "%s\n", readBuffer );
    while (!strNCaseCmp( "NODE_COORD_SECTION", readBuffer, 18 ))
      fscanf( fin, "%s\n", readBuffer );
    
    // Read the node coordinates
    //fscanf( fin, "%i %i %i", &aux1, &aux2, &aux3 );
    fscanf( fin, "%i\n",&aux1);
    fscanf( fin, "%s\n",readBuffer);
    //g->setVertexX( n, aux2 );
    //g->setVertexY( n, aux3 );
    g->setVertexX( n, atof(readBuffer) );
    fscanf( fin, "%s\n",readBuffer);
    g->setVertexY( n, atof(readBuffer) ); 
    for (i = 1; i < n; i++)
    {
      //fscanf( fin, "%i %i %i", &aux1, &aux2, &aux3 );
      //g->setVertexX( i, aux2 );
      //g->setVertexY( i, aux3 );
      fscanf( fin, "%i\n", &aux1 );
      fscanf( fin, "%s\n", readBuffer );
      g->setVertexX( i, atof(readBuffer ) );
      fscanf( fin, "%s\n", readBuffer );
      g->setVertexY( i, atof( readBuffer ) );
    }
    
    // Calculates the distance matrix
    sumAllEdges = 0;
    
    for (i = 2; i <= n; i++)
    {
      for (j = 1; j < i; j++)
      {
        d = ((g->getVertexX(i) - g->getVertexX(j))*(g->getVertexX(i) - g->getVertexX(j))) +
            ((g->getVertexY(i) - g->getVertexY(j))*(g->getVertexY(i) - g->getVertexY(j)));
        
        d = sqrt( d );

        if( params->getFracPrec() > 0 )
        {
           double precision = pow( 10.0, params->getFracPrec() );
           g->setEdgeCost( i, j, floor(d * precision + 0.5) / precision );
//           writeLog( true, "Edge %d, %d, cost = %g\n",
//                i, j, floor(d * precision + 0.5) / precision );
           sumAllEdges += (double) floor(d * precision + 0.5) / (double) precision;
        }
        else
        {
           g->setEdgeCost( i, j, ((int) (d + 0.5)) );
           sumAllEdges += (int) (d + 0.5);
        }
      }
    }
    
    // Sets the format of the file to be printed in tsp view
    tspView = true;
  }
  else
    if (strNCaseCmp( "EXPLICIT", weightType, 8 ))
    {
      // Edge weights
      fscanf( fin, "%s\n", readBuffer );
      while (!strNCaseCmp( "EDGE_WEIGHT_SECTION", readBuffer, 19 ))
        fscanf( fin, "%s\n", readBuffer );
      
      // Calculates the distance matrix
      sumAllEdges = 0;

      for (i = 1; i < n; i++)
      {
        for (j = 0; j < i; j++)
        {
           double auxd;
           fscanf( fin, "%lf", &auxd );
           if( params->isObjInteger() )
           {
              if (j == 0)
                g->setEdgeCost( n, i, (int) (auxd + 0.5) );
              else
                g->setEdgeCost( i, j, (int) (auxd + 0.5) );
              sumAllEdges += (int) (auxd + 0.5);
           }
           else
           {
              if (j == 0)
                 g->setEdgeCost( n, i, auxd );
              else
                g->setEdgeCost( i, j, auxd );
              sumAllEdges += auxd;
           }
        }
      }

      // Sets the format of the file to be printed in tsp view
      tspView = false;
    }
  
  // Find the node demands
  fscanf( fin, "%s\n", readBuffer );
  while (!strNCaseCmp( "DEMAND_SECTION", readBuffer, 14 ) )
    fscanf( fin, "%s\n", readBuffer );
  
  // Allocate the demand list
  listDemands = new int[n + 1];
  
  // Read the demands
  fscanf( fin, "%i %i\n", &aux1, &aux2 );
  listDemands[n] = aux2;
  sumAllDemands  = aux2;
  for (i = 1; i < n; i++)
  {
    fscanf( fin, "%i %i\n", &aux1, &aux2 );
    listDemands[i] = aux2;
    sumAllDemands += aux2;
  }
  
  // Calculate the mcd among the demands and the capacity
  auxMdc = MDC( vehicleCapacity, listDemands[1] );
  
  for (i = 2; (i < n) && (auxMdc > 1); i++) 
    auxMdc = MDC( auxMdc, listDemands[i] );

  if (auxMdc > 1)
  {
    vehicleCapacity /= auxMdc;
    sumAllDemands   /= auxMdc;
    
    for (i = 1; i < n; i++)
      listDemands[i] /= auxMdc;
  }

  // Preprocess the depot edges
  prepDepotEdges();
  
  // Close the input file
  if (fin)
    fclose( fin );
}


//---------------------------------------------------------------------------//
//---------------------------------------------------------------------------//

void CVRPMaster::readDataCarp(char* fname, CVRPFullGraph* g)
{
  register int i;      // counter
  register int j;      // counter
           int n;      // number of vertex
           int auxMdc; // mdc of the demands
           int d;      // auxiliary variable

  CarpIO carp;

  carp.readData( fname );
  if (pearn)
    carp.toCvrpByPearn();
  else
    if (uchoa)
    {
      carp.toCvrpByUchoa( nUchoa );
//      carp.writeCvrpDist();
//      carp.writeCvrpDem();
    }
    else
    {
      carp.toCvrpByLongo();
//      carp.writeCvrpDist();
//      carp.writeCvrpDem();
    }


  // Sets the best upper bound
  // - - - - - - - - - - - - -
  if (carp.carpUB < bestUB)
    bestUB = carp.carpUB;
  // - - - - - - - - - - - - -


  // Find the problem dimension
  // - - - - - - - - - - - - - -
  n = carp.cvrpSize;
  // - - - - - - - - - - - - - -


  // Allocates the graph
  // - - - - - - - - - - - - - - - - -
  g->allocate( n, (n * (n - 1) / 2) );
  // - - - - - - - - - - - - - - - - -


  // Sets the vehicle capacity
  // - - - - - - - - - - - - - - -
  vehicleCapacity = carp.capacity;
  // - - - - - - - - - - - - - - -


  // Node coordinates
  // - - - - - - - - - - - - - - - -
  depot = n;
  
  g->setVertexX( n, 0.0 );
  g->setVertexY( n, 0.0 );

  for (i = 1; i < n; i++)
  {
    g->setVertexX( i, (double) i );
    g->setVertexY( i, (double) i );
  }
  // - - - - - - - - - - - - - - - -


  // Calculates the distance matrix
  // - - - - - - - - - - - - - - - -
  sumAllEdges = 0;

  for (i = 1; i < n; i++)
  {
    d = (int) carp.cvrpDist[i][0];
    g->setEdgeCost( n, i, d );

    if (d > bestUB)
      g->setEdgeFixed( n, i, 0 );
    else
      sumAllEdges += d;
  }

  for (i = 2; i < n; i++)
  {
    for (j = 1; j < i; j++)
    {
      d = (int) carp.cvrpDist[i][j];
      g->setEdgeCost( i, j, d );

      if (d > bestUB)
        g->setEdgeFixed( i, j, 0 );
      else
        sumAllEdges += d;

      if ((uchoa) && ((i % 2) == 0) && (j == (i - 1)))
      {
        g->setEdgeFixed( i, j, 1 );
//        printf( "(%d,%d) = %d     %d\n", i, j, g->getEdgeIndex( i, j ), d );
      }

//      if ((uchoa) && (d == 0))
//        g->setEdgeFixed( i, j, 1 );
    }
  }
  // - - - - - - - - - - - - - - - -


  // Allocate the demand list and read the demands
  // - - - - - - - - - - - - - - - - - - - - - - -
  listDemands = new int[n + 1];
  
  listDemands[n] = (int) carp.cvrpDem[0];
  sumAllDemands  = (int) carp.cvrpDem[0];

  for (i = 1; i < n; i++)
  {
    listDemands[i]  = (int) carp.cvrpDem[i];
    sumAllDemands  += (int) carp.cvrpDem[i];
  }
  // - - - - - - - - - - - - - - - - - - - - - - -


  // Calculate the mdc among the demands and the capacity
  // - - - - - - - - - - - - - - - - - - - - - - - - - - -
  auxMdc = MDC( vehicleCapacity, listDemands[1] );
  
  for (i = 2; (i < n) && (auxMdc > 1); i++) 
    auxMdc = MDC( auxMdc, listDemands[i] );
  
  if (auxMdc > 1)
  {
    vehicleCapacity /= auxMdc;
    sumAllDemands   /= auxMdc;
    
    for (i = 1; i < n; i++)
      listDemands[i] /= auxMdc;
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - -


  // Preprocess the depot edges
  // - - - - - - - - - - - - - -
  prepDepotEdges();
  // - - - - - - - - - - - - - -


  // Sets the format of the file to be printed in tsp view
  // - - - - - - - - - - - - - - - - - - - - - - - - - - -
  tspView = true;
  // - - - - - - - - - - - - - - - - - - - - - - - - - - -

  if (longo)
    computeDPDist( &carp );
}

//---------------------------------------------------------------------------//
//---------------------------------------------------------------------------//
//---------------------------------------------------------------------------//
//---------------------------------------------------------------------------//

void CVRPMaster::prepDepotEdges()
{
   register int i; // counter
   int dem; // auxiliary demand
   int capRes; // resultant capacity
   int count; // auxiliary counter

   minCapVehicle = (int)(sumAllDemands) - ((numberOfVehicles - 1) * vehicleCapacity);
   count = 0;

   if (minCapVehicle < 0)
      minCapVehicle = 0;

   capRes = vehicleCapacity * (numberOfVehicles-1);

   writeLog( true, "\nPreprocessing depot edges...\n" );

   for (i=0;i<gMaster->getM();i++)
   {
      if (gMaster->getEdgeOrg(i) == depot)
      {
         dem=(int)(sumAllDemands) - listDemands[gMaster->getEdgeDst(i)];

         if (dem > capRes)
         {
            count++;
            gMaster->setCanBeTwoEdge(i,false);
         }
      }
      else if (gMaster->getEdgeDst(i) == depot)
      {
         dem=(int)(sumAllDemands) - listDemands[gMaster->getEdgeOrg(i)];

         if (dem > capRes)
         {
            count++;
            gMaster->setCanBeTwoEdge(i,false);
         }
      }
   }

   writeLog( true, "Number of depot edges that cannot be two: %i\n\n",count );
}

//---------------------------------------------------------------------------//
void CVRPMaster::printUsage(char** argv)
{
   fprintf(stderr,"Invalid arguments\n");
   fprintf(stderr,"Usage = %s fname NV UB [options]\n", argv[0]);
   fprintf(stderr,"        fname: Instance file name (.vrp file)\n");
   fprintf(stderr,"        NV:    Number of vehicles\n");
   fprintf(stderr,"        UB:    Upper bound\n");
   fprintf(stderr,"    Options:\n");
   fprintf(stderr,"        -k nn : nn represents the number k of k-cycles to be eliminated (does not check if nn is a valid string describing a number) (default=3)\n");
   fprintf(stderr,"        -root : Only solve the root node\n");
   fprintf(stderr,"        -itercg : Sets iterative k-cycle colgen (default = false)\n");
   fprintf(stderr,"        -revbb  : Reverse order of BBnode choice (default = false)\n");
   fprintf(stderr,"        -dynk   : Choose dynamic choice of k (default= false)\n");
/*   fprintf(stderr,"        -pmin p1 : Sets the pmin parameter to p1 (default=5)\n");*/
/*   fprintf(stderr,"        -pmax p2 : Sets the pmax parameter to p2 (default=8)\n");*/
/*   fprintf(stderr,"        -sbcgiter ni : Sets the numb. of cg iters to perform SB on to be ni (default = 20)\n");*/
/*   fprintf(stderr,"        -carp : solve carp instances using cvrp techniques\n");*/
/*   fprintf(stderr,"        -pearn : use Pearn's carp graph transformation (default = false)\n");*/
/*   fprintf(stderr,"        -longo : use Longos's carp graph transformation (default = false)\n");*/
/*   fprintf(stderr,"        -uchoa nn : use Uchoa's carp graph transformation number nn (default = false)\n");*/
/*   fprintf(stderr,"        -vrptw : solve vrp with time windows instances\n");*/
/*   fprintf(stderr,"        -bimp bb: Estimate (absolute value) on how much a branching improves the LB (default = 0.001 * UB)\n");*/
   fprintf(stderr,"        -stopfath: Stop after first fathoming\n");
/*   fprintf(stderr,"        -rootprox rr: Says what means to be close to the root (UB-LB)/UB (default = 0.005)\n");*/
/*   fprintf(stderr,"        -leafprox ll: Says what means to be close to the leaf (UB-LB)/UB (default = 0.002)\n");*/
   fprintf(stderr,"        -bdtarget tt: Parameter to order sets to try SB (default 2.85)\n");
   fprintf(stderr,"        -comm string: Comment to be added to the logfile\n");
   fprintf(stderr,"        -suff s: Suffix to be added to logfile name\n");
/*   fprintf(stderr,"        -x value: Percentage of gain g that a good SB has to have as score (default = 0.8)\n");*/
   fprintf(stderr,"        -stab: Turns on dual stabilization\n");
/*   fprintf(stderr,"        -sb: Turns on SB\n");*/
   fprintf(stderr,"        -fracprec: Number of digits of fractional accuracy (default = 0, i.e. integer costs)\n");
/*   fprintf(stderr,"        -figsol: Name of figure solution that has UB\n");*/
/*   fprintf(stderr,"        -nodesel n: Node selection strategy (0=DFS, 1=BFS, 2=Best Bound: default=0)\n");*/
}



void CVRPMaster::readData(int argc, char **argv)
{
   int countarg;
   int auxUB; // auxiliary upper bound
   int size; // auxiliary size
   int count; // auxiliary counter
   int auxK; // auxiliary number of vehicles
   register int i; // counter
   char fname[100]=""; // file name
   char fprefix[100]; // prefix of file name
   char figsol[100] = ""; // File name of .fig solution 
   bool flag; // auxiliary flag
   
   if ( argc < 4 )
   {
      printUsage(argv);
      throw(PROMPTINVALIDPARAMNUMEXCEPTION);
   }
   
   // Find the file name 
   count=0;
   size = (int)(strlen(argv[1]));
   
   for (i = size-1; i>=0;i--)
   {
      if ( argv[1][i] =='/' || argv[1][i] =='\\' )
      {
         count = i+1;
         break;
      }
   }
   
   for (i=count;i<size;i++)
      fname[i-count] = argv[1][i];
   
   fname[i-count]='\0';
   
   
   //   strcpy(instanceFileName,fname);
   strcpy(instanceFileName,argv[1]);
   
   // Find the prefix of file name 
   count=0;
   size = (int)(strlen(fname));
   flag = false;
   
   for (i = size-1; i>=0;i--)
   {
      if ( fname[i] =='.')
      {
         flag=true;
         count = i;
         break;
      }
   }
   if (flag)
   {
      for (i=0;i<count;i++)
      {
         fprefix[i]=fname[i];
      }
      fprefix[count]='\0';
   }
   else
   {
      strcpy(fprefix,fname);
   }
   
   // Prefix of solution files
   strcpy( prefixFileName, fprefix );
   
   // Sets the log file
#ifdef WIN32
   strcpy(fname, "log\\");
#else
   strcpy(fname, "log/" );
#endif
   
   char fsuffix[100];
   strcpy( fsuffix, "" );
   strcat(fname, fprefix);
   strcat(fname, ".log");
   printf("Setting log file name = %s\n",fname);
   setLogFile(fname);
   
   auxK = atoi(argv[2]);
   
   if ( auxK <= 0 )
   {
      printUsage(argv);
      throw(INVALIDNUMOFVEHICLESEXCEPTION);
   }
   
   numberOfVehicles = auxK;
   
   auxUB = atoi(argv[3]);
   
   for( countarg = 4; countarg < argc; countarg++ )
   {
      if( strcmp( argv[countarg], "-k" ) == 0 )
      {
         if( argc <= countarg + 1 )
         {
            writeLog( true, "Error. Invalid number of arguments\n" );
            printUsage(argv);
            throw(PROMPTINVALIDPARAMNUMEXCEPTION);
         }
         kcycleparam = atoi( argv[ ++countarg ] );
         
         if( kcycleparam < 0 )
         {
            writeLog( true, "Parameter K can't be negative\n" );
            printUsage(argv);
            throw(PROMPTINVALIDPARAMK);
         }
         
         writeLog( true, "Option: Using kcycle = %d\n", kcycleparam );
      }
      else if( strcmp( argv[countarg],"-root" ) == 0 )
      {
         writeLog( true, "Option: Solving only ROOT Node\n" );
         rootonly = true;
      }
      else if( strcmp( argv[countarg],"-itercg" ) == 0 )
      {
         writeLog( true, "Option: Using iterative kcycle colgen\n" );
         itercg = true;
      }
      else if( strcmp( argv[countarg],"-revbb" ) == 0 )
      {
         writeLog( true, "Option: Using reverse choice of BB node\n" );
         revbb = true;
      }
      else if( strcmp( argv[countarg],"-dynk" ) == 0 )
      {
         writeLog( true, "Option: Using dynamic choice of k\n" );
         dynk = true;
      }
      else if( strcmp( argv[countarg],"-pmin" ) == 0 )
      {
         if( argc <= countarg + 1 )
         {
            writeLog( true, "Error. Invalid number of arguments\n" );
            printUsage(argv);
            throw(PROMPTINVALIDPARAMNUMEXCEPTION);
         }
         pmin =	atoi( argv[++countarg]);
         if( pmin < 0 )
         {
            writeLog( true, "Parameter pmin can't be negative\n" );
            printUsage(argv);
            throw(PROMPTINVALIDPARAMPMIN);
         }
         writeLog( true, "Option: Using pmin=%d\n", pmin );
      }
      else if( strcmp( argv[countarg],"-pmax" ) == 0 )
      {
         if( argc <= countarg + 1 )
         {
            writeLog( true, "Error. Invalid number of arguments\n" );
            printUsage(argv);
            throw(PROMPTINVALIDPARAMNUMEXCEPTION);
         }
         pmax =	atoi( argv[++countarg]);
         if( pmax < 0 )
         {
            writeLog( true, "Parameter pmin can't be negative\n" );
            printUsage(argv);
            throw(PROMPTINVALIDPARAMPMAX);
         }
         writeLog( true, "Option: Using pmax=%d\n", pmax );
      }
      else if( strcmp( argv[countarg],"-sbcgiter" ) == 0 )
      {
         if( argc <= countarg + 1 )
         {
            writeLog( true, "Error. Invalid number of arguments\n" );
            printUsage(argv);
            throw(PROMPTINVALIDPARAMNUMEXCEPTION);
         }
         sbcgiter = atoi( argv[++countarg]);
         if( sbcgiter < 0 )
         {
            writeLog( true, "Parameter pmin can't be negative\n" );
            printUsage(argv);
            throw(PROMPTINVALIDPARAMSBCGITER);
         }
         writeLog( true, "Option: Using sbcgiter=%d\n", sbcgiter );
      }
      else if( strcmp( argv[countarg],"-carp" ) == 0 )
      {
         writeLog( true, "Option: carp instance will be solved\n" );
         carp = true;
      }
      else if( strcmp( argv[countarg],"-pearn" ) == 0 )
      {
         writeLog( true, "Option: Pearns carp graph transformation will be used\n" );
         pearn = true;
      }
      else if( strcmp( argv[countarg],"-longo" ) == 0 )
      {
         writeLog( true, "Option: Longo's carp graph transformation will be used\n");
         longo = true;
      }
      else if( strcmp( argv[countarg],"-uchoa" ) == 0 )
      {
         if( argc <= countarg + 1 )
         {
            writeLog( true, "Error. Invalid number of arguments\n" );
            printUsage(argv);
            throw(PROMPTINVALIDPARAMNUMEXCEPTION);
         }
         nUchoa =	atoi( argv[++countarg]);
         if ((nUchoa < 1) || (nUchoa > 3))
         {
            writeLog( true, "Parameter uchoa must be 1, 2 or 3\n" );
            printUsage(argv);
            throw(PROMPTINVALIDPARAMPMAX);
         }
         writeLog( true, "Option: Uchoas carp graph transformation number %d will be used\n", nUchoa );
         uchoa = true;
      }
      else if( strcmp( argv[countarg],"-vrptw" ) == 0 )
      {
         writeLog( true, "Option: vrp with time windows will be solved\n");
         vrptw = true;
      }
      else if( strcmp( argv[countarg],"-bimp" ) == 0 )
      {
         if( argc <= countarg + 1 )
         {
            writeLog( true, "Error. Invalid number of arguments\n" );
            printUsage(argv);
            throw(PROMPTINVALIDPARAMNUMEXCEPTION);
         }
         bimp = atof( argv[++countarg]);
         writeLog( true, "Option: Branch improvement estimate = %g\n", bimp );
      }
      else if( strcmp( argv[countarg],"-stopfath" ) == 0 )
      {
         writeLog( true, "Option: Stop after first fathoming\n");
         stopfath = true;
      }
      else if( strcmp( argv[countarg],"-rootprox" ) == 0 )
      {
         if( argc <= countarg + 1 )
         {
            writeLog( true, "Error. Invalid number of arguments\n" );
            printUsage(argv);
            throw(PROMPTINVALIDPARAMNUMEXCEPTION);
         }
         rootprox = atof( argv[++countarg]);
         writeLog( true, "Option: rootprox: Close to root means (UB-LB)/UB = %g\n", rootprox );
      }
      else if( strcmp( argv[countarg],"-leafprox" ) == 0 )
      {
         if( argc <= countarg + 1 )
         {
            writeLog( true, "Error. Invalid number of arguments\n" );
            printUsage(argv);
            throw(PROMPTINVALIDPARAMNUMEXCEPTION);
         }
         leafprox = atof( argv[++countarg]);
         writeLog( true, "Option: leafprox: Close to leaves means (UB-LB)/UB = %g\n", leafprox );
      }
      else if( strcmp( argv[countarg],"-bdtarget" ) == 0 )
      {
         if( argc <= countarg + 1 )
         {
            writeLog( true, "Error. Invalid number of arguments\n" );
            printUsage(argv);
            throw(PROMPTINVALIDPARAMNUMEXCEPTION);
         }
         bdtarget = atof( argv[++countarg]);
         writeLog( true, "Option: bdtarget = %g\n", bdtarget );
      }
      else if( strcmp( argv[countarg],"-comm" ) == 0 )
      {
         if( argc <= countarg + 1 )
         {
            writeLog( true, "Error. Invalid number of arguments\n" );
            printUsage(argv);
            throw(PROMPTINVALIDPARAMNUMEXCEPTION);
         }
         writeLog( true, "\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n" );
         writeLog( true, "Comment: %s\n", argv[++countarg] );
         writeLog( true, "\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n" );
      }
      else if( strcmp( argv[countarg],"-suff" ) == 0 )
      {
         if( argc <= countarg + 1 )
         {
            writeLog( true, "Error. Invalid number of arguments\n" );
            printUsage(argv);
            throw(PROMPTINVALIDPARAMNUMEXCEPTION);
         }
         strcpy( fsuffix, argv[++countarg] );
      }
      else if( strcmp( argv[countarg], "-x" ) == 0 )
      {
         if( argc <= countarg + 1 )
         {
            writeLog( true, "Error. Invalid number of arguments\n" );
            printUsage(argv);
            throw(PROMPTINVALIDPARAMNUMEXCEPTION);
         }
         goodfactor = atof( argv[++countarg]);
         writeLog( true, "Option: -x = %g\n", goodfactor );
      }
      else if( strcmp( argv[countarg],"-stab" ) == 0 )
      {
         writeLog( true, "Option: Turning on dual stabilization\n" );
         dostab = true;
      }
      else if( strcmp( argv[countarg],"-sb" ) == 0 )
      {
         writeLog( true, "Option: Turning on SB\n" );
         dosb = true;
      }
      else if( strcmp( argv[countarg],"-fracprec" ) == 0 )
      {
         if( argc <= countarg + 1 )
         {
            writeLog( true, "Error. Invalid number of arguments\n" );
            printUsage(argv);
            throw(PROMPTINVALIDPARAMNUMEXCEPTION);
         }
         int fracprec = atoi( argv[++countarg]);
         params->setFracPrec( fracprec );
         if( fracprec > 0 )
            params->setObjInteger( false );
         writeLog( true, "Option: Fractional precision = %d\n", params->getFracPrec() );
      }
      else if( strcmp( argv[countarg],"-figsol" ) == 0 )
      {
         if( argc <= countarg + 1 )
         {
            writeLog( true, "Error. Invalid number of arguments\n" );
            printUsage(argv);
            throw(PROMPTINVALIDPARAMNUMEXCEPTION);
         }
         strcpy( figsol, argv[++countarg] );
      }
      else if( strcmp( argv[countarg], "-nodesel" ) == 0 )
      {
         if( argc <= countarg + 1 )
         {
            writeLog( true, "Error. Invalid number of arguments\n" );
            printUsage(argv);
            throw(PROMPTINVALIDPARAMNUMEXCEPTION);
         }
         params->setNodeSel( atoi( argv[ ++countarg ] ) );
         
         if( ( params->getNodeSel() < 0 ) || ( params->getNodeSel() > 2 ) )
         {
            writeLog( true, "Unknown node selection parameter\n" );
            printUsage(argv);
            throw(PROMPTINVALIDPARAMNODESEL);
         }
         
         writeLog( true, "Option: Using node selection = %d\n", params->getNodeSel() );
      }
      else
      {
         writeLog( true, "Invalid option\n" );
         printUsage(argv);
         throw(INVALIDINPUTOPTION);
      }
   }
   if( pmax < pmin )
   {
      writeLog( true, "Error: Parameter pmax has to be greater than or equal to pmin\n" );
      throw( PROMPTINVALIDPARAMPMAX );
   }
   
   if (auxUB < bestUB)
      bestUB = auxUB;
   
   // Read the data instance
   if (carp)
   {
      readDataCarp(instanceFileName, gMaster);
      if (pearn)
         bestUB *= 4;
      else
         if (longo)
            kcycleparam = 2;
   }
   else
   {
      if (vrptw)
      {
         //       readDataVrptw(instanceFileName, gMaster);
      }
      else
      {
         readDataCvrp(instanceFileName, gMaster);
      }
   }

   bestlpe = new double[gMaster->getM()];
   if (!bestlpe)
      throw(LOWMEMORYEXCEPTION);

   if( strlen(figsol) > 0 )
   {
      readFig( figsol );
   }
   
   // If no bimp was read, then default
   if( bimp == 0.0 )
      bimp = 0.001 * auxUB;
   setSolSize(gMaster->getN());
   
   // Sets the log file
#ifdef WIN32
   strcpy(endlogfname, "log\\");
#else
   strcpy(endlogfname, "log/" );
#endif
   
   strcat(endlogfname, fprefix);
   strcat(endlogfname, fsuffix);
   strcat(endlogfname, ".log");
   
   strcat( prefixFileName, fsuffix );
}

//---------------------------------------------------------------------------//


void CVRPMaster::writeFeasibleSol( char* prefix, double* xsol,
CVRPFullGraph* g )
{
  int i; // counter
  int countM; // count the edges of the solution
  FILE *fout; // output file
  char figName[100]; // name of the output file
#ifndef WIN32
  strcpy( figName, "fig/");
#else
  strcpy( figName, "fig\\" );
#endif
  strcat( figName, prefix);
  strcat( figName, ".sol");

  int** delta = new int*[ g->getN()+1 ];
  int* countdelta = new int[ g->getN()+1 ];
  char* marked = new char[ g->getN() + 1 ];
  int org, dst;
  double cost;

  for( i = 0; i <= g->getN(); i++ )
  {
     if( i == g->getN() )
     {
        delta[i] = new int[ g->getN() ];
     }
     else
     {
        delta[i] = new int[2];
     }
  }
  memset( countdelta, 0, (g->getN()+1) * sizeof(int) );
  memset( marked, 0, (g->getN()+1) * sizeof(char) );

  fout = fopen(figName,"wt");

  if( !fout )
  {
     throw(FILENOTOPENEXCEPTION);
     return;
  }

  countM = 0;
  cost = 0;
  for( i=0; i < g->getM(); i++)
  {
    if( xsol[i] > params->getIntEps() )
    {
      org = g->getEdgeOrg(i);
      dst = g->getEdgeDst(i);
      delta[org][ countdelta[org]++ ] = dst;
      delta[dst][ countdelta[dst]++ ] = org;
      countM++;
      cost += g->getEdgeCost(i);
    }
  }

  fprintf(fout, "Number of edges: %d\n", countM );

  // Now print the route:
  int next;
  int nroutes=1;
  marked[ g->getN() ] = 1;
  for( i = 0; i < countdelta[ g->getN() ]; i++)
  {
     next = delta[g->getN()][i];
     if( ! marked[ next ] )
     {
        // Then print the route:
	fprintf(fout, "\nRoute #%d: ", nroutes++ );

	while( next != g->getN() )
	{
	   fprintf(fout,"%d ", next );
	   marked[next] = 1;
	   if( ! marked[ delta[next][0] ] )
	   {
	      next = delta[next][0];
	   }
	   else
	   {
	      if( countdelta[next] == 1 )
	      {
	         next = delta[next][0];
	      }
	      else if( ! marked[ delta[next][1] ] )
	      {
	         next = delta[next][1];
	      }
	      else if( ( delta[next][0] != g->getN() ) && 
	               ( delta[next][1] != g->getN() ) )
	      {
	         throw( 100 );
	      }
	      else
	      {
	         next = g->getN();
	      }
	   }
	}
     }
  }
  
  if (getPearn())
    fprintf(fout,"\n\nCost= %g\n", cost / 4 );
  else
    fprintf(fout,"\n\nCost= %g\n", cost );

  fprintf( fout, "\nBEGINDEMAND\n" );

  for (i=1;i<g->getN();i++)
  {
     fprintf(fout, "%d  %d\n", i, listDemands[i]);
  }

  fprintf( fout, "ENDDEMAND\n" );

  if ( fout )
     fclose( fout );
  delete [] countdelta;
  for( i = 0; i <= g->getN(); i++ )
     delete [] delta[i];

  delete [] delta;
  delete [] marked;  
  return;
}

//---------------------------------------------------------------------------//
void CVRPMaster::writeSolTSPView(CVRPBBNode *node, char* prefix, double* xsol, CVRPFullGraph* g )
{
  register int i; // counter
  int countM; // count the edges of the solution
  FILE *fout; // output file
  char figName[100]; // name of the output file
#ifndef WIN32
  strcpy( figName, "fig/");
#else
  strcpy( figName, "fig\\" );
#endif
  strcat( figName, prefix);
  strcat( figName, ".fig");

  

  if( tspView )
  {
     fout = fopen(figName,"wt");

      if( !fout )
      {
         throw(FILENOTOPENEXCEPTION);
         return;
      }

      fprintf(fout, "%d\n", g->getN());
      fprintf(fout, "%d  %d\n", (int) g->getVertexX(g->getN()), (int) g->getVertexY(g->getN()));

      for(i=1; i<g->getN(); i++)
      {
         fprintf(fout, "%d  %d\n", (int) g->getVertexX(i), (int) g->getVertexY(i));
      }
  }
  else
  {
     return;
  }

  countM = 0;

  for( i=0; i < g->getM(); i++)
  {
	  if( xsol[i] > params->getIntEps() )
	  {
        countM++;
     }
  }

  fprintf(fout, "%d\n", countM );

  for(i=0; i < g->getM(); i++)
  {
	  if( xsol[i] > params->getIntEps() )
	  {
        if( tspView )
        {
           if ( g->getEdgeOrg(i) == g->getN())
              fprintf(fout, "%d  %d  %f\n", 0, g->getEdgeDst(i), xsol[i]);
           else if ( g->getEdgeDst(i) == g->getN()  )
              fprintf(fout, "%d  %d  %f\n", g->getEdgeOrg(i), 0, xsol[i]);
           else
              fprintf(fout, "%d  %d  %f\n", g->getEdgeOrg(i), g->getEdgeDst(i), xsol[i]);
        }
	  }
  }

  node->printll(fout);

  fprintf( fout, "\nBEGINDEMAND\n" );

  fprintf(fout, "%d  %d\n", 0, 0);
  for (i=1;i<g->getN();i++)
  {
     fprintf(fout, "%d  %d\n", i, listDemands[i]);
  }

  fprintf( fout, "ENDDEMAND\n" );


  fprintf( fout, "\nROOT 0\n");

  if( tspView )
  {
     if ( fout )
        fclose( fout );
  }
  
  return;
}

void CVRPMaster::readFig( char* fname )
{
   double auxub = 0.0;
   int i, nedges;
   int auxi, auxj;
   double auxd;
   
   FILE* fin = fopen( fname, "rt" );
   if( ! fin )
   {
      throw( FILENOTOPENEXCEPTION );
   }

   fscanf( fin, "%d ", &auxi );

   if( auxi != gMaster->getN() )
   {
      writeLog( true, "Incorrect number of nodes while reading %s\n", fname );
      throw(WRONGFIGFILEEXCEPTION);
   }

   for( i = 0; i < gMaster->getN(); i++ )
   {
      fscanf( fin, "%d %d", &auxi, &auxj );
   }
   
   // Read the edges

   memset( bestlpe, 0, sizeof( double ) * gMaster->getM() );
   
   fscanf( fin, "%d", & nedges );

   for( i = 0; i < nedges; i++ )
   {
      fscanf( fin, "%d %d %lf", &auxi, &auxj, &auxd );

      if( auxi == 0 )
         auxi = gMaster->getN();

      if( auxj == 0 )
         auxj = gMaster->getN();

      if( ( fabs( auxd - 1.0 ) > params->getIntEps() )  &&
          ( fabs( auxd - 2.0 ) > params->getIntEps() ) )
      {
         writeLog( true, "Fig file does not contain an integer solution\n");
         throw(WRONGFIGFILEEXCEPTION);
      }

      //printf("(%d, %d )  ", auxi, auxj );
      //printf("Edge %d / %d\n", gMaster->getEdgeIndex( auxi, auxj ),
      //      gMaster->getM() );
      
      bestlpe[ gMaster->getEdgeIndex( auxi, auxj ) ] = 1.0; 
   
      auxub += gMaster->getEdgeCost( auxi, auxj );
   }

   if( bestUB > auxub )
   {
      writeLog( true, "Setting upper bound read from figfile = %g\n", auxub );
      bestUB = auxub;
   }

   fclose( fin );
}

//---------------------------------------------------------------------------//

bool CVRPMaster::strNCaseCmp (char *a, char *b, int n)
{
   register int i; // counter

   if (((int)strlen(a) < n) || ((int)strlen(b) < n))
      return false;

   for (i=0;i<n;i++)
      if (tolower(a[i])!=tolower(b[i]))
         return false;

   return true;
}

//---------------------------------------------------------------------------//

int CVRPMaster::MDC(int m, int n)
{
   int aux; // auxiliary variable

   if( m < n )
   {
      aux = n;
      n = m;
      m = aux;
   }

   if( n <= 1 ) 
      return 1;

   while( n != 0 )
   {
      aux = m;
      m = n;
      n = aux % n;
   }

   return m;
}

//---------------------------------------------------------------------------//


